//#include "enable_jemalloc.h"

#include <algorithm>
#include <cstdlib>
#include <iostream>
#include <string>

#include "ritch/prepro/dynamic_graph.h"
#include "ritch/prepro/witness_search.h"
#include "ritch/prepro/ordering.h"

bool cmdOptionExists(char** begin, char** end, const std::string& option)
{
	return std::find(begin, end, option) != end;
}

char* getCmdOptionImpl(char** begin, char** end, const std::string & option)
{
	char** itr = std::find(begin, end, option);

	if ( itr != end && ++itr != end )
		return *itr;

	return nullptr;
}

std::string ugly_tchar_to_string(char* foo)
{
	if ( foo == nullptr )
		return std::string("");

	std::string bar(foo);
	std::string result = std::string(bar.begin(), bar.end());
	return result;
}

std::string getCmdOption(char** begin, char** end, const std::string & option)
{
	return ugly_tchar_to_string(getCmdOptionImpl(begin, end, option));
}

std::string getCmdOptionDefault(char** begin, char** end, const std::string & option, const std::string& default_string)
{
	std::string result = ugly_tchar_to_string(getCmdOptionImpl(begin, end, option));

	if ( result.empty() ) return default_string;
	return result;
}

std::string to_text(const bool& b)
{
	if ( b ) return "true";
	return "false";
}

int main(int argc, char* argv[])
{
	//
	// Process command line options.
	//

	const bool wait_at_the_end = cmdOptionExists(argv, argv + argc, "-wait");

	if ( cmdOptionExists(argv, argv + argc, "-help") )
	{
		std::cerr
			<< "\n"
			<< "Show help message:\n"
			<< "-help\n"
			<< "\n"
			<< "Flags:\n"
			<< "-wait\n"
			<< "\n"
			<< "Parameters:\n"
			<< "-m <input graph directory>\n"
			<< "-o <output RITCH_SHORTCUTS and LAYERS file name>\n"
			<< "-b <input default blocks path name> (default: no blocks)\n"
			<< "-t <number of threads>\n";

		if ( wait_at_the_end )
			std::getchar();

		return EXIT_SUCCESS;
	}

	const std::string graph_input_directory_name = getCmdOption(argv, argv + argc, "-m");
	if ( graph_input_directory_name.empty() )
	{
		std::cerr << "expected -m <input graph directory>" << std::endl;
		std::cerr << "To list the available options use option -help" << std::endl;

		if ( wait_at_the_end )
			std::getchar();

		return EXIT_FAILURE;
	}

	std::string output_file_name = getCmdOption(argv, argv + argc, "-o");
	if ( output_file_name.empty() )
	{
		std::cerr << "expected -o <output RITCH_SHORTCUTS and LAYERS file name>" << std::endl;
		std::cerr << "To list the available options use option -help" << std::endl;

		if ( wait_at_the_end )
			std::getchar();

		return EXIT_FAILURE;
	}
	output_file_name = output_file_name.substr(0, output_file_name.find_last_of("."));

	const std::string default_blocks_input_full_file_name = getCmdOption(argv, argv + argc, "-b");


	const std::string shortcuts_output_file_name = output_file_name + ".ritch_shortcuts";
	const std::string layers_output_file_name = output_file_name + ".layers";
	const int n_threads = atoi(getCmdOptionDefault(argv, argv + argc, "-t", "1").c_str());

	std::cout << "STATUS: Input default blocks file is '" << default_blocks_input_full_file_name << "'" << std::endl;
	std::cout << "STATUS: Layers output file is '" << layers_output_file_name << "'" << std::endl;
	std::cout << "STATUS: Shortcuts output file is '" << shortcuts_output_file_name << "'" << std::endl;


	//
	// Do preprocessing
	//
	const auto t_b = ritch::util::time_stamp();
	ritch::DynamicGraph graph(graph_input_directory_name, default_blocks_input_full_file_name);
	std::cout << "STATUS: Graph has " << graph.get_n_nodes() << " nodes and " << graph.get_n_edges() << " edges." << std::endl;

	ritch::Ordering<ritch::WitnessSearch> ordering(
		std::move(graph),
		graph_input_directory_name,
		graph_input_directory_name + ritch::util::PATH_SEPARATOR() + layers_output_file_name,
		graph_input_directory_name + ritch::util::PATH_SEPARATOR() + shortcuts_output_file_name
	);

	ordering.run(n_threads);

	const auto t_e = ritch::util::time_stamp();
	std::cout << "Preprocessing took " << ritch::util::get_duration_in_seconds(t_b, t_e) << " seconds" << std::endl;

	if ( wait_at_the_end )
		std::getchar();

    return EXIT_SUCCESS;
}

