#include <cstdint>
#include <cstdlib>
#include <iostream>
#include <string>
#include <utility>
#include <vector>

#include "ritch/util/misc.h"
#include "ritch/util/timing.h"
#include "ritch/querying/tch_search_graph.h"
#include "ritch/querying/tch_querying-multilabel.h"
#include "ritch/querying/tch_querying.h"
#include "ritch/io/ritch_demands_format.h"

bool cmdOptionExists(char** begin, char** end, const std::string& option)
{
	return std::find(begin, end, option) != end;
}

char* getCmdOptionImpl(char** begin, char** end, const std::string & option)
{
	char** itr = std::find(begin, end, option);
	if ( itr != end && ++itr != end )
		return *itr;

	return nullptr;
}

std::string ugly_tchar_to_string(char* foo)
{
	if ( foo == nullptr )
		return std::string("");

	std::string bar(foo);
	std::string result = std::string(bar.begin(), bar.end());
	return result;
}

std::string getCmdOption(char** begin, char** end, const std::string & option)
{
	return ugly_tchar_to_string(getCmdOptionImpl(begin, end, option));
}

std::string getCmdOptionDefault(char** begin, char** end, const std::string & option, const std::string& default_string)
{
	std::string result = ugly_tchar_to_string(getCmdOptionImpl(begin, end, option));

	if ( result.empty() ) return default_string;
	return result;
}

std::string to_text(const bool& b)
{
	if ( b ) return "true";
	return "false";
}

void run_queries(ritch::TchQueryingML & querying_ml, ritch::TchQuerying & querying_heuristic, std::string input_file, std::string results_file) {
	int32_t number, start, destination, start_time, shift_time;
	std::ifstream file;
	file.open(input_file);
	/* input file is a normal txt file, with on each row a case in the following format:
	 <case_number (int)> <start location (int)> <destination location (int)> <starttime (int)> <endtime (int)>
	*/
	std::ofstream results;
	results.open(results_file);
	if (file.is_open()) {
		while (file.good()) {
			file >> number;
			file >> start;
			file >> destination;
			file >> start_time;
			file >> shift_time;
			const auto t_startq = ritch::util::time_stamp();
			/* run a multilabel search */
			int32_t arrival_time = querying_ml.one_to_one(
				ritch::TchSearchGraph::NodeId(start),
				ritch::TchSearchGraph::NodeId(destination),
				start_time, shift_time, 
				true); /* true = short (4.5h driving - 45 min break - 4.5h driving). false = long (9h driving - 11h break - 9h driving) */
			/* or: run single-label search (latest/optimal break scheduling)
			int32_t arrival_time = querying_heuristic.one_to_one(
				ritch::TchSearchGraph::NodeId(start),
				ritch::TchSearchGraph::NodeId(destination),
				start_time, shift_time
				true); /* true = short (4.5h driving - 45 min break - 4.5h driving). false = long (9h driving - 11h break - 9h driving) 
			*/
			const auto t_endq = ritch::util::time_stamp();
			const double duration = ritch::util::get_duration_in_seconds(t_startq, t_endq);
			results << number << " " << arrival_time << " " << duration << "\n";
		}
		file.close();
		results.close();
	}
}

int main(int argc, char** argv)
{
	const int MAXIMUM_THREADS = -1;
	const int32_t PERIOD = 7*24*60*60;

	const bool wait_at_the_end = cmdOptionExists(argv, argv + argc, "-wait");

	if ( cmdOptionExists(argv, argv + argc, "-help") )
	{
		std::cerr
			<< "\n"
			<< "Show help message:\n"
			<< " -help\n"
			<< "\n"
			<< "Flags:\n"
			<< " -wait\n"
			<< "\n"
			<< "Parameters:\n"
			<< " -m <input graph directory>\n"
			<< " -d <input RITCH_DEMANDS file>\n"
			<< " -s <input RITCH_SHORTCUTS file>\n"
			<< " -b <input default blocks path file name> (default: no blocks)\n";

		if ( wait_at_the_end )
			std::getchar();

		return EXIT_SUCCESS;
	}

	const std::string graph_directory_name = getCmdOption(argv, argv + argc, "-m");
	if ( graph_directory_name.empty() )
	{
		std::cerr << "expected -m <input graph directory>" << std::endl;
		std::cerr << "To list the available options use option -help" << std::endl;

		if ( wait_at_the_end )
			std::getchar();

		return EXIT_FAILURE;
	}

	const std::string ritch_demands_input_file_name = getCmdOption(argv, argv + argc, "-d");
	const std::string ritch_shortcuts_input_file_name = getCmdOption(argv, argv + argc, "-s");
	const std::string default_blocks_input_path_file_name = getCmdOption(argv, argv + argc, "-b");

	RITCH_STATUS("Compiled at " << __TIME__ << " on " << __DATE__ << "\n");

	//removed const from demands
	ritch::TchSearchGraph graph(
		graph_directory_name,
		default_blocks_input_path_file_name,
		graph_directory_name + ritch::util::PATH_SEPARATOR() + ritch_shortcuts_input_file_name,
		PERIOD);
	
	ritch::TchQueryingML queryingml(graph); /* querying object for multilabel algorithm */
	ritch::TchQuerying querying(graph); /* querying object for singlelabel algorithm */
	{
		RITCH_STATUS("Performing queries ...\n");

		size_t max_n_edges = 0;
		size_t total_n_edges = 0;

		const auto t_begin = ritch::util::time_stamp();

		run_queries(queryingml, querying, "test-instances.txt", "querying.txt");

		const auto t_end = ritch::util::time_stamp();
		const double total_time_sec = ritch::util::get_duration_in_seconds(t_begin, t_end);

		RITCH_STATUS("Finished.\n");
		RITCH_STATUS("Total time = " << total_time_sec << " msec.\n");
	}
	if ( wait_at_the_end )
		std::getchar();

	return EXIT_SUCCESS;
}
