#ifndef RITCH_DATASTR_INTERVAL_SET_H_
#define RITCH_DATASTR_INTERVAL_SET_H_

#include <iterator>
#include <limits>
#include <memory>

#include "boost/icl/interval_set.hpp"

namespace ritch
{

class IntervalSet
{
public:
	typedef int32_t Domain;

	class Interval
	{
	private:
		Domain _lower;
		Domain _upper;

	public:
		Interval(const Interval&) = default;
		Interval(Interval&&) = default;

		Interval& operator= (const Interval&) = default;
		Interval& operator= (Interval&&) = default;

		Interval()
		: _lower(std::numeric_limits<Domain>::max()),
		  _upper(std::numeric_limits<Domain>::max())
		{}

		Interval(Domain lower, Domain upper)
		: _lower(lower), _upper(upper)
		{}

		const Domain& get_lower() const { return _lower; }
		const Domain& get_upper() const { return _upper; }

		void set_lower(const Domain& lower) { _lower = lower; }
		void set_upper(const Domain& upper) { _upper = upper; }
	};

	class InputIterator : public std::iterator<std::input_iterator_tag, Interval>
	{
		friend class IntervalSet;

	private:

		class TempInterval
		{
			friend class InputIterator;

		private:
			Interval _interval;
			TempInterval(Domain lower, Domain upper) : _interval(lower, upper) {}

		public:
			const Interval* operator-> () const { return &_interval; }
		};

		typedef boost::icl::interval_set<Domain>::iterator IteratorImpl;
		IteratorImpl _it_impl;

		InputIterator(IteratorImpl it_impl) : _it_impl(it_impl) {}

	public:
		InputIterator() : _it_impl() {}

		InputIterator(const InputIterator&) = default;
		InputIterator(InputIterator&&) = default;

		InputIterator& operator= (const InputIterator&) = default;
		InputIterator& operator= (InputIterator&&) = default;

		InputIterator& operator++ () { ++_it_impl; return *this; }
		InputIterator operator++ (int) { InputIterator tmp(*this); operator++(); return tmp; }

		bool operator== (const InputIterator& other) const { return _it_impl == other._it_impl; }
		bool operator!= (const InputIterator& other) const { return !(*this == other); }

		Interval operator*() const { return Interval(_it_impl->lower(), _it_impl->upper()); }
		TempInterval operator-> () const { return TempInterval(_it_impl->lower(), _it_impl->upper()); }
	};

private:
	boost::icl::interval_set<Domain> _interval_set;

	IntervalSet(const boost::icl::interval_set<Domain>& interval_set)
		: _interval_set(interval_set)
	{}

public:
	IntervalSet() : _interval_set() {}

	IntervalSet(const IntervalSet&) = default;
	IntervalSet(IntervalSet&&) = default;

	IntervalSet& operator= (const IntervalSet&) = default;
	IntervalSet& operator= (IntervalSet&&) = default;

	IntervalSet& add(const Interval& interval)
	{
		_interval_set.insert(
			boost::icl::interval<int32_t>::right_open(interval.get_lower(), interval.get_upper()));

		return *this;
	}

	friend IntervalSet intersect(const IntervalSet& lhs, const IntervalSet& rhs)
	{
		return lhs._interval_set & rhs._interval_set;
	}

	friend IntervalSet unite(const IntervalSet& lhs, const IntervalSet& rhs)
	{
		return lhs._interval_set + rhs._interval_set;
	}

	friend IntervalSet subtract(const IntervalSet& lhs, const IntervalSet& rhs)
	{
		return lhs._interval_set - rhs._interval_set;
	}

	friend IntervalSet intersect(IntervalSet&& lhs, const IntervalSet& rhs)
	{
		lhs._interval_set &= rhs._interval_set;
		return std::move(lhs._interval_set);
	}

	friend IntervalSet unite(IntervalSet&& lhs, const IntervalSet& rhs)
	{
		lhs._interval_set += rhs._interval_set;
		return std::move(lhs._interval_set);
	}

	friend IntervalSet subtract(IntervalSet&& lhs, const IntervalSet& rhs)
	{
		lhs._interval_set -= rhs._interval_set;
		return std::move(lhs._interval_set);
	}

	bool empty() const { return _interval_set.empty(); }
	size_t size() const { return boost::icl::interval_count(_interval_set); }

	InputIterator begin() const { return _interval_set.begin(); }
	InputIterator end() const { return _interval_set.end(); }
};

}

#endif
