#ifndef RITCH_DATASTR_TIME_DIST_H_
#define RITCH_DATASTR_TIME_DIST_H_

#include <algorithm>
#include <iostream>
#include <cstdint>
#include <cstring>
#include <limits>
#include <sstream>
#include <tuple>
#include <vector>

#include "ritch/util/misc.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

	template <int Period = 7 * 60 * 60 * 24>
	class TimeDist_tmpl
	{
	public:

		static constexpr int32_t PERIOD() { return Period; }
		static TimeDist_tmpl INFTY() { return TimeDist_tmpl(); }

		class Segment
		{
		private:
			uint32_t _inverse_slope : 1;
			int32_t _right_x : 31;
			int32_t _right_y;
			int32_t _block_start; //start time of first block on the edge by which this segment is represented
			int32_t _block_end; //end time of first block
			int32_t _default_time;

		public:

			Segment()
				: _inverse_slope(0), _right_x(0), _right_y(0), _block_start(0), _block_end(0), _default_time(0)
			{}

			Segment(int32_t slope, int32_t right_x, int32_t right_y, int32_t block_start, int32_t block_end, int32_t default_time)
				: _inverse_slope(uint32_t(-slope)), _right_x(right_x), _right_y(right_y), _block_start(block_start), _block_end(block_end), _default_time(default_time)
			{
				RITCH_ASSERT(slope == 0 || slope == -1);
				RITCH_ASSERT(_inverse_slope == 0 || _inverse_slope == 1);
				RITCH_ASSERT(RITCH_EQUIV(slope == 0, _inverse_slope == 0));
				RITCH_ASSERT(RITCH_EQUIV(slope == -1, _inverse_slope == 1));
				if (_right_x <= 0) {
					std::cout << _inverse_slope << " " << _right_x << " " << _right_y << " " << _block_start << " " << _block_end << " " << _default_time << std::endl;
				}
				if (_block_start < 0 ) {
					_block_start = _block_start + PERIOD();
				}
				if (_block_end < 0) {
					_block_end = _block_end + PERIOD();
				}
				if (_block_start > PERIOD()) {
					_block_start = _block_start - PERIOD();
				}
				if (_block_end > PERIOD()) {
					_block_end = _block_end - PERIOD();
				}
				RITCH_ASSERT(_block_start >= 0);
				RITCH_ASSERT(_block_end >= 0);
				RITCH_ASSERT(_right_x > 0);
				RITCH_ASSERT(_right_y >= 0);
				RITCH_ASSERT(_right_y >= _default_time);
				RITCH_ASSERT(_default_time >= 0);
				if (block_start == block_end) {
					_block_start = 0;
					_block_end = 0;
				}
			}

			Segment(int32_t slope, int32_t right_x, int32_t right_y, int32_t block_start, int32_t block_end, int32_t default_time, bool checks)
				: _inverse_slope(uint32_t(-slope)), _right_x(right_x), _right_y(right_y), _block_start(block_start), _block_end(block_end), _default_time(default_time)
			{
				RITCH_ASSERT(slope == 0 || slope == -1);
				RITCH_ASSERT(_inverse_slope == 0 || _inverse_slope == 1);
				RITCH_ASSERT(RITCH_EQUIV(slope == 0, _inverse_slope == 0));
				RITCH_ASSERT(RITCH_EQUIV(slope == -1, _inverse_slope == 1));
				if (_right_x <= 0) {
					std::cout << _inverse_slope << " " << _right_x << " " << _right_y << " " << _block_start << " " << _block_end << " " << _default_time << std::endl;
				}
				if (_block_start < 0) {
					_block_start = _block_start + PERIOD();
				}
				if (_block_end < 0) {
					_block_end = _block_end + PERIOD();
				}
				if (_block_start > PERIOD()) {
					_block_start = _block_start - PERIOD();
				}
				if (_block_end > PERIOD()) {
					_block_end = _block_end - PERIOD();
				}
				RITCH_ASSERT(_block_start >= 0);
				RITCH_ASSERT(_block_end >= 0);
				if (_right_x <= 0) {
					std::cout << _inverse_slope << " " << _right_x << " " << _right_y << " " << _block_start << " " << _block_end << " " << _default_time << std::endl;
				}
				RITCH_ASSERT(_right_x > 0);
				RITCH_ASSERT(_right_y >= 0);
				RITCH_ASSERT(_right_y >= _default_time);
				RITCH_ASSERT(_default_time >= 0);
				if (block_start == block_end) {
					_block_start = 0;
					_block_end = 0;
				}
			}

			Segment(const Segment&) = default;
			Segment(Segment&&) = default;

			Segment& operator= (const Segment&) = default;
			Segment& operator= (Segment&&) = default;

			int32_t get_slope() const { return -int32_t(_inverse_slope); }
			int32_t get_right_x() const { return int32_t(_right_x); }
			int32_t get_right_y() const { return int32_t(_right_y); }
			int32_t get_block_start() const { return int32_t(_block_start); }
			int32_t get_block_end() const { return int32_t(_block_end); }
			int32_t get_default_time() const { return int32_t(_default_time); }


			void set_slope(int32_t slope)
			{
				RITCH_ASSERT(slope == -1 || slope == 0);
				RITCH_ASSERT(uint32_t(-slope) == 1 || uint32_t(-slope) == 0);
				_inverse_slope = uint32_t(-slope);
			}

			void set_right_x(int32_t x)
			{
				RITCH_ASSERT(x > 0 && uint32_t(x) <= uint32_t(0x7fffffff));
				_right_x = uint32_t(x);
			}

			void set_right_y(int32_t y)
			{
				RITCH_ASSERT(y > 0 && uint32_t(y) <= uint32_t(0x7fffffff));
				_right_y = uint32_t(y);
			}
			void set_block_start(int32_t bs)
			{
				_block_start = bs;
			}
			void set_block_end(int32_t be)
			{
				_block_end = be;
			}
			void set_default_time(int32_t dt) {
				_default_time = dt;
			}


			friend bool operator== (const Segment& lhs, const Segment& rhs)
			{
				return
					lhs._inverse_slope == rhs._inverse_slope &&
					lhs._right_x == rhs._right_x &&
					lhs._right_y == rhs._right_y;
			}

			friend bool operator!= (const Segment& lhs, const Segment& rhs) { return !(lhs == rhs); }

			friend std::ostream& operator<<(std::ostream& os, const Segment& segment)
			{
				os << "(" << segment.get_slope() << ", " << segment.get_right_x() << ", " << segment.get_right_y() << ")";
				return os;
			}
		};

	private:

		struct SegmentDescr
		{
			int32_t x;
			int32_t y;
			int32_t m;
			int32_t x_left;
			int32_t y_left;
			int32_t block_start;
			int32_t block_end;
			int32_t default_time;
		};

		struct Point { int32_t x; int32_t y; };

		static std::tuple<Point, int32_t, int32_t, int32_t> single_inner_intersection_point(const SegmentDescr& a, const SegmentDescr& b)
		{
			int32_t blockstart = a.block_start;
			int32_t blockend = a.block_end;
			int32_t defaulttime = a.default_time;
			if (a.block_start == 0 && a.block_end == 0 || b.block_start > 0 && b.block_end > 0 && b.block_start - a.y <= a.block_start) {
				blockstart = b.block_start;
				blockend = b.block_end;
				defaulttime = b.default_time;
			}
			else {
				blockstart = a.block_start;
				blockend = a.block_end;
				defaulttime = a.default_time;
			}
			Point max = { std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max() };
			if (a.m == b.m)
				return std::make_tuple(max, blockstart, blockend, defaulttime);

			if (a.y < b.y)
			{
				if (a.m == 0)
					return std::make_tuple(max, blockstart, blockend, defaulttime);

				RITCH_ASSERT(b.m == 0);
				if (a.y_left <= b.y)
					return std::make_tuple(max, blockstart, blockend, defaulttime);

				int32_t x = a.x + (a.y - b.y);
				Point ax = { x, a.y + a.x - x };
				if (b.x_left < x && x < b.x)
					return std::make_tuple(ax, blockstart, blockend, defaulttime);
			}
			else if (b.y < a.y)
			{
				if (b.m == 0)
					return std::make_tuple(max, blockstart, blockend, defaulttime);

				RITCH_ASSERT(a.m == 0);
				if (b.y_left <= a.y)
					return std::make_tuple(max, blockstart, blockend, defaulttime);

				int32_t x = b.x + (b.y - a.y);
				Point bx = { x, b.y + b.x - x };
				if (a.x_left < x && x < a.x)
					return std::make_tuple(bx, blockstart, blockend, defaulttime);
			}

			return std::make_tuple(max, blockstart, blockend, defaulttime);
		}

		static bool is_valid_point(const Point& p)
		{
			return
				p.x != std::numeric_limits<int32_t>::max() &&
				p.y != std::numeric_limits<int32_t>::max();
		}

		SegmentDescr get_time_segment_descr_periodic(const size_t& i, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block) const
		{
			const size_t n = get_n_time_segments();
			const size_t i_mod_right = i % n;
			const size_t i_mod_left = i_mod_right > 0 ? i_mod_right - 1 : n - 1;
			const int32_t offset_right = int32_t(i / n) * PERIOD();
			const int32_t offset_left = i_mod_right > 0 ? offset_right : offset_right - PERIOD();

			const Segment s(get_time_segment(i_mod_right, block));
			const int32_t x_right = s.get_right_x() + offset_right;
			const int32_t y_right = s.get_right_y();

			const int32_t slope = s.get_slope();

			const Segment s_left(get_time_segment(i_mod_left, block));
			const int32_t x_left = s_left.get_right_x() + offset_left;
			int32_t block_diff = x_right - x_left;
			if (block_diff < 0 ) {
				block_diff = (x_right + PERIOD()) - x_left;
			}
			const int32_t y_left = slope == 0 ? y_right : y_right + (block_diff);

			int32_t block_start = s.get_block_start();
			int32_t block_end = s.get_block_end();
			int32_t default_time = s.get_default_time();
			RITCH_ASSERT(default_time <= y_left);
			RITCH_ASSERT(default_time <= y_right);

			return SegmentDescr{ x_right, y_right, slope, x_left, y_left, block_start, block_end, default_time };
		}

		SegmentDescr get_dist_segment_descr_periodic(const size_t& i) const
		{
			const size_t n = get_n_dist_segments();
			const size_t i_mod_right = i % n;
			const size_t i_mod_left = i_mod_right > 0 ? i_mod_right - 1 : n - 1;
			const int32_t offset_right = int32_t(i / n) * PERIOD();
			const int32_t offset_left = i_mod_right > 0 ? offset_right : offset_right - PERIOD();

			const Segment s(get_dist_segment(i_mod_right));
			const int32_t x_right = s.get_right_x() + offset_right;
			const int32_t y_right = s.get_right_y();

			const int32_t slope = s.get_slope();

			const Segment s_left(get_dist_segment(i_mod_left));
			const int32_t x_left = s_left.get_right_x() + offset_left;

			return SegmentDescr{ x_right, y_right, 0, x_left, y_right };
		}

		static inline void append_time_segment(std::vector<Segment>& segments, const Segment& s)
		{
			//RITCH_ASSERT(s.get_right_x() > 0);
			RITCH_ASSERT(s.get_right_x() <= PERIOD());
			if (!segments.empty())
			{
				if (s.get_right_x() <= segments.back().get_right_x()) {
					return;
				}
				RITCH_ASSERT( s.get_right_x() > segments.back().get_right_x() );
				RITCH_ASSERT(s.get_right_y() + (s.get_right_x() - segments.back().get_right_x()) >= segments.back().get_right_y());

				if (s.get_slope() == -1 && segments.back().get_slope() == -1)
				{
					if (segments.back().get_right_y() == s.get_right_y() + (s.get_right_x() - segments.back().get_right_x()))
					{
						segments.back() = s;
						return;
					}
				}
				else if (s.get_slope() == 0 && segments.back().get_slope() == 0)
				{
					if (segments.back().get_right_y() == s.get_right_y())
					{
						segments.back() = s;
						return;
					}
				}
			}

			segments.push_back(s);
		}

		static inline void append_dist_segment(std::vector<Segment>& segments, const Segment& s)
		{
			RITCH_ASSERT(s.get_slope() == 0);
			RITCH_ASSERT(s.get_right_x() > 0);
			RITCH_ASSERT(s.get_right_x() <= PERIOD());

			if (!segments.empty())
			{
				RITCH_ASSERT(s.get_right_x() > segments.back().get_right_x());

				if (segments.back().get_right_y() == s.get_right_y())
				{
					segments.back() = s;
					return;
				}
			}

			segments.push_back(s);
		}

		static void remove_unnecessary_segement_from_end(std::vector<Segment>& segments)
		{
			if (segments.size() < 2)
				return;

			if (segments.front().get_slope() == 0 && segments.back().get_slope() == 0)
			{
				if (segments.front().get_right_y() == segments.back().get_right_y())
					segments.pop_back();
			}
			else if (segments.front().get_slope() == -1 && segments.back().get_slope() == -1)
			{
				const Segment& s_back = segments.back();
				const Segment& s_front = segments.front();

				if (s_back.get_right_y() == s_front.get_right_y() + s_front.get_right_x() + (PERIOD() - s_back.get_right_x()))
					segments.pop_back();
			}
		}

		static void append_middle_node(std::vector<std::pair<int32_t, uint32_t>>& middle_nodes, const int32_t x, const uint32_t middle_node)
		{
			RITCH_ASSERT(x > 0);
			RITCH_ASSERT(x <= PERIOD());

			if (!middle_nodes.empty())
			{
				RITCH_ASSERT(x > middle_nodes.back().first || (x == middle_nodes.back().first && middle_nodes.back().second == middle_node));

				if (middle_nodes.back().second == middle_node)
				{
					middle_nodes.back().first = x;
					return;
				}
			}

			middle_nodes.push_back(std::make_pair(x, middle_node));
		}

		static void remove_unnecessary_middle_node_from_end(std::vector<std::pair<int32_t, uint32_t>>& middle_nodes)
		{
			if (middle_nodes.size() < 2)
				return;

			if (middle_nodes.front().second == middle_nodes.back().second)
				middle_nodes.pop_back();
		}

		static const std::pair<int32_t, uint32_t> get_middle_node_periodic(const std::vector<std::pair<int32_t, uint32_t>>& middle_nodes, const size_t& i)
		{
			const size_t n = middle_nodes.size();
			const size_t i_mod_right = i % n;
			const int32_t offset = int32_t(i / n) * PERIOD();
			int32_t x = middle_nodes[i_mod_right].first + offset;
			uint32_t middle_node = middle_nodes[i_mod_right].second;

			return std::make_pair(x, middle_node);
		}

		int32_t left_y_time_at_zero(const std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>> & block) const
		{
			const Segment& segment = get_time_segment(0, block);

			if (segment.get_slope() == 0)
				return segment.get_right_y();

			return segment.get_right_y() + segment.get_right_x();
		}

		static std::vector<Segment> link_time(const int32_t& c, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block)
		{
			const size_t n = rhs.get_n_time_segments();

			std::vector<Segment> result_time_segments;
			result_time_segments.reserve(n);

			for (size_t i = 0; i != n; ++i)
			{
				const Segment& s = rhs.get_time_segment(i, block);
				result_time_segments.push_back(Segment(s.get_slope(), s.get_right_x(), s.get_right_y() + c, s.get_block_start(), s.get_block_end(), s.get_default_time() + c, true)); //TODO!!
			}

			return result_time_segments;
		}

		static std::vector<Segment> link_time(const TimeDist_tmpl& lhs, const int32_t& c, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block)
		{
			const size_t n = lhs.get_n_time_segments();
			const int32_t c_mod = util::modulo(c, PERIOD());

			std::vector<Segment> result_time_segments;
			result_time_segments.reserve(n);

			size_t k = 0;
			//Changed < c_mod to <= c_mod
			while (k < n && lhs.get_time_segment(k, block).get_right_x() < c_mod)

				++k;

			for (size_t i = k; i != n; ++i)
			{
				const Segment& s = lhs.get_time_segment(i, block);

				if (s.get_block_start() == s.get_block_end()) {
					if (s.get_default_time() == 0) {
						if (s.get_right_x() - c_mod <= 0) {
							append_time_segment(result_time_segments,
								Segment(s.get_slope(), s.get_right_x() - c_mod + PERIOD(), s.get_right_y() + c, 0, 0, s.get_right_y() + c, true));
						}else{
							append_time_segment(result_time_segments, 
								Segment(s.get_slope(), s.get_right_x() - c_mod, s.get_right_y() + c, 0, 0, s.get_right_y() + c, true));
						}
					}
					else {
						if (s.get_right_x() - c_mod <= 0) {
							append_time_segment(result_time_segments,
								Segment(s.get_slope(), s.get_right_x() - c_mod + PERIOD(), s.get_right_y() + c, 0, 0, s.get_default_time() + c, true));
						}
						else {
							append_time_segment(result_time_segments,
								Segment(s.get_slope(), s.get_right_x() - c_mod, s.get_right_y() + c, 0, 0, s.get_default_time() + c, true));
						}
					}
				}
				else {
					if(s.get_right_x() - c_mod <= 0) {
						append_time_segment(result_time_segments,Segment(s.get_slope(), s.get_right_x() - c_mod + PERIOD(), s.get_right_y() + c, s.get_block_start() - c_mod, s.get_block_end() - c_mod, s.get_default_time() + c, true));
					}
					else {
						append_time_segment(result_time_segments,
							Segment(s.get_slope(), s.get_right_x() - c_mod, s.get_right_y() + c, s.get_block_start() - c_mod, s.get_block_end() - c_mod, s.get_default_time() + c, true));
					}
				}
			}

			for (size_t i = 0; i != k; ++i)
			{
				const Segment& s = lhs.get_time_segment(i, block);
				if (s.get_right_x() + (PERIOD() - c_mod) > s.get_block_start() + (PERIOD() - c_mod)) {
					//result_time_segments.push_back(
						//Segment(s.get_slope(), s.get_right_x() + (PERIOD() - c_mod), s.get_right_y() + c));
						//Segment(s.get_slope(), s.get_right_x() + (PERIOD() - c_mod), s.get_right_y() + c, 0, 0, s.get_default_time() + c, true));
					append_time_segment(result_time_segments, Segment(s.get_slope(), s.get_right_x() + (PERIOD() - c_mod), s.get_right_y() + c, 0, 0, s.get_default_time() + c, true));
				}
				else {
					append_time_segment(result_time_segments, Segment(s.get_slope(), s.get_right_x() + (PERIOD() - c_mod), s.get_right_y() + c, s.get_block_start() + (PERIOD() - c_mod), s.get_block_end() + (PERIOD() - c_mod), s.get_default_time() + c, true));

					//result_time_segments.push_back(
					//	Segment(s.get_slope(), s.get_right_x() + (PERIOD() - c_mod), s.get_right_y() + c, s.get_block_start() + (PERIOD() - c_mod), s.get_block_end() + (PERIOD() - c_mod), s.get_default_time() + c, true));
				}
			}

			return result_time_segments;
		}

		static std::vector<Segment> link_time(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR)
		{
			std::vector<Segment> result_time_segments;
			if (lhs.has_const_time(blockL) && rhs.has_const_time(blockR))
			{
				result_time_segments.emplace_back(0, PERIOD(), lhs.get_const_time(blockL) + rhs.get_const_time(blockR), 0, 0, lhs.get_const_time(blockL) + rhs.get_const_time(blockR));
				return result_time_segments;
			}

			if (lhs.has_const_time(blockL))
			{
				result_time_segments = link_time(lhs.get_const_time(blockL), rhs, blockR);
				return result_time_segments;
			}

			if (rhs.has_const_time(blockR))
			{
				result_time_segments = link_time(lhs, rhs.get_const_time(blockR), blockL);
				return result_time_segments;
			}

			result_time_segments.reserve(lhs.get_n_time_segments() + rhs.get_n_time_segments());

			size_t i = 0;
			size_t j = 0;

			auto g = [&lhs, &blockL](const size_t& i) { return lhs.get_time_segment_descr_periodic(i, blockL); };
			auto f = [&rhs, &blockR](const size_t& i) { return rhs.get_time_segment_descr_periodic(i, blockR); };
			
			while (g(i).x < rhs.left_y_time_at_zero(blockR)) ++i;

			//int32_t prevFY = f(j).y_left;
			while (true)
			{
				while (g(i).x < f(j).x_left + f(j).y_left) ++i;

				int32_t x;
				int32_t y;
				int32_t m;
				int32_t blockstart;
				int32_t blockend;
				int32_t defaulttime;

				//while (g(i).x < f(j).x_left + prevFY) ++i;

				blockstart = 0;
				blockend = 0;
				if (f(j).block_start == 0 && f(j).block_end == 0 || g(i).block_start > 0 && g(i).block_end > 0  && g(i).block_start - f(j).y <= f(j).block_start) {
					blockstart = g(i).block_start;
					blockend = g(i).block_end;
				}
				else {
					blockstart = f(j).block_start;
					blockend = f(j).block_end;
				}
				defaulttime = g(i).default_time + f(j).default_time;
				if (g(i).default_time == 0) {
					defaulttime = defaulttime  + lhs.get_min_time();
				}
				if (f(j).default_time == 0) {
					defaulttime = defaulttime + rhs.get_min_time();
				}
				//prevFY = f(j).y;

				//start time of segment lhs is exactly at rhs+tt
				if (g(i).x == f(j).x + f(j).y)
				{
					// new x = rhs.x, new y = tt rhs + lhs, new m = als f=0, g(i).m, anders -1
					x = f(j).x;
					y = f(j).y + g(i).y;
					m = (f(j).m == 0) ? g(i).m : -1;

					++j;
					if (f(j).m == 0) ++i;
				}
				else if (g(i).x < f(j).x + f(j).y)
				{
					RITCH_ASSERT(f(j).m == 0);

					if (g(i).x == f(j).x_left + f(j).y_left)
					{
						++i;
						continue;
					}
					x = g(i).x - f(j).y_left;
					y = g(i).y + f(j).y_left;
					m = g(i).m;

					++i;
				}
				else
				{
					RITCH_ASSERT(g(i).x > f(j).x + f(j).y);

					x = f(j).x;
					y = (g(i).m == 0) ? (g(i).y + f(j).y) : (g(i).y + g(i).x - f(j).x);
					m = (f(j).m == 0) ? g(i).m : -1;
					
					++j;
				}
				if (x > PERIOD()) break;
				if (x == 0) continue;

				if (blockstart == 0 && blockend == 0 && result_time_segments.size() > 0 && y <= result_time_segments.back().get_block_start() ) {
					blockstart = result_time_segments.back().get_block_start();
					blockend = result_time_segments.back().get_block_end();
				}

				RITCH_ASSERT(defaulttime <= y);

				append_time_segment(result_time_segments, Segment(m, x, y, blockstart, blockend, defaulttime, true));

			}

			return result_time_segments;
		}

		static std::vector<Segment> link_dist(const int32_t& c, const TimeDist_tmpl& rhs)
		{
			const size_t n = rhs.get_n_dist_segments();

			std::vector<Segment> result_dist_segments;
			result_dist_segments.reserve(n);

			for (size_t i = 0; i != n; ++i)
			{
				const Segment& s = rhs.get_dist_segment(i);
				result_dist_segments.push_back(Segment(0, s.get_right_x(), s.get_right_y() + c, 0, 0, 0));
			}

			return result_dist_segments;
		}

		static std::vector<Segment> link_dist(const TimeDist_tmpl& lhs, const int32_t& const_rhs_time, const int32_t& const_rhs_dist)
		{
			const size_t n = lhs.get_n_dist_segments();
			const int32_t t_mod = util::modulo(const_rhs_time, PERIOD());
			const int32_t& c = const_rhs_dist;

			std::vector<Segment> result_dist_segments;
			result_dist_segments.reserve(n);

			size_t k = 0;
			while (k < n && lhs.get_dist_segment(k).get_right_x() < t_mod)
				++k;

			for (size_t i = k; i != n; ++i)
			{
				const Segment& s = lhs.get_dist_segment(i);

				result_dist_segments.push_back(
					Segment(s.get_slope(), s.get_right_x() - t_mod, s.get_right_y() + c, 0, 0, 0));
			}

			for (size_t i = 0; i != k; ++i)
			{
				const Segment& s = lhs.get_dist_segment(i);

				result_dist_segments.push_back(
					Segment(s.get_slope(), s.get_right_x() + (PERIOD() - t_mod), s.get_right_y() + c, 0, 0, 0));
			}

			return result_dist_segments;
		}

		static std::vector<Segment> link_dist(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR)
		{
			std::vector<Segment> result_dist_segments;

			if (!lhs.has_dist() || !rhs.has_dist())
				return result_dist_segments;

			if (lhs.has_const_dist() && rhs.has_const_dist())
			{
				result_dist_segments.assign(1, Segment(0, PERIOD(), lhs.get_const_dist() + rhs.get_const_dist(), 0, 0, 0));
				return result_dist_segments;
			}

			if (lhs.has_const_dist())
			{
				result_dist_segments = link_dist(lhs.get_const_dist(), rhs);
				return result_dist_segments;
			}

			if (rhs.has_const_time(blockR) && rhs.has_const_dist())
			{
				result_dist_segments = link_dist(lhs, rhs.get_const_time(blockR), rhs.get_const_dist());
				return result_dist_segments;
			}

			result_dist_segments.reserve(lhs.get_n_dist_segments() + rhs.get_n_dist_segments());

			auto f = [&rhs, &blockR](const size_t& i) { return rhs.get_time_segment_descr_periodic(i, blockR); };
			auto c = [&rhs](const size_t& i) { return rhs.get_dist_segment_descr_periodic(i); };
			auto d = [&lhs](const size_t& i) { return lhs.get_dist_segment_descr_periodic(i); };

			size_t i = 0;
			size_t j = 0;
			size_t k_right = 0;
			size_t k_left = 0;

			while (d(i).x < rhs.left_y_time_at_zero(blockR)) ++i;

			while (true)
			{
				while (f(k_left).x + f(k_left).y <= d(i).x_left)
					++k_left;

				k_right = std::max(k_right, k_left);
				while (f(k_right).x + f(k_right).y < d(i).x)
					++k_right;

				if (f(k_left).x + f(k_left).y > d(i).x && f(k_left).x_left + f(k_left).y_left >= d(i).x)
				{
					++i;
					continue;
				}

				
				int32_t x_left;
				if (f(k_left).x + f(k_left).y >= d(i).x_left)
					x_left = f(k_left).x_left;
				else
					x_left = d(i).x_left - f(k_left).y;
				
				int32_t x_right;
				if (f(k_right).x + f(k_right).y == d(i).x)
					x_right = f(k_right).x;
				else if (f(k_right).x_left + f(k_right).y_left >= d(i).x) {
					x_right = f(k_right -1).x;
				}
				else
					x_right = d(i).x - f(k_right).y;
			
				while (c(j).x <= x_left) ++j;

				int32_t x;
				int32_t distance = d(i).y + c(j).y;

				//c(j).x == d(i).x*f(x)
				if (c(j).x == x_right)
				{
					x = c(j).x;
					++j; 
					++i;
				}
				else if (c(j).x < x_right)
				{
					x = c(j).x;
					++j;
				}
				else
				{
					x = x_right;
					++i;
				}

				if (x > PERIOD()) break;
				if (x == 0) continue;
				append_dist_segment(result_dist_segments, Segment(0, std::min(x, PERIOD()), distance, 0, 0,0));
			}

			return result_dist_segments;
		}

		static std::vector<Segment> merge_time(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR)
		{
			std::vector<Segment> result_time_segments;

			if (lhs.has_const_time(blockL) && rhs.has_const_time(blockR))
			{
				result_time_segments.assign(1, Segment(0, PERIOD(), std::min(lhs.get_const_time(blockL), rhs.get_const_time(blockR)), 0, 0, std::min(lhs.get_const_time(blockL), rhs.get_const_time(blockR))));
				return result_time_segments;
			}

			const size_t n = lhs.get_n_time_segments();
			const size_t m = rhs.get_n_time_segments();

			result_time_segments.reserve(1 + 2 * (n + m));

			size_t i = 0;
			size_t j = 0;

			auto g = [&lhs, &blockL](const size_t& i) { return lhs.get_time_segment_descr_periodic(i, blockL); };
			auto f = [&rhs, &blockR](const size_t& i) { return rhs.get_time_segment_descr_periodic(i, blockR); };

			while (i < n || j < m)
			{
				const std::tuple<Point, int32_t, int32_t, int32_t> p = single_inner_intersection_point(g(i), f(j));
				if (is_valid_point(std::get<0>(p)) && std::get<0>(p).x > 0) {
					append_time_segment(result_time_segments, Segment(0, std::get<0>(p).x, std::get<0>(p).y, std::get<1>(p), std::get<2>(p), std::min(std::get<0>(p).y, std::get<3>(p))));
				}
				int32_t gdef = g(i).default_time;
				int32_t fdef = f(j).default_time;
				if (g(i).default_time == 0) {
					gdef = lhs.get_min_time();
				}
				if (f(j).default_time == 0) {
					fdef = rhs.get_min_time();
				}

				if (g(i).x == f(j).x)
				{
					if (g(i).y == f(j).y)
					{
						if (g(i).m == 0 || f(j).m == 0) {
							//in favour of g, thus add g(i).blockstart and blockend
							append_time_segment(result_time_segments, Segment(0, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						}
						else {
							append_time_segment(result_time_segments, Segment(-1, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						}
					}
					else if (g(i).y < f(j).y)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
					}
					else
					{
						RITCH_ASSERT(g(i).y > f(j).y);
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
					}

					++i;
					++j;
				}
				else if (g(i).x < f(j).x)
				{
					const int32_t y_f = f(j).m == 0 ? f(j).y : f(j).y + f(j).x - g(i).x;

					if (g(i).m == f(j).m && g(i).y == y_f)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
					}
					else if (g(i).y < y_f || (g(i).y == y_f && g(i).m == 0))
					{
						RITCH_ASSERT(RITCH_IMPLIES(g(i).y == y_f, f(j).m != 0));
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
					}
					else
					{
						//constant section f(j), slope section g(i). y_f == g(i).y = f(j).y + f(j).x - g(i).x
						RITCH_ASSERT(y_f < g(i).y || (y_f == g(i).y && g(i).m != 0 && f(j).m == 0));
						append_time_segment(result_time_segments, Segment(f(j).m, g(i).x, y_f, f(j).block_start, f(j).block_end, fdef)); //changed defaulttime y_f to fdef
					}

					++i;
				}
				else
				{
					RITCH_ASSERT(f(j).x < g(i).x);
					const int32_t y_g = g(i).m == 0 ? g(i).y : g(i).y + g(i).x - f(j).x;

					if (f(j).m == g(i).m && f(j).y == y_g)
					{
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
					}
					else if (f(j).y < y_g || (f(j).y == y_g && f(j).m == 0))
					{
						RITCH_ASSERT(RITCH_IMPLIES(f(j).y == y_g, g(i).m != 0));
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
					}
					else
					{
						RITCH_ASSERT(y_g < f(j).y || (y_g == f(j).y && f(j).m != 0 && g(i).m == 0));
						append_time_segment(result_time_segments, Segment(g(i).m, f(j).x, y_g, g(i).block_start, g(i).block_end, gdef));//changed defaulttime y_g to gdef
					}

					++j;
				}
			}

			const std::tuple<Point, int32_t, int32_t, int32_t> tp = single_inner_intersection_point(g(n), f(m));
			const Point p = std::get<0>(tp);
			if (is_valid_point(p) && p.x > 0 && p.x <= PERIOD()) {
				append_time_segment(result_time_segments, Segment(0, p.x, p.y, std::get<1>(tp), std::get<2>(tp), std::min(p.y, std::get<3>(tp))));
				//std::cout << "12";
			}

			return result_time_segments;
		}

		static bool dbg_middle_node_is_ok
		(
			const TimeDist_tmpl& lhs,
			const TimeDist_tmpl& rhs,
			const std::vector<std::pair<int32_t, uint32_t>>& lhs_middle_nodes,
			const uint32_t single_rhs_middle_node,
			const std::vector<std::pair<int32_t, uint32_t>>& result_middle_nodes,
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL,
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR
		)
		{
			const size_t N_RANDOM_POINTS = 100;

			std::vector<int32_t> x_values;
			x_values.push_back(0);
			x_values.push_back(PERIOD());

			for (size_t i = 0; i != lhs.get_n_time_segments(); ++i)
				x_values.push_back(lhs.get_time_segment(i, blockL).get_right_x());

			for (size_t i = 0; i != rhs.get_n_time_segments(); ++i)
				x_values.push_back(rhs.get_time_segment(i, blockR).get_right_x());

			for (size_t i = 0; i != lhs_middle_nodes.size(); ++i)
				x_values.push_back(lhs_middle_nodes[i].first);

			for (size_t i = 0; i != N_RANDOM_POINTS; ++i)
				x_values.push_back(util::random_int32(0, PERIOD() - 1));

			const size_t n_x_values = x_values.size();
			for (size_t i = 0; i != n_x_values; ++i)
			{
				x_values.push_back(x_values[i] + PERIOD());
				x_values.push_back(x_values[i] - PERIOD());
			}

			auto get_middle_node =
				[](const int32_t x, const std::vector<std::pair<int32_t, uint32_t>>& middle_nodes)
			{
				const int32_t x_mod = util::modulo(x - 1, PERIOD()) + 1;

				for (const auto& descr : middle_nodes)
					if (descr.first >= x_mod)
						return descr.second;

				return middle_nodes.front().second;
			};

			for (const int32_t& x : x_values)
			{
				int32_t time_rhs = rhs.eval_time(x);
				int32_t time_lhs = lhs.eval_time(x);
				int32_t dist_rhs = rhs.eval_dist(x);
				int32_t dist_lhs = lhs.eval_dist(x);

				uint32_t middle_node;

				if (time_lhs < time_rhs)
					middle_node = get_middle_node(x, lhs_middle_nodes);
				else if (time_lhs > time_rhs)
					middle_node = single_rhs_middle_node;
				
				else if (dist_lhs < dist_rhs)
					middle_node = get_middle_node(x, lhs_middle_nodes);
				else
					middle_node = single_rhs_middle_node;
					
				if (middle_node != get_middle_node(x, result_middle_nodes))
				{
					std::cerr << "\n\nx = " << x << std::endl;
					std::cerr << "time_lhs = " << time_lhs << std::endl;
					std::cerr << "time_rhs = " << time_rhs << std::endl;
					std::cerr << "dist_lhs = " << dist_lhs << std::endl;
					std::cerr << "dist_rhs = " << dist_rhs << std::endl;
					std::cerr << "\nlhs = " << lhs << std::endl;
					std::cerr << "\nrhs = " << rhs << std::endl;
					std::cerr << "\nlhs_middle_nodes = " << util::to_text(lhs_middle_nodes) << std::endl;
					std::cerr << "\nsingle_rhs_middle_node = " << single_rhs_middle_node << std::endl;
					std::cerr << "\nresult = " << util::to_text(result_middle_nodes) << std::endl;

					return false;
				}
			}

			return true;
		}

		static bool dbg_link_is_ok(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const TimeDist_tmpl& result, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockNew)
		{
			const size_t N_RANDOM_POINTS = 100;

			std::vector<int32_t> x_values;
			x_values.push_back(0);
			x_values.push_back(PERIOD());

			for (size_t i = 0; i != lhs.get_n_time_segments(); ++i)
				x_values.push_back(lhs.get_time_segment(i, blockL).get_right_x());

			for (size_t i = 0; i != rhs.get_n_time_segments(); ++i)
				x_values.push_back(rhs.get_time_segment(i, blockR).get_right_x());

			for (size_t i = 0; i != result.get_n_time_segments(); ++i)
				x_values.push_back(result.get_time_segment(i, blockNew).get_right_x());

			for (size_t i = 0; i != lhs.get_n_dist_segments(); ++i)
				x_values.push_back(lhs.get_dist_segment(i).get_right_x());

			for (size_t i = 0; i != rhs.get_n_dist_segments(); ++i)
				x_values.push_back(rhs.get_dist_segment(i).get_right_x());

			for (size_t i = 0; i != result.get_n_dist_segments(); ++i)
				x_values.push_back(result.get_dist_segment(i).get_right_x());

			for (size_t i = 0; i != N_RANDOM_POINTS; ++i)
				x_values.push_back(util::random_int32(0, PERIOD() - 1));

			const size_t n_x_values = x_values.size();
			for (size_t i = 0; i != n_x_values; ++i)
			{
				x_values.push_back(x_values[i] + PERIOD());
				x_values.push_back(x_values[i] - PERIOD());
			}

			for (const int32_t& x : x_values)
			{
				int32_t time_rhs = rhs.eval_time(x);
				int32_t time_lhs_rhs = lhs.eval_time(time_rhs + x) + time_rhs;
				int32_t time_result = result.eval_time(x);

				if (time_lhs_rhs != time_result)
				{
					std::cerr << "\n\nx = " << x << std::endl;
					std::cerr << "time_rhs = " << time_rhs << std::endl;
					std::cerr << "time_lhs = " << lhs.eval_time(time_rhs + x) << std::endl;
					std::cerr << "time_lhs_rhs = " << time_lhs_rhs << std::endl;
					std::cerr << "time_result = " << time_result << std::endl;
					std::cerr << "\nlhs = " << lhs << std::endl;
					std::cerr << "\nrhs = " << rhs << std::endl;
					std::cerr << "\nresult = " << result << std::endl;

					return false;
				}
				/*
				if (!rhs.has_dist() || !lhs.has_dist())
					continue;

				if (!result.has_dist())
					return false;
				
				int32_t dist_rhs = rhs.eval_dist(x);
				int32_t dist_lhs_rhs = lhs.eval_dist(time_rhs + x) + dist_rhs;
				int32_t dist_result = result.eval_dist(x);

				if (dist_lhs_rhs != dist_result)
				{
					std::cerr << "\n\nx = " << x << std::endl;
					std::cerr << "dist_lhs_rhs = " << dist_lhs_rhs << std::endl;
					std::cerr << "dist_result = " << dist_result << std::endl;
					std::cerr << "\nlhs = " << lhs << std::endl;
					std::cerr << "\nrhs = " << rhs << std::endl;
					std::cerr << "\nresult = " << result << std::endl;

					return false;
				}*/
			}
			for (int i = 0; i < blockNew.size(); i++) {
				assert(result.get_time_segment(i, blockNew).get_right_y() >= std::get<3>(blockNew[i]));
			}

			return true;
		}

		static bool dbg_merge_is_ok(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const TimeDist_tmpl& result, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockNew)
		{
			const size_t N_RANDOM_POINTS = 100;

			std::vector<int32_t> x_values;
			x_values.push_back(0);
			x_values.push_back(PERIOD());

			for (size_t i = 0; i != lhs.get_n_time_segments(); ++i)
				x_values.push_back(lhs.get_time_segment(i, blockL).get_right_x());

			for (size_t i = 0; i != rhs.get_n_time_segments(); ++i)
				x_values.push_back(rhs.get_time_segment(i, blockR).get_right_x());

			for (size_t i = 0; i != result.get_n_time_segments(); ++i)
				x_values.push_back(result.get_time_segment(i, blockNew).get_right_x());

			for (size_t i = 0; i != rhs.get_n_dist_segments(); ++i)
				x_values.push_back(rhs.get_dist_segment(i).get_right_x());

			for (size_t i = 0; i != result.get_n_dist_segments(); ++i)
				x_values.push_back(result.get_dist_segment(i).get_right_x());

			for (size_t i = 0; i != N_RANDOM_POINTS; ++i)
				x_values.push_back(util::random_int32(0, PERIOD() - 1));

			for (size_t i = 0; i != N_RANDOM_POINTS; ++i)
				x_values.push_back(util::random_int32(0, PERIOD() - 1));

			const size_t n_x_values = x_values.size();
			for (size_t i = 0; i != n_x_values; ++i)
			{
				x_values.push_back(x_values[i] + PERIOD());
				x_values.push_back(x_values[i] - PERIOD());
			}

			for (const int32_t& x : x_values)
			{
				int32_t time_lhs = lhs.eval_time(x);
				int32_t time_rhs = rhs.eval_time(x);
				int32_t time_result = result.eval_time(x);

				if (std::min(time_lhs, time_rhs) != time_result)
				{
					std::cerr << "\n\nx = " << x << std::endl;
					std::cerr << "time_lhs = " << time_lhs << std::endl;
					std::cerr << "time_rhs = " << time_rhs << std::endl;
					std::cerr << "time_result = " << time_result << std::endl;
					std::cerr << "\nlhs =\n" << lhs << std::endl;
					std::cerr << "\nrhs =\n" << rhs << std::endl;
					std::cerr << "\nresult\n= " << result << std::endl;

					return false;
				}
				/*
				if (!rhs.has_dist() || !lhs.has_dist())
					continue;
				
				if (!result.has_dist())
					return false;
				
				int32_t dist_lhs = lhs.eval_dist(x);
				int32_t dist_rhs = rhs.eval_dist(x);
				int32_t dist_result = result.eval_dist(x);

				if
					(
					(dist_lhs != dist_result && dist_rhs != dist_result)
						|| (time_lhs < time_rhs && dist_lhs != dist_result)
						|| (time_lhs > time_rhs && dist_rhs != dist_result)
						|| (time_lhs == time_rhs && dist_lhs < dist_rhs && dist_lhs != dist_result)
						|| (time_lhs == time_rhs && dist_rhs < dist_lhs && dist_rhs != dist_result)
						)
				{
					std::cerr << "\n\nx = " << x << std::endl;
					std::cerr << "time_lhs = " << time_lhs << std::endl;
					std::cerr << "time_rhs = " << time_rhs << std::endl;
					std::cerr << "dist_lhs = " << dist_lhs << std::endl;
					std::cerr << "dist_rhs = " << dist_rhs << std::endl;
					std::cerr << "dist_result = " << dist_result << std::endl;
					std::cerr << "\nlhs =\n" << lhs << std::endl;
					std::cerr << "\nrhs =\n" << rhs << std::endl;
					std::cerr << "\nresult =\n" << result << std::endl;

					return false;
				}
				*/
			}

			for (int i = 0; i < blockNew.size(); i++) {
				assert(result.get_time_segment(i, blockNew).get_right_y() >= std::get<3>(blockNew[i]));
			}

			return true;
		}

		static constexpr uint32_t UNDEFINED_BELOW = 0;
		static constexpr uint32_t G_BELOW = 1;
		static constexpr uint32_t F_BELOW = 2;

		bool is_ptr() const
		{
			return ((_data & 1) == 0);
		}

		const uint16_t* get_ptr() const
		{
			RITCH_ASSERT(is_ptr());
			return reinterpret_cast<const uint16_t*>(_data);
		}

		uint16_t* get_ptr()
		{
			RITCH_ASSERT(is_ptr());
			return reinterpret_cast<uint16_t*>(_data);
		}

		void set_ptr(uint16_t* ptr)
		{
			_data = reinterpret_cast<uint64_t>(ptr);
			RITCH_ASSERT(is_ptr());
		}

		bool has_inplace_time() const
		{
			return !is_ptr();
		}

		bool has_allocated_time() const
		{
			return is_ptr();
		}

		bool has_inplace_dist() const
		{
			return (_data & 1) == 1 && (_data & 0xffffffff) != 0xffffffff;
		}

		bool has_allocated_dist() const
		{
			return
				is_ptr() &&
				get_ptr() &&
				(((*reinterpret_cast<const uint32_t*>(get_ptr())) & 1) != 0);
		}

		int32_t get_inplace_time() const
		{
			RITCH_ASSERT(has_inplace_time());
			return int32_t(_data >> 32);
		}

		int32_t get_inplace_dist() const
		{
			RITCH_ASSERT(has_inplace_dist());
			return int32_t((_data & 0xffffffff) >> 1);
		}

		void set_inplace_time_and_dist(uint32_t time, uint32_t dist)
		{
			RITCH_ASSERT((uint64_t(dist) << 1) < uint64_t(1) << 32);
			_data = (uint64_t(time) << 32) | (uint64_t(dist) << 1) | uint64_t(1);
			RITCH_ASSERT(!is_ptr());
		}

		void set_inplace_time_and_unset_dist(uint32_t time)
		{
			_data = (uint64_t(time) << 32) | uint64_t(0xffffffff);
			RITCH_ASSERT(has_inplace_time());
			RITCH_ASSERT(!has_inplace_dist());
			RITCH_ASSERT(!has_allocated_dist());
		}

		const uint16_t* get_time_segment_base() const
		{
			RITCH_ASSERT(is_ptr());
			bool has_non_const_distance = ((*reinterpret_cast<const uint32_t*>(get_ptr())) & 1) != 0;

			size_t offset = 6;
			if (has_non_const_distance)
				offset += 2;

			return get_ptr() + offset;
		}

		const uint16_t* get_dist_segment_base() const
		{
			RITCH_ASSERT(is_ptr());
			if (((*reinterpret_cast<const uint32_t*>(get_ptr())) & 1) == 0)
				return nullptr;

			uint32_t first_word = *reinterpret_cast<const uint32_t*>(get_ptr());
			size_t n_time_segments = first_word >> 1;
			return get_ptr() + 8 + 3 * n_time_segments;
		}

		uint64_t _data;

		static uint16_t get_x_lo(const uint16_t* ptr) { return ptr[0]; }
		static uint16_t get_x_hi(const uint16_t* ptr) { return (ptr[2] >> 8) & uint16_t(0x7f); }
		static uint16_t get_y_lo(const uint16_t* ptr) { return ptr[1]; }
		static uint16_t get_y_hi(const uint16_t* ptr) { return ptr[2] & uint16_t(0xff); }
		static int32_t get_slope(const uint16_t* ptr) { return (ptr[2] & uint16_t(0x8000)) == uint16_t(0) ? 0 : -1; }

		static int32_t get_x(const uint16_t* ptr) { return (uint32_t(get_x_hi(ptr)) << 16) | uint32_t(get_x_lo(ptr)); }
		static int32_t get_y(const uint16_t* ptr) { return (uint32_t(get_y_hi(ptr)) << 16) | uint32_t(get_y_lo(ptr)); }

		static void set_slope(uint16_t* ptr, int32_t slope)
		{
			if (slope == 0)
				ptr[2] &= uint16_t(0x7fff);
			else
				ptr[2] |= uint16_t(0x8000);
		}

		static void set_x_lo(uint16_t* ptr, uint16_t x_lo) { ptr[0] = x_lo; }
		static void set_y_lo(uint16_t* ptr, uint16_t y_lo) { ptr[1] = y_lo; }

		static void set_x_hi(uint16_t* ptr, uint16_t x_hi)
		{
			RITCH_ASSERT(x_hi <= uint16_t(0x7f));
			ptr[2] &= uint16_t(0x80ff);
			ptr[2] |= (x_hi << 8);
		}

		static void set_y_hi(uint16_t* ptr, uint16_t y_hi)
		{
			RITCH_ASSERT(y_hi <= uint16_t(0xff));
			ptr[2] &= uint16_t(0xff00);
			ptr[2] |= y_hi;
		}

		static void set_x(uint16_t* ptr, uint32_t x)
		{
			RITCH_ASSERT(x <= 0x7fffff);
			set_x_lo(ptr, uint16_t(x & 0xffff));
			set_x_hi(ptr, uint16_t(x >> 16));
		}

		static void set_y(uint16_t* ptr, uint32_t y)
		{
			RITCH_ASSERT(y <= 0xffffff);
			set_y_lo(ptr, uint16_t(y & 0xffff));
			set_y_hi(ptr, uint16_t(y >> 16));
		}

	public:

		TimeDist_tmpl()
			: _data(0) 
		{}

		TimeDist_tmpl(int32_t const_time)
			: _data(0)
		{
			set_inplace_time_and_unset_dist(const_time);
		}

		TimeDist_tmpl(int32_t const_time, int32_t const_dist)
			: _data(0) 
		{
			set_inplace_time_and_dist(const_time, const_dist);
		}


		template <typename TimeSegIt>
		TimeDist_tmpl(const TimeSegIt& time_segments_begin, const TimeSegIt& time_segments_end, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> &blockinfo)
			: TimeDist_tmpl(time_segments_begin, time_segments_end, time_segments_begin, time_segments_begin, blockinfo, true)
		{}
		template <typename TimeSegIt>
		TimeDist_tmpl(const TimeSegIt& time_segments_begin, const TimeSegIt& time_segments_end, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> &blockinfo, bool checks)
			: TimeDist_tmpl(time_segments_begin, time_segments_end, time_segments_begin, time_segments_begin, blockinfo, checks)
		{}

		//template <typename TimeSegmentIt, typename DistSegmentIt>
		//TimeDist_tmpl(const TimeSegmentIt& time_seg_begin, const TimeSegmentIt& time_seg_end, const DistSegmentIt& dist_seg_begin, const DistSegmentIt& dist_seg_end, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block, bool checks)
		//	: TimeDist_tmpl(time_seg_begin, time_seg_end, dist_seg_begin, dist_seg_end,block, checks)
		//{}

		template <typename TimeSegmentIt, typename DistSegmentIt>
		TimeDist_tmpl(const TimeSegmentIt& time_seg_begin, const TimeSegmentIt& time_seg_end, const DistSegmentIt& dist_seg_begin, const DistSegmentIt& dist_seg_end, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockinfo, bool checks)
			: TimeDist_tmpl()
		{
			bool add_block_info = true;
			if (blockinfo.size() != 0) {
				add_block_info = false;
			}
			const size_t n_time_segments = std::distance(time_seg_begin, time_seg_end);
			const size_t n_dist_segments = std::distance(dist_seg_begin, dist_seg_end);

			if (n_time_segments == 0)
				return;

			RITCH_ASSERT(
				time_seg_begin->get_right_y() + time_seg_begin->get_right_x() + (PERIOD() - (time_seg_end - 1)->get_right_x())
				>= (time_seg_end - 1)->get_right_y()
			);

			if (n_time_segments == 1 && n_dist_segments == 0 && time_seg_begin->get_slope() == 0)
			{
				set_inplace_time_and_unset_dist(time_seg_begin->get_right_y());
				return;
			}

			if (n_time_segments == 1 && n_dist_segments == 1 && time_seg_begin->get_slope() == 0)
			{
				RITCH_ASSERT(dist_seg_begin->get_slope() == 0);

				set_inplace_time_and_dist(time_seg_begin->get_right_y(), dist_seg_begin->get_right_y());
				return;
			}

			size_t n_elements = (6 + 3 * n_time_segments);
			if (n_dist_segments != 0)
				n_elements += (2 + 3 * n_dist_segments);

			uint16_t* data_ptr = new uint16_t[n_elements];
			reinterpret_cast<uint32_t*>(data_ptr)[0] = uint32_t(n_time_segments) << 1;

			size_t offset = 0;
			if (n_dist_segments != 0)
			{
				reinterpret_cast<uint32_t*>(data_ptr)[0] |= 1;
				reinterpret_cast<uint32_t*>(data_ptr)[1] = uint32_t(n_dist_segments);
				offset += 2;
			}

			int32_t& min_time = *reinterpret_cast<int32_t*>(data_ptr + 2 + offset);
			int32_t& max_time = *reinterpret_cast<int32_t*>(data_ptr + 4 + offset);

			min_time = std::numeric_limits<int32_t>::max();
			max_time = std::numeric_limits<int32_t>::min();
			uint16_t* ptr = data_ptr + 6 + offset;

			int32_t previous_right_x = (time_seg_end - 1)->get_right_x() - PERIOD();
			for (TimeSegmentIt it = time_seg_begin; it != time_seg_end; ++it)
			{
				RITCH_ASSERT(it->get_right_x() > 0);
				RITCH_ASSERT(it->get_right_x() <= PERIOD());

				set_slope(ptr, it->get_slope());
				set_x(ptr, it->get_right_x());
				set_y(ptr, it->get_right_y());
				RITCH_ASSERT(get_slope(ptr) == it->get_slope());
				RITCH_ASSERT(get_x(ptr) == it->get_right_x());
				RITCH_ASSERT(get_y(ptr) == it->get_right_y());

				if (add_block_info) {
					blockinfo.push_back(std::make_tuple(it->get_right_x(), it->get_block_start(), it->get_block_end(), it->get_default_time()));
					RITCH_ASSERT(it->get_default_time() <= it->get_right_y());
				}
				ptr += 3;

				min_time = std::min(min_time, it->get_right_y());

				if (it->get_slope() == 0)
					max_time = std::max(max_time, it->get_right_y());
				else
					max_time = std::max(max_time, it->get_right_y() + (it->get_right_x() - previous_right_x));

				previous_right_x = it->get_right_x();
			}
			if (add_block_info && checks) {
				bool onceearlier = false;
				for (int i = 0; i < blockinfo.size(); i++) {
					if (!onceearlier) {
						if (min_time >= std::get<3>(blockinfo[i])) {
							onceearlier = true;
						}
					}
				}
				RITCH_ASSERT(onceearlier);
			}


			for (DistSegmentIt it = dist_seg_begin; it != dist_seg_end; ++it)
			{
				RITCH_ASSERT(it->get_right_x() > 0);
				RITCH_ASSERT(it->get_right_x() <= PERIOD());

				set_slope(ptr, 0);
				set_x(ptr, it->get_right_x());
				set_y(ptr, it->get_right_y());
				RITCH_ASSERT(get_slope(ptr) == 0);
				RITCH_ASSERT(get_x(ptr) == it->get_right_x());
				RITCH_ASSERT(get_y(ptr) == it->get_right_y());

				ptr += 3;
			}

			reinterpret_cast<uint16_t*&>(_data) = data_ptr;
			RITCH_ASSERT(blockinfo.size() == get_n_time_segments());
			RITCH_ASSERT(get_n_time_segments() == n_time_segments);
			RITCH_ASSERT(get_n_dist_segments() == n_dist_segments);
		}

		TimeDist_tmpl(const std::vector<Segment>& time_segments, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block, bool checks)
			: TimeDist_tmpl(time_segments.begin(), time_segments.end(), block, checks)
		{}
		//TimeDist_tmpl(const std::vector<Segment>& time_segments, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockinfo, bool checks)
		//	: TimeDist_tmpl(time_segments.begin(), time_segments.end(), blockinfo, checks)
		//{}

		TimeDist_tmpl(const std::vector<Segment>& time_segments, const std::vector<Segment>& dist_segments, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block, bool checks)
			: TimeDist_tmpl(time_segments.begin(), time_segments.end(), dist_segments.begin(), dist_segments.end(), block, checks)
		{}
		//TimeDist_tmpl(const std::vector<Segment>& time_segments, const std::vector<Segment>& dist_segments, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockinfo, bool checks)
		//	: TimeDist_tmpl(time_segments.begin(), time_segments.end(), dist_segments.begin(), dist_segments.end(), blockinfo, checks)
		//{}

		TimeDist_tmpl(const std::vector<Segment>& time_segments, const int32_t distance, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block, bool checks)
			: TimeDist_tmpl(time_segments, std::vector<Segment>(1, Segment(0, PERIOD(), distance, 0, 0,0)), block , checks)
		{}

		//TimeDist_tmpl(const std::vector<Segment>& time_segments, const int32_t distance, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockinfo, bool checks)
		//	: TimeDist_tmpl(time_segments, std::vector<Segment>(1, Segment(0, PERIOD(), distance, 0, 0,0)), blockinfo, checks)
		//{}

		TimeDist_tmpl(const int32_t time, const std::vector<Segment>& dist_segments, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockinfo, bool checks)
			: TimeDist_tmpl(std::vector<Segment>(1, Segment(0, PERIOD(), time, 0, 0,time)), dist_segments, blockinfo, checks)
		{}
		
		TimeDist_tmpl(const TimeDist_tmpl& other)
		{
			if (other.is_ptr())
			{
				size_t n_elements = (6 + 3 * other.get_n_time_segments());
				if (other.get_n_dist_segments() != 0)
					n_elements += (2 + 3 * other.get_n_dist_segments());

				reinterpret_cast<uint16_t*&>(_data) = new uint16_t[n_elements];

				memcpy(
					reinterpret_cast<uint16_t*&>(_data),
					reinterpret_cast<const uint16_t*>(other._data),
					n_elements * sizeof(uint16_t));
			}
			else
			{
				_data = other._data;
			}
		}
		

		TimeDist_tmpl(const TimeDist_tmpl& other, const std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>> & blockO, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockT)
		{
			if (other.is_ptr())
			{
				size_t n_elements = (6 + 3 * other.get_n_time_segments());
				if (other.get_n_dist_segments() != 0)
					n_elements += (2 + 3 * other.get_n_dist_segments());

				reinterpret_cast<uint16_t*&>(_data) = new uint16_t[n_elements];

				memcpy(
					reinterpret_cast<uint16_t*&>(_data),
					reinterpret_cast<const uint16_t*>(other._data),
					n_elements * sizeof(uint16_t));
			}
			else
			{
				_data = other._data;
			}
			blockT = blockO;
		}
		
		TimeDist_tmpl(TimeDist_tmpl&& other)
		{
			_data = other._data;
			other._data = 0;

		}
		
		~TimeDist_tmpl()
		{
			if (is_ptr())
				delete[] reinterpret_cast<uint16_t*>(_data);

			_data = 0;
		}


		
		TimeDist_tmpl& operator= (const TimeDist_tmpl& other)
		{
			if (this != &other)
			{
				if (is_ptr())
					delete[] reinterpret_cast<uint16_t*>(_data);

				if (!other.is_ptr())
				{
					if (!other.has_inplace_dist())
						set_inplace_time_and_unset_dist(other.get_inplace_time());
					else
						set_inplace_time_and_dist(other.get_inplace_time(), other.get_inplace_dist());
				}
				else
				{
					size_t n_elements = (6 + 3 * other.get_n_time_segments());
					if (other.get_n_dist_segments() != 0)
						n_elements += (2 + 3 * get_n_dist_segments());

					reinterpret_cast<uint16_t*&>(_data) = new uint16_t[n_elements];

					memcpy(
						reinterpret_cast<uint16_t*&>(_data),
						reinterpret_cast<const uint16_t*>(other._data),
						n_elements * sizeof(uint16_t));
				}
			}

			return *this;
		}
		
		TimeDist_tmpl& operator= (TimeDist_tmpl&& other)
		{
			if (this != &other)
			{
				if (is_ptr())
					delete[] reinterpret_cast<uint16_t*>(_data);
				_data = other._data;
				other._data = 0;
			}

			return *this;
		}
		
		bool is_infty() const { return _data == 0; }

		bool has_distance_info() const
		{
			return has_inplace_dist() || has_allocated_dist();
		}

		size_t get_time_complexity() const
		{
			if (!is_ptr())
				return 0;

			const size_t n_segements = get_n_time_segments();
			return n_segements <= 1 ? 0 : n_segements;
		}

		size_t get_dist_complexity() const
		{
			if (!is_ptr())
				return 0;

			const size_t n_segements = get_n_dist_segments();
			return n_segements <= 1 ? 0 : n_segements;
		}

		size_t get_n_time_segments() const
		{
			if (!is_ptr())
				return 1;

			if (get_ptr() == nullptr)
				return 0;

			const uint32_t& first_word = *reinterpret_cast<const uint32_t*>(get_ptr());
			return (first_word >> 1);
		}

		size_t get_n_dist_segments() const
		{
			if (has_inplace_dist())
				return 1;

			if (!has_allocated_dist())
				return 0;

			const uint32_t& second_word = reinterpret_cast<const uint32_t*>(get_ptr())[1];
			return second_word;
		}

		std::tuple<int32_t, int32_t> get_min_max_time() const
		{
			if (!is_ptr())
				return std::make_tuple(get_inplace_time(), get_inplace_time());

			if (get_ptr() == nullptr)
				return std::make_tuple(
					std::numeric_limits<int32_t>::max(),
					std::numeric_limits<int32_t>::max()
				);

			size_t offset = 0;
			if (has_allocated_dist()) ++offset;

			return
				std::make_tuple(
					reinterpret_cast<const int32_t*>(get_ptr())[1 + offset],
					reinterpret_cast<const int32_t*>(get_ptr())[2 + offset]
				);
		}

		int32_t get_min_time() const { return std::get<0>(get_min_max_time()); }
		int32_t get_max_time() const { return std::get<1>(get_min_max_time()); }

		Segment get_time_segment(size_t index, const std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>> & blockinfo) const
		{
			if (!is_ptr())
				return Segment(0, PERIOD(), get_inplace_time(), 0, 0, get_inplace_time());

			RITCH_ASSERT(index < get_n_time_segments());
			const uint16_t* ptr = get_time_segment_base() + 3 * index;

			RITCH_ASSERT(get_x(ptr) > 0);
			RITCH_ASSERT(get_x(ptr) <= PERIOD());
			RITCH_ASSERT(get_y(ptr) >= 0);
			if (blockinfo.size() == 0) {
				return Segment{ get_slope(ptr), get_x(ptr), get_y(ptr),0, 0, get_min_time(), true };
			}
			if (index < blockinfo.size()) {
				return Segment{ get_slope(ptr), get_x(ptr), get_y(ptr),std::get<1>(blockinfo.at(index)), std::get<2>(blockinfo.at(index)), std::get<3>(blockinfo.at(index)), true };
			}
			else {
				return Segment{ get_slope(ptr), get_x(ptr), get_y(ptr),std::get<1>(blockinfo.at(blockinfo.size() - 1)), std::get<2>(blockinfo.at(blockinfo.size() - 1)), std::get<3>(blockinfo.at(blockinfo.size() - 1)), true };
			}
		}


		Segment get_dist_segment(size_t index) const
		{
			if (has_inplace_dist())
				return Segment(0, PERIOD(), get_inplace_dist(), 0, 0,0);

			RITCH_ASSERT(is_ptr());
			RITCH_ASSERT(has_allocated_dist());
			RITCH_ASSERT(index < get_n_dist_segments());
			const uint16_t* ptr = get_dist_segment_base() + 3 * index;

			RITCH_ASSERT(get_x(ptr) > 0);
			RITCH_ASSERT(get_x(ptr) <= PERIOD());
			RITCH_ASSERT(get_y(ptr) >= 0);
			return Segment(0, get_x(ptr), get_y(ptr), 0, 0,0);
		}

		int32_t eval_time(int32_t x) const
		{
			if (has_inplace_time())
				return get_inplace_time();

			RITCH_ASSERT(is_ptr());

			if (get_ptr() == nullptr)
				return std::numeric_limits<int32_t>::max();

			const int32_t x_mod = util::modulo(x - 1, PERIOD()) + 1;
			uint16_t const * const ptr_begin = get_time_segment_base();
			uint16_t const * const ptr_end = ptr_begin + 3 * get_n_time_segments();

			for (const uint16_t* ptr = ptr_begin; ptr != ptr_end; ptr += 3)
			{
				if (get_x(ptr) >= x_mod)
				{
					if (get_slope(ptr) == 0)
						return get_y(ptr);

					return get_y(ptr) + (get_x(ptr) - x_mod);
				}
			}

			if (get_slope(ptr_begin) == 0)
				return get_y(ptr_begin);

			return get_y(ptr_begin) + get_x(ptr_begin) + (PERIOD() - x_mod);
		}

		int32_t eval_dist(int32_t x) const
		{
			if (!has_inplace_dist() && !has_allocated_dist())
				return std::numeric_limits<int32_t>::max();

			if (has_inplace_dist())
				return get_inplace_dist();

			const int32_t x_mod = util::modulo(x - 1, PERIOD()) + 1;
			uint16_t const * const ptr_begin = get_dist_segment_base();
			uint16_t const * const ptr_end = ptr_begin + 3 * get_n_dist_segments();

			for (const uint16_t* ptr = ptr_begin; ptr != ptr_end; ptr += 3)
				if (get_x(ptr) >= x_mod)
					return get_y(ptr);

			return get_y(ptr_begin);
		}

		std::tuple<int32_t, int32_t> eval_time_dist(int32_t x) const
		{
			return std::make_tuple(eval_time(x), eval_dist(x));
		}

		bool time_undercuts_strictly(const TimeDist_tmpl& other, const int32_t& this_offset = 0, const int32_t& other_offset = 0) const
		{
			if (get_max_time() + this_offset < other.get_min_time() + other_offset)
				return true;

			if (other.get_max_time() + other_offset <= get_min_time() + this_offset)
				return false;

			if (has_const_time(std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>()) && get_const_time(std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>()) + this_offset < other.get_max_time() + other_offset)
				return true;

			if (other.has_const_time(std::vector<std::tuple<int32_t,int32_t, int32_t, int32_t>>()) && get_min_time() + this_offset < other.get_const_time(std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>()) + other_offset)
				return true;

			const size_t n = get_n_time_segments();
			const size_t m = other.get_n_time_segments();

			size_t i = 0;
			size_t j = 0;

			auto g = [this, &this_offset](const size_t& i)
			{
				auto segment = this->get_time_segment_descr_periodic(i, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>());
				segment.y += this_offset;
				segment.y_left += this_offset;

				return segment;
			};

			auto f = [&other, &other_offset](const size_t& i)
			{
				auto segment = other.get_time_segment_descr_periodic(i, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>());
				segment.y += other_offset;
				segment.y_left += other_offset;

				return segment;
			};

			while (i < n || j < m)
			{
				//const Point p = single_inner_intersection_point(g(i), f(j));
				const std::tuple<Point, int32_t, int32_t, int32_t> tupleP = single_inner_intersection_point(g(i), f(j));
				const Point p = std::get<0>(tupleP);

				if (is_valid_point(p))
					return true;

				if (g(i).x == f(j).x)
				{
					if (g(i).y == f(j).y && g(i).m == 0 && f(j).m != 0)
						return true;
					if (g(i).y < f(j).y)
						return true;

					++i;
					++j;
				}
				else if (g(i).x < f(j).x)
				{
					if (f(j).m == 0)
					{
						if (g(i).y < f(j).y)
							return true;
					}
					else
					{
						const int32_t y_f = f(j).y + f(j).x - g(i).x;

						if (g(i).y < y_f || (g(i).y == y_f && g(i).m == 0))
							return true;
					}

					++i;
				}
				else
				{
					RITCH_ASSERT(f(j).x < g(i).x);

					if (g(i).m == 0)
					{
						if (g(i).y < f(j).y)
							return true;
					}
					else
					{
						const int32_t y_g = g(i).y + g(i).x - f(j).x;
						if (y_g < f(j).y) return true;
					}

					++j;
				}
			}

			const std::tuple<Point, int32_t, int32_t, int32_t> tupleP = single_inner_intersection_point(g(n), f(m));
			const Point p = std::get<0>(tupleP);
			if (is_valid_point(p))
				return true;

			return false;
		}

		bool time_undercuts_weakly(const TimeDist_tmpl& other, const int32_t weaking_offset = 1) const
		{
			return time_undercuts_strictly(other, 0, weaking_offset);
		}

		bool has_const_time(const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block) const
		{
			return get_n_time_segments() == 1 && get_time_segment(0, block).get_slope() == 0;
		}

		bool has_dist() const
		{
			return get_n_dist_segments() > 0;
		}

		bool has_const_dist() const
		{
			return get_n_dist_segments() == 1;
		}

		int32_t get_const_time(const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block) const
		{
			RITCH_ASSERT(has_const_time(block));
			return get_time_segment(0, block).get_right_y();
		}

		int32_t get_const_dist() const
		{
			RITCH_ASSERT(has_const_dist());
			return get_dist_segment(0).get_right_y();
		}

		/* link also updates information of blocknew! */
		friend TimeDist_tmpl link(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockNew, bool checks)
		{
			TimeDist_tmpl result;

			if (lhs.has_const_time(blockL) && rhs.has_const_time(blockR))
			{
				if (!lhs.has_dist() || !rhs.has_dist())
					result = TimeDist_tmpl(lhs.get_const_time(blockL) + rhs.get_const_time(blockR));

				else if (lhs.has_const_dist() && rhs.has_const_dist())

					result =
					TimeDist_tmpl(
						lhs.get_const_time(blockL) + rhs.get_const_time(blockR),
						lhs.get_const_dist() + rhs.get_const_dist());

				else {
					std::vector<Segment> result_dist_segments = link_dist(lhs, rhs, blockL, blockR);
					result = TimeDist_tmpl(lhs.get_const_time(blockL) + rhs.get_const_time(blockR), result_dist_segments, blockNew, checks);
				}
			}
			else {


				std::vector<Segment> result_time_segments = link_time(lhs, rhs, blockL, blockR);
				remove_unnecessary_segement_from_end(result_time_segments);

				if (lhs.get_n_dist_segments() == 0 || rhs.get_n_dist_segments() == 0)
					result = TimeDist_tmpl(result_time_segments.begin(), result_time_segments.end(), blockNew, checks);

				std::vector<Segment> result_dist_segments = link_dist(lhs, rhs, blockL, blockR);
				remove_unnecessary_segement_from_end(result_dist_segments);
				result = TimeDist_tmpl(
					result_time_segments.begin(), result_time_segments.end(),
					result_dist_segments.begin(), result_dist_segments.end(), blockNew, checks);
			}
			//if (checks) {
			assert(
				dbg_link_is_ok(lhs, rhs, result, blockL, blockR, blockNew
				)
				);
			//}

			return result;
				
		}




		friend TimeDist_tmpl merge(const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> &blockL, const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> &blockR, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> &blockNew, bool checks)
		{
			if (lhs.get_max_time() < rhs.get_min_time())
			{
				assert(dbg_merge_is_ok(lhs, rhs, lhs, blockL, blockR, blockNew));
				return TimeDist_tmpl(lhs, blockL, blockNew);
			}

			if (rhs.get_max_time() < lhs.get_min_time())
			{
				assert(dbg_merge_is_ok(lhs, rhs, rhs, blockL, blockR, blockNew));
				return TimeDist_tmpl(rhs, blockR, blockNew);
			}

			if (lhs.get_n_dist_segments() == 0 || rhs.get_n_dist_segments() == 0)
			{
				if (lhs.get_max_time() == rhs.get_min_time())
				{
					assert(dbg_merge_is_ok(lhs, rhs, lhs, blockL, blockR, blockNew));
					return TimeDist_tmpl(lhs, blockL, blockNew);
				}
				if (rhs.get_max_time() == lhs.get_min_time())
				{
					assert(dbg_merge_is_ok(lhs, rhs, rhs, blockL, blockR, blockNew));
 					return TimeDist_tmpl(rhs, blockR, blockNew);
				}

				std::vector<Segment> result_time_segments = merge_time(lhs, rhs, blockL, blockR);
				remove_unnecessary_segement_from_end(result_time_segments);

				assert(dbg_merge_is_ok(lhs, rhs, TimeDist_tmpl(result_time_segments, blockNew, false), blockL, blockR, blockNew));
				return TimeDist_tmpl(result_time_segments, blockNew, checks);
			}

			const size_t n_g_time = lhs.get_n_time_segments();
			const size_t n_g_dist = lhs.get_n_dist_segments();
			auto g = [&lhs, blockL](const size_t& i) { return lhs.get_time_segment_descr_periodic(i, blockL); };
			auto d = [&lhs](const size_t& i) { return lhs.get_dist_segment_descr_periodic(i); };
			size_t i = 0;
			size_t k = 0;

			const size_t n_f_time = rhs.get_n_time_segments();
			const size_t n_f_dist = rhs.get_n_dist_segments();
			auto f = [&rhs, blockR](const size_t& i) { return rhs.get_time_segment_descr_periodic(i, blockR); };
			auto c = [&rhs](const size_t& i) { return rhs.get_dist_segment_descr_periodic(i); };
			size_t j = 0;
			size_t l = 0;

			std::vector<Segment> result_time_segments;
			result_time_segments.reserve(1 + 2 * (n_g_time + n_f_time));

			std::vector<Segment> result_dist_segments;
			result_dist_segments.reserve(1 + 2 * (n_g_time + n_f_time) + n_g_dist + n_f_dist);

			auto emit_distance_segments_until =
				[&](const int32_t& x_right, const size_t& below)
			{
				if (below == F_BELOW)
				{
					for (; c(l).x <= x_right; ++l)
						append_dist_segment(result_dist_segments, Segment(0, c(l).x, c(l).y, 0, 0,0));

					if (result_dist_segments.empty() || x_right > result_dist_segments.back().get_right_x())
						append_dist_segment(result_dist_segments, Segment(0, x_right, c(l).y, 0, 0,0));

					while (d(k).x <= x_right) ++k;
				}
				else if (below == G_BELOW)
				{
					for (; d(k).x <= x_right; ++k)
						append_dist_segment(result_dist_segments, Segment(0, d(k).x, d(k).y, 0, 0,0));

					if (result_dist_segments.empty() || x_right > result_dist_segments.back().get_right_x())
						append_dist_segment(result_dist_segments, Segment(0, x_right, d(k).y, 0, 0,0));

					while (c(l).x <= x_right) ++l;
				}
				else
				{
					int32_t x, y;
					while (d(k).x <= x_right || c(l).x <= x_right)
					{
						x = std::min(d(k).x, c(l).x);
						y = std::min(d(k).y, c(l).y);

						if (d(k).x == c(l).x) { ++k; ++l; }
						else if (d(k).x < c(l).x) { ++k; }
						else { ++l; }

						append_dist_segment(result_dist_segments, Segment(0, x, y, 0, 0,0));
					}

					if (result_dist_segments.empty() || x_right > result_dist_segments.back().get_right_x())
						append_dist_segment(result_dist_segments, Segment(0, x_right, std::min(d(k).y, c(l).y), 0, 0,0));
				}
			};

			auto emit_minimum_distance_at =
				[&](const int32_t x)
			{
				append_dist_segment(result_dist_segments, Segment(0, x, std::min(d(k).y, c(l).y), 0, 0,0));
				while (c(l).x <= x) ++l;
				while (d(k).x <= x) ++k;
			};

			while (i < n_g_time || j < n_f_time)
			{
				int32_t gdef = g(i).default_time;
				int32_t fdef = f(j).default_time;
				if (g(i).default_time == 0) {
					gdef = lhs.get_min_time();
				}
				if (f(j).default_time == 0) {
					fdef = rhs.get_min_time();
				}
				const std::tuple<Point, int32_t, int32_t, int32_t> tp = single_inner_intersection_point(g(i), f(j));
				const Point p = std::get<0>(tp);
				if (is_valid_point(p) && p.x > 0)
				{
					append_time_segment(result_time_segments, Segment(0, p.x, p.y, std::get<1>(tp), std::get<2>(tp), std::get<3>(tp)));

					if (g(i).y < f(j).y)
						emit_distance_segments_until(p.x - 1, F_BELOW);
					else
						emit_distance_segments_until(p.x - 1, G_BELOW);

					emit_minimum_distance_at(p.x);
				}

				if (g(i).x == f(j).x)
				{
					if (g(i).y == f(j).y)
					{
						if (g(i).m == 0 || f(j).m == 0)
							append_time_segment(result_time_segments, Segment(0, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						else
							append_time_segment(result_time_segments, Segment(-1, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));

						if (g(i).m == f(j).m)
							emit_distance_segments_until(g(i).x - 1, UNDEFINED_BELOW);
						else if (g(i).m == 0)
							emit_distance_segments_until(g(i).x - 1, G_BELOW);
						else if (f(j).m == 0)
							emit_distance_segments_until(g(i).x - 1, F_BELOW);

						emit_minimum_distance_at(g(i).x);
					}
					else if (g(i).y < f(j).y)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						emit_distance_segments_until(g(i).x, G_BELOW);
					}
					else
					{
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
						emit_distance_segments_until(f(j).x, F_BELOW);
					}

					++i;
					++j;
				}
				else if (g(i).x < f(j).x)
				{
					const int32_t y_f = f(j).m == 0 ? f(j).y : f(j).y + f(j).x - g(i).x;

					if (g(i).m == f(j).m && g(i).y == y_f)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						emit_distance_segments_until(g(i).x, UNDEFINED_BELOW);
					}
					else if (g(i).y < y_f)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
						emit_distance_segments_until(g(i).x, G_BELOW);
					}
					else if (y_f < g(i).y)
					{
						append_time_segment(result_time_segments, Segment(f(j).m, g(i).x, y_f, f(j).block_start, f(j).block_end, fdef));
						emit_distance_segments_until(g(i).x, F_BELOW);
					}
					else
					{
						RITCH_ASSERT(g(i).m != f(j).m);
						RITCH_ASSERT(g(i).y == y_f);

						if (g(i).m == 0)
						{
							append_time_segment(result_time_segments, Segment(g(i).m, g(i).x, g(i).y, g(i).block_start, g(i).block_end, gdef));
							emit_distance_segments_until(g(i).x - 1, G_BELOW);
						}
						else
						{
							RITCH_ASSERT(f(j).m == 0);
							append_time_segment(result_time_segments, Segment(f(j).m, g(i).x, y_f, f(j).block_start, f(j).block_end, fdef));
							emit_distance_segments_until(g(i).x - 1, F_BELOW);
						}

						emit_minimum_distance_at(g(i).x);
					}

					++i;
				}
				else
				{
					RITCH_ASSERT(f(j).x < g(i).x);
					const int32_t y_g = g(i).m == 0 ? g(i).y : g(i).y + g(i).x - f(j).x;

					if (f(j).m == g(i).m && f(j).y == y_g)
					{
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
						emit_distance_segments_until(f(j).x, UNDEFINED_BELOW);
					}
					else if (f(j).y < y_g)
					{
						append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
						emit_distance_segments_until(f(j).x, F_BELOW);
					}
					else if (y_g < f(j).y)
					{
						append_time_segment(result_time_segments, Segment(g(i).m, f(j).x, y_g, g(i).block_start, g(i).block_end, gdef));
						emit_distance_segments_until(f(j).x, G_BELOW);
					}
					else
					{
						RITCH_ASSERT(g(i).m != f(j).m);
						RITCH_ASSERT(f(j).y == y_g);

						if (f(j).m == 0)
						{
							append_time_segment(result_time_segments, Segment(f(j).m, f(j).x, f(j).y, f(j).block_start, f(j).block_end, fdef));
							emit_distance_segments_until(f(j).x - 1, F_BELOW);
						}
						else
						{
							RITCH_ASSERT(g(i).m == 0);
							append_time_segment(result_time_segments, Segment(g(i).m, f(j).x, y_g, g(i).block_start, g(i).block_end, gdef));
							emit_distance_segments_until(f(j).x - 1, G_BELOW);
						}

						emit_minimum_distance_at(f(j).x);
					}

					++j;
				}
			}

			const std::tuple<Point, int32_t, int32_t, int32_t> tupleP = single_inner_intersection_point(g(n_g_time), f(n_f_time));
			const Point p = std::get<0>(tupleP);

			RITCH_ASSERT(g(n_g_time).x > PERIOD());
			RITCH_ASSERT(f(n_f_time).x > PERIOD());

			if (is_valid_point(p))
			{
				RITCH_ASSERT(p.x > 0);
				RITCH_ASSERT(g(n_g_time).y != f(n_f_time).y);

				if (p.x <= PERIOD())
				{
					append_time_segment(result_time_segments, Segment(0, p.x, p.y, std::get<1>(tupleP), std::get<2>(tupleP), std::get<3>(tupleP)));

					if (g(n_g_time).y < f(n_f_time).y)
						emit_distance_segments_until(p.x - 1, F_BELOW);
					else
						emit_distance_segments_until(p.x - 1, G_BELOW);

					if (result_dist_segments.back().get_right_x() < p.x)
						emit_minimum_distance_at(p.x);

					if (p.x < PERIOD())
					{
						if (g(n_g_time).y < f(n_f_time).y)
							emit_distance_segments_until(PERIOD(), G_BELOW);
						else
							emit_distance_segments_until(PERIOD(), F_BELOW);
					}
				}
				else
				{
					if (g(n_g_time).y < f(n_f_time).y)
						emit_distance_segments_until(PERIOD(), F_BELOW);
					else
						emit_distance_segments_until(PERIOD(), G_BELOW);
				}
			}
			else
			{
				int32_t y_g;
				int32_t y_f;

				if (g(n_g_time).x <= f(n_f_time).x)
				{
					y_g = g(n_g_time).y;
					y_f = f(n_f_time).m == 0 ? f(n_f_time).y : f(n_f_time).y + (f(n_f_time).x - g(n_g_time).x);
				}
				else
				{
					y_f = f(n_f_time).y;
					y_g = g(n_g_time).m == 0 ? g(n_g_time).y : g(n_g_time).y + (g(n_g_time).x - f(n_f_time).x);
				}

				if (y_g == y_f && g(n_g_time).m == f(n_f_time).m)
				{
					emit_distance_segments_until(PERIOD(), UNDEFINED_BELOW);
				}
				else if (y_g < y_f || (y_g == y_f && g(n_g_time).m == 0))
				{
					RITCH_ASSERT(RITCH_IMPLIES(y_g == y_f, f(n_f_time).m != 0));
					emit_distance_segments_until(PERIOD(), G_BELOW);
				}
				else
				{
					RITCH_ASSERT(y_f < y_g || (y_f == y_g && f(n_f_time).m == 0 && g(n_g_time).m != 0));
					emit_distance_segments_until(PERIOD(), F_BELOW);
				}

			}

			remove_unnecessary_segement_from_end(result_time_segments);
			remove_unnecessary_segement_from_end(result_dist_segments);
			assert(dbg_merge_is_ok(lhs, rhs, TimeDist_tmpl(result_time_segments, result_dist_segments, std::vector<std::tuple<int32_t,int32_t,int32_t,int32_t>>(), false), blockL, blockR, blockNew));

			return TimeDist_tmpl(result_time_segments, result_dist_segments, blockNew, checks);
		}

		friend std::vector<std::pair<int32_t, uint32_t>> merge_middle_nodes
		(
			const TimeDist_tmpl& lhs,
			const TimeDist_tmpl& rhs,
			const std::vector<std::pair<int32_t, uint32_t>>& lhs_middle_nodes,
			const uint32_t rhs_single_middle_node,
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockL,
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & blockR

		)
		{
			std::vector<std::pair<int32_t, uint32_t>> result;

			if (lhs.get_max_time() < rhs.get_min_time())
			{
				result = lhs_middle_nodes;
				return result;
			}

			if (rhs.get_max_time() < lhs.get_min_time())
			{
				result.assign(1, std::make_pair(PERIOD(), rhs_single_middle_node));
				return result;
			}

			if (lhs.get_n_dist_segments() == 0 || rhs.get_n_dist_segments() == 0)
			{
				throw std::runtime_error("nerge_middle_nodes(): not implemented for TimeDist_tmpl without distances.");
			}

			const size_t n_g_time = lhs.get_n_time_segments();
			const size_t n_g_nodes = lhs_middle_nodes.size();
			const size_t n_g_dist = lhs.get_n_dist_segments();
			auto g = [&lhs, &blockL](const size_t& i) { return lhs.get_time_segment_descr_periodic(i, blockL); };
			auto d = [&lhs](const size_t& i) { return lhs.get_dist_segment_descr_periodic(i); };
			auto middle_nodes_g = [&lhs_middle_nodes](const size_t& i) { return get_middle_node_periodic(lhs_middle_nodes, i); };
			size_t h = 0;
			size_t i = 0;
			size_t k = 0;

			const size_t n_f_time = rhs.get_n_time_segments();
			const size_t n_f_dist = rhs.get_n_dist_segments();
			auto f = [&rhs, &blockR](const size_t& i) { return rhs.get_time_segment_descr_periodic(i, blockR); };
			auto c = [&rhs](const size_t& i) { return rhs.get_dist_segment_descr_periodic(i); };
			size_t j = 0;
			size_t l = 0;

			auto emit_middle_nodes_until =
				[&](const int32_t& x_right, const size_t& below)
			{
				if (below == F_BELOW)
				{
					append_middle_node(result, x_right, rhs_single_middle_node);

					while (c(l).x <= x_right) ++l;
					while (d(k).x <= x_right) ++k;
					while (middle_nodes_g(h).first <= x_right) ++h;
				}
				else if (below == G_BELOW)
				{
					for (; middle_nodes_g(h).first <= x_right; ++h)
						append_middle_node(result, middle_nodes_g(h).first, middle_nodes_g(h).second);

					if (result.empty() || x_right > result.back().first)
						append_middle_node(result, x_right, middle_nodes_g(h).second);

					while (c(l).x <= x_right) ++l;
					while (d(k).x <= x_right) ++k;
				}
				else
				{
					while (d(k).x <= x_right || c(l).x <= x_right || middle_nodes_g(h).first <= x_right)
					{
						int32_t x = util::min(d(k).x, c(l).x, middle_nodes_g(h).first);

						if (d(k).y < c(l).y)
							append_middle_node(result, x, middle_nodes_g(h).second);
						else
							append_middle_node(result, x, rhs_single_middle_node);

						if (middle_nodes_g(h).first == x) ++h;
						if (d(k).x == x) ++k;
						if (c(l).x == x) ++l;
					}

					if (result.empty() || x_right > result.back().first)
					{
						uint32_t middle_node = d(k).y < c(l).y ? middle_nodes_g(h).second : rhs_single_middle_node;
						append_middle_node(result, x_right, middle_node);
					}
				}
			};

			auto emit_middle_node_at =
				[&](const int32_t x)
			{
				append_middle_node(result, x, d(k).y < c(l).y ? middle_nodes_g(h).second : rhs_single_middle_node);
				while (middle_nodes_g(h).first <= x) ++h;
				while (d(k).x <= x) ++k;
				while (c(l).x <= x) ++l;
			};

			while (i < n_g_time || j < n_f_time)
			{
				const std::tuple<Point, int32_t, int32_t, int32_t> tupleP = single_inner_intersection_point(g(i), f(j));
				const Point p = std::get<0>(tupleP);

				if (is_valid_point(p) && p.x > 0)
				{
					if (g(i).y < f(j).y)
						emit_middle_nodes_until(p.x - 1, F_BELOW);
					else
						emit_middle_nodes_until(p.x - 1, G_BELOW);

					emit_middle_node_at(p.x);
				}

				if (g(i).x == f(j).x)
				{
					if (g(i).y == f(j).y)
					{
						if (g(i).m == f(j).m)
							emit_middle_nodes_until(g(i).x - 1, UNDEFINED_BELOW);
						else if (g(i).m == 0)
							emit_middle_nodes_until(g(i).x - 1, G_BELOW);
						else if (f(j).m == 0)
							emit_middle_nodes_until(g(i).x - 1, F_BELOW);

						emit_middle_node_at(g(i).x);
					}
					else if (g(i).y < f(j).y)
						emit_middle_nodes_until(g(i).x, G_BELOW);
					else
						emit_middle_nodes_until(f(j).x, F_BELOW);

					++i;
					++j;
				}
				else if (g(i).x < f(j).x)
				{
					const int32_t y_f = f(j).m == 0 ? f(j).y : f(j).y + f(j).x - g(i).x;

					if (g(i).m == f(j).m && g(i).y == y_f)
						emit_middle_nodes_until(g(i).x, UNDEFINED_BELOW);
					else if (g(i).y < y_f)
						emit_middle_nodes_until(g(i).x, G_BELOW);
					else if (y_f < g(i).y)
						emit_middle_nodes_until(g(i).x, F_BELOW);
					else
					{
						RITCH_ASSERT(g(i).m != f(j).m);
						RITCH_ASSERT(g(i).y == y_f);

						if (g(i).m == 0)
							emit_middle_nodes_until(g(i).x - 1, G_BELOW);
						else
							emit_middle_nodes_until(g(i).x - 1, F_BELOW);

						emit_middle_node_at(g(i).x);
					}

					++i;
				}
				else
				{
					RITCH_ASSERT(f(j).x < g(i).x);
					const int32_t y_g = g(i).m == 0 ? g(i).y : g(i).y + g(i).x - f(j).x;

					if (f(j).m == g(i).m && f(j).y == y_g)
						emit_middle_nodes_until(f(j).x, UNDEFINED_BELOW);
					else if (f(j).y < y_g)
						emit_middle_nodes_until(f(j).x, F_BELOW);
					else if (y_g < f(j).y)
						emit_middle_nodes_until(f(j).x, G_BELOW);
					else
					{
						RITCH_ASSERT(g(i).m != f(j).m);
						RITCH_ASSERT(f(j).y == y_g);

						if (f(j).m == 0)
							emit_middle_nodes_until(f(j).x - 1, F_BELOW);
						else
							emit_middle_nodes_until(f(j).x - 1, G_BELOW);

						emit_middle_node_at(f(j).x);
					}

					++j;
				}
			}

			const std::tuple<Point, int32_t, int32_t, int32_t> tupleP = single_inner_intersection_point(g(n_g_time), f(n_f_time));
			const Point p = std::get<0>(tupleP);

			RITCH_ASSERT(g(n_g_time).x > PERIOD());
			RITCH_ASSERT(f(n_f_time).x > PERIOD());

			if (is_valid_point(p))
			{
				RITCH_ASSERT(p.x > 0);
				RITCH_ASSERT(g(n_g_time).y != f(n_f_time).y);

				if (p.x <= PERIOD())
				{
					if (g(n_g_time).y < f(n_f_time).y)
						emit_middle_nodes_until(p.x - 1, F_BELOW);
					else
						emit_middle_nodes_until(p.x - 1, G_BELOW);

					if (result.back().first < p.x)
						emit_middle_node_at(p.x);

					if (p.x < PERIOD())
					{
						if (g(n_g_time).y < f(n_f_time).y)
							emit_middle_nodes_until(PERIOD(), G_BELOW);
						else
							emit_middle_nodes_until(PERIOD(), F_BELOW);
					}
				}
				else if (result.back().first < PERIOD())
				{
					if (g(n_g_time).y < f(n_f_time).y)
						emit_middle_nodes_until(PERIOD(), F_BELOW);
					else
						emit_middle_nodes_until(PERIOD(), G_BELOW);
				}
			}
			else if (result.back().first < PERIOD())
			{
				int32_t y_g;
				int32_t y_f;

				if (g(n_g_time).x <= f(n_f_time).x)
				{
					y_g = g(n_g_time).y;
					y_f = f(n_f_time).m == 0 ? f(n_f_time).y : f(n_f_time).y + (f(n_f_time).x - g(n_g_time).x);
				}
				else
				{
					y_f = f(n_f_time).y;
					y_g = g(n_g_time).m == 0 ? g(n_g_time).y : g(n_g_time).y + (g(n_g_time).x - f(n_f_time).x);
				}

				if (y_g == y_f && g(n_g_time).m == f(n_f_time).m)
				{
					emit_middle_nodes_until(PERIOD(), UNDEFINED_BELOW);
				}
				else if (y_g < y_f || (y_g == y_f && g(n_g_time).m == 0))
				{
					RITCH_ASSERT(RITCH_IMPLIES(y_g == y_f, f(n_f_time).m != 0));
					emit_middle_nodes_until(PERIOD(), G_BELOW);
				}
				else
				{
					RITCH_ASSERT(y_f < y_g || (y_f == y_g && f(n_f_time).m == 0 && g(n_g_time).m != 0));
					emit_middle_nodes_until(PERIOD(), F_BELOW);
				}

			}

			remove_unnecessary_middle_node_from_end(result);
			assert(dbg_middle_node_is_ok(lhs, rhs, lhs_middle_nodes, rhs_single_middle_node, result, blockL, blockR));
			return result;
		}

		std::string to_text(std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block) const
		{
			std::stringstream ss;

			ss << "{";
			for (size_t i = 0; i != get_n_time_segments(); ++i)
				ss << " " << get_time_segment(i, block);

			ss << " }, {";
			for (size_t i = 0; i != get_n_dist_segments(); ++i)
				ss << " " << get_dist_segment(i);

			ss << " }";

			return ss.str();
		}

		friend bool operator== (const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs)
		{
			if (lhs.get_n_time_segments() != rhs.get_n_time_segments())
				return false;

			if (lhs.get_n_dist_segments() != rhs.get_n_dist_segments())
				return false;

			for (size_t i = 0; i != lhs.get_n_time_segments(); ++i)
				if (lhs.get_time_segment(i, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>()) != rhs.get_time_segment(i, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>()))
					return false;

			for (size_t i = 0; i != lhs.get_n_dist_segments(); ++i)
				if (lhs.get_dist_segment(i) != rhs.get_dist_segment(i))
					return false;

			return true;
		}

		friend bool operator!= (const TimeDist_tmpl& lhs, const TimeDist_tmpl& rhs)
		{
			return !(lhs == rhs);
		}

		friend std::ostream& operator<<(std::ostream& os, const TimeDist_tmpl& time_dist)
		{
			os << time_dist.to_text(std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>>());
			return os;
		}
	};

}

#endif // RITCH_DATASTR_TIME_DIST_H_
