#ifndef RITCH_IO_BINARY_H_
#define RITCH_IO_BINARY_H_

#include <cassert>
#include <cstdint>
#include <fstream>
#include <vector>

namespace ritch
{
namespace io
{

template <typename T>
static inline void read_value(T& value, std::ifstream& is)
{
	is.read((char*) &value, sizeof(value));
}

static inline uint32_t read_uint32(std::ifstream& is)
{
	uint32_t value;
	read_value(value, is);
	return value;
}

static inline uint8_t read_uint8(std::ifstream& is)
{
	uint8_t value;
	read_value(value, is);
	return value;
}

static inline int32_t read_int32(std::ifstream& is)
{
	int32_t value;
	read_value(value, is);
	return value;
}
static inline double read_double(std::ifstream& is)
{
	double value;
	read_value(value, is);
	return value;
}

template<typename T>
static inline void write_value(const T value, std::ofstream& os)
{
	os.write((const char*) &value, sizeof(value));
}

static inline void write_uint8(const uint8_t value, std::ofstream& os)
{
	write_value(value, os);
}

static inline void write_uint32(const uint32_t value, std::ofstream& os)
{
	write_value(value, os);
}

static inline void write_int32(const int32_t value, std::ofstream& os)
{
	write_value(value, os);
}

static inline void write_double(const double value, std::ofstream& os)
{
	write_value(value, os);
}

template <typename T>
static inline uint64_t chunk_size(T) { return sizeof(T); }

template <typename T, uint64_t N>
static inline uint64_t chunk_size(std::array<T,N>) { return sizeof(T)*N; }

template <typename T>
static inline void read_chunk(T& value, std::ifstream& in)
{
	in.read((char*) &value, sizeof(value));
	assert(! in.eof());
}

template <typename T, size_t N>
static inline void read_chunk(std::array<T,N>& array, std::ifstream& in)
{
	in.read((char*) &array, sizeof(T)*N);
	assert(! in.eof());
}

template <typename T>
static inline void read_all(std::vector<T>& vector, std::ifstream& in)
{
	in.seekg(0, in.end);
	const uint64_t size(in.tellg());
	in.seekg(0);

	vector.reserve(vector.size() + size);

	uint64_t i = 0;
	while ( i != size )
	{
		T chunk;
		read_chunk(chunk, in);
		vector.push_back(chunk);

		i += chunk_size(chunk);
	}
}

template <typename T>
static inline void write_chunk(const T& value, std::ofstream& out)
{
	out.write((const char*) &value, sizeof(value));
}

template <typename T, uint64_t N>
static inline void write_chunk(const std::array<T,N>& array, std::ofstream& out)
{
	out.write((const char*) &array, sizeof(T)*N);
}

template <typename T>
class InputFileIterator
{
private:
	std::string _name;
	std::ifstream _in;
	uint64_t _size_in_byte;
	uint64_t _pos;

public:
	InputFileIterator(const std::string& input_path_name)
	: _name(input_path_name), _in(std::ifstream(input_path_name, std::ios_base::binary)), _size_in_byte(0), _pos(0)
	{
		_in.seekg(0, _in.end);
		_size_in_byte = uint64_t(_in.tellg());
		_in.seekg(0);
	}

	InputFileIterator(const std::string& input_path_name, const uint64_t start_pos, const uint64_t end_pos)
	: _name(input_path_name), _in(std::ifstream(input_path_name, std::ios_base::binary)), _size_in_byte(0), _pos(0)
	{
		_size_in_byte = end_pos >= start_pos ? end_pos - start_pos : 0;
		_in.seekg(start_pos);
	}

	uint64_t get_size_in_byte() const
	{
		return _size_in_byte;
	}

	bool finished() const
	{
		return _pos >= _size_in_byte;
	}

	T read()
	{
		if ( finished() )
			throw std::runtime_error("File '" + _name + "' too short.");

		T chunk;
		read_chunk(chunk, _in);
		_pos += chunk_size(chunk);

		return chunk;
	}

	void close()
	{
		_name = "";
		_in.close();
		_size_in_byte = 0;
		_pos = 0;
	}
};

template <typename T>
static inline InputFileIterator<T> get_input_file_iterator(const std::string& path_file_name)
{
	const std::string name(path_file_name);
	io::InputFileIterator<T> file_it(name);

	if ( file_it.get_size_in_byte() > std::numeric_limits<size_t>::max() )
		throw std::runtime_error(("File " + path_file_name + " too long.").c_str());

	return file_it;
}

}
}

#endif // RITCH_IO_BINARY_H_
