#ifndef RITCH_IO_HERE_TIME_PARSER_H_
#define RITCH_IO_HERE_TIME_PARSER_H_

#include <cassert>
#include <cstdint>
#include <string>

#include <vector>

#include "ritch/datastr/interval_set.h"
#include "ritch/util/misc.h"

namespace ritch
{
namespace io
{

class HereTimeParser
{
	static_assert(
		util::is_chain_without_gaps('0', '1', '2', '3', '4', '5', '6', '7', '8', '9'),
		"HereTimeParser: Expected char to be fulfill '0' + 1 = '1' && '1' + 1 = '2' && ... && '8' + 1 = '9'.");

public:

	static constexpr int32_t ONE_SECOND = 1;
	static constexpr int32_t ONE_MINUTE = 60 * ONE_SECOND;
	static constexpr int32_t ONE_HOUR = 60 * ONE_MINUTE;
	static constexpr int32_t ONE_DAY = 24 * ONE_HOUR;
	static constexpr int32_t ONE_WEEK = 7 * ONE_DAY;

	static constexpr int32_t UNTIL_FOREVER = std::numeric_limits<int32_t>::max();
	static constexpr int32_t SINCE_FOREVER = std::numeric_limits<int32_t>::min();

private:
	
	static constexpr int32_t UNSPECIFIED_DATE = std::numeric_limits<int32_t>::max();

	static constexpr int32_t UNSPECIFIED = std::numeric_limits<int32_t>::max();
	static constexpr int32_t ANY = std::numeric_limits<int32_t>::max() - 1;

	static constexpr size_t RECURSION_LIMIT() noexcept { return 20; }

	IntervalSet make_interval_set(int32_t starting_time, int32_t duration)
	{
		IntervalSet result;
		result.add(IntervalSet::Interval(starting_time, starting_time + duration));
		return result;
	}

	bool finished() noexcept
	{
		return _it >= _end;
	}

	void check_not_finished()
	{
		if ( finished() )
			throw std::runtime_error("HereTimeParser: Reached end of string.");
	}

	template <typename T, typename S>
	void check_is_one_of(const T& item, const std::initializer_list<S>& list)
	{
		if ( ! util::is_one_of(item, list) )
			throw std::runtime_error("HereTimeParser: Expected one of "
				+ util::to_text(list) + ", got '" + util::to_text(item) + "'.");
	}

	template <typename T>
	void check_is_in_range(const T& item, const T& lower, const T& upper)
	{
		if ( ! util::is_in_range(item, lower, upper) )
			throw std::runtime_error("HereTimeParser: Expected one from ["
				+ util::to_text(lower) + ".." + util::to_text(upper) + "], got '" + util::to_text(item) + "'.");
	}

	void jump_over_all_spaces() noexcept
	{
		while ( ! finished() && util::is_one_of(*_it, {' ','\n','\r','\t'}) )
			++_it;
	}

	char lookahead()
	{
		check_not_finished();
		return *_it;
	}

	void skip()
	{
		check_not_finished();
		++_it;
		jump_over_all_spaces();
	}

	int32_t parse_number()
	{
		int32_t value = 0;
		int32_t sign = 1;

		if ( lookahead() == '-' )
		{
			sign = -1;
			skip();
		}

		if ( ! util::is_in_range(lookahead(), '0', '9') )
			throw std::runtime_error("HereTimeParser: Expected a decimal digit.");

		while ( ! finished() && util::is_in_range(*_it, '0', '9') )
		{
			value = 10 * value + (*_it - '0');
			++_it;
		}

		jump_over_all_spaces();
		return sign * value;
	}

	// skip one character but throws exception if skipped character is other than specified
	void consume(const char c)
	{
		check_is_one_of(lookahead(), {c});

		++_it;
		jump_over_all_spaces();
	}

	std::vector<int32_t> parse_starting_date()
	{
		std::vector<int32_t> result;

		bool date_not_supported = false;
		std::vector<std::pair<char,int32_t>> specifiers;
		specifiers.reserve(10);

		while ( ! finished() && util::is_one_of(*_it, {'y','M','w','d','f','l','t','h','m','s'}) )
		{
			const char type = *_it;
			skip();

			int32_t number = parse_number();
			specifiers.push_back(std::make_pair(type, number));

			switch ( type )
			{
				case 'y': // year, WARNING: not supported yet
				case 'M': // month, WARNING: not supported yet
				case 'w': // specific week of a year, WARNING: not supported yet
				case 'd': // specific day of a month, WARNING: not supported yet
				case 'f': // specific first week day of a month, WARNING: not supported yet
				case 'l': // specific last week day of a month, WARNING: not supported yet
					date_not_supported = true;
					break;

				case 't': // day of week, Sun, Mon, ..., Sat
					check_is_in_range(number, 1, 7);
					break;

				case 'h':
					check_is_in_range(number, 0, 23);
					break;

				case 'm':
					check_is_in_range(number, 0, 59);
					break;

				case 's':
					check_is_in_range(number, 0, 59);
					break;

				default:
					throw std::logic_error("HereTimeParser: This code should never be reached.");
			}
		}

		if ( date_not_supported || specifiers.empty() )
		{
			result = std::vector<int32_t>(1, UNSPECIFIED_DATE);
			return result;
		}

		// check whether sorted with respect to 't' < 'h' < 'm' < 's'
		if (
			! std::is_sorted(specifiers.begin(), specifiers.end(),
				[](std::pair<char,int32_t>& lhs, std::pair<char,int32_t>& rhs) {
					if ( rhs.first == 't' ) return false;
					if ( lhs.first == 't' ) return true;
					else return lhs.first < rhs.first;
				}
			)
		)
			throw std::runtime_error("HereTimeParser: Type specifiers in starting date "
				"failed to be in order 't' < 'h' < 'm' < 's'.");
		
		if ( ! util::sorted_sequence_is_unique(specifiers.begin(), specifiers.end()) )
			throw std::runtime_error("HereTimeParser: repeated occurrence of time type in starting date.");

		int32_t second = ANY;
		int32_t minute = ANY;
		int32_t hour = ANY;
		int32_t day = ANY;

		if ( specifiers.back().first == 's' )
		{
			second = specifiers.back().second;
		}
		else if ( specifiers.back().first == 'm' )
		{
			second = 0;
			minute =  specifiers.back().second;
		}
		else if ( specifiers.back().first == 'h' )
		{
			second = 0;
			minute = 0;
			hour = specifiers.back().second;
		}
		else if ( specifiers.back().first == 't' )
		{
			second = 0;
			minute = 0;
			hour = 0;
			day = specifiers.back().second;
		}

		specifiers.pop_back();

		auto set_specified_value = [&specifiers] (char type, int32_t& value)
		{
			if ( specifiers.empty() ) return;
			if ( specifiers.back().first != type ) return;

			value = specifiers.back().second;
			specifiers.pop_back();
		};

		set_specified_value('m', minute);
		set_specified_value('h', hour);
		set_specified_value('t', day);
	
		auto range_begin = [](int32_t value, int32_t first, int32_t last)
		{
			if ( value == ANY ) return first;
			else return value;
		};

		auto range_end = [](int32_t value, int32_t first, int32_t last)
		{
			if ( value == ANY ) return last + 1;
			else return value + 1;
		};

		for ( int32_t d = range_begin(day, 1, 7) ; d != range_end(day, 1, 7) ; ++d )
			for ( int32_t h = range_begin(hour, 0, 23) ; h != range_end(hour, 0, 23) ; ++h )
				for ( int32_t m = range_begin(minute, 0, 59) ; m != range_end(minute, 0, 59) ; ++m )
					for ( int32_t s = range_begin(second, 0, 59) ; s != range_end(second, 0, 59) ; ++s )
						result.push_back((d-1)*ONE_DAY + h*ONE_HOUR + m*ONE_MINUTE + s*ONE_SECOND);

		return result;
	}

	int32_t parse_time_duration()
	{
		check_not_finished();
		int32_t result = 0;

		while ( ! finished() && util::is_one_of(*_it, {'y', 'M', 'w', 'd', 'h', 'm', 's'}) )
		{
			const char type = *_it;
			skip();
			int32_t number = parse_number();

			if ( number == 0 )
				throw std::runtime_error("HereTimeParser: Expected a number other than 0.");

			if ( result == UNTIL_FOREVER || result == SINCE_FOREVER )
				continue;

			switch (type)
			{
				case 'y': // year, WARNING: not fully supported yet
				case 'M': // month, WARNING: not fully supported yet
				case 'w': // week, WARNING: not fully supported yet
					if ( number > 0 )
						result = UNTIL_FOREVER;
					else if ( number < 0 )
						result = SINCE_FOREVER;
					break;

				case 'd':
					result += number * ONE_DAY;
					break;

				case 'h':
					result += number * ONE_HOUR;
					break;

				case 'm':
					result += number * ONE_MINUTE;
					break;

				case 's':
					result += number * ONE_SECOND;
					break;

				default:
					throw std::logic_error("HereTimeParser: This code should never be reached.");
			}
		}

		return result;
	}

	char parse_set_operator()
	{
		const char op = lookahead();
		check_is_one_of(op, {'*', '+', '-'});
		skip();
		return op;
	}
	

	IntervalSet parse_expression_tail(IntervalSet&& lhs)
	{
		IntervalSet result;

		if ( lookahead() == ']' )
		{
			result = IntervalSet();
			return result;
		}
		
		const char op = parse_set_operator();
		IntervalSet rhs = parse_expression();

		if ( lhs.empty() )
		{
			if ( op == '-' )
				result = IntervalSet();
			else
				result = std::move(rhs);
		}
		else
		{
			if ( rhs.empty() )
				result = std::move(lhs);
			else if ( op == '*' )
				result = intersect(std::move(lhs), rhs);
			else if ( op == '+' )
				result = unite(std::move(lhs), rhs);
			else if ( op == '-' )
				result = subtract(std::move(lhs), rhs);
			else
				throw std::logic_error("HereTimeParser: This code should never be reached.");
		}

		return result;
	}

	IntervalSet create_interval_set(const std::vector<int32_t>& starting_times, const int32_t duration)
	{
		IntervalSet result;

		if ( duration == 0 || duration == UNTIL_FOREVER || duration == SINCE_FOREVER )
		{
			result = IntervalSet();
		}
		else
		{
			for ( const auto& starting_time : starting_times )
			{
				if ( starting_time == UNSPECIFIED_DATE )
					continue;

				IntervalSet::Interval new_interval;

				if ( duration > 0 )
					new_interval = IntervalSet::Interval(starting_time, starting_time + duration);
				else if ( duration < 0 )
					new_interval = IntervalSet::Interval(std::max(0, starting_time - duration), starting_time);
				else
					throw std::logic_error("HereTimeParser: This code should never be reached.");

				if ( new_interval.get_lower() == _period )
				{
					new_interval.set_lower(new_interval.get_lower() - _period);
					new_interval.set_upper(new_interval.get_upper() - _period);
				}

				if ( new_interval.get_lower() >= new_interval.get_upper() )
					throw std::logic_error("HereTimeParser: Ill-formed interval created.");

				if ( new_interval.get_lower() < 0 )
					throw std::logic_error("HereTimeParser: Left interval limit less than 0.");

				if ( new_interval.get_lower() >= _period )
					throw std::logic_error("HereTimeParser: Interval outside period interval.");

				if ( new_interval.get_upper() > _period )
				{
					result.add(IntervalSet::Interval(new_interval.get_lower(), _period));
					result.add(IntervalSet::Interval(0, std::min(_period, new_interval.get_upper() - _period)));
				}
				else
				{
					result.add(new_interval);
				}
			}
		}
	
		return result;
	}

	IntervalSet parse_expression()
	{
		IntervalSet result;

		if ( ++_recursion_depth > RECURSION_LIMIT() )
			throw std::runtime_error("HereTimeParser: Recursion depth exceeded limit.");

		consume('[');

		if ( lookahead() != ']' )
		{
			if ( lookahead() == '(' )
			{
				consume('(');

				const std::vector<int32_t> starting_times =
					lookahead() == ')' ? std::vector<int32_t>(1, UNSPECIFIED_DATE) : parse_starting_date();

				consume(')');
				consume('{');
				const int32_t duration = lookahead() == '}' ? 0 : parse_time_duration();
				consume('}');

				result = create_interval_set(starting_times, duration);
			}
			else
			{
				IntervalSet result_head = parse_expression();
				result = parse_expression_tail(std::move(result_head));
			}
		}

		consume(']');
		return result;

	}

	std::string::const_iterator _begin;
	std::string::const_iterator _end;
	std::string::const_iterator _it;
	int32_t _period;
	size_t _recursion_depth;

public:

	HereTimeParser(const std::string& expression, int32_t period)
	: _begin(expression.begin()), _end(expression.end()), _it(_begin), _period(period), _recursion_depth(0)
	{}

	IntervalSet parse()
	{
		if ( _begin == _end )
			return IntervalSet();

		_it = _begin;
		_recursion_depth = 0;

		jump_over_all_spaces();
		IntervalSet result = parse_expression();
		jump_over_all_spaces();

		if ( ! finished() )
			throw std::runtime_error("HereTimeParser: Expected end of input.");

		return result;
	}
};

IntervalSet  parse_here_time_expr(const std::string& expression)
{
	return HereTimeParser(expression, HereTimeParser::ONE_WEEK).parse();
}

}
}

#endif // RITCH_IO_HERE_TIME_PARSER_H_
