
#ifndef RITCH_IO_LAYERS_FORMAT_H_
#define RITCH_IO_LAYERS_FORMAT_H_

#include <algorithm>
#include <fstream>
#include <string>
#include <limits>
#include <vector>

#include "ritch/io/binary.h"
#include "ritch/util/misc.h"

namespace ritch
{
namespace io
{
namespace layers_format
{

const uint32_t FILE_SEPARATOR = 0x04030201;
const uint32_t FILE_TERMINATOR = 0x08070605;
const uint32_t UINT32_MAX_VALUE = std::numeric_limits<uint32_t>::max();

class OutputFile
{

private:

    std::ofstream _file_output_stream;
    std::string _file_name;
    uint32_t _n_nodes;
    uint32_t _current_n_layers;

public:

    OutputFile(const std::string& file_name, const uint32_t& n_nodes)
    : _file_output_stream(file_name, std::fstream::binary),
      _file_name(file_name),
      _n_nodes(n_nodes),
	  _current_n_layers(0)
    {
        if ( _file_name.empty() )
            RITCH_ERROR("Empty output file name given.\n");

        if ( _file_output_stream.good() )
        {
            // write format tag
            _file_output_stream.write("LAYERS\r\n", 8);

            // write version number
            write_uint32(1, _file_output_stream);

            // write number of nodes
            write_uint32(n_nodes, _file_output_stream);

            // write a place holder for the number of layers
            write_uint32(UINT32_MAX_VALUE, _file_output_stream);
			
            // write a place holder for the level info
            for ( uint32_t node_id = 0; node_id < n_nodes; ++node_id )
                write_uint32(UINT32_MAX_VALUE, _file_output_stream);
        }
        else
        {
            RITCH_ERROR("MMTCH output file '" << _file_name << "' not good.\n");
        }
    }

    ~OutputFile()
    {
        if ( _file_output_stream.is_open() )
        {
            RITCH_WARNING("Automatically closing non-closed LAYERS output file on destruction.\n");
            close();
        }
    }

	void write_level_info(const std::vector<uint32_t>& node_id_to_level)
	{
		if ( ! _file_output_stream.good() )
		{
			RITCH_ERROR("RITCH_SHORTCUTS output file '" << _file_name << "' not good.\n");
			return;
		}

		if ( _n_nodes != node_id_to_level.size() )
		{
			RITCH_ERROR("Node level information has wrong size.\n");
			return;
		}

		assert(util::set_is_unique(node_id_to_level.begin(), node_id_to_level.end()));

		auto old_pos = _file_output_stream.tellp();
		_file_output_stream.seekp(8 + 3*sizeof(uint32_t));

		for ( uint32_t node_id = 0 ; node_id < _n_nodes ; ++node_id )
		{
			assert(node_id < node_id_to_level.size());
			const uint32_t level = node_id_to_level[node_id];

			if ( level >= _n_nodes )
			{
				RITCH_ERROR("Level exceeds number of nodes (" << level << " >= " << _n_nodes << ").\n");
				return;
			}

			write_uint32(level, _file_output_stream);
		}

		_file_output_stream.seekp(old_pos);
	}

	void write_level(uint32_t node_id, uint32_t level)
    {
        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("LAYERS output file '" << _file_name << "' not good.\n");
            return;
        }

		if ( level >= _n_nodes )
		{
			RITCH_ERROR("Level exceeds number of nodes (" << level << " >= " << _n_nodes << ").\n");
			return;
		}

		auto old_pos = _file_output_stream.tellp();
		_file_output_stream.seekp(8 + (node_id + 3) * sizeof(uint32_t));
		write_uint32(level, _file_output_stream);
        _file_output_stream.seekp(old_pos);
    }

    template <typename Iterator>
    void write_layer(const Iterator& nodes_in_layer_begin, const Iterator& nodes_in_layer_end)
    {
        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("LAYERS output file '" << _file_name << "' not good.\n");
            return;
        }

		// write a separator at the beginning of each layers data
		write_uint32(FILE_SEPARATOR, _file_output_stream);

		// write number of nodes in layer
		write_uint32(uint32_t(std::distance(nodes_in_layer_begin, nodes_in_layer_end)), _file_output_stream);

		// write id of all nodes in layer
		for ( auto it = nodes_in_layer_begin ; it != nodes_in_layer_end ; ++it )
			write_uint32(*it, _file_output_stream);

        ++_current_n_layers;
    }

    void close()
    {
        if ( ! _file_output_stream.is_open() )
        {
            RITCH_ERROR("Trying to close a non-open LAYERS file '" << _file_name << "'.");
            return;
        }

        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("LAYERS output file '" << _file_name << "' not good.\n");
            return;
        }

        // append file terminator
        write_uint32(FILE_TERMINATOR, _file_output_stream);

        // store number of layers
        _file_output_stream.seekp( 8 + 2 * sizeof(uint32_t) );
        write_uint32(_current_n_layers, _file_output_stream);

        _file_output_stream.close();
    }


};


std::vector<uint32_t> read_levels(const std::string& input_file_name)
{
	std::vector<uint32_t> result;

	if (input_file_name == "")
	{
		RITCH_ERROR("Empty input file name given.\n");
		result.clear();
		return result;
	}

	std::cout << "STATUS: Reading level information from LAYERS file '" << input_file_name << "'..." << std::flush;

	std::ifstream input_layers_file(input_file_name, std::ios_base::binary);

	if ( ! input_layers_file.is_open() )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Unable to open file.\n");
		result.clear();
		return result;
	}

	if ( ! input_layers_file.good() )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Something went wrong.\n");
		result.clear();
		return result;
	}

	char* buffer = new char[8];
	input_layers_file.read(buffer, 8);
	if ( std::string(buffer, 8) != "LAYERS\r\n" )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Not a valid LAYERS file.\n");
		result.clear();
		return result;
	}

	const uint32_t version = read_uint32(input_layers_file);

	if ( version != 1 )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("file has wrong version (expected 1).\n");
		result.clear();
		return result;
	}

	const uint32_t n_nodes = read_uint32(input_layers_file);
	const uint32_t n_layers = read_uint32(input_layers_file);

	result.reserve(n_nodes);

	for ( size_t i = 0 ; i != n_nodes ; ++i )
		result.push_back(read_uint32(input_layers_file));

	if ( ! util::set_is_unique(result.begin(), result.end()) )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Not a valid LAYERS file.\n");
		result.clear();
		return result;
	}

	if ( n_layers == 0 ) return result;

	const uint32_t separator = read_uint32(input_layers_file);

	if ( separator != FILE_SEPARATOR )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Not a valid LAYERS file.\n");
		result.clear();
		return result;
	}

	RITCH_CONTINUE_STATUS(" OK\n");
	return result;
}


std::vector<std::vector<uint32_t>> read_layers(const std::string& input_file_name)
{
	std::vector<std::vector<uint32_t>> result;

	if (input_file_name == "")
	{
		RITCH_ERROR("Empty input file name given.\n");
		result.clear();
		return result;
	}

	std::cout << "STATUS: Reading layer information from LAYERS file '" << input_file_name << "'..." << std::flush;

	std::ifstream input_layers_file(input_file_name, std::ios_base::binary);

	if ( ! input_layers_file.is_open() )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Unable to open file.\n");
		result.clear();
		return result;
	}

	if ( ! input_layers_file.good() )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Something went wrong.\n");
		result.clear();
		return result;
	}

	char* buffer = new char[8];
	input_layers_file.read(buffer, 8);
	if ( std::string(buffer, 8) != "LAYERS\r\n" )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Not a valid LAYERS file.\n");
		result.clear();
		return result;
	}

	const uint32_t version = read_uint32(input_layers_file);

	if ( version != 1 )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("file has wrong version (expected 1).\n");
		result.clear();
		return result;
	}

	const uint32_t n_nodes = read_uint32(input_layers_file);
	const uint32_t n_layers = read_uint32(input_layers_file);

	{
		std::vector<uint32_t> levels;
		levels.reserve(n_nodes);

		for ( size_t i = 0 ; i != n_nodes ; ++i )
			levels.push_back(read_uint32(input_layers_file));

		if ( ! util::set_is_unique(levels.begin(), levels.end()) )
		{
			RITCH_CONTINUE_STATUS(" ABORT\n");
			RITCH_ERROR("Not a valid LAYERS file.\n");
			result.clear();
			return result;
		}
	}

	if ( n_layers == 0 )
	{
		result.clear();
		result.clear();
		return result;
	}

	result.reserve(n_layers);

	for ( size_t i = 0 ; i != n_layers ; ++i )
	{
		const uint32_t separator = read_uint32(input_layers_file);

		if ( separator != FILE_SEPARATOR )
		{
			RITCH_CONTINUE_STATUS(" ABORT\n");
			RITCH_ERROR("Not a valid LAYERS file.\n");
			result.clear();
			return result;
		}

		const uint32_t n_nodes_in_layer = read_uint32(input_layers_file);

		if ( n_nodes_in_layer == 0 )
		{
			RITCH_CONTINUE_STATUS(" ABORT\n");
			RITCH_ERROR("Not a valid LAYERS file.\n");
			result.clear();
			return result;
		}

		result.emplace_back();
		result.back().reserve(n_layers);

		for ( size_t j = 0 ; j != n_nodes_in_layer ; ++j )
		{
			const uint32_t node_id = read_uint32(input_layers_file);
			result.back().push_back(node_id);
		}
	}

	const uint32_t terminator = read_uint32(input_layers_file);

	if ( terminator != FILE_TERMINATOR )
	{
		RITCH_CONTINUE_STATUS(" ABORT\n");
		RITCH_ERROR("Not a valid LAYERS file.\n");
		result.clear();
		return result;
	}

	RITCH_CONTINUE_STATUS(" OK\n");
	return result;
}

}
}
}

#endif // RITCH_IO_LAYERS_FORMAT_H_
