#ifndef RITCH_IO_ORTEC_FORMAT_H_
#define RITCH_IO_ORTEC_FORMAT_H_

#include <array>
#include <cstdint>
#include <cassert>
#include <iostream>
#include <string>
#include <unordered_map>
#include <vector>

#include "ritch/io/binary.h"
#include "ritch/util/misc.h"

namespace ritch
{
namespace io
{
namespace ortec_format
{

static inline uint32_t road_type_2_default_speed(const uint8_t road_type)
{
	const std::array<uint32_t, 19> ROAD_TYPE_2_SPEED = { 78, 75, 69, 65, 56, 54, 47, 45, 36, 33, 24, 20, 15, 11, 9, 7, 4, 1, 2 };

	assert(road_type < ROAD_TYPE_2_SPEED.size());
	return ROAD_TYPE_2_SPEED[road_type];
}

template <typename T>
static bool road_type_is_motorway_or_alike(T road_type)
{
	return
		   road_type == T(0)
		|| road_type == T(1)
		|| road_type == T(2);
}

static size_t read_n_edges(const std::string& graph_directory_name)
{
	auto naar_net_it = get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "naar.net");
	size_t n_edges = naar_net_it.get_size_in_byte() / 4;
	naar_net_it.close();

	return n_edges;
}

static size_t read_n_nodes(const std::string& graph_directory_name)
{
	auto first_net_it = get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "first.net");
	size_t n_nodes = first_net_it.get_size_in_byte() / 4 - 1;
	first_net_it.close();

	return n_nodes;
}

static inline InputFileIterator<uint32_t> get_target_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "naar.net");
}

static inline InputFileIterator<uint32_t> get_source_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "fromnodes.net");
}

static inline InputFileIterator<uint32_t> get_distance_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "lengtelong.net");
}

static inline InputFileIterator<uint8_t> get_road_type_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint8_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "type.net");
}

static inline InputFileIterator<uint32_t> get_first_edge_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "first.net");
}

static inline InputFileIterator<uint32_t> get_reverse_first_edge_iterator(const std::string& graph_directory_name)
{
	return get_input_file_iterator<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "fromnodeindex.net");
}

template <typename T>
static inline std::vector<T> read_graph_data(const std::string& file_path_name)
{
	std::cerr << "STATUS: Reading '" << file_path_name << "'... " << std::flush;
	std::ifstream in(file_path_name, std::ios_base::binary);

	std::vector<T> result;

	if ( ! in.good() )
	{
		std::cerr << "ABORT" << std::endl << "ERROR: Maybe file does not exist." << std::endl;
		return result;
	}

	read_all(result, in);
	std::cerr << "OK" << std::endl;

	return result;
}

template <typename T>
void write_graph_data(const std::string& file_path_name, std::vector<T>& data)
{
	std::cerr << "STATUS: Writing '" << file_path_name << "... " << std::flush;
	std::ofstream out_file(file_path_name, std::ios_base::binary);

	if ( ! out_file.good() )
	{
		std::cerr << "ABORT" << std::endl << "ERROR: Failed to open/create that file." << std::endl;
		return;
	}

	for ( const auto& chunk : data )
		write_chunk(chunk, out_file);

	out_file.close();
	std::cerr << "OK" << std::endl;
}

static inline std::vector<uint32_t> read_map_node_rec_no_2_first_edge_rec_no(const std::string& graph_directory_name)
{
	return read_graph_data<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "first.net");
}

static inline std::vector<uint32_t> read_map_edge_rec_no_2_target(const std::string& graph_directory_name)
{
	return read_graph_data<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "naar.net");
}

static inline std::vector<uint32_t> read_map_edge_rec_no_2_distance(const std::string& graph_directory_name)
{
	return read_graph_data<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "lengtelong.net");
}

static inline std::vector<uint8_t> read_map_edge_rec_no_2_road_type(const std::string& graph_directory_name)
{
	return read_graph_data<uint8_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "type.net");
}

static inline std::vector<uint16_t> read_map_edge_rec_no_2_country_phone_code(const std::string& graph_directory_name)
{
	return read_graph_data<uint16_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "lnkcou.net");
}

std::vector<uint32_t> read_map_node_rec_no_2_id(const std::string& graph_directory_name)
{
	return read_graph_data<uint32_t>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "org_nr.net");
}

std::vector<std::array<uint32_t, 7>> read_map_edge_rec_no_2_link_ref(const std::string& graph_directory_name)
{
	return read_graph_data<std::array<uint32_t, 7>>(
		graph_directory_name + util::PATH_SEPARATOR() + "Network"+ util::PATH_SEPARATOR() + "NTP_LinkReferenced.net");
}

std::vector<std::array<uint8_t, 96>> read_speed_patterns(const std::string& graph_directory_name)
{
	return read_graph_data<std::array<uint8_t, 96>>(
		graph_directory_name + util::PATH_SEPARATOR() + "Network"+ util::PATH_SEPARATOR() + "NTP_SpeedPatterns.bin");
}

std::unordered_map<uint32_t, std::pair<int32_t, int32_t>> read_map_node_id_2_lat_long(const std::string& graph_directory_name)
{
	std::vector<std::array<uint8_t, 18>> data =
		read_graph_data<std::array<uint8_t, 18>>(graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "arcs.net");

	std::unordered_map<uint32_t, std::pair<int32_t, int32_t>> result;

	for ( const auto& record : data )
	{
		uint32_t node_id = *(reinterpret_cast<const int32_t*>(&record[4]));
		int32_t longitude = *(reinterpret_cast<const int32_t*>(&record[9]));
		int32_t latitude = *(reinterpret_cast<const int32_t*>(&record[13]));

		result[node_id] = std::make_pair(latitude, longitude);
	}

	return result;
}

std::vector<std::pair<uint32_t, uint32_t>> read_topology(const std::string& graph_directory_name)
{
	//
	// Read target nodes
	//

	io::InputFileIterator<uint32_t> target_it = io::ortec_format::get_target_iterator(graph_directory_name);
	std::cerr << "STATUS: Reading '" << (graph_directory_name + util::PATH_SEPARATOR() + "Network" + util::PATH_SEPARATOR() + "naar.net'... ") << std::flush;

	std::vector<std::pair<uint32_t, uint32_t>> result;
	result.reserve(target_it.get_size_in_byte() / 4);

	while ( ! target_it.finished() )
	{
		result.emplace_back();
		result.back().second = target_it.read();
	}

	target_it.close();
	std::cerr << "OK" << std::endl;

	//
	// Read source nodes
	//

	std::vector<uint32_t> first = read_map_node_rec_no_2_first_edge_rec_no(graph_directory_name);

	if ( first.empty() )
	{
		result.clear();
		return result;
	}

	for ( size_t i = 0, j = 0 ; i != result.size() ; ++i )
	{
		assert(j+1 < first.size());
		while ( i >= first[j+1] )
		{
			++j;
			assert(j+1 < first.size());
		}

		if ( first[j] >= result.size() )
		{
			std::cerr << "ABORT" << std::endl << "ERROR: Source node index too large (see " << j << " >= " << result.size() << ")." << std::endl;
			return result;
		}

		result[i].first = uint32_t(j);
	}

	return result;
}


}
}
}

#endif // RITCH_IO_ORTEC_FORMAT_H_
