#ifndef RITCH_IO_DEMANDS_FORMAT_H_
#define RITCH_IO_DEMANDS_FORMAT_H_

#include <cstdint>
#include <limits>
#include <string>
#include <vector>
#include <fstream>

#include "ritch/io/binary.h"

namespace ritch
{
namespace io
{
namespace ritch_demands_format
{

constexpr uint32_t FILE_TERMINATOR = 0x11220033;

struct Demand
{
	uint32_t start_node;
	uint32_t dest_node;
	int32_t departure_time;
	int32_t arrival_time;
	uint32_t travel_distance;
	uint32_t dijkstra_rank;

	Demand()
	: start_node(std::numeric_limits<decltype(start_node)>::max()),
	  dest_node(std::numeric_limits<decltype(dest_node)>::max()),
	  departure_time(std::numeric_limits<decltype(departure_time)>::max()),
	  arrival_time(std::numeric_limits<decltype(arrival_time)>::max()),
	  travel_distance(std::numeric_limits<decltype(travel_distance)>::max()),
	  dijkstra_rank(std::numeric_limits<decltype(dijkstra_rank)>::max())
	{}
};

void write(const std::string& file_path_name, const std::vector<Demand>& demands)
{

	std::ofstream out(file_path_name, std::fstream::binary);

	// write format tag
	out.write("ritch_demands\r\n", 15);

	// write version number
	write_uint32(1, out);

	// write number of demands
	write_uint32(uint32_t(demands.size()), out);
	
	// write demand data
	for ( const auto& demand : demands )
		write_chunk(demand, out);

	// write terminator
	write_uint32(FILE_TERMINATOR, out);
}

std::vector<Demand> read(const std::string& file_path_name)
{
	std::vector<Demand> demands;
	std::ifstream in(file_path_name, std::fstream::binary);

	if ( file_path_name == "" )
	{
		RITCH_ERROR("Empty input file name given.\n");
		return demands;
	}

	if ( ! in.is_open() )
	{
		RITCH_ERROR("Unable to open file '" << file_path_name << "'.\n");
		return demands;
	}

	if ( ! in.good() )
	{
		RITCH_ERROR("Something went wrong with file '" << file_path_name << "'.\n");
		return demands;
	}

	char* buffer = new char[20];
	in.read(buffer, 15);
	if ( std::string(buffer, 15) != "ritch_demands\r\n" )
	{
		RITCH_ERROR("File '" << file_path_name << "' is not a valid RITCH_DEMANDS file.\n");
		return demands;
	}

	const uint32_t version = read_uint32(in);

	if ( version != 1 )
	{
		RITCH_ERROR("File '" << file_path_name << "' has wrong version (expected 1).\n");
		return demands;
	}

	const uint32_t n_demands = read_uint32(in);

	demands.reserve(n_demands);
	for ( size_t i = 0 ; i != n_demands ; ++i )
	{
		demands.emplace_back();
		read_chunk(demands.back(), in);
	}

	uint32_t terminator = read_uint32(in);
	if ( terminator != FILE_TERMINATOR )
	{
		RITCH_ERROR("File '" << file_path_name << "' is not a valid RITCH_SHORTCUTS file.\n");
		demands.clear();
		return demands;
	}

	return demands;
}


}
}
}

#endif // RITCH_IO_DEMANDS_FORMAT_H_
