#ifndef RITCH_IO_RITCH_SHORTCUTS_FORMAT_H_
#define RITCH_IO_RITCH_SHORTCUTS_FORMAT_H_

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstdint>
#include <fstream>
#include <string>
#include <limits>
#include <memory>
#include <vector>

#include "ritch/io/binary.h"
#include "ritch/prepro/dynamic_graph.h"
#include "ritch/util/misc.h"

namespace ritch
{
namespace io
{
namespace ritch_shortcuts_format
{

constexpr uint32_t FILE_SEPARATOR() { return 0x12341234; }
constexpr uint32_t FILE_TERMINATOR() { return 0x43214321; }
constexpr uint32_t UINT32_MAX_VALUE() { return std::numeric_limits<uint32_t>::max(); }

class OutputFile
{

private:

	typedef DynamicGraph::TimeDist TimeDist;

    std::ofstream _file_output_stream;
    std::string _file_name;
    uint32_t _n_nodes;
    uint32_t _current_n_edges;
	uint32_t _period;

public:

    OutputFile(const std::string& file_name, const uint32_t& n_nodes, const uint32_t& period)
    : _file_output_stream( file_name, std::fstream::binary ),
      _file_name(file_name),
      _n_nodes(n_nodes),
      _current_n_edges(0),
	  _period(period)
    {
        if ( _file_name.empty() )
            RITCH_ERROR("Empty output file name given.\n");

        if ( _file_output_stream.good() )
        {
            // write format tag
            _file_output_stream.write("RITCH_SHORTCUTS\r\n", 17);

            // write version number
            write_uint32(1, _file_output_stream);

            // write number of nodes
            write_uint32(n_nodes, _file_output_stream);

            // write a place holder for the number of edges
            write_uint32(UINT32_MAX_VALUE(), _file_output_stream);
			
			// write period
			write_uint32(_period, _file_output_stream);

			// write a place holder for the level info
            for ( uint32_t node_id = 0; node_id < n_nodes; ++node_id )
                write_uint32(UINT32_MAX_VALUE(), _file_output_stream);

            // write separator between level info and edge info
            write_uint32(FILE_SEPARATOR(), _file_output_stream);
        }
        else
        {
            RITCH_ERROR("RITCH_SHORTCUTS output file '" << _file_name << "' not good.\n");
        }
    }

    ~OutputFile()
    {
        if ( _file_output_stream.is_open() )
        {
            RITCH_WARNING("Automatically closing non-closed RITCH_SHORTCUTS output file on destruction.\n");
            close();
        }
    }

	const std::string& get_file_name() const
	{
		return _file_name;
	}

    void write_level_info(const std::vector<uint32_t>& node_id_to_level)
    {
        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("RITCH_SHORTCUTS output file '" << _file_name << "' not good.\n");
            return;
        }

        if ( _n_nodes != node_id_to_level.size() )
        {
            RITCH_ERROR("Node level information has wrong size.\n");
            return;
        }

        assert(util::set_is_unique(node_id_to_level));

        auto old_pos = _file_output_stream.tellp();
        _file_output_stream.seekp(17 + 4 * sizeof(uint32_t));

        for ( uint32_t node_id = 0 ; node_id < _n_nodes ; ++node_id )
        {
            assert(node_id < node_id_to_level.size());
            const uint32_t level = node_id_to_level[node_id];

            if ( level >= _n_nodes )
            {
                RITCH_ERROR("Level exceeds number of nodes (" << level << " >= " << _n_nodes << ").\n");
                return;
            }

            write_uint32(level, _file_output_stream);
        }

		_file_output_stream.seekp( old_pos );
    }

	template <typename TimeDist>
	void write_edge
	(
		const uint32_t src,
		const uint32_t tgt,
		const TimeDist& time_dist,
		const std::vector<std::pair<int32_t, uint32_t>>& middle_node_info,
		const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>& edge_block
	)
    {
		if ( _period != TimeDist::PERIOD() )
		{
			RITCH_ERROR("ERROR: Inconsistent TTF periods.\n");
			return;
		}
		
        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("RITCH_SHORTCUTS output file '" << _file_name << "' not good.\n");
            return;
        }

        write_uint32(src, _file_output_stream);
        write_uint32(tgt, _file_output_stream);

		//
		// write middle node time-dependent information
		//

		write_uint32(uint32_t(middle_node_info.size()), _file_output_stream);

		for ( const auto& elem : middle_node_info )
		{
			write_int32(elem.first, _file_output_stream);
			write_uint32(elem.second, _file_output_stream);
		}

		//
		// write TTF
		//

		if ( time_dist.get_n_time_segments() == 0 )
		{
			RITCH_ERROR("Shortcut without travel time information encountered.\n");
			return;
		}

		write_uint32(uint32_t(time_dist.get_n_time_segments()), _file_output_stream);

		for ( size_t i = 0 ; i < time_dist.get_n_time_segments() ; ++i )
		{
			const auto segment = time_dist.get_time_segment(i, edge_block);
			write_uint8(uint8_t(segment.get_slope() == 0 ? 0 : 1), _file_output_stream);
			write_int32(segment.get_right_x(), _file_output_stream);
			write_int32(segment.get_right_y(), _file_output_stream);
			write_int32(segment.get_block_start(), _file_output_stream);
			write_int32(segment.get_block_end(), _file_output_stream);
			write_int32(segment.get_default_time(), _file_output_stream);

		}

		//
		// write TDF
		//

		if ( time_dist.get_n_dist_segments() == 0 )
		{
			RITCH_ERROR("Shortcut without travel distance information encountered.\n");
			return;
		}

		write_uint32(uint32_t(time_dist.get_n_dist_segments()), _file_output_stream);

		for ( size_t i = 0 ; i < time_dist.get_n_dist_segments() ; ++i )
		{
			const auto segment = time_dist.get_dist_segment(i);
			write_int32(segment.get_right_x(), _file_output_stream);
			write_int32(segment.get_right_y(), _file_output_stream);
		}

		//
		// write edge block info
		//
		write_uint32(uint32_t(edge_block.size()), _file_output_stream);
		for (size_t i = 0; i < edge_block.size(); i++) {
			const std::tuple<int32_t, int32_t, int32_t, int32_t> block = edge_block.at(i);
			write_int32(std::get<0>(block), _file_output_stream);
			write_int32(std::get<1>(block), _file_output_stream);
			write_int32(std::get<2>(block), _file_output_stream);
			write_int32(std::get<3>(block), _file_output_stream);

		}


        ++_current_n_edges;
    }

    void close()
    {
        if ( ! _file_output_stream.is_open() )
        {
            RITCH_ERROR("Trying to close a non-open RITCH_SHORTCUTS file '" << _file_name << "'.");
            return;
        }

        if ( ! _file_output_stream.good() )
        {
            RITCH_ERROR("ATCH output file '" << _file_name << "' not good.\n");
            return;
        }

        // append file terminator
        write_uint32(FILE_TERMINATOR(), _file_output_stream);

        // store number of edges
        _file_output_stream.seekp( 17 + 2 * sizeof(uint32_t) );
        write_uint32(_current_n_edges, _file_output_stream);

        _file_output_stream.close();
    }


};

class InputFile
{

private:

	typedef DynamicGraph::TimeDist TimeDist;

	std::ifstream _file_input_stream;
	std::string _file_name;
	uint32_t _n_nodes;
	uint32_t _n_edges;
	uint32_t _period;
	uint32_t _n_edges_read;

public:

	InputFile(const std::string& input_file_name)
	: _file_input_stream(input_file_name, std::ios_base::binary),
	  _file_name(input_file_name),
	  _n_nodes(0),
	  _n_edges(0),
	  _period(0),
	  _n_edges_read(0)
	{
		if ( input_file_name == "" )
		{
			RITCH_ERROR("Empty input file name given.\n");
			return;
		}

		if ( ! _file_input_stream.is_open() )
		{
			RITCH_ERROR("Unable to open file '" << _file_name << "'.\n");
			return;
		}

		if ( ! _file_input_stream.good() )
		{
			RITCH_ERROR("Something went wrong.\n");
			return;
		}

		char* buffer = new char[20];
		_file_input_stream.read(buffer, 17);
		if ( std::string(buffer, 17) != "RITCH_SHORTCUTS\r\n" )
		{
			RITCH_ERROR("Not a valid RITCH_SHORTCUTS file.\n");
			return;
		}

		const uint32_t version = read_uint32(_file_input_stream);

		if ( version != 1 )
		{
			RITCH_ERROR("File has wrong version (expected 1).\n");
			return;
		}

		_n_nodes = read_uint32(_file_input_stream);
		_n_edges = read_uint32(_file_input_stream);
		_period = read_uint32(_file_input_stream);

		_file_input_stream.seekg(17 + (4 + _n_nodes) * sizeof(uint32_t));

		const uint32_t separator = read_uint32(_file_input_stream);
		if ( separator != FILE_SEPARATOR() )
		{
			RITCH_ERROR("Not a valid RITCH_SHORTCUTS file.\n");
			return;
		}
	}

	size_t get_n_edges()
	{
		return _n_edges;
	}

	std::vector<uint32_t> read_map_node_id_2_level()
	{
		std::cout << "STATUS: Reading level information from RITCH_SHORTCUTS file '" << _file_name << "'..." << std::flush;

		std::vector<uint32_t> result;
		result.reserve(_n_nodes);

		auto old_pos = _file_input_stream.tellg();
		_file_input_stream.seekg(17 + 4 * sizeof(uint32_t));

		for ( uint32_t i = 0 ; i != _n_nodes ; ++i )
			result.push_back(read_uint32(_file_input_stream));

		const uint32_t separator = read_uint32(_file_input_stream);

		if ( separator != FILE_SEPARATOR() )
		{
			RITCH_CONTINUE_STATUS(" ABORT\n");
			RITCH_ERROR("Not a valid RITCH_SHORTCUTS file.\n");
			result.clear();
			return result;
		}

		_file_input_stream.seekg(old_pos);
		RITCH_CONTINUE_STATUS(" OK\n");

		return result;
	}

	bool has_next_edge()
	{
		if ( _n_edges_read == _n_edges )
		{
			const uint32_t terminator = read_uint32(_file_input_stream);
			if ( terminator != FILE_TERMINATOR() )
				RITCH_ERROR("Not a valid RITCH_SHORTCUTS file.\n");

			++_n_edges_read;
			return false;
		}

		if ( _n_edges_read >= _n_edges )
			return false;

		return true;
	}

	void read_next_edge(uint32_t& o_source, uint32_t& o_target, std::vector<std::pair<int32_t, uint32_t>>& o_middle_node_data, TimeDist& o_time_dist, std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>> &blockinfo)
	{
		if ( ! has_next_edge() )
			throw std::runtime_error("The RITCH_SHORTCUTS file '" + _file_name + "' has not enough edges.");

		++_n_edges_read;

		o_source = read_uint32(_file_input_stream);
		o_target = read_uint32(_file_input_stream);

		const uint32_t middle_data_size = read_uint32(_file_input_stream);
		o_middle_node_data.clear();
		o_middle_node_data.reserve(middle_data_size);

		for ( uint32_t i = 0 ; i != middle_data_size ; ++i )
		{
			const int32_t x_right = read_int32(_file_input_stream);
			const uint32_t node_id = read_uint32(_file_input_stream);
			o_middle_node_data.push_back(std::make_pair(x_right, node_id));
		}

		const uint32_t n_time_segments = read_uint32(_file_input_stream);
		std::vector<TimeDist::Segment> time_segments;
		time_segments.reserve(n_time_segments);

		for ( uint32_t i = 0 ; i != n_time_segments ; ++i )
		{
			const int32_t slope = read_uint8(_file_input_stream) == 0 ? 0 : -1;
			const int32_t x_right = read_int32(_file_input_stream);
			const int32_t y_right = read_int32(_file_input_stream);
			const int32_t blockstart = read_int32(_file_input_stream);
			const int32_t blockend = read_int32(_file_input_stream);
			const int32_t defaulttime = read_int32(_file_input_stream);
			time_segments.push_back(TimeDist::Segment(slope, x_right, y_right, blockstart, blockend, defaulttime));

		}

		const uint32_t n_dist_segments = read_uint32(_file_input_stream);
		std::vector<TimeDist::Segment> dist_segments;
		dist_segments.reserve(n_dist_segments);

		for ( uint32_t i = 0 ; i != n_dist_segments ; ++i )
		{
			const int32_t x_right = read_int32(_file_input_stream);
			const int32_t distance = read_int32(_file_input_stream);

			dist_segments.push_back(TimeDist::Segment(0, x_right, distance, 0, 0, 0));
		}

		const uint32_t n_edgeblock = read_uint32(_file_input_stream);
		blockinfo.reserve(n_edgeblock);

		for (uint32_t i = 0; i != n_edgeblock; ++i)
		{
			const int32_t x = read_int32(_file_input_stream);
			const int32_t blockstart = read_int32(_file_input_stream);
			const int32_t blockend = read_int32(_file_input_stream);
			const int32_t defaulttime = read_int32(_file_input_stream);
			if (blockend < blockstart) {
				blockinfo.push_back(std::make_tuple(x, blockstart-(24 * 7 * 60 * 60), blockend, defaulttime));
			}
			else {
				blockinfo.push_back(std::make_tuple(x, blockstart, blockend, defaulttime));
			}
		}

		o_time_dist = TimeDist(time_segments, dist_segments, blockinfo, true);
	}

	void close()
	{
		_file_input_stream.close();
	}

};

size_t read_n_edges(const std::string& input_file_name)
{
	InputFile input_file(input_file_name);
	const size_t n_shortcut_edges = input_file.get_n_edges();
	input_file.close();

	return n_shortcut_edges;
}

}
}
}

#endif // RITCH_IO_RITCH_SHORTCUTS_FORMAT_H_

