#ifndef RITCH_PREPRO_DYNAMIC_GRAPH_H_
#define RITCH_PREPRO_DYNAMIC_GRAPH_H_

#include <assert.h>
#include <cstdint>
#include <vector>
#include <limits>

#include "ritch/datastr/time_dist.h"
#include "ritch/io/hwy_blocks_format.h"
#include "ritch/io/ortec_format.h"
#include "ritch/util/misc.h"

namespace ritch
{

class DynamicGraph
{

public:
	
	typedef TimeDist_tmpl<24*7*60*60> TimeDist;

	RITCH_DEFINE_STRONG_ID(NodeId)
	RITCH_DEFINE_STRONG_ID(FwEdgeId)
	RITCH_DEFINE_STRONG_ID(BwEdgeId)

//private:

	static constexpr uint16_t INVALID_BLOCKING_PROFILE_ID() { return std::numeric_limits<uint16_t>::max(); }
	static int32_t calculate_default_travel_time(uint32_t speed, uint32_t distance) noexcept
	{
		return std::max(1, int32_t((36*distance) / speed + 5) / 10);
	}

	struct Node
	{
		FwEdgeId _out_edges_begin;
		BwEdgeId _in_edges_begin;
		uint16_t _n_out_edges;
		uint16_t _n_in_edges;

		Node()
		: _out_edges_begin(FwEdgeId::INVALID()),
		  _in_edges_begin(BwEdgeId::INVALID()),
		  _n_out_edges(0),
		  _n_in_edges(0)
		{}
	};

	static_assert(sizeof(Node) == 12, "");

	class FwEdge
	{
	public:
		static constexpr uint8_t INVALID_ROAD_TYPE() { return std::numeric_limits<uint8_t>::max(); }
		static constexpr uint16_t INVALID_BLOCKING_ID() { return std::numeric_limits<uint16_t>::max(); }

	private:

		std::array<uint32_t,4> _data;
		
		static void do_copy(const FwEdge& from, FwEdge& to)
		{
			if ( from.has_time_dist() )
			{
				to._data[0] = from._data[0];
				to._data[1] = from._data[1];

				new (reinterpret_cast<TimeDist*>(&(to._data[2]))) TimeDist(from.get_time_dist());
			}
			else
			{
				to._data = from._data;
			}
		}
		
		static void do_move(FwEdge&& from, FwEdge& to)
		{
			to._data = from._data;

			from._data = { 0xfffffffe, 1, 0, 0 };
		}

	public:
		FwEdge() { _data = { 0xfffffffe, 1, 0, 0 }; }

		FwEdge(const FwEdge& other) { do_copy(other, *this); }
		FwEdge(FwEdge&& other) { do_move(std::move(other), *this); }

		FwEdge& operator= (const FwEdge& other)
		{
			if ( this != &other )
			{
				if ( has_time_dist() )
					get_time_dist().TimeDist_tmpl::~TimeDist();

				do_copy(other, *this);
			}

			return *this;
		}

		FwEdge& operator= (FwEdge&& other)
		{
			if ( this != &other )
			{
				if ( has_time_dist() )
					get_time_dist().TimeDist_tmpl::~TimeDist();

				do_move(std::move(other), *this);
			}

			return *this;
		}

		~FwEdge()
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data = { 0xfffffffe, 1, 0, 0 };
		}

		void set_target(NodeId node_id) noexcept { _data[0] = (node_id << 1); }
		NodeId get_target() const noexcept { return NodeId(_data[0] >> 1); }

		bool has_time_dist() const noexcept { return (_data[0] & 1) == 1; }

		void turn_into_dummy() noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data = { 0xfffffffe, 1, 0, 0 };
		}

		bool is_dummy() const noexcept { return _data[0] == 0xfffffffe; }

		void set_n_original_edges(uint32_t n_original_edges) noexcept { _data[1] = n_original_edges; }
		uint32_t get_n_original_edges() const noexcept { return _data[1]; }

		void set_time_dist(TimeDist&& time_dist) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] |= 1;
			new (reinterpret_cast<TimeDist*>(&(_data[2]))) TimeDist(std::move(time_dist));
		}

		void set_time_dist(const TimeDist& time_dist) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] |= 1;

			new (reinterpret_cast<TimeDist*>(&(_data[2]))) TimeDist(time_dist);

		}

		const TimeDist & get_time_dist() const noexcept
		{
			assert(has_time_dist());
			return (*reinterpret_cast<const TimeDist*>(&(_data[2])));
		}

		TimeDist& get_time_dist() noexcept

		{
			return const_cast<TimeDist&>(reinterpret_cast<const FwEdge&>(*this).get_time_dist());
		}

		void set_blocking_id(uint16_t blocking_id)
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			reinterpret_cast<uint16_t*>(&(_data[2]))[0] = blocking_id;
		}

		uint16_t get_blocking_id() const noexcept
		{
			assert(!has_time_dist());
			return reinterpret_cast<const uint16_t*>(&(_data[2]))[0];
		}

		void set_road_type(uint8_t road_type) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			reinterpret_cast<uint16_t*>(&(_data[2]))[1] = road_type;
		}

		uint8_t get_road_type() const noexcept
		{
			assert(!has_time_dist());
			return uint8_t(reinterpret_cast<const uint16_t*>(&(_data[2]))[1]);
		}

		void set_distance(uint32_t distance) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			_data[3] = distance;
		}

		uint32_t get_distance() const noexcept
		{
			assert(!has_time_dist());
			return _data[3];
		}

		std::string to_text() const noexcept
		{
			std::stringstream ss;

			ss
				<< "{target=" << get_target()
				<< ", n_original=" << get_n_original_edges()
				<< ", has_time_dist=" << util::to_text(has_time_dist());

			if ( has_time_dist() )
				ss << ", time_dist=" << get_time_dist();
			else
				ss
				<< ", road_type=" << uint32_t(get_road_type())
				<< ", block_id=" << get_blocking_id()
				<< ", distance=" << get_distance()
				<< "}";

			return ss.str();
		}
	};

	struct BwEdge
	{
	private:
		NodeId _source;

	public:
		void set_source(NodeId node_id) noexcept { _source = node_id; }
		NodeId get_source() const noexcept { return _source; }

		bool is_dummy() const noexcept { return _source == NodeId::INVALID(); }
		void turn_into_dummy() noexcept { _source = NodeId::INVALID(); }
	};

	bool dbg_created_time_dist_is_ok(const TimeDist& time_dist, FwEdgeId e, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> & block) const
	{
		const size_t N_RANDOM_TEST_CASES = 100;

		std::vector<uint32_t> x_values;
		x_values.push_back(0);
		x_values.push_back(TimeDist::PERIOD());

		for ( size_t i = 0 ; i != time_dist.get_n_time_segments() ; ++i )
			x_values.push_back(time_dist.get_time_segment(i, block).get_right_x());

		for ( size_t i = 0 ; i != N_RANDOM_TEST_CASES ; ++i )
			x_values.push_back(util::random_int32(0, TimeDist::PERIOD()));


		for ( int32_t x : x_values )
		{
			if ( time_dist.eval_time(x) != eval_time(e, x) )
				return false;


			if ( time_dist.eval_dist(x) != eval_dist(e, x) )
				return false;
		}

		return true;
	}

	const Node& get_node(NodeId node_id) const noexcept
	{
		assert(node_id < _nodes.size());
		return _nodes[node_id.the_value()];
	}

	Node& get_node(NodeId n) noexcept
	{
		return const_cast<Node&>(reinterpret_cast<const DynamicGraph&>(*this).get_node(n));
	}
	public:
	const FwEdge& get_forward_edge(FwEdgeId e) const noexcept
	{
		assert(e < _fw_edges.size());
		return _fw_edges[e.the_value()];
	}

	FwEdge& get_forward_edge(FwEdgeId e) noexcept
	{
		return const_cast<FwEdge&>(reinterpret_cast<const DynamicGraph&>(*this).get_forward_edge(e));
	}
	private:
	const BwEdge& get_backward_edge(BwEdgeId e) const noexcept
	{
		assert(e < _bw_edges.size());
		return _bw_edges[e.the_value()];
	}

	BwEdge& get_backward_edge(BwEdgeId e) noexcept
	{
		return const_cast<BwEdge&>(reinterpret_cast<const DynamicGraph&>(*this).get_backward_edge(e));
	}
	
	bool is_dummy(FwEdgeId e) const noexcept
	{
		return get_forward_edge(e).is_dummy();
	}

	void turn_into_dummy(FwEdgeId e) noexcept
	{
		get_forward_edge(e).turn_into_dummy();
	}

	bool is_dummy(BwEdgeId e) const noexcept
	{
		return get_backward_edge(e).is_dummy();
	}

	void turn_into_dummy(BwEdgeId e) noexcept
	{
		get_backward_edge(e).turn_into_dummy();
	}

	bool has_forward_edge(NodeId source, NodeId target)
	{
		return find_forward_edge_id(source, target) != FwEdgeId::INVALID();
	}

	bool has_backward_edge(NodeId source, NodeId target)
	{
		return find_backward_edge_id(source, target) != BwEdgeId::INVALID();
	}

	void remove_backward_edge(NodeId source, NodeId target) noexcept
	{
		assert(has_backward_edge(source, target));

		BwEdgeId e_remove = find_backward_edge_id(source, target);
		auto& remove_edge = get_backward_edge(e_remove);

		BwEdgeId e_last = in_edges_end(target) - 1;
		auto& last_edge = get_backward_edge(e_last);

		if ( e_remove != e_last )
			remove_edge = std::move(last_edge);

		last_edge.turn_into_dummy();
		--get_node(target)._n_in_edges;
	}

	void remove_forward_edge(NodeId source, NodeId target) noexcept
	{
		assert(has_forward_edge(source, target));

		FwEdgeId e_remove = find_forward_edge_id(source, target);
		auto& remove_edge = get_forward_edge(e_remove);

		FwEdgeId e_last = out_edges_end(source) - 1;
		auto& last_edge = get_forward_edge(e_last);

		if ( e_remove != e_last )
			remove_edge = std::move(last_edge);

		--get_node(source)._n_out_edges;
	}

	template <typename EdgeType, typename EdgeIdType>
	auto insert_edge_impl(std::vector<EdgeType>& edges, EdgeIdType& begin, uint16_t& n_edges) -> EdgeIdType 
	{
		assert(begin == EdgeIdType::INVALID() || begin < edges.size());

		EdgeIdType edge_insert_id;

		if ( begin == EdgeIdType::INVALID() )
		{
			edge_insert_id = EdgeIdType(edges.size());
			edges.push_back(EdgeType());
			n_edges = 1;
		}
		else if ( begin + n_edges == edges.size() )
		{
			edge_insert_id = edges.size();
			edges.push_back(EdgeType());
			++n_edges;
		}
		else if ( edges[begin + n_edges].is_dummy() )
		{
			edge_insert_id = begin + n_edges;
			++n_edges;
		}
		else if ( 0u < begin && edges[begin - 1].is_dummy() )
		{
			--begin;
			edge_insert_id = begin;
			++n_edges;
		}
		else
		{
			const size_t old_n_edges = n_edges;
			EdgeIdType begin_new(edges.size());

			edges.reserve(edges.size() + old_n_edges);
			for ( EdgeIdType e = begin ; e != begin + n_edges ; ++e )
			{
				edges.emplace_back(std::move(edges[e]));
				edges[e].turn_into_dummy();
			}

			edge_insert_id = edges.size();
			edges.push_back(EdgeType());
			++n_edges;

			begin = begin_new;
		}
		return edge_insert_id;
	}

	std::vector<Node> _nodes;
	std::vector<FwEdge> _fw_edges;
	std::vector<BwEdge> _bw_edges;
	std::vector<uint32_t> _blocking_info;
	size_t _n_nodes;
	size_t _n_edges;


public:

	DynamicGraph(const std::string& graph_directory_name, const std::string& hwy_blocks_file_name)
	: _nodes(), _fw_edges(), _bw_edges(), _blocking_info(), _n_nodes(0), _n_edges(0)
	{
		io::InputFileIterator<uint32_t> target_it = io::ortec_format::get_target_iterator(graph_directory_name);
		io::InputFileIterator<uint32_t> distance_it = io::ortec_format::get_distance_iterator(graph_directory_name);
		io::InputFileIterator<uint8_t> type_it = io::ortec_format::get_road_type_iterator(graph_directory_name);

		const size_t target_size_in_byte = target_it.get_size_in_byte();

		if ( distance_it.get_size_in_byte() != target_size_in_byte )
			throw std::runtime_error("Files 'lengtelong.net' and 'naar_net' have different number of edges.");

		if ( type_it.get_size_in_byte() != target_size_in_byte / 4 )
			throw std::runtime_error("Files 'type.net' and 'naar_net' have different number of edges.");

		_n_edges = target_size_in_byte / 4;
		_fw_edges.resize(_n_edges);
		_bw_edges.resize(_n_edges);

		std::cerr << "STATUS: Graph directory is '" << graph_directory_name + util::PATH_SEPARATOR() + "'" << std::endl;

		//
		// read target nodes
		//

		std::cerr << "STATUS: Reading 'naar.net'... " << std::flush;
		for ( size_t pos = 0 ; ! target_it.finished() ; ++pos )
			_fw_edges[pos].set_target(NodeId(target_it.read()));

		target_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read road types
		//

		std::cerr << "STATUS: Reading 'type.net'... " << std::flush;
		for ( size_t pos = 0 ; ! type_it.finished() ; ++pos )
			_fw_edges[pos].set_road_type(type_it.read());

		type_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read distances
		//

		std::cerr << "STATUS: Reading 'lengtelong.net'... " << std::flush;
		for ( size_t pos = 0 ; ! distance_it.finished() ; ++pos )
			_fw_edges[pos].set_distance(distance_it.read());

		distance_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read and setup highway blockings, which includes reading country phone codes
		//

		if ( ! hwy_blocks_file_name.empty() )
		{
			std::vector<uint16_t> map_edge_rec_no_2_phone_code =
				io::ortec_format::read_map_edge_rec_no_2_country_phone_code(graph_directory_name);

			typedef io::hwy_blocks_format::BlockingDescr Blocking;
			std::vector<Blocking> blockings = io::hwy_blocks_format::read_blockings(hwy_blocks_file_name);

			for ( size_t i = blockings.size() ; i > 0 ; --i )
			{
				if ( blockings[i-1]._start_time == 0 && blockings[i-1]._end_time == 0 )
				{
					swap(blockings[i-1], blockings.back());
					blockings.pop_back();
				}
			}

			std::sort(blockings.begin(), blockings.end(),
				[](Blocking& lhs, Blocking& rhs)
				{
					if ( lhs._phone_code < rhs._phone_code )
						return true;

					if ( lhs._phone_code == rhs._phone_code )
						if ( lhs._start_time < rhs._start_time )
							return true;

					return false;
				}
			);

			size_t max_phone_code = 0;
			for ( const auto& blocking : blockings )
				max_phone_code = std::max(max_phone_code, size_t(blocking._phone_code));

			if ( max_phone_code >= size_t(std::numeric_limits<uint16_t>::max()) )
				throw std::runtime_error("Phone code too high to be encoded with 16 bits.");

			std::vector<uint16_t> map_phone_code_2_blocking_info_pos(max_phone_code + 1, 0xffff);
			uint16_t current_phone_code = std::numeric_limits<uint16_t>::max();
			uint16_t pos_n_blockings = 0;

			for ( size_t i = 0 ; i < blockings.size() ; ++i )
			{
				if ( blockings[i]._end_time < blockings[i]._start_time )
					blockings[i]._end_time += TimeDist::PERIOD();

				if ( ! RITCH_IMPLIES(blockings[i]._phone_code == current_phone_code, _blocking_info.size() > 1) )
					throw std::runtime_error("Something went wrong.");

				if ( blockings[i]._phone_code != current_phone_code )
				{
					current_phone_code = blockings[i]._phone_code;

					if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
						throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

					pos_n_blockings = uint16_t(_blocking_info.size());
					map_phone_code_2_blocking_info_pos[current_phone_code] = pos_n_blockings;
					util::append(_blocking_info, {0u, 0u}); // store size and max duration
				}
				else
				{
					uint32_t last_end_time = _blocking_info[_blocking_info.size() - 2] + _blocking_info[_blocking_info.size() - 1];

					if ( uint32_t(blockings[i]._start_time) <= last_end_time )
					{
						if ( i == 0 ) throw std::runtime_error("Something went wrong.");

						uint32_t duration = blockings[i]._end_time - blockings[i-1]._start_time;
						_blocking_info.back() = std::max(_blocking_info.back(), duration);

						continue;
					}
				}

				_blocking_info.push_back(blockings[i]._start_time);
				uint32_t duration = blockings[i]._end_time - blockings[i]._start_time;
				_blocking_info.push_back(duration);

				_blocking_info[pos_n_blockings] += 2; // increase size
				_blocking_info[pos_n_blockings + 1] = std::max(_blocking_info[pos_n_blockings + 1], duration); // update max duration
			}

			if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
				throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

			for ( size_t i = 0 ; i < _fw_edges.size() ; ++i )
			{
				_fw_edges[i].set_blocking_id(0xffff);

				uint16_t phone_code = map_edge_rec_no_2_phone_code[i];

				if ( phone_code >= map_phone_code_2_blocking_info_pos.size() )
					continue;

				_fw_edges[i].set_blocking_id(map_phone_code_2_blocking_info_pos[phone_code]);

			}
		}
		else
		{
			for ( size_t i = 0 ; i < _fw_edges.size() ; ++i )
				_fw_edges[i].set_blocking_id(0xffff);
		}

		//
		// read first edge information
		//

		io::InputFileIterator<uint32_t> first_edge_it = io::ortec_format::get_first_edge_iterator(graph_directory_name);
		const size_t first_edge_file_size = first_edge_it.get_size_in_byte();
		_n_nodes = std::max(size_t(1), first_edge_it.get_size_in_byte() / size_t(4)) - size_t(1);
		_nodes.resize(_n_nodes);

		std::cerr << "STATUS: Reading 'first.net'... " << std::flush;
		_nodes[0]._out_edges_begin = first_edge_it.read();
		for ( size_t pos = 1 ; pos != _n_nodes ; ++pos )
		{
			assert(!first_edge_it.finished());

			_nodes[pos]._out_edges_begin = first_edge_it.read();
			_nodes[pos-1]._n_out_edges = distance(_nodes[pos]._out_edges_begin, _nodes[pos-1]._out_edges_begin);
		}

		if ( _fw_edges.size() > size_t(std::numeric_limits<uint16_t>::max()) + size_t(_nodes.back()._out_edges_begin))
			throw std::runtime_error("number of outgoing edges too high to be encoded with 16 bits.");
		
		_nodes.back()._n_out_edges = uint16_t(_fw_edges.size() - size_t(_nodes.back()._out_edges_begin));

		first_edge_it.close();
		std::cerr << "OK" << std::endl;

		//
		// setup backward edges
		//

		io::InputFileIterator<uint32_t> source_it = io::ortec_format::get_source_iterator(graph_directory_name);

		if ( source_it.get_size_in_byte() != target_size_in_byte )
			throw std::runtime_error("Files 'fromnodes.net' and 'naar_net' have different number of edges.");

		_bw_edges.resize(_n_edges);

		std::cerr << "STATUS: Reading 'fromnodes.net'... " << std::flush;
		for ( size_t pos = 0 ; ! source_it.finished() ; ++pos )
			_bw_edges[pos].set_source(NodeId(source_it.read()));

		source_it.close();
		std::cerr << "OK" << std::endl;

		//
		// setup nodes information about backward edges
		//

		io::InputFileIterator<uint32_t> reverse_first_edge_it = io::ortec_format::get_reverse_first_edge_iterator(graph_directory_name);

		if ( reverse_first_edge_it.get_size_in_byte() != first_edge_file_size )
			throw std::runtime_error("Files 'fromnodeindex.net' and 'first_net' have different number of edges.");

		std::cerr << "STATUS: Reading 'fromnodeindex.net'... " << std::flush;
		_nodes[0]._in_edges_begin = reverse_first_edge_it.read();
		for ( size_t pos = 1 ; pos != _n_nodes ; ++pos )
		{
			assert(!reverse_first_edge_it.finished());

			_nodes[pos]._in_edges_begin = reverse_first_edge_it.read();
			_nodes[pos-1]._n_in_edges = distance(_nodes[pos]._in_edges_begin, _nodes[pos-1]._in_edges_begin);
		}

		if ( _bw_edges.size() - size_t(_nodes.back()._in_edges_begin) > std::numeric_limits<uint16_t>::max() )
			throw std::runtime_error("number of outgoing edges too high to be encoded with 16 bits.");

		_nodes.back()._n_in_edges = uint16_t(_bw_edges.size() - size_t(_nodes.back()._in_edges_begin));

		reverse_first_edge_it.close();
		std::cerr << "OK" << std::endl;

	}


	size_t get_n_edges() const noexcept
	{
		return _n_edges;
	}

	size_t get_n_fw_edges() const noexcept
	{
		return _fw_edges.size();
	}



	size_t get_n_nodes() const noexcept
	{
		return _nodes.size();
	}

	NodeId nodes_begin()  const noexcept
	{
		return NodeId(0);
	}

	NodeId nodes_end()  const noexcept
	{
		return NodeId(_nodes.size());
	}

	FwEdgeId out_edges_begin(NodeId node_id) const noexcept
	{
		return get_node(node_id)._out_edges_begin;
	}

	FwEdgeId out_edges_end(NodeId node_id) const noexcept
	{
		return get_node(node_id)._out_edges_begin + get_node(node_id)._n_out_edges;
	}

	uint32_t get_n_out_edges(NodeId node_id) const noexcept
	{
		return get_node(node_id)._n_out_edges;
	}

	BwEdgeId in_edges_begin(NodeId node_id) const noexcept
	{
		return get_node(node_id)._in_edges_begin;
	}

	BwEdgeId in_edges_end(NodeId node_id) const noexcept
	{
		return get_node(node_id)._in_edges_begin + get_node(node_id)._n_in_edges;
	}

	uint32_t get_n_in_edges(NodeId node_id) const noexcept
	{
		return get_node(node_id)._n_in_edges;
	}

	NodeId get_target(FwEdgeId e) const noexcept
	{
		return get_forward_edge(e).get_target();
	}
	
	NodeId get_source(BwEdgeId e) const noexcept
	{
		return get_backward_edge(e).get_source();
	}

	uint32_t get_n_original_edges(FwEdgeId e) const noexcept
	{
		return get_forward_edge(e).get_n_original_edges();
	}

	void set_n_original_edges(FwEdgeId e, uint32_t n_original_edge) noexcept
	{
		get_forward_edge(e).set_n_original_edges(n_original_edge);
	}

	BwEdgeId find_backward_edge_id(NodeId source, NodeId target) const noexcept
	{
		for ( BwEdgeId e = in_edges_begin(target) ; e != in_edges_end(target) ; ++e )
			if ( get_source(e) == source )
				return e;

		return BwEdgeId::INVALID();
	}

	FwEdgeId find_forward_edge_id(NodeId source, NodeId target) const noexcept
	{
		for ( FwEdgeId e = out_edges_begin(source) ; e != out_edges_end(source) ; ++e )
			if ( get_target(e) == target )
				return e;

		return FwEdgeId::INVALID();
	}

	bool has_time_dist(FwEdgeId e) const noexcept
	{
		return get_forward_edge(e).has_time_dist();
	}

	FwEdgeId find_forward_counterpart(BwEdgeId e, NodeId target) const noexcept
	{
		NodeId source = get_backward_edge(e).get_source();
		for ( FwEdgeId e_fw = out_edges_begin(source) ; e_fw != out_edges_end(source) ; ++e_fw )
			if ( get_forward_edge(e_fw).get_target() == target )
				return e_fw;

		return FwEdgeId::INVALID();
	}

	BwEdgeId find_backward_counterpart(FwEdgeId e, NodeId source_id) const noexcept
	{
		NodeId target_id = get_forward_edge(e).get_target();
		for ( BwEdgeId e_bw = in_edges_begin(target_id) ; e_bw != in_edges_end(target_id) ; ++e_bw )
			if ( get_backward_edge(e_bw).get_source() == source_id )
				return e_bw;

		return BwEdgeId::INVALID();
	}

	const TimeDist & get_time_dist(FwEdgeId e) const noexcept
	{
		return get_forward_edge(e).get_time_dist();
	}

	void set_time_dist(FwEdgeId e, TimeDist&& time_dist) noexcept
	{
		get_forward_edge(e).set_time_dist(std::move(time_dist));
	}


	uint32_t eval_dist(FwEdgeId e, int32_t departure_time) const noexcept
	{
		const auto& edge = get_forward_edge(e);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().eval_dist(departure_time);

		return edge.get_distance();
	}

	TimeDist create_time_dist(FwEdgeId e) const noexcept {
		return create_time_dist(e, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>());
	}


	TimeDist create_time_dist(FwEdgeId e , std::vector<std::tuple<int32_t,int32_t,int32_t,int32_t>> & block) const noexcept
	{
		block = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();
		
		const auto& edge = get_forward_edge(e);
		assert(!edge.has_time_dist());

		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);

		if ( edge.get_blocking_id() == INVALID_BLOCKING_PROFILE_ID() )
			return TimeDist(default_time, distance);

		const uint16_t blocking_profile_id = edge.get_blocking_id();

		const uint32_t blocking_profile_size = _blocking_info[blocking_profile_id];
		const uint32_t blocking_info_begin = blocking_profile_id + 2; //blocking_profile_id = size, +1 = max duration, +2 = first entry
		const uint32_t blocking_info_end = blocking_info_begin + blocking_profile_size;

		std::vector<TimeDist::Segment> time_segments;
		time_segments.reserve(blocking_profile_size + blocking_profile_size/2 + 3);

		//if last block start + last block duration > timedistPeriod() 
		if ( _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] > uint32_t(TimeDist::PERIOD()) )
		{
			if (_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD() > _blocking_info[blocking_info_end]) {
				time_segments.emplace_back(
					-1, _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), default_time,0,0, default_time);
			}
			else {
				time_segments.emplace_back(
					-1, _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), default_time, _blocking_info[blocking_info_end], _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), default_time);
			}
		}

		for ( uint32_t i = blocking_info_begin ; i < blocking_info_end ; i += 2 )
		{
			const int32_t& start_time = _blocking_info[i];
			const int32_t& duration = _blocking_info[i+1];

			//constant travel time
			if ( start_time - default_time > 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), start_time - default_time > time_segments.back().get_right_x()));
				if (start_time - default_time > start_time) {
					time_segments.emplace_back(0, start_time - default_time, default_time, 0, 0, default_time);
				}
				else {
					time_segments.emplace_back(0, start_time - default_time, default_time, start_time, start_time + duration, default_time);
				}
			}

			//full waiting time (+constant travel time)
			if ( start_time > 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), start_time > time_segments.back().get_right_x()));
				time_segments.emplace_back(0, start_time, default_time + duration, start_time, start_time + duration, default_time);
			}

			//slope part (during duration of block)
			if ( start_time + duration > 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), start_time + duration > time_segments.back().get_right_x()));
				time_segments.emplace_back(-1, start_time + duration, default_time, start_time, start_time + duration, default_time);
			}
		}

		assert(!time_segments.empty());

		if ( time_segments.back().get_right_x() > TimeDist::PERIOD() )
		{
			time_segments.pop_back();
		}
		else //fill the complete period, until TimeDist::PERIOD()
		{
			const int32_t& first_start_time = _blocking_info[blocking_info_begin];
			const int32_t& first_duration = _blocking_info[blocking_info_begin + 1];

			if ( first_start_time - default_time <= 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), first_start_time - default_time + TimeDist::PERIOD() > time_segments.back().get_right_x()));
				if (first_start_time - default_time + TimeDist::PERIOD() > first_start_time) {
					time_segments.emplace_back(0, first_start_time - default_time + TimeDist::PERIOD(), default_time, 0, 0, default_time);
				}
				else {
					time_segments.emplace_back(0, first_start_time - default_time + TimeDist::PERIOD(), default_time, first_start_time, first_start_time + first_duration, default_time);
				}
			}

			if ( first_start_time <= 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), first_start_time + TimeDist::PERIOD() > time_segments.back().get_right_x()));
				time_segments.emplace_back(0, first_start_time + TimeDist::PERIOD(), default_time + first_duration, first_start_time + TimeDist::PERIOD(), first_start_time + first_duration + TimeDist::PERIOD(), default_time);
			}

			if ( first_start_time + first_duration <= 0 )
			{
				assert(RITCH_IMPLIES(!time_segments.empty(), first_start_time + first_duration > time_segments.back().get_right_x()));

				time_segments.emplace_back(-1, first_start_time + first_duration + TimeDist::PERIOD(), default_time, first_start_time + TimeDist::PERIOD(), first_start_time + TimeDist::PERIOD() + first_duration, default_time);
			}
		}

		assert( dbg_created_time_dist_is_ok(TimeDist(time_segments, distance, block, false), e, block) );
		
		return TimeDist(time_segments, distance, block, true);;
	}

	size_t get_time_complexity(FwEdgeId e) const noexcept
	{
		const auto& edge = get_forward_edge(e);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().get_time_complexity();

		if ( edge.get_blocking_id() == INVALID_BLOCKING_PROFILE_ID() )
			return 0;

		const uint16_t blocking_profile_id = edge.get_blocking_id();
		const uint32_t blocking_profile_size = _blocking_info[blocking_profile_id];

		if ( blocking_profile_size <= 2 )
			return 0;

		return blocking_profile_size + blocking_profile_size / 2;
	}

	int32_t eval_time(FwEdgeId e, int32_t departure_time) const noexcept
	{
		const auto& edge = get_forward_edge(e);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().eval_time(departure_time);
		
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);

		const uint16_t blocking_index = edge.get_blocking_id();
		if ( edge.get_blocking_id() == INVALID_BLOCKING_PROFILE_ID() )
			return default_time;

		departure_time = util::modulo(departure_time, TimeDist::PERIOD());

		const uint32_t size = _blocking_info[blocking_index];
		const uint32_t begin = blocking_index + 2;
		const uint32_t end = begin + size;

		const int32_t last_start_time = _blocking_info[end - 2];
		const int32_t last_duration = _blocking_info[end - 1];
		if ( departure_time < last_start_time + last_duration - TimeDist::PERIOD() )
			return default_time + last_start_time +  last_duration - TimeDist::PERIOD() - departure_time;

		for ( uint32_t i = begin ; i < end ; i += 2 )
		{
			const int32_t start_time = _blocking_info[i];
			const int32_t duration = _blocking_info[i+1];

			if ( departure_time <= start_time - default_time )
				return default_time;
			if ( departure_time <= start_time )
				return default_time + duration;
			if ( departure_time <= start_time + duration )
				return default_time + duration + (start_time - departure_time);
		}

		const int32_t first_start_time = _blocking_info[begin];
		const int32_t first_duration = _blocking_info[begin + 1];
		if ( departure_time > first_start_time + TimeDist::PERIOD() - default_time )
			return default_time + first_duration;
	
		return default_time;
	}

	int32_t get_min_time(FwEdgeId e) const
	{
		const auto& edge = get_forward_edge(e);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().get_min_time();

		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();

		return calculate_default_travel_time(speed, distance);
	}

	std::tuple<int32_t,int32_t> get_min_max_time(BwEdgeId e, NodeId target) const
	{
		NodeId source = get_source(e);
		FwEdgeId e_fw = find_forward_edge_id(source, target);
		return get_min_max_time(e_fw);
	}

	std::tuple<int32_t,int32_t> get_min_max_time(FwEdgeId e) const
	{
		const auto& edge = get_forward_edge(e);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().get_min_max_time();

		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);

		if ( edge.get_blocking_id() == INVALID_BLOCKING_PROFILE_ID() )
			return std::make_tuple(default_time, default_time);

		const uint16_t blocking_profile_id = edge.get_blocking_id();
		const uint32_t max_duration = _blocking_info[blocking_profile_id + 1];

		int32_t max_time = default_time + max_duration;
		return std::make_tuple(default_time, max_time);
	}

	bool has_edge(NodeId source, NodeId target)
	{
		bool result = has_backward_edge(source, target);
		assert(result == has_forward_edge(source, target));

		return result;
	}

	std::tuple<FwEdgeId, BwEdgeId> insert_edge(NodeId source, NodeId target)
	{
		assert(!has_edge(source, target));
		const FwEdgeId fw_insert_id = insert_edge_impl(_fw_edges, get_node(source)._out_edges_begin, get_node(source)._n_out_edges);
		const BwEdgeId bw_insert_id = insert_edge_impl(_bw_edges, get_node(target)._in_edges_begin, get_node(target)._n_in_edges);

		get_forward_edge(fw_insert_id).set_target(target);
		get_backward_edge(bw_insert_id).set_source(source);

		++_n_edges;

		assert(dbg_forward_edge_is_ok(source, fw_insert_id));
		assert(dbg_backward_edge_is_ok(target, bw_insert_id));
		assert(dbg_node_is_ok(source));
		assert(dbg_node_is_ok(target));
		assert(dbg_is_outgoing_edge_of(fw_insert_id, source));
		assert(dbg_is_incoming_edge_of(bw_insert_id, target));

		assert(has_edge(source, target));

		return std::make_tuple(fw_insert_id, bw_insert_id);
	}

	void remove_all_edges_of_node(NodeId node_id)
	{
		#ifndef NDEBUG
		std::vector<NodeId> dbg_concerned_nodes({node_id});
		#endif

		assert(_n_edges >= get_node(node_id)._n_in_edges);
		_n_edges -= get_node(node_id)._n_in_edges;

		assert(_n_edges >= get_node(node_id)._n_out_edges);
		_n_edges -= get_node(node_id)._n_out_edges;

		for ( FwEdgeId e = out_edges_begin(node_id) ; e != out_edges_end(node_id) ; ++e )
		{
			#ifndef NDEBUG
				dbg_concerned_nodes.push_back(get_target(e));
			#endif

			remove_backward_edge(node_id, get_target(e));
			assert(find_backward_edge_id(node_id, get_target(e)) == BwEdgeId::INVALID());
			turn_into_dummy(e);
		}

		for ( BwEdgeId e = in_edges_begin(node_id) ; e != in_edges_end(node_id) ; ++e )
		{
			#ifndef NDEBUG
				dbg_concerned_nodes.push_back(get_source(e));
			#endif

			remove_forward_edge(get_source(e), node_id);
			assert(find_forward_edge_id(get_source(e), node_id) == FwEdgeId::INVALID());
			turn_into_dummy(e);
		}

		get_node(node_id)._n_in_edges = 0;
		get_node(node_id)._n_out_edges = 0;

		#ifndef NDEBUG
			for ( const auto& id : dbg_concerned_nodes )
				assert(dbg_node_is_ok(id));
		#endif
	}

	bool dbg_forward_edge_is_ok(NodeId source_id, FwEdgeId e) const noexcept
	{
		if ( find_forward_edge_id(source_id, get_target(e)) != e )
			return false;

		BwEdgeId e_bw = find_backward_edge_id(source_id, get_target(e));

		if ( e_bw == BwEdgeId::INVALID() )
			return false;

		if ( get_source(e_bw) != source_id )
			return false;

		if ( find_forward_counterpart(e_bw, get_target(e)) != e )
			return false;

		return true;
	}

	bool dbg_is_outgoing_edge_of(FwEdgeId e, NodeId source) const noexcept
	{
		if ( e < out_edges_begin(source) )
			return false;
		if ( out_edges_end(source) < e + 1 )
			return false;

		return true;
	}

	bool dbg_is_incoming_edge_of(BwEdgeId e, NodeId target) const noexcept
	{
		if ( e < in_edges_begin(target) )
			return false;
		if ( in_edges_end(target) < e + 1 )
			return false;

		return true;
	}

	bool dbg_backward_edge_is_ok(NodeId target_id, BwEdgeId e) const noexcept
	{
		if ( find_backward_edge_id(get_source(e), target_id) != e )
			return false;

		FwEdgeId e_fw = find_forward_edge_id(get_source(e), target_id);

		if ( e_fw == FwEdgeId::INVALID() )
			return false;

		if ( get_target(e_fw) != target_id )
			return false;

		if ( find_backward_counterpart(e_fw, get_source(e)) != e )
			return false;
		
		return true;
	}

	bool dbg_node_is_ok(NodeId node_id) const noexcept
	{
		for ( FwEdgeId e_fw = out_edges_begin(node_id) ; e_fw != out_edges_end(node_id) ; ++e_fw )
			if ( ! dbg_forward_edge_is_ok(node_id, e_fw) )
				return false;

		for ( BwEdgeId e_bw = in_edges_begin(node_id) ; e_bw != in_edges_end(node_id) ; ++e_bw )
			if ( ! dbg_backward_edge_is_ok(node_id, e_bw) )
				return false;

		return true;
	}

	bool dbg_graph_is_ok() const noexcept
	{
		for ( NodeId node_id = nodes_begin() ; node_id != nodes_end() ; ++node_id )
			if ( ! dbg_node_is_ok(node_id) )
				return false;

		return true;
	}

	std::string edge_to_text(FwEdgeId e) const
	{
		return get_forward_edge(e).to_text();
	}

	size_t get_capacity() const noexcept
	{
		return _fw_edges.capacity();
	}

	void compactify()
	{
		// TODO
		RITCH_WARNING("DynamicGraph::compactify() is not implemented yet.");
	}

	std::vector < std::tuple<int32_t, int32_t, int32_t, int32_t>> get_edge_block(NodeId source, NodeId target){
		FwEdgeId e = find_forward_edge_id(source, target);
		const auto& edge = get_forward_edge(e);
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);
		std::vector < std::tuple<int32_t, int32_t, int32_t, int32_t>> edgeblock;

		if (edge.get_blocking_id() == INVALID_BLOCKING_PROFILE_ID()) {
			edgeblock.emplace_back(TimeDist::PERIOD(), 0, 0, default_time);
		}
		else {
			const uint16_t blocking_profile_id = edge.get_blocking_id();
			const uint32_t blocking_profile_size = _blocking_info[blocking_profile_id];
			const uint32_t blocking_info_begin = blocking_profile_id + 2; //blocking_profile_id = size, +1 = max duration, +2 = first entry
			const uint32_t blocking_info_end = blocking_info_begin + blocking_profile_size;


			if (_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] > uint32_t(TimeDist::PERIOD()))
			{
				if (_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD() > _blocking_info[blocking_info_end]) {
					edgeblock.emplace_back(std::make_tuple(_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), 0, 0, default_time));
				}
				else {
					edgeblock.emplace_back(
						_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), _blocking_info[blocking_info_end], _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), default_time);
				}
			}

			for (uint32_t i = blocking_info_begin; i < blocking_info_end; i += 2)
			{
				const int32_t& start_time = _blocking_info[i];
				const int32_t& duration = _blocking_info[i + 1];

				//constant travel time
				if (start_time - default_time > 0)
				{
					if (start_time - default_time > start_time) {
						edgeblock.emplace_back(start_time - default_time, 0, 0, default_time);
					}
					else {
						edgeblock.emplace_back(start_time - default_time, start_time, start_time + duration, default_time);
					}
				}

				//full waiting time (+constant travel time)
				if (start_time > 0)
				{
					edgeblock.emplace_back(start_time, start_time, start_time + duration, default_time);
				}

				//slope part (during duration of block)
				if (start_time + duration > 0)
				{
					edgeblock.emplace_back(start_time + duration, start_time, start_time + duration, default_time);
				}
			}

			if (std::get<0>(edgeblock.back()) > TimeDist::PERIOD())
			{
				edgeblock.pop_back();
			}
			else //fill the complete period, until TimeDist::PERIOD()
			{
				const int32_t& first_start_time = _blocking_info[blocking_info_begin];
				const int32_t& first_duration = _blocking_info[blocking_info_begin + 1];

				if (first_start_time - default_time <= 0)
				{
					if (first_start_time - default_time + TimeDist::PERIOD() > first_start_time) {
						edgeblock.emplace_back(first_start_time - default_time + TimeDist::PERIOD(), 0, 0, default_time);
					}
					else {
						edgeblock.emplace_back(first_start_time - default_time + TimeDist::PERIOD(), first_start_time, first_start_time + first_duration, default_time);
					}
				}

				if (first_start_time <= 0)
				{
					edgeblock.emplace_back(first_start_time + TimeDist::PERIOD(), first_start_time + TimeDist::PERIOD(), first_start_time + first_duration + TimeDist::PERIOD(), default_time);
				}

				if (first_start_time + first_duration <= 0)
				{
					edgeblock.emplace_back(first_start_time + first_duration + TimeDist::PERIOD(), first_start_time + TimeDist::PERIOD(), first_start_time + TimeDist::PERIOD() + first_duration, default_time);
				}
			}

	}
		return edgeblock;
	}

};

}

#endif /* RITCH_PREPRO_DYNAMIC_GRAPH_H_ */
