#ifndef RITCH_PREPRO_ORDERING_H_
#define RITCH_PREPRO_ORDERING_H_

#include <array>
#include <algorithm>
#include <cassert>
#include <iomanip>
#include <memory>
#include <omp.h>
#include <queue>
#include <unordered_map>
#include <unordered_set>
#include <utility>
#include <vector>

#include "ritch/io/layers_format.h"
#include "ritch/io/ritch_shortcuts_format.h"
#include "ritch/prepro/dynamic_graph.h"
#include "ritch/util/misc.h"
#include "ritch/util/timing.h"

namespace ritch
{

template <typename WitnessSearchType>
class Ordering
{
public:

	static const int MAXIMUM_THREADS = -1;

private:

	typedef DynamicGraph::NodeId NodeId;
	typedef DynamicGraph::FwEdgeId FwEdgeId;
	typedef DynamicGraph::BwEdgeId BwEdgeId;
	typedef DynamicGraph::TimeDist TimeDist;
	static const bool SIMULATE = true;
	static const bool DO_NOT_SIMULATE = false;

	typedef typename WitnessSearchType::ShortcutStatus ShortcutStatus;

	struct FwEdgeIdHash { size_t operator() (FwEdgeId e) const noexcept { return std::hash<uint32_t>()(uint32_t(e)); } };
	struct BwEdgeIdHash { size_t operator() (BwEdgeId e) const noexcept { return std::hash<uint32_t>()(uint32_t(e)); } };
	struct NodeIdHash { size_t operator() (NodeId e) const noexcept { return std::hash<uint32_t>()(uint32_t(e)); } };

	struct NodeIdPairHash {
		size_t operator() (const std::pair<NodeId, NodeId>& p) const noexcept {
			return std::hash<uint32_t>()(uint32_t(p.first)) ^ std::hash<uint32_t>()(uint32_t(p.second));
		}
	};

	class WitnessCacheEntry
	{
	private:
		std::array<uint32_t,3> _data;

	public:
		WitnessCacheEntry()
		: WitnessCacheEntry(NodeId::INVALID(), NodeId::INVALID(), WitnessSearchType::UNDECIDED, 0x3fffffff)
		{}

		WitnessCacheEntry(NodeId source_id, NodeId target_id, ShortcutStatus status, uint32_t complexity)
		: _data({ source_id, target_id, (status & 3) | (complexity << 2) })
		{
			assert(status <= 2);
			assert(complexity <= 0x3fffffff);
		}

		ShortcutStatus get_shortcut_status() const noexcept { return _data[2] & 3; }
		void set_shortcut_status(ShortcutStatus status) noexcept { assert(status <= 3); _data[2] = (_data[2] | 3) & status; }

		uint32_t get_shortcut_complexity() const noexcept { return _data[2] >> 2; }
		void set_shortcut_complexity(uint32_t complexity) noexcept { assert(complexity <= 0x3ffffffF); _data[2] = complexity << 2; }

		NodeId get_source_id() const noexcept { return NodeId(_data[0]); }
		void set_source_id(NodeId source_id) noexcept { _data[0] = source_id; }
		
		NodeId get_target_id() const noexcept { return NodeId(_data[1]); }
		void set_target_id(NodeId target_id) noexcept { _data[1] = target_id; }

		friend bool operator== (const WitnessCacheEntry& lhs, const WitnessCacheEntry& rhs)
		{
			bool is_equal = (memcmp(&lhs._data, &rhs._data, sizeof(_data)) == 0);
			return is_equal;
		}

		friend bool operator!= (const WitnessCacheEntry& lhs, const WitnessCacheEntry& rhs)
		{
			return !(lhs == rhs);
		}
	};

	class WitnessCache
	{
	private:
		std::vector<std::vector<WitnessCacheEntry>> _data;

	public:

		WitnessCache(size_t n_nodes)
		: _data(n_nodes)
		{}

		static const WitnessCacheEntry INVALID_ENTRY() { return WitnessCacheEntry(); }

		const WitnessCacheEntry lookup(NodeId source_id, NodeId middle_id, NodeId target_id) const
		{
			for ( const auto& cache_entry : _data[middle_id] )
				if ( cache_entry.get_source_id() == source_id && cache_entry.get_target_id() == target_id )
					return cache_entry;

			return INVALID_ENTRY();
		}

		bool contains(NodeId source_id, NodeId middle_id, NodeId target_id) const
		{
			const auto& entry = lookup(source_id, middle_id, target_id);
			assert(RITCH_IMPLIES(entry != INVALID_ENTRY(), entry.get_source_id() == source_id));
			assert(RITCH_IMPLIES(entry != INVALID_ENTRY(), entry.get_target_id() == target_id));

			return entry != INVALID_ENTRY();
		}

		void insert(NodeId middle_id, WitnessCacheEntry cache_entry)
		{
			assert( ! contains(cache_entry.get_source_id(), middle_id, cache_entry.get_target_id()) );
			_data[middle_id].push_back(cache_entry);
		}

		void remove(NodeId middle_id)
		{
			_data[middle_id].clear();
			_data[middle_id].shrink_to_fit();
		}

		void remove(NodeId source_id, NodeId middle_id, NodeId target_id)
		{
			assert(middle_id != NodeId::INVALID());

			if ( source_id == NodeId::INVALID() && target_id == NodeId::INVALID() )
			{
				remove(middle_id);
			}
			else if ( source_id == NodeId::INVALID() )
			{
				for ( size_t i = 0 ; i < _data[middle_id].size() ; ++i )
				{
					if ( target_id == _data[middle_id][i].get_target_id() )
					{
						_data[middle_id][i] = _data[middle_id].back();
						_data[middle_id].pop_back();
						--i;
					}
				}
			}
			else if ( target_id == NodeId::INVALID() )
			{
				for ( size_t i = 0 ; i < _data[middle_id].size() ; ++i )
				{
					if ( source_id == _data[middle_id][i].get_source_id() )
					{
						_data[middle_id][i] = _data[middle_id].back();
						_data[middle_id].pop_back();
						--i;
					}
				}
			}
			else
			{
				for ( size_t i = 0 ; i < _data[middle_id].size() ; ++i )
				{
					if ( source_id == _data[middle_id][i].get_source_id() && target_id == _data[middle_id][i].get_target_id() )
					{
						_data[middle_id][i] = _data[middle_id].back();
						_data[middle_id].pop_back();
						--i;
					}
				}
			}
		}

		void shrink_to_fit()
		{
			for ( auto& bucket : _data )
				bucket.shrink_to_fit();
		}
	};

	struct LocalThreadData
	{
		std::vector<EdgeDescr> _edges_to_insert;

		std::vector<WitnessCacheEntry> _witnesses_to_insert_to_cache;
		std::vector<NodeId> _middle_nodes_of_witnesses;
		WitnessSearchType _witness_search;


		LocalThreadData(const LocalThreadData&) = delete;
		LocalThreadData& operator=(const LocalThreadData&) = delete;

		LocalThreadData(LocalThreadData&&) = default;
		LocalThreadData& operator=(LocalThreadData&&) = default;

		LocalThreadData()
			: _edges_to_insert(),
		  _witnesses_to_insert_to_cache(),
		  _middle_nodes_of_witnesses(),
		  _witness_search(nullptr)
		{}

		LocalThreadData(const DynamicGraph& graph)
		: _edges_to_insert(),
		  _witnesses_to_insert_to_cache(),
		  _middle_nodes_of_witnesses(),
		  _witness_search(&graph)
		{}
	};

	struct MiddleNodeDataElem
	{
		int32_t _x_right;
		uint32_t _middle_id;
	};

	struct MiddleNodeData
	{
	private:
		uint64_t _data;

		bool is_ptr() const noexcept { return (_data & 1) == 0; }

		uint32_t const * const & the_ptr() const noexcept
		{
			assert(is_ptr());
			return reinterpret_cast<uint32_t const * const &>(_data);
		}

		uint32_t*& the_ptr() noexcept
		{
			assert(is_ptr());
			return reinterpret_cast<uint32_t*&>(_data);
		}

		bool has_zero_middle_nodes() const noexcept { return _data == 0xffffffffffffffff; }
		bool has_single_middle_node() const noexcept { return !has_zero_middle_nodes() && !is_ptr(); }

		static void do_copy(MiddleNodeData& from, MiddleNodeData& to) noexcept
		{
			if ( other.is_ptr() && other.the_ptr() != nullptr )
			{
				const size_t n_elems = 2*from.size() + 1
				to.the_ptr() = new (std::nothrow) uint32_t[n_elems];

				memcpy(to.the_ptr(), from.the_ptr(), n_elems * sizeof(uint32_t));
				to._data = reinterpret_cast<uint64_t>(ptr);
			}
			else
			{
				to._data = from._data;
			}
		}

		static void do_move(MiddleNodeData& from, MiddleNodeData& to) noexcept
		{
			to._data = from._data;
			from._data = 0xffffffffffffffff;
		}

	public:
		MiddleNodeData(const MiddleNodeData& other) { do_copy(other, *this); }
		MiddleNodeData(MiddleNodeData&& other) { do_move(other, *this); }

		MiddleNodeData& operator= (const MiddleNodeData& other)
		{
			if ( this != &other )
			{
				if ( is_ptr() && the_ptr() != nullptr )
					delete[] the_ptr();

				do_copy(other, *this);
			}
		}

		MiddleNodeData& operator= (MiddleNodeData&& other)
		{

			if ( this != &other )
			{
				if ( is_ptr() && the_ptr() != nullptr )
					delete[] the_ptr();

				do_move(other, *this);
			}
		}

		MiddleNodeData() : _data(0xffffffffffffffff) {}

		~MiddleNodeData()
		{
			if ( is_ptr() && the_ptr() != nullptr )
				delete[] the_ptr();

			_data = 0xffffffffffffffff;
		}

		uint32_t size() const noexcept
		{
			if ( has_zero_middle_nodes() ) return 1;
			if ( has_single_middle_node() ) return 1;

			return the_ptr()[0];
		}

		NodeId get_single_middle_node() const noexcept
		{
			assert(!has_zero_middle_nodes());
			assert(!is_ptr());
			return NodeId(reinterpret_cast<const uint32_t*>(&_data)[1]);
		}

		MiddleNodeDataElem get_middle_node_data_elem(size_t index) const noexcept
		{
			if ( has_zero_middle_nodes() ) return MiddleNodeDataElem{ TimeDist::PERIOD(), NodeId::INVALID() };
			if ( has_single_middle_node() ) return MiddleNodeDataElem{  TimeDist::PERIOD(), get_single_middle_node() };

			int32_t x_right = reinterpret_cast<const int32_t&>(the_ptr()[2*index + 1]);
			uint32_t middle_id = the_ptr()[2*index + 2];

			return MiddleNodeDataElem{ x_right, middle_id };
		}

		void store_single_middle_node(uint32_t middle_id) noexcept
		{
			if ( is_ptr() && the_ptr() != nullptr )
				delete[] the_ptr();

			reinterpret_cast<uint32_t*>(&_data)[0] = 1;
			reinterpret_cast<uint32_t*>(&_data)[1] = middle_id;
		}

		void store_multiple_middle_nodes(const std::vector<std::pair<int32_t,uint32_t>>& node_info) noexcept
		{
			if ( node_info.size() == 1 )
			{
				store_single_middle_node(node_info.front().second);
				return;
			}

			if ( is_ptr() && the_ptr() != nullptr )
				delete[] the_ptr();

			uint32_t* ptr = new (std::nothrow) uint32_t[2*node_info.size() + 1];
			_data = reinterpret_cast<uint64_t>(ptr);

			ptr[0] = uint32_t(node_info.size());
			for ( size_t i = 0 ; i != node_info.size() ; ++i )
			{
				reinterpret_cast<int32_t&>(ptr[2*i+1]) = node_info[i].first;
				ptr[2*i+2] = node_info[i].second;
			}
		}

		std::vector<std::pair<int32_t, uint32_t>> obtain_content() const noexcept
		{
			const uint32_t n_elems = size();
			std::vector<std::pair<int32_t, uint32_t>> result;
			result.reserve(n_elems);

			for ( uint32_t i = 0 ; i != n_elems ; ++i )
			{
				const auto& elem = get_middle_node_data_elem(i);
				result.push_back(std::make_pair(elem._x_right, elem._middle_id));
			}

			return result;
		}
	};

	bool dbg_middle_node_info_is_ok(const std::vector<std::pair<int32_t, uint32_t>>& middle_node_info, NodeId src, NodeId tgt) const
	{

		for ( const auto& item : middle_node_info )
		{
			NodeId node_id(item.second);

			bool out_edge_ok = false;
			for ( FwEdgeId e = _graph.out_edges_begin(node_id) ; e != _graph.out_edges_end(node_id) && !out_edge_ok ; ++e )
				if ( _graph.get_target(e) == tgt )
					out_edge_ok = true;

			bool in_edge_ok = false;
			for ( BwEdgeId e = _graph.in_edges_begin(node_id) ; e != _graph.in_edges_end(node_id) && !in_edge_ok ; ++e )
				if ( _graph.get_source(e) == src )
					in_edge_ok = true;

			if ( !in_edge_ok || !out_edge_ok )
				return false;
		}

		return true;
	}

	void store_single_middle_node(NodeId source, NodeId target, NodeId middle_id) noexcept
	{
		_middle_node_info[std::make_pair(source, target)].store_single_middle_node(middle_id);
	}

	void store_multiple_middle_nodes(NodeId source, NodeId target, const std::vector<std::pair<int32_t, uint32_t>>& middle_node_info) noexcept
	{
		_middle_node_info[std::make_pair(source, target)].store_multiple_middle_nodes(middle_node_info);
	}

	std::vector<std::pair<int32_t, uint32_t>> obtain_middle_nodes(NodeId source, NodeId target) const noexcept
	{
		auto it = _middle_node_info.find(std::make_pair(source, target));

		if ( it == _middle_node_info.end() )
			return std::vector<std::pair<int32_t, uint32_t>>();

		 const MiddleNodeData& data = it->second;
		 return data.obtain_content();
	}

	std::vector<std::pair<int32_t, uint32_t>> obtain_and_remove_middle_nodes(NodeId source, NodeId target) noexcept
	{
		std::vector<std::pair<int32_t, uint32_t>> result;
		auto it = _middle_node_info.find(std::make_pair(source, target));

		if ( it == _middle_node_info.end() )
			return result;

		result = it->second.obtain_content();
		_middle_node_info.erase(it);
		return result;
	}

	bool has_middle_node(NodeId source, NodeId target) const noexcept
	{
		return _middle_node_info.count(std::make_pair(source, target)) != 0;
	}

	bool has_smaller_cost(NodeId lhs, NodeId rhs) const
	{
		assert( lhs < _nodes.size() );
		assert( rhs < _nodes.size() );
		assert( lhs < _random_permutation.size() );
		assert( rhs < _random_permutation.size() );

		if ( _node_cost[lhs] < _node_cost[rhs] )
			return true;

		if ( _node_cost[lhs] == _node_cost[rhs] )
			if ( _random_permutation[lhs] < _random_permutation[rhs] )
				return true;

		return false;
	}

	bool is_minimum_in_neighborhood(NodeId node_id, int hop_radius) const
	{
		assert( node_id < _graph.get_n_nodes() );
		assert( ! has_smaller_cost(node_id, node_id) );

		std::queue<NodeId> Q;
		Q.push(node_id);

		std::unordered_map<NodeId,int, NodeIdHash> n_hops;
		n_hops[node_id] = 0;

		while ( ! Q.empty() )
		{
			const NodeId source_id = Q.front();
			Q.pop();

			for ( FwEdgeId e = _graph.out_edges_begin(source_id) ; e != _graph.out_edges_end(source_id) ; ++e )
			{
				const NodeId target_id = _graph.get_target(e);
				if ( n_hops.find(target_id) != n_hops.end() ) continue;

				assert( has_smaller_cost(target_id, node_id) || has_smaller_cost(node_id, target_id) );
				if ( has_smaller_cost(target_id, node_id) ) return false;

				n_hops[target_id] = n_hops[source_id] + 1;

				if ( n_hops[target_id] >= hop_radius ) continue;
				Q.push(target_id);
			}

			for ( BwEdgeId e = _graph.in_edges_begin(source_id) ; e != _graph.in_edges_end(source_id) ; ++e )
			{
				const NodeId target_id = _graph.get_source(e);
				if ( n_hops.find(target_id) != n_hops.end() ) continue;

				assert( has_smaller_cost(target_id, node_id) || has_smaller_cost(node_id, target_id) );
				if ( has_smaller_cost(target_id, node_id) ) return false;

				n_hops[target_id] = n_hops[source_id] + 1;

				if ( n_hops[target_id] >= hop_radius ) continue;
				Q.push(target_id);
			}
		}

		return true;
	}

	void parallelly_choose_nodes_to_be_contracted_next()
	{
		#pragma omp parallel
		{
			#pragma omp for schedule(dynamic)
			for ( int i = int(_working_nodes_end) ; i < int(_nodes.size()) ; ++i )
			{
				const NodeId middle_id = _nodes[i];
				if ( is_minimum_in_neighborhood(middle_id, 2) )
					_node_selected_for_contraction[middle_id] = 1;
			}
		}

		auto first_unselected_node_id =
			std::stable_partition(_nodes.begin() + _working_nodes_end, _nodes.end(),
				[&](NodeId x) { return _node_selected_for_contraction[x]; } );

		_working_nodes_end = first_unselected_node_id - _nodes.begin();
	}

	bool is_shortcut_necessary(BwEdgeId e_in, FwEdgeId e_in_fw, std::vector<std::tuple<int32_t,int32_t,int32_t,int32_t>> block_in, NodeId middle_id, FwEdgeId e_out, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_out, bool simulate, EdgeDescr& o_insert_edge) const
	{
		const NodeId source_id = _graph.get_source(e_in);
		const NodeId target_id = _graph.get_target(e_out);

		assert( source_id != target_id );
		assert( source_id != middle_id );
		assert( target_id != middle_id );

		// use already cached witness if present
		WitnessCacheEntry witness;
		assert( witness.get_shortcut_status() == WitnessSearchType::UNDECIDED );

		if ( _witness_cache.contains(source_id, middle_id, target_id) )
		{
			witness = _witness_cache.lookup(source_id, middle_id, target_id);
			assert( witness.get_shortcut_status() != WitnessSearchType::UNDECIDED );
		}

		const uint32_t n_original_edges = _graph.get_n_original_edges(e_in_fw) + _graph.get_n_original_edges(e_out);
		const uint32_t complexity_e_in = uint32_t(_graph.get_time_complexity(e_in_fw));
		const uint32_t complexity_e_out = uint32_t(_graph.get_time_complexity(e_out));

		const uint32_t shortcut_complexity =
			witness.get_shortcut_status() == WitnessSearchType::UNDECIDED
				? complexity_e_in + complexity_e_out
				: witness.get_shortcut_complexity();

		o_insert_edge = EdgeDescr(source_id, target_id, middle_id, shortcut_complexity, n_original_edges);

		// return if witness is already cached
		if ( witness.get_shortcut_status() != WitnessSearchType::UNDECIDED )
			return witness.get_shortcut_status() == WitnessSearchType::NECESSARY;

		assert( omp_get_thread_num() < int( _local_thread_data_ptrs.size() ) );
		LocalThreadData& thread_data = *_local_thread_data_ptrs[omp_get_thread_num()];

		// perform witness search if witness is not cached yet
		//removed const_edge_block
		ShortcutStatus shortcut_status = thread_data._witness_search.run(source_id, e_in, e_in_fw, block_in, middle_id, e_out, block_out, target_id, o_insert_edge);
		assert( shortcut_status != WitnessSearchType::UNDECIDED );

		if ( simulate )
		{
			thread_data._witnesses_to_insert_to_cache.emplace_back(
				source_id, target_id, shortcut_status, o_insert_edge._complexity);

			thread_data._middle_nodes_of_witnesses.push_back(middle_id);
		}

		return shortcut_status == WitnessSearchType::NECESSARY;
	}

	void simulate_contraction_and_update_cost_of_node(NodeId middle_node)
	{

		uint32_t n_insert_edges = 0;

		uint32_t n_original_edges_of_inserted_edges = 0;
		uint32_t n_original_edges_of_removed_edges = 0;

		uint32_t total_complexity_of_inserted_edges = 0;
		uint32_t total_complexity_of_removed_edges = 0;

		uint32_t n_removed_edges = 0;

		for ( FwEdgeId e = _graph.out_edges_begin(middle_node) ; e != _graph.out_edges_end(middle_node) ; ++e )
		{
			++n_removed_edges;
			n_original_edges_of_removed_edges += _graph.get_n_original_edges(e);
			total_complexity_of_removed_edges += uint32_t(_graph.get_time_complexity(e));
		}

		for ( BwEdgeId e = _graph.in_edges_begin(middle_node) ; e != _graph.in_edges_end(middle_node) ; ++e )
		{
			FwEdgeId e_fw = _graph.find_forward_counterpart(e, middle_node);
				
			++n_removed_edges;
			n_original_edges_of_removed_edges += _graph.get_n_original_edges(e_fw);
			total_complexity_of_removed_edges += uint32_t(_graph.get_time_complexity(e_fw));
		}

		for ( BwEdgeId e_in = _graph.in_edges_begin(middle_node) ; e_in != _graph.in_edges_end(middle_node) ; ++e_in )
		{
			const FwEdgeId e_in_fw = _graph.find_forward_counterpart(e_in, middle_node);

			for ( FwEdgeId e_out = _graph.out_edges_begin(middle_node) ; e_out != _graph.out_edges_end(middle_node) ; ++e_out )
			{
				if ( _graph.get_source(e_in) == _graph.get_target(e_out) )
					continue;

				EdgeDescr insert_edge;

				std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_in = get_edge_block(_graph.get_source(e_in), middle_node);
				std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_out = get_edge_block(middle_node, _graph.get_target(e_out));

				bool shortcut_necessary = is_shortcut_necessary(e_in, e_in_fw, block_in, middle_node, e_out, block_out, SIMULATE, insert_edge);

				if ( shortcut_necessary )
				{
					++n_insert_edges;

					n_original_edges_of_inserted_edges += insert_edge._n_original_edges;
					total_complexity_of_inserted_edges += insert_edge._complexity;
				}

			}
		}

		double edges_quotient = double(n_insert_edges) / std::max(1.0, double(n_removed_edges));

		double original_edges_quotient =
			double(n_original_edges_of_inserted_edges) / std::max(1.0, double(n_original_edges_of_removed_edges));

		double complexity_quotient =
			double(total_complexity_of_inserted_edges) / std::max(1.0, double(total_complexity_of_removed_edges));

		double new_node_cost =
			2.0 * edges_quotient +
			_hierarchie_depth[middle_node] +
			original_edges_quotient +
			2.0 * complexity_quotient;

		_node_cost[middle_node] = new_node_cost;
	}

	void contract(const BwEdgeId e_in, const NodeId middle_id, const FwEdgeId e_out, bool simulate = DO_NOT_SIMULATE)
	{
		const FwEdgeId e_in_fw = _graph.find_forward_counterpart(e_in, middle_id);

		EdgeDescr insert_edge;
		
		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockNew = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();

		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_in = get_edge_block(_graph.get_source(e_in), middle_id);
		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_out = get_edge_block(middle_id, _graph.get_target(e_out));
		
		const bool shortcut_necessary = is_shortcut_necessary(e_in, e_in_fw, block_in, middle_id, e_out, block_out, simulate, insert_edge);

		if ( ! shortcut_necessary ) return;
		
		if (insert_edge._time_dist.is_infty()) {
			insert_edge._time_dist = link(
				_graph.has_time_dist(e_out) ? _graph.get_time_dist(e_out) : _graph.create_time_dist(e_out, block_out),
				_graph.has_time_dist(e_in_fw) ? _graph.get_time_dist(e_in_fw) : _graph.create_time_dist(e_in_fw, block_in),
				block_out, block_in, blockNew, true);
			insert_edge._block = blockNew; /* new block info is filled in link operaion, stored in ._block*/
		}
		
		assert( omp_get_thread_num() < int(_local_thread_data_ptrs.size()) );
		_local_thread_data_ptrs[omp_get_thread_num()]->_edges_to_insert.push_back(std::move(insert_edge));

	}

	void parallelly_contract_nodes()
	{
		#pragma omp parallel
		{
			#pragma omp for schedule(dynamic)
			for ( int i = int(_working_nodes_begin) ; i < int(_working_nodes_end) ; ++i )
			{
				NodeId middle_id = _nodes[i];

				for ( BwEdgeId e_in = _graph.in_edges_begin(middle_id) ; e_in != _graph.in_edges_end(middle_id) ; ++e_in )
					for ( FwEdgeId e_out = _graph.out_edges_begin(middle_id) ; e_out != _graph.out_edges_end(middle_id) ; ++e_out )
						if (_graph.get_source(e_in) != _graph.get_target(e_out)) {
							contract(e_in, middle_id, e_out);
						}
			}
		}

		for ( size_t i = _working_nodes_begin ; i < _working_nodes_end ; ++i )
		{
			NodeId middle_id = _nodes[i];
			_witness_cache.remove(middle_id);

			for ( BwEdgeId e_in = _graph.in_edges_begin(middle_id) ; e_in != _graph.in_edges_end(middle_id) ; ++e_in )
			{
				NodeId source_id = _graph.get_source(e_in);

				for ( FwEdgeId e_out = _graph.out_edges_begin(middle_id) ; e_out != _graph.out_edges_end(middle_id) ; ++e_out )
				{
					NodeId target_id = _graph.get_target(e_out);

					_witness_cache.remove(NodeId::INVALID(), source_id, middle_id);
					_witness_cache.remove(middle_id, target_id, NodeId::INVALID());
				}
			}
		}
	}



	DynamicGraph _graph;

	size_t _working_nodes_begin;
	size_t _working_nodes_end;
	std::vector<NodeId> _nodes;
	std::vector<char> _node_selected_for_contraction;

	std::vector<double> _node_cost;
	std::vector<uint32_t> _random_permutation;
	std::vector<uint16_t> _hierarchie_depth;
	std::unordered_map<std::pair<NodeId,NodeId>, MiddleNodeData, NodeIdPairHash> _middle_node_info;
	std::vector<uint32_t> _node_to_level;

	mutable WitnessCache _witness_cache;
	std::vector<std::unique_ptr<LocalThreadData>> _local_thread_data_ptrs;

	io::ritch_shortcuts_format::OutputFile _ritch_shortcuts_output_file;
	io::layers_format::OutputFile _layers_output_file;

	/*<startnode, endnode> : list of <>*/
	std::unordered_map<std::pair<NodeId, NodeId>, std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>>, NodeIdPairHash> _edge_block_info;


public:

	Ordering
	(
		DynamicGraph&& graph,
		const std::string& graph_directory_name,
		const std::string& layers_output_file_name,
		const std::string& ritch_shortcuts_output_file_name
	)
		: _graph(std::move(graph)),
		_working_nodes_begin(0),
		_working_nodes_end(0),
		_nodes(_graph.get_n_nodes(), NodeId::INVALID()),
		_node_selected_for_contraction(_graph.get_n_nodes(), 0),
		_node_cost(_graph.get_n_nodes(), 0),
		_random_permutation(_graph.get_n_nodes(), NodeId::INVALID()),
		_hierarchie_depth(_graph.get_n_nodes(), 1u),
		_middle_node_info(),
		_node_to_level(_graph.get_n_nodes(), std::numeric_limits<uint32_t>::max()),
		_witness_cache(_graph.get_n_nodes()),
		_local_thread_data_ptrs(),
		_ritch_shortcuts_output_file(ritch_shortcuts_output_file_name, uint32_t(_graph.get_n_nodes()), uint32_t(TimeDist::PERIOD())),
		_layers_output_file(layers_output_file_name, uint32_t(_graph.get_n_nodes())),
		_edge_block_info(0)
	{
		std::generate(_nodes.begin(), _nodes.end(),
			[]() { static uint32_t counter(0); return counter++; });

		std::generate(_random_permutation.begin(), _random_permutation.end(),
			[](){ static uint32_t counter(0); return counter++; });

		std::random_shuffle(_random_permutation.begin(), _random_permutation.end());
	}

	void run(int n_threads = MAXIMUM_THREADS)
	{
		auto t_begin = util::time_stamp();

		if ( n_threads == MAXIMUM_THREADS )
			n_threads = omp_get_max_threads();
		else
			n_threads = std::min(n_threads, omp_get_max_threads());

		omp_set_num_threads(n_threads);
		_local_thread_data_ptrs.resize(n_threads);

		#pragma omp parallel
		{
			_local_thread_data_ptrs[omp_get_thread_num()] = std::make_unique<LocalThreadData>(_graph);
		}

		RITCH_STATUS("TCH ordering running with " << std::min(n_threads, omp_get_max_threads()) << " threads\n");

		//
		// set up initial tentative node costs
		//

		RITCH_STATUS("Computing initial node costs... ");
		auto t_init_begin = util::time_stamp();

		const int nodes_begin(_graph.nodes_begin());
		const int nodes_end(_graph.nodes_end());

		#pragma omp parallel
		{
			#pragma omp for schedule(dynamic)
			for ( int i = nodes_begin ; i < nodes_end ; ++i )
				simulate_contraction_and_update_cost_of_node(NodeId(i));

			LocalThreadData& thread_data = *_local_thread_data_ptrs[omp_get_thread_num()];
			
			#pragma omp critical
			{
				assert(
					   thread_data._witnesses_to_insert_to_cache.size()
					== thread_data._middle_nodes_of_witnesses.size()
				);

				for ( size_t i = 0 ; i != thread_data._witnesses_to_insert_to_cache.size(); ++i )
				{
					const NodeId& middle_node_id = thread_data._middle_nodes_of_witnesses[i];
					const WitnessCacheEntry& witness_cache_entry = thread_data._witnesses_to_insert_to_cache[i];

					_witness_cache.insert(middle_node_id, witness_cache_entry);
				}

				thread_data._witnesses_to_insert_to_cache.clear();
				thread_data._witnesses_to_insert_to_cache.shrink_to_fit();

				thread_data._middle_nodes_of_witnesses.clear();
				thread_data._middle_nodes_of_witnesses.shrink_to_fit();
			}
		}


		auto t_init_end = util::time_stamp();
		RITCH_CONTINUE_STATUS("OK, took " << util::get_duration_in_seconds(t_init_begin, t_init_end) << "\n");

		//
		// repeatedly compute independent sets and contract their nodes,
		// merging necessary shortcuts with already present edges
		//

		size_t old_n_edges = _graph.get_n_edges();

		RITCH_STATUS("Contract independent node sets of size ");
		while ( _working_nodes_begin < _nodes.size() )
		{
			parallelly_choose_nodes_to_be_contracted_next();
			RITCH_CONTINUE_STATUS( (_working_nodes_end - _working_nodes_begin) << "/" << (_nodes.size() - _working_nodes_begin));
			parallelly_contract_nodes();
			RITCH_CONTINUE_STATUS(" ");

			std::vector<NodeId> nodes_to_update;
			for ( size_t i = _working_nodes_begin ; i < _working_nodes_end ; ++i )
			{
				assert( i < _nodes.size() );
				const NodeId node_id = _nodes[i];
				_node_to_level[node_id] = uint32_t(i);

				for ( FwEdgeId e = _graph.out_edges_begin(node_id) ; e != _graph.out_edges_end(node_id) ; ++e )
				{
					nodes_to_update.push_back(_graph.get_target(e));
					_hierarchie_depth[_graph.get_target(e)] =
						std::max(_hierarchie_depth[_graph.get_target(e)], util::plus_one_till_max(_hierarchie_depth[node_id]));

					if ( !_graph.has_time_dist(e) && !has_middle_node(node_id, _graph.get_target(e)) )
						continue;

					auto middle_node_info = obtain_and_remove_middle_nodes(node_id, _graph.get_target(e));
					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockE = get_edge_block(node_id, _graph.get_target(e));

					_ritch_shortcuts_output_file.write_edge(node_id, _graph.get_target(e), _graph.get_time_dist(e), std::move(middle_node_info), blockE);


				}

				for ( BwEdgeId e = _graph.in_edges_begin(node_id) ; e != _graph.in_edges_end(node_id) ; ++e )
				{
					nodes_to_update.push_back(_graph.get_source(e));
					_hierarchie_depth[_graph.get_source(e)] =
						std::max(_hierarchie_depth[_graph.get_source(e)], util::plus_one_till_max(_hierarchie_depth[node_id]));

					FwEdgeId e_fw = _graph.find_forward_counterpart(e, node_id);

					if ( !_graph.has_time_dist(e_fw) && !has_middle_node(_graph.get_source(e), node_id) )
						continue;

					auto middle_node_info = obtain_and_remove_middle_nodes(_graph.get_source(e), node_id);
					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockEFW = get_edge_block(_graph.get_source(e), node_id); //TODO!
					_ritch_shortcuts_output_file.write_edge(_graph.get_source(e), node_id, _graph.get_time_dist(e_fw), middle_node_info, blockEFW); //TODO!
				}

				_graph.remove_all_edges_of_node(node_id);
			}

			assert(_graph.dbg_graph_is_ok());

			_layers_output_file.write_layer(_nodes.begin() + _working_nodes_begin, _nodes.begin() + _working_nodes_end);

			std::sort(nodes_to_update.begin(), nodes_to_update.end());
			util::make_sorted_sequence_unique(nodes_to_update);

			RITCH_CONTINUE_STATUS("(u=" << nodes_to_update.size() << ", d=" << std::setprecision(3) << (double(_graph.get_n_edges()) / double(_nodes.size() - _working_nodes_begin)) << ") ");
			RITCH_CONTINUE_STATUS("-");

			//
			// insert/merge new shortcut edges; also, remove merged edges from cache (otherwise computation will be wrong)
			//

			for ( auto& thread_data_ptr : _local_thread_data_ptrs )
			{
				if ( ! thread_data_ptr ) continue;

				for ( auto& edge_descr : thread_data_ptr->_edges_to_insert )
				{
					assert( ! edge_descr._time_dist.is_infty() );
					FwEdgeId e = _graph.find_forward_edge_id(edge_descr._source_id, edge_descr._target_id);

					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockEdgeDescr = edge_descr._block;

if ( edge_descr._time_dist == TimeDist() )
{
	std::cout << "X X X X X X X X " << std::endl;
}
					if ( e == FwEdgeId::INVALID() )
					{
						FwEdgeId e_new = std::get<0>(_graph.insert_edge(edge_descr._source_id, edge_descr._target_id));
						store_single_middle_node(edge_descr._source_id, edge_descr._target_id, edge_descr._middle_id);
						_graph.set_time_dist(e_new, std::move(edge_descr._time_dist));
						_graph.set_n_original_edges(e_new, edge_descr._n_original_edges);
						update_edge_block(e_new, edge_descr._source_id, edge_descr._target_id, blockEdgeDescr);
					}
					else
					{
						std::vector<std::pair<int32_t, uint32_t>> old_middle_node_info = obtain_middle_nodes(edge_descr._source_id, edge_descr._target_id);

						std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockE = get_edge_block(edge_descr._source_id, edge_descr._target_id); //original edgeblock
						std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockNew = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(); //new edgeblock

						if ( old_middle_node_info.empty() )
							old_middle_node_info.push_back(std::make_pair(TimeDist::PERIOD(), NodeId::INVALID()));

						auto merged_middle_node_info =
							merge_middle_nodes(
								_graph.has_time_dist(e) ? _graph.get_time_dist(e) : _graph.create_time_dist(e, blockE),
								edge_descr._time_dist, old_middle_node_info, edge_descr._middle_id, blockE, blockEdgeDescr);

						assert(!merged_middle_node_info.empty());

						if ( merged_middle_node_info.size() > 1 || merged_middle_node_info.front().second != NodeId::INVALID() )
							store_multiple_middle_nodes(edge_descr._source_id, edge_descr._target_id, merged_middle_node_info);

						/* merge writes info to blockNew */
						TimeDist merged_time_dist = merge(
							_graph.has_time_dist(e) ? _graph.get_time_dist(e) : _graph.create_time_dist(e, blockE), edge_descr._time_dist,
							blockE, blockEdgeDescr, blockNew, true);

						_graph.set_time_dist(e, std::move(merged_time_dist));
						_graph.set_n_original_edges(e, std::max(_graph.get_n_original_edges(e), edge_descr._n_original_edges));
						/* store updated info in list of edgeblock (kept completely out of timedist) */
						update_edge_block(e, edge_descr._source_id, edge_descr._target_id, blockNew);

						_witness_cache.remove(NodeId::INVALID(), edge_descr._source_id, edge_descr._target_id);
						_witness_cache.remove(edge_descr._source_id, edge_descr._target_id, NodeId::INVALID());

					}
				}
				
				thread_data_ptr->_edges_to_insert.clear();
				thread_data_ptr->_edges_to_insert.shrink_to_fit();
			}

			assert(_graph.dbg_graph_is_ok());

			RITCH_CONTINUE_STATUS(" ");

			//
			// update cost of nodes adjacent to contracted nodes
			//

			#pragma omp parallel
			{
				#pragma omp for schedule(dynamic)
				for (int i = 0; i < (int)nodes_to_update.size(); ++i)
				{
					NodeId x = nodes_to_update[i];
					simulate_contraction_and_update_cost_of_node(x);
				}
			}

			for ( auto& thread_data_ptr : _local_thread_data_ptrs )
			{
				if ( ! thread_data_ptr ) continue;

				for ( size_t i = 0 ; i != thread_data_ptr->_witnesses_to_insert_to_cache.size() ; ++i )
				{
					_witness_cache.insert(
						thread_data_ptr->_middle_nodes_of_witnesses[i],
						thread_data_ptr->_witnesses_to_insert_to_cache[i]);
				}

				thread_data_ptr->_witnesses_to_insert_to_cache.clear();
				thread_data_ptr->_witnesses_to_insert_to_cache.shrink_to_fit();

				thread_data_ptr->_middle_nodes_of_witnesses.clear();
				thread_data_ptr->_middle_nodes_of_witnesses.shrink_to_fit();
			}

			//
			// mark all nodes in the current independent set as contracted
			//

			_working_nodes_begin = _working_nodes_end;
			_witness_cache.shrink_to_fit();

/*
			if ( _graph.get_n_edges() < _graph.get_capacity() / 2 && _graph.get_n_edges() >= 10000 )
			{
				RITCH_CONTINUE_STATUS(" ***** ");
				_graph.compactify(); // < NOT IMPLEMENTED YET
			}
*/
		}

		RITCH_CONTINUE_STATUS("finished\n");

		auto t7 = util::time_stamp();

		RITCH_STATUS("Write node level information to RITCH_SHORTCUTS file... ");
		_ritch_shortcuts_output_file.write_level_info(_node_to_level);
		RITCH_CONTINUE_STATUS("OK\n");
		RITCH_STATUS("Finalize RITCH_SHORTCUTS file... ");
		_ritch_shortcuts_output_file.close();
		RITCH_CONTINUE_STATUS("OK\n");
		RITCH_STATUS("RITCH_SHORTCUTS file is '" << _ritch_shortcuts_output_file.get_file_name() << "'\n");

		RITCH_STATUS("Finalize LAYERS file... ");
		_layers_output_file.write_level_info(_node_to_level);
		_layers_output_file.close();
		RITCH_CONTINUE_STATUS("OK\n");

		auto t8 = util::time_stamp();
//		total_io_time += util::get_duration_in_seconds(t7, t8);

		auto t_end = util::time_stamp();
		double total_ordering_time = util::get_duration_in_seconds(t_begin, t_end);

		RITCH_STATUS("Ordering took " << total_ordering_time << " sec in total.\n");
//		RITCH_STATUS("Writing hierarchy to file took " << total_io_time << " sec in total.\n");
//		RITCH_STATUS("Ordering took " << (total_ordering_time - total_io_time) << " sec without writing to file.\n");
		
		std::cout << std::endl << std::endl << std::endl;
	}

	/* list of <boundary_of_segment, starttime_block, endtime_block, freeflow_traveltime>*/
	std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> get_edge_block(NodeId source, NodeId target) {
		auto res = _edge_block_info.find(std::make_pair(source,target));
		if (res == _edge_block_info.end()) {
			assert(!_graph.has_time_dist(_graph.find_forward_edge_id(source, target)));
			_edge_block_info[std::make_pair(source, target)] = _graph.get_edge_block(source, target);
			return _edge_block_info[std::make_pair(source, target)];
		}
		else {
			return res->second;
		}

	}

	bool has_edge_block(NodeId source, NodeId target) {
		auto res = _edge_block_info.find(std::make_pair(source, target));
		return (res == _edge_block_info.end());
	}

	/* update new block information to the map of all blocks*/
	void  update_edge_block(FwEdgeId e, NodeId source, NodeId target, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block) {
		for (int i = 0; i < block.size(); i++) {
			RITCH_ASSERT(std::get<3>(block[i]) <= _graph.eval_time(e,std::get<0>(block[i])));
		}
		_edge_block_info[std::make_pair(source,target)] = block;
	}

};

}

#endif // RITCH_PREPRO_ORDERING_H_
