#ifndef RITCH_WITNESS_SEARCH_H_
#define RITCH_WITNESS_SEARCH_H_

#include <cstdint>
#include <algorithm>
#include <iterator>
#include <limits>
#include <tuple>
#include <unordered_map>
#include <vector>

#include <boost/heap/pairing_heap.hpp>

#include "ritch/datastr/time_dist.h"
#include "ritch/prepro/dynamic_graph.h"
#include "ritch/prepro/edge_descr.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

class WitnessSearch
{

private:

	typedef DynamicGraph Graph;
	typedef Graph::BwEdgeId BwEdgeId;
	typedef Graph::FwEdgeId FwEdgeId;
	typedef Graph::NodeId NodeId;

	RITCH_DEFINE_STRONG_ID(SearchNodeId)

public:

	static constexpr uint32_t UNDECIDED = 0;
	static constexpr uint32_t NECESSARY = 1;
	static constexpr uint32_t NOT_NECESSARY = 2;

	static constexpr bool DO_AVOID_MIDDLE_NODE = true;
	static constexpr bool DO_NOT_AVOID_MIDDLE_NODE = false;

	typedef uint32_t ShortcutStatus;
	typedef Graph::TimeDist TimeDist;

private:

	struct SearchGraphEdge
	{
		FwEdgeId _edge_id;
		SearchNodeId _search_node_id;

		SearchGraphEdge()
		: _edge_id(FwEdgeId::INVALID()),
		  _search_node_id(std::numeric_limits<SearchNodeId>::max())
		{}

		SearchGraphEdge(const FwEdgeId& e, const SearchNodeId& id)
		: _edge_id(e), _search_node_id(id)
		{}
	};
	
	struct Interval
	{
	private:
		int32_t _lower;
		int32_t _upper;

	public:
		Interval()
		: Interval(std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max())
		{}

		Interval(const int32_t lower, const int32_t upper)
		: _lower(lower), _upper(upper)
		{}

		int32_t get_lower() const { return _lower; }
		int32_t get_upper() const { return _upper; }

		bool is_infty() const noexcept
		{
			assert(RITCH_IMPLIES(_lower == std::numeric_limits<int32_t>::max(), _upper == std::numeric_limits<int32_t>::max()));
			return _lower == std::numeric_limits<int32_t>::max();
		}

		friend bool operator== (const Interval& lhs, const Interval& rhs) noexcept
		{
			return lhs._lower == rhs._lower && lhs._upper == rhs._upper;
		}

		friend bool operator!= (const Interval& lhs, const Interval& rhs) noexcept
		{
			return !(lhs == rhs);
		}
	};

	template <typename HeapHandleT>
	struct SearchNode_tmpl
	{
		HeapHandleT _heap_handle;
		Interval _interval;
		TimeDist _time_dist;
		NodeId _node_id;
		uint32_t _enqueued : 1;
		uint32_t _visited : 1;
		uint32_t _n_hops : 8;

		std::pair<SearchGraphEdge,SearchGraphEdge> _predecessors;
		std::vector<SearchGraphEdge> _outgoing;

		SearchNode_tmpl(const SearchNode_tmpl&) = default;
		SearchNode_tmpl& operator=(const SearchNode_tmpl&) = default;

		SearchNode_tmpl(SearchNode_tmpl&&) = default;
		SearchNode_tmpl& operator=(SearchNode_tmpl&&) = default;

		SearchNode_tmpl()
		: _heap_handle(),
			_node_id(NodeId::INVALID()),
			_enqueued(0),
			_visited(0),
			_n_hops(0),
			_interval(),
			_predecessors(), _outgoing()
		{}
	};

	class SearchContext
	{

	private:

		struct HeapItem
		{
			SearchNodeId _search_node_id;
			int32_t _priority;

			HeapItem(const SearchNodeId search_node_id, const int32_t priority)
				: _search_node_id(search_node_id), _priority(priority)
			{}

			friend inline bool operator< (const HeapItem& lhs, const HeapItem& rhs)
			{
				return rhs._priority < lhs._priority;
			}
		};

		struct NodeIdHash
		{
			size_t operator() (NodeId node_id) const noexcept
			{
				return std::hash<uint32_t>()(uint32_t(node_id));
			}
		};

		typedef boost::heap::pairing_heap<HeapItem> HeapImpl;
		typedef HeapImpl::handle_type HeapHandle;
		typedef std::unordered_map<NodeId, SearchNodeId, NodeIdHash> HashTable;

		HeapImpl _heap;
		HashTable _hash_table;
		std::vector<SearchNode_tmpl<HeapHandle>> _search_nodes;

	public:

		typedef SearchNode_tmpl<HeapHandle> SearchNode;

		SearchContext()
			: _heap(), _hash_table(), _search_nodes()
		{}

		bool pq_empty() const
		{
			return _heap.empty();
		}

		bool pq_contains(const NodeId& node_iterator) const
		{
			assert( _hash_table.count(node_iterator) < 2 );

			const auto it = _hash_table.find(node_iterator);
			if ( it == _hash_table.end() ) return false;

			return _search_nodes[it->second]._enqueued != 0;
		}

		bool pq_contains(const SearchNode& search_node) const
		{
			return search_node._enqueued != 0;
		}

		bool reached(const NodeId& node_iterator) const
		{
			return _hash_table.find(node_iterator) != _hash_table.end();
		}

		const SearchNode& get_search_node(const NodeId& node_iterator) const
		{
			assert( _hash_table.find(node_iterator) != _hash_table.end() );
			return _search_nodes[_hash_table.find(node_iterator)->second];
		}

		SearchNode& get_search_node(const NodeId& node_iterator)
		{
			//assert( _hash_table.find(node_iterator) != _hash_table.end() );
			return _search_nodes[_hash_table.find(node_iterator)->second];
		}

		SearchNode& get_search_node_from_id(const SearchNodeId& search_node_id)
		{
			//assert( search_node_id < _search_nodes.size() );
			return _search_nodes[search_node_id];
		}

		const SearchNode& get_search_node_from_id(const SearchNodeId& search_node_id) const
		{
			//assert(search_node_id < _search_nodes.size());
			return _search_nodes[search_node_id];
		}

		SearchNodeId get_search_node_id(const SearchNode& search_node) const
		{
			assert( reached(search_node._node_id) );

			const SearchNodeId search_node_id = SearchNodeId(std::distance(&(*(_search_nodes.begin())), &search_node));
			return search_node_id;
		}

		const SearchNode& get_min() const
		{
			assert( ! pq_empty() );
			return _search_nodes[_heap.top()._search_node_id];
		}

		const int32_t& get_min_priority() const
		{
			assert( ! pq_empty() );
			return _heap.top()._priority;
		}

		SearchNode& insert(const NodeId& node_iterator, const int32_t& priority)
		{
			assert( ! reached(node_iterator) );
			assert( ! pq_contains(node_iterator) );

			const SearchNodeId search_node_id(_search_nodes.size());

			_search_nodes.emplace_back();
			const HeapHandle heap_handle = _heap.push(HeapItem(search_node_id, priority));

			_hash_table[node_iterator] = search_node_id;
			_search_nodes.back()._node_id = node_iterator;
			_search_nodes.back()._enqueued = 1;
			_search_nodes.back()._heap_handle = heap_handle;

			return _search_nodes.back();
		}

		void pq_re_insert(SearchNode& search_node, const int32_t& priority)
		{
			assert( reached(search_node._node_id) );
			assert( ! pq_contains(search_node._node_id) );

			const SearchNodeId search_node_id = get_search_node_id(search_node);
			const HeapHandle heap_handle = _heap.push(HeapItem(search_node_id, priority));

			search_node._heap_handle = heap_handle;
			search_node._enqueued = 1;
		}

		void pq_decrease(SearchNode& search_node, const int32_t& priority)
		{
			assert( reached(search_node._node_id) );
			assert( pq_contains(search_node._node_id) );

			const SearchNodeId search_node_id = get_search_node_id(search_node);
			const HeapHandle heap_handle = search_node._heap_handle;

			assert( (*heap_handle)._priority >= priority );
			_heap.update(heap_handle, HeapItem(search_node_id, priority));
		}

		void pq_update(SearchNode& search_node, const int32_t& priority)
		{
			assert(reached(search_node._node_id));
			assert(pq_contains(search_node._node_id));

			const SearchNodeId search_node_id = get_search_node_id(search_node);
			const HeapHandle heap_handle = search_node._heap_handle;

			_heap.update(heap_handle, HeapItem(search_node_id, priority));
		}

		SearchNode& pq_delete_min()
		{
			assert( ! pq_empty() );

			SearchNode& search_node = _search_nodes[_heap.top()._search_node_id];
			assert( _hash_table.find(search_node._node_id) != _hash_table.end() );
			_heap.pop();

			search_node._enqueued = 0;
			return search_node;
		}

		int32_t pq_get_priority(const SearchNode& search_node) const
		{
			assert(reached(search_node._node_id));
			assert(pq_contains(search_node._node_id));
			return (*(search_node._heap_handle))._priority;
		}

		void clear_pq()
		{
			for ( const auto& heap_item : _heap )
				_search_nodes[heap_item._search_node_id]._enqueued = 0;

			_heap.clear();
		}

		void clear_all()
		{
			_heap.clear();
			_hash_table.clear();
			_search_nodes.clear();
		}
	};

	typedef SearchContext::SearchNode SearchNode;

	const Graph* _graph;
	SearchContext _context;

	void interval_search(NodeId start_id, NodeId middle_id, NodeId destination_id, bool avoid_middle_node = false, size_t hop_limit = 16)
	{
		SearchNode& s = _context.insert(start_id, 0);
		s._n_hops = 0;
		s._interval = Interval(0, 0);

		while ( ! _context.pq_empty() )
		{
			SearchNode& u = _context.pq_delete_min();

			if ( u._node_id == destination_id )
				return;

			if ( u._n_hops >= hop_limit )
				continue;

			const NodeId u_id = u._node_id;
			const int32_t u_lower = u._interval.get_lower();
			const int32_t u_upper = u._interval.get_upper();
			const uint32_t u_hops = u._n_hops;
			const SearchNodeId u_sid = _context.get_search_node_id(u);

			for ( auto e = _graph->out_edges_begin(u_id) ; e != _graph->out_edges_end(u_id) ; ++e )
			{
				const NodeId v_id = _graph->get_target(e);
				
				if ( avoid_middle_node )
					if ( v_id == middle_id )
						continue;

				auto min_max_e = _graph->get_min_max_time(e);
				const int32_t min_e = std::get<0>(min_max_e);
				const int32_t max_e = std::get<1>(min_max_e);

				const Interval interval_v_new(u_lower + min_e, u_upper + max_e);

				if ( ! _context.reached(v_id) )
				{
					SearchNode& v = _context.insert(v_id, interval_v_new.get_lower());

					v._n_hops = u_hops + 1;
					v._interval = interval_v_new;
					v._predecessors.first = SearchGraphEdge(e, u_sid);
					v._predecessors.second = SearchGraphEdge(e, u_sid);
				}
				else
				{
					assert( _context.reached(v_id) );

					SearchNode& v = _context.get_search_node(v_id);
					assert(!v._interval.is_infty());

					if (
						interval_v_new.get_lower() >= v._interval.get_lower() &&
						interval_v_new.get_upper() >= v._interval.get_upper()
					)
						continue;

					if ( interval_v_new.get_upper() < v._interval.get_upper() )
						v._predecessors.first = SearchGraphEdge(e, u_sid);

					if ( interval_v_new.get_lower() < v._interval.get_lower() )
						v._predecessors.second = SearchGraphEdge(e, u_sid);

					v._n_hops = std::max(u_hops + 1, v._n_hops);

					 const Interval interval_merged(
								std::min(v._interval.get_lower(), interval_v_new.get_lower()),
								std::min(v._interval.get_upper(), interval_v_new.get_upper()));

					v._interval = interval_merged;
					if ( ! _context.pq_contains(v) )
						_context.pq_re_insert(v, interval_merged.get_lower());
					else
						_context.pq_decrease(v, interval_merged.get_lower());
				}
			}
		}

		return;
	}

	void copy_predecessor_graph(const SearchNodeId start_search_id, const SearchNodeId dest_search_id)
	{
		std::vector<SearchNodeId> visited_nodes;
		std::queue<SearchNodeId> Q;

		SearchNode& s = _context.get_search_node_from_id(start_search_id);
		SearchNode& d = _context.get_search_node_from_id(dest_search_id);

		d._visited = 1;
		Q.push(dest_search_id);

		while ( ! Q.empty() )
		{
			const SearchNodeId u_search_id = Q.front();
			Q.pop();

			SearchNode& u = _context.get_search_node_from_id(u_search_id);
			if ( u._predecessors.first._edge_id == u._predecessors.second._edge_id )
				u._predecessors.second._edge_id = FwEdgeId::INVALID();

			for ( const auto& pred : { u._predecessors.first, u._predecessors.second } )
			{
				const FwEdgeId& e_uv = pred._edge_id;
				if ( e_uv == FwEdgeId::INVALID() ) continue;

				const SearchNodeId& v_search_id = pred._search_node_id;
				SearchNode& v = _context.get_search_node_from_id(v_search_id);
				v._outgoing.emplace_back(e_uv, u_search_id);

				if ( v._visited != 0 ) continue;

				v._visited = 1;
				visited_nodes.push_back(v_search_id);
				Q.push(v_search_id);
			}
		}

	}

	void profile_search(NodeId start_id, const NodeId middle_id, NodeId destination_id, bool avoid_middle_node = false, size_t hop_limit = 16)
	{
		_context.clear_pq();

		SearchNode& s = _context.get_search_node(start_id);
		s._time_dist = TimeDist(0);
		_context.pq_re_insert(s, s._interval.get_lower());

		const SearchNode& d = _context.get_search_node(destination_id);

		while ( ! _context.pq_empty() )
		{
//			if ( ! d._time_dist.is_infty() )
//				if ( _context.get_min_priority() > d._time_dist.get_max_time() )
//					return;

			SearchNode& u = _context.pq_delete_min();
			if ( u._node_id == destination_id ) return;

			for ( const auto& pred : u._outgoing )
			{
				FwEdgeId e = pred._edge_id;
				if ( e == FwEdgeId::INVALID() ) continue;

				SearchNode& v = _context.get_search_node_from_id(pred._search_node_id);
				assert(!v._interval.is_infty());

				if ( avoid_middle_node )
					if ( v._node_id == middle_id )
						continue;

				const int32_t min_time_e = _graph->get_min_time(e);
				const int32_t min_time_u = u._time_dist.get_min_time();
				int32_t max_time_v = std::numeric_limits<int32_t>::max();

				if ( ! v._time_dist.is_infty() )
				{
					max_time_v = v._time_dist.get_max_time();

					if ( min_time_u + min_time_e >= max_time_v ) 
						continue;

					if ( min_time_u + min_time_e + v._interval.get_lower() > max_time_v + v._interval.get_upper() )
						continue;
				}

				std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockDummyA = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(0);
				std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockDummyB = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(0);
				std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockDummy = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(0);


				TimeDist f_v_new = link(_graph->has_time_dist(e) ? _graph->get_time_dist(e) : _graph->create_time_dist(e, blockDummyA), u._time_dist, blockDummyA, blockDummyB, blockDummy, false);
				const int32_t f_v_new_min_time = f_v_new.get_min_time();

				if ( v._time_dist.is_infty() )
				{
					_context.pq_re_insert(v, f_v_new_min_time + v._interval.get_lower());
					v._time_dist = std::move(f_v_new);

				}
				else
				{
					if ( f_v_new_min_time >= max_time_v ) continue;
					if ( ! f_v_new.time_undercuts_strictly(v._time_dist) ) continue;

					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockDummyX = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(0);
					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockDummyY = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(0);

					TimeDist f_v_merge = merge(f_v_new, v._time_dist, blockDummy, blockDummyX, blockDummyY, false);
					int32_t priority_v = f_v_merge.get_min_time();
					v._time_dist = std::move(f_v_merge);

					if ( ! _context.pq_contains(v) )
						_context.pq_re_insert(v, priority_v);
					else
						_context.pq_decrease(v, priority_v);
				}
			}
		}

		return;
	}

public:

	WitnessSearch(const Graph* graph)
	: _graph(graph), _context()
	{}

	ShortcutStatus run(NodeId source_id, BwEdgeId e_in, FwEdgeId e_in_fw, std::vector<std::tuple<int32_t,int32_t,int32_t,int32_t>> block_in, NodeId middle_id, FwEdgeId e_out, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block_out, NodeId target_id, EdgeDescr& insert_edge)
	{
		_context.clear_all();

		//
		// forward interval search
		//

		interval_search(source_id, middle_id, target_id);
		if ( ! _context.reached(target_id) ) return NECESSARY;

		const SearchNode& target = _context.get_search_node(target_id);
		const Interval& interval = target._interval;

		if ( interval.is_infty() )
			return NECESSARY;

		int32_t min_time_in;
		int32_t max_time_in;
		std::tie(min_time_in, max_time_in) = _graph->get_min_max_time(e_in_fw);

		int32_t min_time_out;
		int32_t max_time_out;
		std::tie(min_time_out, max_time_out) = _graph->get_min_max_time(e_out);

//		if ( interval.get_upper() <= min_time_in + min_time_out )
		if ( interval.get_upper() < min_time_in + min_time_out )
			return NOT_NECESSARY;

//		if ( interval.get_lower() > max_time_in + max_time_out )
		if ( interval.get_lower() >= max_time_in + max_time_out )
			return NECESSARY;

		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> blockRes = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();

		insert_edge._time_dist = link(
			_graph->has_time_dist(e_out) ? _graph->get_time_dist(e_out) : _graph->create_time_dist(e_out, block_out),
			_graph->has_time_dist(e_in_fw) ? _graph->get_time_dist(e_in_fw) : _graph->create_time_dist(e_in_fw, block_in),
			block_out, block_in, blockRes, false);
		
		insert_edge._block = std::move(blockRes);

		const TimeDist& f_shortcut = insert_edge._time_dist;

		if ( interval.get_upper() < f_shortcut.get_min_time() )
			return NOT_NECESSARY;

		//
		// forward profile search in thinned predecessor graph of backward interval search
		//
	
		const SearchNode& source = _context.get_search_node(source_id);
		copy_predecessor_graph(_context.get_search_node_id(source), _context.get_search_node_id(target));
		profile_search(source_id, middle_id, target_id);
		const TimeDist& f_target = target._time_dist;

		if ( f_target.is_infty() )
			return NECESSARY;

		if ( f_shortcut.get_min_time() > f_target.get_max_time() )
			return NOT_NECESSARY;

		if ( f_shortcut.time_undercuts_weakly(f_target ))
			return NECESSARY;

		return NOT_NECESSARY;
	}
};

}

#endif // RITCH_WITNESS_SEARCH_H_
