#ifndef RITCH_FLAT_QUERYING_H_
#define RITCH_FLAT_QUERYING_H_

#include <cassert>
#include <cstdint>
#include <limits>
#include <vector>

#include "ritch/querying/flat_search_graph.h"

#include <boost/heap/pairing_heap.hpp>

namespace ritch
{

class FlatQuerying
{

public:

	typedef FlatSearchGraph Graph;
	typedef Graph::NodeId NodeId;
	typedef Graph::EdgeId EdgeId;

	class Path
	{
		// TODO
	};

private:

	typedef uint32_t SearchNodeId;
	static constexpr SearchNodeId INVALID_SEARCH_NODE_ID() { return std::numeric_limits<uint32_t>::max(); };

	template <typename HeapHandleT>
	struct SearchNode_tmpl
	{
		HeapHandleT _heap_handle;
		int32_t _arrival_time;
		uint32_t _distance : 31;
		uint32_t _enqueued : 1;
		NodeId _node_id;
		SearchNodeId _pred_search_node_id;

		SearchNode_tmpl(const SearchNode_tmpl&) = default;
		SearchNode_tmpl& operator=(const SearchNode_tmpl&) = default;

		SearchNode_tmpl(SearchNode_tmpl&&) = default;
		SearchNode_tmpl& operator=(SearchNode_tmpl&&) = default;

		SearchNode_tmpl()
		: _heap_handle(),
		  _arrival_time(std::numeric_limits<int32_t>::max()),
		  _distance(0),
		  _node_id(Graph::INVALID_NODE_ID()),
		  _enqueued(0),
		  _pred_search_node_id(INVALID_SEARCH_NODE_ID())
		{}
	};

	class SearchContext
	{
	private:

		struct HeapItem
		{
			SearchNodeId _search_node_id;
			int32_t _priority;

			HeapItem(SearchNodeId search_node_id, int32_t priority)
			: _search_node_id(search_node_id), _priority(priority)
			{}

			friend inline bool operator< (const HeapItem& lhs, const HeapItem& rhs)
			{
				return rhs._priority < lhs._priority;
			}
		};

		typedef boost::heap::pairing_heap<HeapItem> HeapImpl;
		typedef HeapImpl::handle_type HeapHandle;

		HeapImpl _heap;
		std::vector<SearchNodeId> _table;
		std::vector<SearchNode_tmpl<HeapHandle>> _search_nodes;

	public:

		typedef SearchNode_tmpl<HeapHandle> SearchNode;

		SearchContext(size_t n_nodes)
		: _heap(), _table(n_nodes, INVALID_SEARCH_NODE_ID()), _search_nodes()
		{}

		bool pq_empty() const
		{
			return _heap.empty();
		}

		bool pq_contains(const NodeId node_id) const
		{
			if ( _table[node_id] == INVALID_SEARCH_NODE_ID() )
				return false;

			SearchNodeId search_node_id = _table[node_id];
			assert(search_node_id < _search_nodes.size());
			return _search_nodes[search_node_id]._enqueued == 1;
		}

		bool pq_contains(const SearchNode& search_node) const
		{
			return search_node._enqueued == 1;
		}

		bool reached(const NodeId node_id) const
		{
			return _table[node_id] != INVALID_SEARCH_NODE_ID();
		}

		const SearchNode& get_search_node(const NodeId node_id) const
		{
			assert(reached(node_id));
			return _search_nodes[_table[node_id]];
		}

		SearchNode& get_search_node(const NodeId node_id)
		{
			assert(reached(node_id));
			return _search_nodes[_table[node_id]];
		}

		SearchNode& get_search_node_from_id(const SearchNodeId search_node_id)
		{
			assert(search_node_id < _search_nodes.size());
			return _search_nodes[search_node_id];
		}

		const SearchNode& get_search_node_from_id(const SearchNodeId search_node_id) const
		{
			assert(search_node_id < _search_nodes.size());
			return _search_nodes[search_node_id];
		}

		SearchNodeId get_search_node_id(const SearchNode& search_node) const
		{
			return SearchNodeId(std::distance(&(_search_nodes.front()), &search_node));
		}

		const SearchNode& pq_min_search_node() const
		{
			assert(! pq_empty());
			return _search_nodes[_heap.top()._search_node_id];
		}

		const int32_t pq_min_priority() const
		{
			assert(! pq_empty());
			return _heap.top()._priority;
		}

		SearchNode& pq_insert(const NodeId node_id, const int32_t priority)
		{
			assert(! reached(node_id));
			assert(! pq_contains(node_id));
			const SearchNodeId search_node_id = SearchNodeId(_search_nodes.size());

			_search_nodes.emplace_back();
			const HeapHandle heap_handle = _heap.push(HeapItem(search_node_id, priority));

			_table[node_id] = search_node_id;
			_search_nodes.back()._node_id = node_id;
			_search_nodes.back()._enqueued = 1;
			_search_nodes.back()._heap_handle = heap_handle;

			return _search_nodes.back();
		}

		void pq_re_insert(SearchNode& search_node, const int32_t priority)
		{
			assert(reached(search_node._node_id));
			assert(! pq_contains(search_node));

			const SearchNodeId search_node_id = get_search_node_id(search_node);
			const HeapHandle heap_handle = _heap.push(HeapItem(search_node_id, priority));

			search_node._heap_handle = heap_handle;
			search_node._enqueued = 1;
		}

		void pq_update_priority(SearchNode& search_node, const int32_t priority)
		{
			assert(reached(search_node._node_id));
			assert(pq_contains(search_node._node_id));

			const SearchNodeId search_node_id = get_search_node_id(search_node);
			const HeapHandle heap_handle = search_node._heap_handle;
			_heap.update(heap_handle, HeapItem(search_node_id, priority));
		}

		SearchNode& pq_delete_min()
		{
			assert(! pq_empty());

			SearchNode& search_node = _search_nodes[_heap.top()._search_node_id];
			assert(_table[search_node._node_id] != INVALID_SEARCH_NODE_ID());
			_heap.pop();

			search_node._enqueued = 0;
			return search_node;
		}

		void clear_pq()
		{
			for ( const auto& heap_item : _heap )
				_search_nodes[heap_item._search_node_id]._enqueued = 0;

			_heap.clear();
		}

		void clear_all()
		{
			_heap.clear();
			_table.assign(_table.size(), Graph::INVALID_NODE_ID());
			_search_nodes.clear();
		}
	};

	typedef SearchContext::SearchNode SearchNode;

	void search()
	{
		SearchNode& s = _context.pq_insert(_start_node_id, _departure_time);
		s._arrival_time = _departure_time;

		while ( ! _context.pq_empty() )
		{
			if ( _context.pq_min_priority() > _arrival_time )
				return;

			const SearchNode& u = _context.pq_delete_min();
			const NodeId u_id = u._node_id;
			SearchNodeId u_search_node_id = _context.get_search_node_id(u);
			int32_t u_arrival_time = u._arrival_time;
			uint32_t u_distance = u._distance;

			++_dijkstra_rank;

			if ( u_id == _dest_node_id )
			{
				assert(_arrival_time >= u_arrival_time);
				assert(_travel_distance >= u_distance);

				_arrival_time = u_arrival_time;
				_travel_distance = u_distance;
			}

			for ( EdgeId e = _graph->edges_begin(u_id) ; e != _graph->edges_end(u_id) ; ++e )
			{
				const NodeId v_id = _graph->get_target(e);
				const int32_t arrival_time_v_new = u_arrival_time + _graph->get_default_travel_time(e) + _graph->eval_waiting_time(e, u_arrival_time);
				const uint32_t distance_v_new = u_distance + _graph->get_distance(e);

				if ( ! _context.reached(v_id) )
				{
					SearchNode& v = _context.pq_insert(v_id, arrival_time_v_new);
					v._arrival_time = arrival_time_v_new;
					v._pred_search_node_id = u_search_node_id;
					v._distance = distance_v_new;
				}
				else
				{
					SearchNode& v = _context.get_search_node(v_id);

					if ( arrival_time_v_new < v._arrival_time   || arrival_time_v_new == v._arrival_time && distance_v_new < v._distance )
					{
						v._arrival_time = arrival_time_v_new;
						v._pred_search_node_id = u_search_node_id;
						v._distance = distance_v_new;

						if ( _context.pq_contains(v) )
							_context.pq_update_priority(v, arrival_time_v_new);
						else
							_context.pq_re_insert(v, arrival_time_v_new);
					}
				}
			}
		}
	}

	Graph const * const _graph;

	NodeId _start_node_id;
	NodeId _dest_node_id;
	int32_t _departure_time;
	int32_t _arrival_time;
	uint32_t _travel_distance;
	uint32_t _dijkstra_rank;
	SearchContext _context;

public:

	class FlatQuerying(const FlatSearchGraph& graph)
	: _graph(&graph),
	  _start_node_id(Graph::INVALID_NODE_ID()),
	  _dest_node_id(Graph::INVALID_NODE_ID()),
	  _departure_time(std::numeric_limits<int32_t>::max()),
	  _arrival_time(std::numeric_limits<int32_t>::max()),
	  _travel_distance(std::numeric_limits<uint32_t>::max()),
	  _dijkstra_rank(0),
	  _context(_graph->get_n_nodes())
	{}

	void run_one_to_one(NodeId start_node_id, NodeId dest_node_id, int32_t departure_time) noexcept
	{
		_context.clear_all();

		_start_node_id = start_node_id;
		_dest_node_id = dest_node_id;
		_departure_time = departure_time;
		_arrival_time = std::numeric_limits<int32_t>::max();
		_travel_distance = std::numeric_limits<uint32_t>::max();
		_dijkstra_rank = 0;

		search();
	}

	uint32_t get_travel_distance() const noexcept
	{
		return _travel_distance;
	}

	int32_t get_arrival_time() const noexcept
	{
		return _arrival_time;
	}

	uint32_t get_dijkstra_rank() const noexcept
	{
		return _dijkstra_rank;
	}

	Path get_path() const
	{
		// TODO

		throw std::runtime_error("Path extraction not implemented yet.");
	}
};

}

#endif // RITCH_FLAT_QUERYING_H_
