#ifndef RITCH_FLAT_SEARCH_GRAPH_H_
#define RITCH_FLAT_SEARCH_GRAPH_H_

#include <cassert>
#include <cstdint>
#include <limits>
#include <string>
#include <vector>

#include "ritch/datastr/blocking_profile.h"
#include "ritch/io/binary.h"
#include "ritch/io/ortec_format.h"

namespace ritch
{

class FlatSearchGraph
{

public:

	typedef uint32_t NodeId;
	typedef uint32_t EdgeId;

	static constexpr NodeId INVALID_NODE_ID() { return std::numeric_limits<NodeId>::max(); }

private:

	struct Node
	{
		EdgeId _first_edge;
	};

	struct Edge
	{
		NodeId _target;
		uint32_t _distance;
		uint32_t _blocking_info_pos : 16;
		uint32_t _road_type : 8;

		Edge()
		: _target(std::numeric_limits<NodeId>::max()),
		  _distance(std::numeric_limits<uint32_t>::max()),
		  _blocking_info_pos(0xffff),
		  _road_type(0xff)
		{}
	};

	struct BlockingDescr
	{
		int32_t _start_time;
		int32_t _duration;
	};

	const Edge& get_edge(EdgeId edge_id) const
	{
		assert(edge_id + 1 < _edges.size());
		return _edges[edge_id];
	}

	static int32_t calculate_default_travel_time(uint32_t speed, uint32_t distance) noexcept
	{
		//MT: Why this +5 ??
		return std::max(1, int32_t((36 * distance) / speed + 5) / 10);
		//return std::max(1, int32_t(((36 * distance) / speed) / 10));
	}

	int32_t _period;
	std::vector<Node> _nodes;
	std::vector<Edge> _edges;
	std::vector<uint32_t> _blocking_info;

public:

	FlatSearchGraph(const std::string& graph_directory_name, const std::string& default_blocks_file_name, int32_t period)
	: _period(period), _nodes(), _edges(), _blocking_info()
	{
		io::InputFileIterator<uint32_t> target_it = io::ortec_format::get_target_iterator(graph_directory_name);
		io::InputFileIterator<uint32_t> distance_it = io::ortec_format::get_distance_iterator(graph_directory_name);
		io::InputFileIterator<uint8_t> type_it = io::ortec_format::get_road_type_iterator(graph_directory_name);

		size_t target_file_size = target_it.get_size_in_byte();

		if ( distance_it.get_size_in_byte() != target_file_size )
			throw std::runtime_error("Files lengtelong.net and naar_net have different number of edges.");

		if ( type_it.get_size_in_byte() != target_file_size / 4 )
			throw std::runtime_error("Files type.net and naar_net have different number of edges.");

		_edges.resize(target_file_size / 4);

		std::cerr << "STATUS: Graph directory is '" << graph_directory_name + util::PATH_SEPARATOR() + "'" << std::endl;

		//
		// read target nodes
		//

		std::cerr << "STATUS: Reading 'naar.net'... " << std::flush;
		for ( size_t pos = 0 ; ! target_it.finished() ; ++pos )
			_edges[pos]._target = target_it.read();

		target_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read distances
		//

		std::cerr << "STATUS: Reading 'lengtelong.net'... " << std::flush;
		for ( size_t pos = 0 ; ! distance_it.finished() ; ++pos )
			_edges[pos]._distance = distance_it.read();

		distance_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read road types
		//

		std::cerr << "STATUS: Reading 'type.net'... " << std::flush;
		for ( size_t pos = 0 ; ! type_it.finished() ; ++pos )
			_edges[pos]._road_type = type_it.read();

		type_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read first edge information
		//

		io::InputFileIterator<uint32_t> first_edge_it = io::ortec_format::get_first_edge_iterator(graph_directory_name);
		_nodes.resize(first_edge_it.get_size_in_byte() / 4);

		std::cerr << "STATUS: Reading 'first.net'... " << std::flush;
		for ( size_t pos = 0 ; ! first_edge_it.finished() ; ++pos )
			_nodes[pos]._first_edge = first_edge_it.read();

		first_edge_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read and setup highway blockings, which includes reading country phone codes
		//

		if ( ! default_blocks_file_name.empty() )
		{
			std::vector<uint16_t> map_edge_rec_no_2_phone_code =
				io::ortec_format::read_map_edge_rec_no_2_country_phone_code(graph_directory_name);

			typedef io::hwy_blocks_format::BlockingDescr Blocking;
			std::vector<Blocking> blockings = io::hwy_blocks_format::read_blockings(default_blocks_file_name);

			for ( size_t i = blockings.size() ; i > 0 ; --i )
			{
				if ( blockings[i-1]._start_time == 0 && blockings[i-1]._end_time == 0 )
				{
					swap(blockings[i-1], blockings.back());
					blockings.pop_back();
				}
			}

			std::sort(blockings.begin(), blockings.end(),
				[](Blocking& lhs, Blocking& rhs)
				{
					if ( lhs._phone_code < rhs._phone_code )
						return true;

					if ( lhs._phone_code == rhs._phone_code )
						if ( lhs._start_time < rhs._start_time )
							return true;

					return false;
				}
			);

			size_t max_phone_code = 0;
			for ( const auto& blocking : blockings )
				max_phone_code = std::max(max_phone_code, size_t(blocking._phone_code));

			if ( max_phone_code >= size_t(std::numeric_limits<uint16_t>::max()) )
				throw std::runtime_error("Phone code too high to be encoded with 16 bits.");

			std::vector<uint16_t> map_phone_code_2_blocking_info_pos(max_phone_code + 1, 0xffff);
			uint16_t current_phone_code = std::numeric_limits<uint16_t>::max();
			uint16_t pos_n_blockings = 0;

			for ( size_t i = 0 ; i < blockings.size() ; ++i )
			{
				if ( blockings[i]._end_time < blockings[i]._start_time )
					blockings[i]._end_time += _period;

				if ( ! RITCH_IMPLIES(blockings[i]._phone_code == current_phone_code, _blocking_info.size() > 1) )
					throw std::runtime_error("Something went wrong.");

				if ( blockings[i]._phone_code != current_phone_code )
				{
					current_phone_code = blockings[i]._phone_code;

					if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
						throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

					pos_n_blockings = uint16_t(_blocking_info.size());
					map_phone_code_2_blocking_info_pos[current_phone_code] = pos_n_blockings;
					util::append(_blocking_info, {0u, 0u}); // store size and max duration
				}
				else
				{
					uint32_t last_end_time = _blocking_info[_blocking_info.size() - 2] + _blocking_info[_blocking_info.size() - 1];

					if ( uint32_t(blockings[i]._start_time) <= last_end_time )
					{
						if ( i == 0 ) throw std::runtime_error("Something went wrong.");

						uint32_t duration = blockings[i]._end_time - blockings[i-1]._start_time;
						_blocking_info.back() = std::max(_blocking_info.back(), duration);

						continue;
					}
				}

				_blocking_info.push_back(blockings[i]._start_time);
				uint32_t duration = blockings[i]._end_time - blockings[i]._start_time;
				_blocking_info.push_back(duration);

				_blocking_info[pos_n_blockings] += 2; // increase size
				_blocking_info[pos_n_blockings + 1] = std::max(_blocking_info[pos_n_blockings + 1], duration); // update max duration
			}

			if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
				throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

			for ( size_t i = 0 ; i < _edges.size() ; ++i )
			{
				_edges[i]._blocking_info_pos = 0xffff;

				uint16_t phone_code = map_edge_rec_no_2_phone_code[i];

				if ( phone_code >= map_phone_code_2_blocking_info_pos.size() )
					continue;

				if ( util::is_one_of(phone_code, { io::hwy_blocks_format::PHONE_CODE_CZ(), io::hwy_blocks_format::PHONE_CODE_SK() }) )
					if ( ! io::ortec_format::road_type_is_motorway_or_alike(_edges[i]._road_type) )
						continue;

				_edges[i]._blocking_info_pos = map_phone_code_2_blocking_info_pos[phone_code];


//				uint16_t phone_code = 43;
//				//				if ( io::ortec_format::road_type_is_motorway_or_alike(_edges[i].get_road_type()) )
//				_edges[i]._blocking_info_pos = map_phone_code_2_blocking_info_pos[phone_code];
			}
		}
		else
		{
			for ( EdgeId edge_id = 0 ; edge_id < _edges.size() ; ++edge_id )
				_edges[edge_id]._blocking_info_pos = 0xffff;
		}

		_edges.emplace_back();
		_edges.back()._blocking_info_pos = uint16_t(_blocking_info.size());
	}

	NodeId get_n_nodes() const
	{
		if ( _nodes.empty() ) return 0;
		return NodeId(_nodes.size() - 1);
	}

	EdgeId get_n_edges() const
	{
		if ( _edges.empty() ) return 0;
		return NodeId(_edges.size() - 1);
	}

	EdgeId edges_begin(NodeId node_id) const noexcept
	{
		assert( node_id + 1 < _nodes.size() );
		return _nodes[node_id]._first_edge;
	}

	EdgeId edges_end(NodeId node_id) const noexcept
	{
		assert( node_id + 1 < _nodes.size() );
		return _nodes[node_id + 1]._first_edge;
	}

	NodeId get_target(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id)._target;
	}

	uint32_t get_distance(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id)._distance;
	}

	uint8_t get_road_type(EdgeId edge_id) const noexcept
	{
		assert(edge_id + 1 < _edges.size());
		return get_edge(edge_id)._road_type;
	}

	int32_t eval_waiting_time(EdgeId edge_id, int32_t departure_time) const noexcept
	{
		const Edge& edge = get_edge(edge_id);

		if ( edge._blocking_info_pos == 0xffff )
			return 0;

		const uint32_t speed = io::ortec_format::road_type_2_default_speed(get_road_type(edge_id));
		const uint32_t distance = get_distance(edge_id);
		const int32_t default_travel_time = calculate_default_travel_time(speed, distance);

		departure_time = util::modulo(departure_time, _period);

		const uint32_t size = _blocking_info[edge._blocking_info_pos];
		const uint32_t begin = edge._blocking_info_pos + 2;
		const uint32_t end = begin + size;

		const int32_t last_start_time = _blocking_info[end - 2];
		const int32_t last_duration = _blocking_info[end - 1];
		if ( departure_time < last_start_time + last_duration - _period )
			return last_start_time + last_duration - _period - departure_time;

		for ( uint32_t i = begin ; i < end ; i += 2 )
		{
			 const int32_t& start_time = _blocking_info[i];
			 const int32_t& duration = _blocking_info[i+1];

			 if ( departure_time <= start_time -  default_travel_time)
				 return 0;
			 if ( departure_time <= start_time )
				 return duration;
			 if ( departure_time <= start_time + duration )
				 return start_time + duration - departure_time;
		}

		const int32_t first_start_time = _blocking_info[begin];
		const int32_t first_duration = _blocking_info[begin + 1];
		if ( departure_time > first_start_time + _period - default_travel_time )
			return first_duration;

		return 0;
	}

	int32_t get_default_travel_time(EdgeId edge_id) const noexcept
	{
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(get_road_type(edge_id)); //in km/h
		const uint32_t distance = get_distance(edge_id); //in meters
		return calculate_default_travel_time(speed, distance);
	}
};

}

#endif // RITCH_FLAT_SEARCH_GRAPH_H_
