#ifndef RITCH_QUERYING_PATH_H_
#define RITCH_QUERYING_PATH_H_

#include <cassert>
#include <cstdint>
#include <vector>

namespace ritch
{

template <typename NodeIdType, typename EdgeIdType, typename LocationType>
class Path_tmpl
{
public:
	typedef NodeIdType NodeId;
	typedef EdgeIdType EdgeId;
	typedef LocationType Location;

private:

	Location _start;
	Location _dest;

	int32_t _departure_time;
	int32_t _arrival_time;
	int32_t _distance;

	std::vector<NodeId> _node_ids;
	std::vector<EdgeId> _edge_ids;
	std::vector<int32_t> _arrival_times;
	std::vector<int32_t> _shift_times;
	std::vector<int32_t> _distances;

public:

	Path_tmpl(const Path_tmpl&) = default;
	Path_tmpl(Path_tmpl&&) = default;

	Path_tmpl& operator= (const Path_tmpl&) = default;
	Path_tmpl& operator= (Path_tmpl&&) = default;

	Path_tmpl()
	: _start(), _dest(),
	  _departure_time(std::numeric_limits<int32_t>::max()),
	  _arrival_time(std::numeric_limits<int32_t>::max()),
	  _distance(std::numeric_limits<int32_t>::max()),
	  _node_ids(), _edge_ids(), _arrival_times(), _distances()
	{}

	template <typename NodeIdIter, typename EdgeIdIter, typename ArrivalTimeIter, typename DistanceIter>
	Path_tmpl
	(
		const Location& start,
		const Location& dest,
		const int32_t departure_time,
		const int32_t arrival_time,
		const int32_t distance,
		NodeIdIter node_ids_begin,
		NodeIdIter node_ids_end,
		EdgeIdIter edge_ids_begin,
		EdgeIdIter edge_ids_end,
		ArrivalTimeIter arrival_times_begin,
		ArrivalTimeIter arrival_times_end,
		DistanceIter distances_begin,
		DistanceIter distances_end
	)
	: _start(start),
	  _dest(dest),
	  _departure_time(departure_time),
	  _arrival_time(arrival_time),
	  _distance(distance),
	  _node_ids(node_ids_begin, node_ids_end),
	  _edge_ids(edge_ids_begin, edge_ids_end),
	  _distances(distances_begin, distances_end)
	{
		assert(_node_ids.size() == _edge_ids.size() + 1);
		assert(_arrival_times.size() == _node_ids.size());
		assert(_distances.size() == _node_ids.size());
	}

	Path_tmpl
	(
		const Location& start,
		const Location& dest,
		int32_t departure_time,
		int32_t arrival_time,
		int32_t distance,
		std::vector<NodeId>&& node_ids,
		std::vector<EdgeId>&& edge_ids,
		std::vector<int32_t>&& arrival_times,
		std::vector<int32_t>&& distances
	)
	: _start(start),
	  _dest(dest),
	  _departure_time(departure_time),
	  _arrival_time(arrival_time),
	  _distance(distance),
	  _node_ids(std::move(node_ids)),
	  _edge_ids(std::move(edge_ids)),
	  _arrival_times(std::move(arrival_times)),
	  _distances(std::move(distances))
	{
		assert(_node_ids.size() == _edge_ids.size() + 1);
		assert(_arrival_times.size() == _node_ids.size());
		assert(_distances.size() == _node_ids.size());
	}


	Path_tmpl
	(
		const Location& start,
		const Location& dest,
		int32_t departure_time,
		int32_t arrival_time,
		int32_t distance,
		std::vector<NodeId>&& node_ids,
		std::vector<EdgeId>&& edge_ids,
		std::vector<int32_t>&& arrival_times,
		std::vector<int32_t>&& shift_times,
		std::vector<int32_t>&& distances
	)
		: _start(start),
		_dest(dest),
		_departure_time(departure_time),
		_arrival_time(arrival_time),
		_distance(distance),
		_node_ids(std::move(node_ids)),
		_edge_ids(std::move(edge_ids)),
		_arrival_times(std::move(arrival_times)),
		_shift_times(std::move(shift_times)),
		_distances(std::move(distances))
	{
		assert(_node_ids.size() == _edge_ids.size() + 1);
		assert(_arrival_times.size() == _node_ids.size());
		assert(_distances.size() == _node_ids.size());
	}

	const Location& get_start() const { return _start; }
	const Location& get_destination() const { return _dest; }
	int32_t get_departure_time() const{ return _departure_time; }
	int32_t get_arrival_time() const { return _arrival_time; }

	bool empty() const
	{
		return _node_ids.empty();
	}

	size_t get_n_nodes() const
	{
		return _node_ids.size();
	}

	NodeId get_node_id(size_t index) const
	{
		assert(index < _node_ids.size());
		return _node_ids[index];
	}

	int32_t get_arrival_time(size_t index) const
	{
		assert(index < _arrival_times.size());
		return _arrival_times[index];
	}

	int32_t get_distance(size_t index) const
	{
		assert(index < _distances.size());
		return _distances[index];
	}

	size_t get_n_edges() const
	{
		return _edge_ids.size();
	}
	
	EdgeId get_edge_id(size_t index) const
	{
		assert(index < _edge_ids.size());
		return _edge_ids[index];
	}

	std::tuple<NodeId,EdgeId,NodeId> get_edge(size_t index) const
	{
		assert(index < _edge_ids.size());
		assert(index + 1 < _node_ids.size());
		return std::make_tuple(_node_ids[index], _edge_ids[index], _node_ids[index + 1]);
	}

	void printNodes() const
	{
		int arrival = 0;
		std::cout << std::endl << "TCH  takes " << _arrival_time << " or " << _arrival_times[_arrival_times.size()-1] << " seconds" << std::endl;
		for (int i = _node_ids.size() - 1; i >= 0; i--) {
			//if (_arrival_times[i]) {
				std::cout << _node_ids[i] << ":" << _arrival_times[i] << ":" << _shift_times[i] <<",";
				//arrival = _arrival_times[i];
			//}
		}
	}

	void printPath() const
	{
		for (int i = (_node_ids.size() - 1); i >= 0 ; i--) {
			std::cout << _node_ids[i] << ",";
		}
	}

};


}

#endif
