#ifndef RITCH_QUERYING_TCH_QUERYING_H_
#define RITCH_QUERYING_TCH_QUERYING_H_

#include <stack>
#include <tuple>
#include <vector>

#include <boost/heap/pairing_heap.hpp>

#include "ritch/querying/tch_search_graph.h"
#include "ritch/querying/location.h"
#include "ritch/querying/path.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

class TchQuerying
{

public:

	typedef TchSearchGraph::NodeId NodeId;
	typedef TchSearchGraph::EdgeId EdgeId;
	typedef TchSearchGraph::TimeDist TimeDist;
	typedef Location_tmpl<NodeId, TimeDist> Location;
	typedef Path_tmpl<NodeId,EdgeId,Location> Path;

private:

	RITCH_DEFINE_STRONG_ID(SearchId)

	static constexpr uint32_t FORWARD() { return 0; };
	static constexpr uint32_t BACKWARD() { return 1; };
	static constexpr bool DO_STORE_PREDECESSORS = true;
	static constexpr bool DONT_STORE_PREDECESSORS = false;

	struct SubgraphEdge
	{
		EdgeId _edge_id;
		SearchId _search_id;

		SubgraphEdge()
		: _edge_id(EdgeId::INVALID()), _search_id(SearchId::INVALID())
		{}

		SubgraphEdge(EdgeId edge_id, SearchId search_id)
		: _edge_id(edge_id), _search_id(search_id)
		{}
	};

	struct Interval
	{
	private:
		int32_t _lower;
		int32_t _upper;

	public:
		Interval()
		: Interval(std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max())
		{}

		Interval(const int32_t lower, const int32_t upper)
			: _lower(lower), _upper(upper)
		{}

		static Interval INFTY() { return Interval(); }

		int32_t get_lower() const { return _lower; }
		int32_t get_upper() const { return _upper; }

		bool is_infty() const noexcept
		{
			assert(
				RITCH_EQUIV(
					_lower == std::numeric_limits<int32_t>::max(),
					_upper == std::numeric_limits<int32_t>::max()
				)
			);

			return _lower == std::numeric_limits<int32_t>::max();
		}

		friend bool operator== (const Interval& lhs, const Interval& rhs) noexcept
		{
			return lhs._lower == rhs._lower && lhs._upper == rhs._upper;
		}

		friend bool operator!= (const Interval& lhs, const Interval& rhs) noexcept
		{
			return !(lhs == rhs);
		}
	};

	template <typename HeapHandleT>
	struct SearchNode_tmpl
	{
		std::array<HeapHandleT, 2> _heap_handle;
		NodeId _node_id;
		uint32_t _flags;
		int32_t _time_fw;
		uint32_t _dist_fw;
		Interval _interval_bw;
		std::array<int32_t, 2> _upper_stall;
		SubgraphEdge _predecessor_fw;
		std::vector<SubgraphEdge> _predecessors_bw;

		SearchNode_tmpl(const SearchNode_tmpl&) = delete;
		SearchNode_tmpl& operator=(const SearchNode_tmpl&) = delete;

		SearchNode_tmpl(SearchNode_tmpl&&) = default;
		SearchNode_tmpl& operator=(SearchNode_tmpl&&) = default;

		SearchNode_tmpl()
		: _heap_handle(),
		  _node_id(NodeId::INVALID()),
		  _flags(0),
		  _time_fw(std::numeric_limits<int32_t>::max()),
		  _dist_fw(std::numeric_limits<uint32_t>::max()),
		  _interval_bw(Interval::INFTY()),
		  _upper_stall({std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max()}),
		  _predecessor_fw(),
		  _predecessors_bw()
		{}

		bool is_enqueued(const size_t direction) const
		{
			assert(direction == 0 || direction == 1);
			return (_flags & (1 << direction)) != 0;
		}

		void set_enqueued(bool enqueued, uint32_t direction)
		{
			assert(direction == 0 || direction == 1);

			if ( enqueued )
				_flags |= (1 << direction);
			else
				_flags &= ~(1 << direction);
		}
	};

	class SearchContext
	{
	private:

		struct HeapItem
		{
			SearchId _search_id;
			int32_t _priority;

			HeapItem(SearchId search_id, int32_t priority)
			: _search_id(search_id), _priority(priority)
			{}

			friend inline bool operator< (const HeapItem& lhs, const HeapItem& rhs) { return rhs._priority < lhs._priority; }
		};

		struct NodeIdHash
		{
			size_t operator() (NodeId node_id) const noexcept { return std::hash<uint32_t>()(uint32_t(node_id)); }
		};

		typedef boost::heap::pairing_heap<HeapItem> HeapImpl;
		typedef HeapImpl::handle_type HeapHandle;

		std::array<HeapImpl, 2> _heap;
		std::unordered_map<NodeId, SearchId, NodeIdHash> _table;
		std::vector<SearchNode_tmpl<HeapHandle>> _search_nodes;

	public:

		typedef SearchNode_tmpl<HeapHandle> SearchNode;

		SearchContext()
		: _heap(),
		  _table(),
		  _search_nodes()
		{}

		bool pq_empty(const uint32_t direction) const noexcept
		{
			assert(direction < _heap.size());
			return _heap[direction].empty();
		}

		bool pq_contains(const NodeId node_id, const uint32_t direction) const noexcept
		{
			auto it = _table.find(node_id);
			if ( it == _table.end() ) return false;

			const SearchId search_id = it->second;
			//assert(search_id < _search_nodes.size());

			return _search_nodes[search_id.the_value()].is_enqueued(direction);
		}

		bool pq_contains(const SearchNode& search_node, const uint32_t direction) const noexcept
		{
			return search_node.is_enqueued(direction);
		}

		bool touched(const NodeId node_id) const noexcept
		{
			return _table.count(node_id) > 0;
		}

		const SearchNode& get_search_node(const NodeId node_id) const noexcept
		{
			assert(_table.count(node_id) > 0);
			return _search_nodes[_table.find(node_id)->second.the_value()];
		}

		SearchNode& get_search_node(const NodeId node_id) noexcept
		{
			assert(_table.count(node_id) > 0);
			return _search_nodes[_table.find(node_id)->second.the_value()];
		}

		SearchNode& get_or_create_search_node(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table[node_id];

			if ( search_id == SearchId::INVALID() )
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
			}

			return _search_nodes[search_id.the_value()];
		}

		SearchNode& get_search_node_from_sid(const SearchId search_id)
		{
			//assert(search_id < _search_nodes.size());
			return _search_nodes[search_id.the_value()];
		}

		const SearchNode& get_search_node_from_sid(const SearchId search_id) const
		{
			//assert(search_id < _search_nodes.size());
			return _search_nodes[search_id.the_value()];
		}

		SearchId get_search_id(const SearchNode& search_node) const
		{
			//assert(touched(search_node._node_id));
			const SearchId search_id(std::distance(&(_search_nodes[0]), &search_node));
			return search_id;
		}

		const SearchNode& pq_min_search_node(const uint32_t direction) const
		{
			// assert(!pq_empty(direction));
			return _search_nodes[_heap[direction].top()._search_id.the_value()];
		}

		int32_t pq_min_priority(const uint32_t direction) const
		{
			assert(!pq_empty(direction));
			return _heap[direction].top()._priority;
		}

		void pq_insert_or_update(SearchNode& search_node, const int32_t priority, const uint32_t direction)
		{
			const SearchId search_id = get_search_id(search_node);

			if ( ! search_node.is_enqueued(direction) )
			{
				const HeapHandle heap_handle = _heap[direction].push(HeapItem(search_id, priority));
				search_node._heap_handle[direction] = heap_handle;
				search_node.set_enqueued(true, direction);
			}
			else
			{
				const HeapHandle heap_handle = search_node._heap_handle[direction];
				_heap[direction].update(heap_handle, HeapItem(search_id, priority));
			}
		}

		SearchNode& pq_delete_min(const uint32_t direction)
		{
			assert(!pq_empty(direction));

			SearchNode& search_node = _search_nodes[_heap[direction].top()._search_id.the_value()];
			assert(_table.count(search_node._node_id) > 0);
			_heap[direction].pop();

			search_node.set_enqueued(false, direction);
			return search_node;
		}

		void pq_clear(const uint32_t direction)
		{
			for ( const auto& heap_item : _heap[direction] )
				_search_nodes[heap_item._search_id.the_value()].set_enqueued(false, direction);

			_heap[direction].clear();
		}

		void clear_all()
		{
			_heap[FORWARD()].clear();
			_heap[BACKWARD()].clear();
			_table.clear();
			_search_nodes.clear();
		}
	};

	typedef SearchContext::SearchNode SearchNode;

	static uint32_t inverse(const uint32_t direction)
	{
		if ( direction == FORWARD() )
			return BACKWARD();

		return FORWARD();
	}

	void bidirectional_search(const Location& start, const Location& destination, const int32_t departure_time)
	{
		_context.clear_all();

		for ( size_t i = 0 ; i != start.size() ; ++i )
		{
			const NodeId& node_id = start.get_node_id(i);
			assert(!_context.pq_contains(node_id, FORWARD()));

			const TimeDist& time_dist = start.get_time_dist(i);
			const int32_t departure_time_on_node = time_dist.eval_time(departure_time) + departure_time;

			SearchNode& search_node = _context.get_or_create_search_node(node_id);
			_context.pq_insert_or_update(search_node, departure_time_on_node, FORWARD());
			search_node._time_fw = departure_time_on_node;
			search_node._dist_fw = 0;
		}

		for ( size_t i = 0 ; i != destination.size() ; ++i )
		{
			const NodeId& node_id = destination.get_node_id(i);
			assert(!_context.pq_contains(node_id, BACKWARD()));

			const TimeDist& time_dist = destination.get_time_dist(i);

			SearchNode& search_node = _context.get_or_create_search_node(node_id);
			_context.pq_insert_or_update(search_node, time_dist.get_min_time(), BACKWARD());
			search_node._interval_bw = Interval(time_dist.get_min_time(), time_dist.get_max_time());
		}

		_candidate_search_ids.clear();
		_upper_bound = std::numeric_limits<int32_t>::max();

		uint32_t direction = BACKWARD();
		std::array<bool, 2> finished = {false, false};

		while ( ! _context.pq_empty(FORWARD()) || ! _context.pq_empty(BACKWARD()) )
		{
			if ( finished[FORWARD()] && finished[BACKWARD()] ) break;
			if ( ! _context.pq_empty(inverse(direction)) ) direction = inverse(direction);
			if ( finished[FORWARD()] ) direction = BACKWARD();
			if ( finished[BACKWARD()] ) direction = FORWARD();

			if ( _context.pq_empty(direction) )
			{
				finished[direction] = true;
				continue;
			}

			//
			// take node with minimal label out of PQ
			//
			SearchNode& u = _context.pq_delete_min(direction);

			//
			// finish search if current minimal label exceeds upper bound of up-down-paths see so far
			//

			if ( direction == FORWARD() && u._time_fw > _upper_bound )
			{
				finished[FORWARD()] = true;
				continue;
			}
			else if ( direction == BACKWARD() && u._interval_bw.get_lower() > _upper_bound )
			{
				finished[BACKWARD()] = true;
				continue;
			}

			//
			// store candidates and update upper bound of up-down-paths seen so far
			//

			if ( u._time_fw != std::numeric_limits<int32_t>::max() && !u._interval_bw.is_infty() )
			{
				if ( u._time_fw + u._interval_bw.get_lower() <= _upper_bound )
					_candidate_search_ids.push_back(_context.get_search_id(u));

				_upper_bound = std::min(_upper_bound, u._time_fw + u._interval_bw.get_upper());
			}

			//
			// relax edges
			//

			const SearchId u_sid = _context.get_search_id(u);
			const NodeId u_id = u._node_id;
			const int32_t u_time_fw = u._time_fw;
			const uint32_t u_dist_fw = u._dist_fw;
			const Interval u_interval_bw = u._interval_bw;

			for ( EdgeId e = _graph->edges_begin(u_id) ; e != _graph->edges_end(u_id) ; ++e )
			{
				if ( direction == FORWARD() && ! _graph->is_directed_upward(e) ) continue;
				else if ( direction == BACKWARD() && ! _graph->is_directed_downward(e) ) continue;

				const NodeId v_id = _graph->get_other_node(e);
				SearchNode& v = _context.get_or_create_search_node(v_id);

				if ( direction == FORWARD() )
				{
					const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
					const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;

					if ( v_time_fw_new < v._time_fw || (v_time_fw_new == v._time_fw && v_dist_fw_new < v._dist_fw) )
					{
						v._time_fw = v_time_fw_new;
						v._dist_fw = v_dist_fw_new;
						v._predecessor_fw = SubgraphEdge(e, u_sid);
						_context.pq_insert_or_update(v, v._time_fw, FORWARD());
					}
				}
				else
				{
					int32_t min_e, max_e;
					std::tie(min_e, max_e) = _graph->get_min_max_time(e);
					const Interval v_interval_bw_new(u_interval_bw.get_lower() + min_e, u_interval_bw.get_upper() + max_e);

					if ( v._interval_bw.is_infty() )
					{
						v._interval_bw = v_interval_bw_new;
						v._predecessors_bw.emplace_back(e, u_sid);
					}
					else
					{
						if ( v_interval_bw_new.get_lower() > v._interval_bw.get_upper() )
							continue;

						if ( v_interval_bw_new.get_upper() < v._interval_bw.get_lower() )
							v._predecessors_bw.clear();

						v._predecessors_bw.emplace_back(e, u_sid);

						if ( v_interval_bw_new.get_lower() >= v._interval_bw.get_lower() )
							if ( v_interval_bw_new.get_upper() >= v._interval_bw.get_upper() )
								continue;

						v._interval_bw =
							Interval(
								std::min(v._interval_bw.get_lower(), v_interval_bw_new.get_lower()),
								std::min(v._interval_bw.get_upper(), v_interval_bw_new.get_upper())
							);
					}

					_context.pq_insert_or_update(v, v._interval_bw.get_lower(), BACKWARD());
				}
			}
		}
	}

	void downward_search()
	{
		util::make_unique(_candidate_search_ids);
		_context.pq_clear(FORWARD());

		for ( const SearchId& x_id : _candidate_search_ids )
		{
			SearchNode& x = _context.get_search_node_from_sid(x_id);

			if ( x._time_fw + x._interval_bw.get_lower() > _upper_bound ) continue;
//			if ( x._time_fw > x._upper_stall[FORWARD()] ) continue;

			_context.pq_insert_or_update(x, x._time_fw, FORWARD());
		}

		while ( ! _context.pq_empty(FORWARD()) )
		{
			const SearchNode& u = _context.pq_delete_min(FORWARD());
			const SearchId u_sid = _context.get_search_id(u);

			for ( const auto& pred : u._predecessors_bw )
			{
				const EdgeId e = pred._edge_id;
				assert(_graph->is_directed_downward(e));
				SearchNode& v = _context.get_search_node_from_sid(pred._search_id);

				const int32_t time_v_new = _graph->eval_time(e, u._time_fw) + u._time_fw;
				const uint32_t dist_v_new = _graph->eval_distance(e, u._time_fw) + u._dist_fw;

				if ( time_v_new < v._time_fw || (time_v_new == v._time_fw && dist_v_new < v._dist_fw) )
				{
					v._time_fw = time_v_new;
					v._dist_fw = dist_v_new;

					v._predecessor_fw = SubgraphEdge(e, u_sid);
					_context.pq_insert_or_update(v, v._time_fw, FORWARD());
				}
			}
		}
	}

	void extract_up_down_path_info(std::vector<NodeId>& o_reverse_node_ids, std::vector<EdgeId>& o_reverse_edge_ids) const noexcept
	{
		assert(o_reverse_node_ids.empty());
		assert(o_reverse_edge_ids.empty());

		if ( _arrival_node_id == NodeId::INVALID() )
			return;

		SearchId search_id = _context.get_search_id(_context.get_search_node(_arrival_node_id));

		while ( true )
		{
			const SearchNode& search_node = _context.get_search_node_from_sid(search_id);
			o_reverse_node_ids.push_back(search_node._node_id);

			if ( !o_reverse_edge_ids.empty() && search_node._predecessor_fw._search_id != SearchId::INVALID() )
			{
				bool this_is_downward = _graph->is_directed_downward(o_reverse_edge_ids.back());
				bool pred_is_upward = _graph->is_directed_upward(search_node._predecessor_fw._edge_id);
//				assert(pred_is_upward || this_is_downward);
			}

			search_id = search_node._predecessor_fw._search_id;
			if ( search_id == SearchId::INVALID() )
				break;

			EdgeId edge_id = search_node._predecessor_fw._edge_id;
			assert(edge_id != EdgeId::INVALID());

			o_reverse_edge_ids.push_back(edge_id);
		}

		return;
	}

	Location _start;
	Location _dest;
	int32_t _departure_time;

	TchSearchGraph const * const _graph;
	SearchContext _context;
	int32_t _upper_bound;
	std::vector<SearchId> _candidate_search_ids;
	int32_t _arrival_time;
	NodeId _arrival_node_id;
	uint32_t _travel_distance;

public:

	TchQuerying(const TchSearchGraph& graph)
	: _start(),
	  _dest(),
	  _departure_time(std::numeric_limits<int32_t>::max()),
	  _graph(&graph),
	  _context(),
	  _upper_bound(std::numeric_limits<int32_t>::max()),
	  _arrival_time(std::numeric_limits<uint32_t>::max())
	{}

	int32_t one_to_one(const Location& start, const Location& destination, const int32_t departure_time)
	{
		_start = start;
		_dest = destination;
		_departure_time = departure_time;

		_arrival_time = std::numeric_limits<int32_t>::max();
		_arrival_node_id = NodeId::INVALID();
		_travel_distance = std::numeric_limits<uint32_t>::max();

		bidirectional_search(start, destination, departure_time);

		if ( _upper_bound == std::numeric_limits<int32_t>::max() )
			return std::numeric_limits<int32_t>::max();

		downward_search();

		for ( size_t i = 0 ; i != destination.size() ; ++i )
		{
			const NodeId node_id = destination.get_node_id(i);
			const SearchNode& search_node = _context.get_search_node(node_id);

			if ( search_node._time_fw == std::numeric_limits<int32_t>::max() )
				continue;

			if ( _arrival_time < search_node._time_fw )
				continue;

			if ( _arrival_time == search_node._time_fw && _travel_distance <= search_node._dist_fw )
				continue;

			_arrival_node_id = node_id;
			_arrival_time = search_node._time_fw + destination.get_time_dist(i).eval_time(search_node._time_fw);
			_travel_distance = search_node._dist_fw + destination.get_time_dist(i).eval_dist(search_node._time_fw);
		}

		return _arrival_time;
	}

	int32_t get_arrival_time() const noexcept
	{
		return _arrival_time;
	}

	uint32_t get_travel_distance() const noexcept
	{
		return _travel_distance;
	}

	Path retrieve_result_path() const
	{
		if ( _arrival_node_id == NodeId::INVALID() )
			return Path();

		//
		// extract up-down-path from predecessor information
		//

		std::vector<NodeId> reverse_nodes;
		std::vector<EdgeId> reverse_edges;

		extract_up_down_path_info(reverse_nodes, reverse_edges);

		//
		// expand the up-down-path using the time-dependent middle node information computed
		// during preprocessing to obtain the result path in the original road network
		//

		std::stack<std::tuple<NodeId,EdgeId,NodeId>> S;
		const SearchNode& first_node_of_path = _context.get_search_node(reverse_nodes.back());
		int32_t current_time = first_node_of_path._time_fw;
		int32_t current_distance = first_node_of_path._dist_fw;

		std::vector<NodeId> result_nodes({reverse_nodes.back()});
		std::vector<int32_t> result_arrival_times({current_time});
		std::vector<int32_t> result_distances({current_distance});
		std::vector<EdgeId> result_edges;

		for ( size_t i = reverse_edges.size() ; i > 0 ; --i )
		{
			assert(i < reverse_nodes.size());
			assert(current_time == _context.get_search_node(reverse_nodes[i])._time_fw);
			assert(current_distance == _context.get_search_node(reverse_nodes[i])._dist_fw);

			{
				const NodeId& src = reverse_nodes[i];
				const NodeId& tgt = reverse_nodes[i-1];
				const EdgeId& e = reverse_edges[i-1];
				assert(_graph->eval_time(e, current_time) + current_time == _context.get_search_node(tgt)._time_fw);

				const uint32_t dist_e = _graph->eval_distance(e, current_time);
//				assert(_graph->eval_distance(e, current_time) + current_distance == _context.get_search_node(tgt)._dist_fw);

				assert(S.empty());
				S.push(std::make_tuple(src, e, tgt));
			}

			while ( ! S.empty() )
			{
				const auto& stack_top = S.top();
				NodeId src = std::get<0>(stack_top);
				EdgeId e_shortcut = std::get<1>(stack_top);
				NodeId tgt = std::get<2>(stack_top);
				S.pop();

				NodeId middle = _graph->get_middle_node_descr_from_departure_time(e_shortcut, current_time)._node_id;
				if ( middle == NodeId::INVALID() )
				{
 
					current_distance += _graph->eval_distance(e_shortcut, current_time);
					current_time += _graph->eval_time(e_shortcut, current_time);

					result_edges.push_back(e_shortcut);
					result_nodes.push_back(tgt);
					result_arrival_times.push_back(current_time);
					result_distances.push_back(current_distance);

					continue;
				}

				EdgeId e_up;
				EdgeId e_down;
				for ( EdgeId e = _graph->edges_begin(middle) ; e != _graph->edges_end(middle) ; ++e )
				{
					if ( _graph->is_directed_upward(e) && _graph->get_other_node(e) == tgt ) e_up = e;
					if ( _graph->is_directed_downward(e) && _graph->get_other_node(e) == src ) e_down = e;

					if ( e_up != EdgeId::INVALID() && e_down != EdgeId::INVALID() )
						break;
				}

				assert(e_up != EdgeId::INVALID());
				assert(e_down != EdgeId::INVALID());

				#ifndef NDEBUG
				{
					int32_t time_at_middle = _graph->eval_time(e_down, current_time) + current_time;
					int32_t time_at_tgt_via_middle = _graph->eval_time(e_up, time_at_middle) +  time_at_middle;
					int32_t time_at_tgt_via_shortcut = _graph->eval_time(e_shortcut, current_time) + current_time;
					assert(time_at_tgt_via_middle == time_at_tgt_via_shortcut);

					int32_t dist_at_middle = _graph->eval_distance(e_down, current_time) + current_distance;
					int32_t dist_at_tgt_via_middle = _graph->eval_distance(e_up, time_at_middle) + dist_at_middle;
					int32_t dist_at_tgt_via_shortcut = _graph->eval_distance(e_shortcut, current_time) + current_distance;
					assert(dist_at_tgt_via_middle == dist_at_tgt_via_shortcut);
				}
				#endif

				S.push(std::make_tuple(middle, e_up, tgt));
				S.push(std::make_tuple(src, e_down, middle));
			}
		}

		return Path(
			_start, _dest, _departure_time, _arrival_time, _travel_distance,
				std::move(result_nodes), std::move(result_edges), std::move(result_arrival_times), std::move(result_distances));
	}
};



}

#endif // RITCH_QUERYING_TCH_QUERYING_H_
