#ifndef RITCH_QUERYING_TCH_QUERYING_ML_H_
#define RITCH_QUERYING_TCH_QUERYING_ML_H_

#include <stack>
#include <tuple>
#include <vector>

#include <boost/heap/pairing_heap.hpp>

#include "ritch/querying/tch_search_graph.h"
#include "ritch/querying/location.h"
#include "ritch/querying/path.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

	class TchQueryingML
	{

	public:

		typedef TchSearchGraph::NodeId NodeId;
		typedef TchSearchGraph::EdgeId EdgeId;
		typedef TchSearchGraph::TimeDist TimeDist;
		typedef Location_tmpl<NodeId, TimeDist> Location;
		typedef Path_tmpl<NodeId, EdgeId, Location> Path;

	private:

		RITCH_DEFINE_STRONG_ID(SearchId)

		static constexpr uint32_t FORWARD() { return 0; };
		static constexpr uint32_t BACKWARD() { return 1; };
		static constexpr bool DO_STORE_PREDECESSORS = true;
		static constexpr bool DONT_STORE_PREDECESSORS = false;
		int SHIFT_THRESHOLD;
		int BREAK_TIME;

		struct SubgraphEdge
		{
			EdgeId _edge_id;
			SearchId _search_id;

			SubgraphEdge()
				: _edge_id(EdgeId::INVALID()), _search_id(SearchId::INVALID())
			{}

			SubgraphEdge(EdgeId edge_id, SearchId search_id)
				: _edge_id(edge_id), _search_id(search_id)
			{}
		};

		struct Interval
		{
		private:
			int32_t _lower;
			int32_t _upper;

		public:
			Interval()
				: Interval(std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max())
			{}

			Interval(const int32_t lower, const int32_t upper)
				: _lower(lower), _upper(upper)
			{}

			static Interval INFTY() { return Interval(); }

			int32_t get_lower() const { return _lower; }
			int32_t get_upper() const { return _upper; }

			bool is_infty() const noexcept
			{
				assert(
					RITCH_EQUIV(
						_lower == std::numeric_limits<int32_t>::max(),
						_upper == std::numeric_limits<int32_t>::max()
					)
				);

				return _lower == std::numeric_limits<int32_t>::max();
			}

			friend bool operator== (const Interval& lhs, const Interval& rhs) noexcept
			{
				return lhs._lower == rhs._lower && lhs._upper == rhs._upper;
			}

			friend bool operator!= (const Interval& lhs, const Interval& rhs) noexcept
			{
				return !(lhs == rhs);
			}
		};

		template <typename HeapHandleT>
		struct SearchNode_tmpl
		{
			std::array<HeapHandleT, 2> _heap_handle;
			NodeId _node_id;
			uint32_t _flags;
			int32_t _shift_time;
			bool _breakTaken;
			int32_t _time_fw;
			uint32_t _dist_fw;
			Interval _interval_bw;
			std::array<int32_t, 2> _upper_stall;
			SubgraphEdge _predecessor_fw;
			std::vector<SubgraphEdge> _predecessors_bw;

			SearchNode_tmpl(const SearchNode_tmpl&) = delete;
			SearchNode_tmpl& operator=(const SearchNode_tmpl&) = delete;

			SearchNode_tmpl(SearchNode_tmpl&&) = default;
			SearchNode_tmpl& operator=(SearchNode_tmpl&&) = default;

			SearchNode_tmpl()
				: _heap_handle(),
				_node_id(NodeId::INVALID()),
				_flags(0),
				_shift_time(0),
				_breakTaken(false),

				_time_fw(std::numeric_limits<int32_t>::max()),
				_dist_fw(std::numeric_limits<uint32_t>::max()),
				_interval_bw(Interval::INFTY()),
				_upper_stall({ std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max() }),
				_predecessor_fw(),
				_predecessors_bw()
			{}

			bool is_enqueued(const size_t direction) const
			{
				assert(direction == 0 || direction == 1);
				return (_flags & (1 << direction)) != 0;
			}

			void set_enqueued(bool enqueued, uint32_t direction)
			{
				assert(direction == 0 || direction == 1);

				if (enqueued)
					_flags |= (1 << direction);
				else
					_flags &= ~(1 << direction);
			}
		};

		class SearchContext
		{
		private:

			struct HeapItem
			{
				SearchId _search_id;
				int32_t _priority;

				HeapItem(SearchId search_id, int32_t priority)
					: _search_id(search_id), _priority(priority)
				{}

				friend inline bool operator< (const HeapItem& lhs, const HeapItem& rhs) { return rhs._priority < lhs._priority; }
			};

			struct NodeIdHash
			{
				size_t operator() (NodeId node_id) const noexcept { return std::hash<uint32_t>()(uint32_t(node_id)); }
			};

			typedef boost::heap::pairing_heap<HeapItem> HeapImpl;
			typedef HeapImpl::handle_type HeapHandle;

			std::array<HeapImpl, 2> _heap;
			std::unordered_map<NodeId, std::vector<SearchId>, NodeIdHash> _table;
			std::unordered_map<NodeId, SearchId, NodeIdHash> _table_backward;

			std::unordered_map<NodeId, std::vector<SearchId>, NodeIdHash> _table_break;
			std::unordered_map<NodeId, std::vector<SearchId>, NodeIdHash> _table_breakdown;

			std::vector<SearchNode_tmpl<HeapHandle>> _search_nodes;

		public:

			typedef SearchNode_tmpl<HeapHandle> SearchNode;

			SearchContext()
				: _heap(),
				_table(),
				_table_backward(),
				_table_break(),
				_table_breakdown(),
				_search_nodes()
			{}

			bool pq_empty(const uint32_t direction) const noexcept
			{
				assert(direction < _heap.size());
				return _heap[direction].empty();
			}

			bool pq_contains(const NodeId node_id, const uint32_t direction) const noexcept
			{
				bool nobreak = false;
				bool breaktaken = false;
				bool breakdown = false;
				bool backward = false;
				auto it = _table.find(node_id);
				if (it != _table.end()) {
					for (int i = 0; i < it->second.size(); i++) {
						if (!nobreak) {
							nobreak = _search_nodes[it->second[i]].is_enqueued(direction);
						}
					}
				}
				it = _table_break.find(node_id);
				if (it != _table_break.end()) {
					for (int i = 0; i < it->second.size(); i++) {
						if (!breaktaken) {
							breaktaken = _search_nodes[it->second[i]].is_enqueued(direction);
						}
					}
				}
				it = _table_breakdown.find(node_id);
				if (it != _table_breakdown.end()) {
					for (int i = 0; i < it->second.size(); i++) {
						if (!breakdown) {
							breakdown = _search_nodes[it->second[i]].is_enqueued(direction);
						}
					}
				}

				auto itb = _table_backward.find(node_id);
				if (itb != _table_backward.end()) {
					const SearchId search_id = itb->second;
					backward = _search_nodes[search_id.the_value()].is_enqueued(direction);
				}


				return nobreak || breaktaken || breakdown || backward;
			}

			bool pq_contains(const SearchNode& search_node, const uint32_t direction) const noexcept
			{
				return search_node.is_enqueued(direction);
			}

			bool touched(const NodeId node_id) const noexcept
			{
				return _table.count(node_id) > 0;
			}

			bool touched(const NodeId node_id, bool breaktaken, bool down) {
				if (breaktaken) {
					if (down) {
						return _table_breakdown.count(node_id) > 0;
					}
					else {
						return _table_break.count(node_id) > 0;
					}
				}
				else {
					return (_table.count(node_id) > 0);
				}
			}

			bool touched_break(const NodeId node_id) const noexcept
			{
				return _table_break.count(node_id) > 0;
			}

			bool touched_break_down(const NodeId node_id) const noexcept
			{
				return _table_breakdown.count(node_id) > 0;
			}

			bool touched_break_down() const noexcept
			{
				return _table_breakdown.size() > 0;
			}

			bool touched_backward(const NodeId node_id) const noexcept
			{
				return _table_backward.count(node_id) > 0;
			}

			const std::vector<SearchId>& get_search_ids(const NodeId node_id, bool breaktaken, bool down) const noexcept
			{
				if (breaktaken) {
					if (down) {
						assert(_table_breakdown.count(node_id) > 0);
						return _table_breakdown.find(node_id)->second;
					}
					else {
						assert(_table_break.count(node_id) > 0);
						return _table_break.find(node_id)->second;
					}
				}
				else {
					assert(_table.count(node_id) > 0);
					return _table.find(node_id)->second;
				}
			}

			SearchNode& create_search_node(const NodeId node_id, bool breaktaken, bool down) noexcept
			{
				SearchId search_id = SearchId(_search_nodes.size());
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
				if (breaktaken) {
					if (down)	_table_breakdown[node_id].push_back(search_id);
					else		_table_break[node_id].push_back(search_id);
				}
				else {
					_table[node_id].push_back(search_id);
				}
				return _search_nodes[search_id.the_value()];
			}

			SearchNode& get_or_create_search_node_backwardsearch(const NodeId node_id) {
				SearchId& search_id = _table_backward[node_id];
				if (search_id == SearchId::INVALID())
				{
					search_id = _search_nodes.size();
					_search_nodes.emplace_back();
					_search_nodes.back()._node_id = node_id;
				}

				return _search_nodes[search_id.the_value()];
			}

			SearchNode& get_search_node_backwardsearch(const NodeId node_id) {
				assert(_table_backward.count(node_id) > 0);
				return _search_nodes[_table_backward.find(node_id)->second.the_value()];
			}

			SearchNode& get_search_node_from_sid(const SearchId search_id)
			{
				return _search_nodes[search_id.the_value()];
			}

			const SearchNode& get_search_node_from_sid(const SearchId search_id) const
			{
				return _search_nodes[search_id.the_value()];
			}

			SearchId get_search_id(const SearchNode& search_node) const
			{
				const SearchId search_id(std::distance(&(_search_nodes[0]), &search_node));
				return search_id;
			}

			const SearchNode& pq_min_search_node(const uint32_t direction) const
			{
				return _search_nodes[_heap[direction].top()._search_id.the_value()];
			}

			int32_t pq_min_priority(const uint32_t direction) const
			{
				assert(!pq_empty(direction));
				return _heap[direction].top()._priority;
			}

			void pq_insert(SearchNode& search_node, const int32_t priority, const uint32_t direction)
			{
				const SearchId search_id = get_search_id(search_node);
				const HeapHandle heap_handle = _heap[direction].push(HeapItem(search_id, priority));
				search_node._heap_handle[direction] = heap_handle;
				search_node.set_enqueued(true, direction);
			}


			void pq_insert_or_update_backwardsearch(SearchNode& search_node, const int32_t priority, const uint32_t direction)
			{
				const SearchId search_id = get_search_id(search_node);
				assert(_table_backward[search_node._node_id] == search_id);
				if (!search_node.is_enqueued(direction))
				{
					const HeapHandle heap_handle = _heap[direction].push(HeapItem(search_id, priority));
					search_node._heap_handle[direction] = heap_handle;
					search_node.set_enqueued(true, direction);
				}
				else
				{
					const HeapHandle heap_handle = search_node._heap_handle[direction];
					_heap[direction].update(heap_handle, HeapItem(search_id, priority));
				}
			}

			void pq_remove(SearchNode& search_node, const uint32_t direction) {
				search_node.set_enqueued(false, direction);
			}

			bool pq_remove_not_enqueued(const uint32_t direction) {
				SearchId sid = _heap[direction].top()._search_id;
				while (!_search_nodes[sid.the_value()].is_enqueued(direction)) {
					_heap[direction].pop();
					if (_heap[direction].empty()) {
						return false;
					}
					sid = _heap[direction].top()._search_id.the_value();
				}
				return true;
			}

			SearchNode& pq_delete_min(const uint32_t direction)
			{
				assert(!pq_empty(direction));
				SearchNode& search_node = _search_nodes[_heap[direction].top()._search_id.the_value()];
				_heap[direction].pop();
				search_node.set_enqueued(false, direction);
				return search_node;
			}

			void pq_clear(const uint32_t direction)
			{
				for (const auto& heap_item : _heap[direction])
					_search_nodes[heap_item._search_id.the_value()].set_enqueued(false, direction);

				_heap[direction].clear();
			}

			void clear_all()
			{
				_heap[FORWARD()].clear();
				_heap[BACKWARD()].clear();
				_table.clear();
				_table_backward.clear();
				_table_break.clear();
				_table_breakdown.clear();
				_search_nodes.clear();
			}
		};

		typedef SearchContext::SearchNode SearchNode;

		static uint32_t inverse(const uint32_t direction)
		{
			if (direction == FORWARD())
				return BACKWARD();

			return FORWARD();
		}

		void bidirectional_search(const Location& start, const Location& destination, const int32_t departure_time, const int32_t shift_start)
		{
			_context.clear_all();

			//Add start node to forward pq
			for (size_t i = 0; i != start.size(); ++i)
			{
				const NodeId& node_id = start.get_node_id(i);
				assert(!_context.pq_contains(node_id, FORWARD()));

				const TimeDist& time_dist = start.get_time_dist(i);
				const int32_t departure_time_on_node = time_dist.eval_time(departure_time) + departure_time;
				assert(departure_time_on_node < 604800);
				const uint32_t distance_on_node = time_dist.eval_dist(departure_time);

				SearchNode& search_node = _context.create_search_node(node_id, false, false );
				_context.pq_insert(search_node, departure_time_on_node, FORWARD());
				search_node._time_fw = departure_time_on_node;
				search_node._dist_fw = distance_on_node;
				search_node._shift_time = shift_start;
			}

			//Add destination node to backward pq
			for (size_t i = 0; i != destination.size(); ++i)
			{
				const NodeId& node_id = destination.get_node_id(i);
				assert(!_context.pq_contains(node_id, BACKWARD()));

				const TimeDist& time_dist = destination.get_time_dist(i);

				SearchNode& search_node = _context.get_or_create_search_node_backwardsearch(node_id);
				_context.pq_insert(search_node, time_dist.get_min_time(), BACKWARD());
				search_node._interval_bw = Interval(time_dist.get_min_time(), time_dist.get_max_time());
			}

			_candidate_search_ids.clear();
			_upper_bound = std::numeric_limits<int32_t>::max();
			_upper_bound_break = std::numeric_limits<int32_t>::max();


			uint32_t direction = BACKWARD();
			std::array<bool, 2> finished = { false, false };

			while (!_context.pq_empty(FORWARD()) || !_context.pq_empty(BACKWARD()))
			{
				//alternate forward/backward direction
				if (finished[FORWARD()] && finished[BACKWARD()]) break;
				if (!_context.pq_empty(inverse(direction))) direction = inverse(direction);
				if (finished[FORWARD()]) direction = BACKWARD();
				if (finished[BACKWARD()]) direction = FORWARD();

				if (_context.pq_empty(direction))
				{
					finished[direction] = true;
					continue;
				}

				//
				// take node with minimal label out of PQ
				//
				if (!_context.pq_remove_not_enqueued(direction)) {
					finished[direction] = true;
					continue;
				}
				SearchNode& u = _context.pq_delete_min(direction);

				//If direction = forward, check if backward touched
				if (direction == FORWARD()) {
					if (_context.touched_backward(u._node_id)) {
						SearchNode& back = _context.get_search_node_backwardsearch(u._node_id);
						if (u._time_fw != std::numeric_limits<int32_t>::max() && !back._interval_bw.is_infty())
						{
							if (u._breakTaken) {
								if (u._time_fw + back._interval_bw.get_lower() <= _upper_bound_break) {
									_candidate_search_ids.push_back(_context.get_search_id(u));
								}
								_upper_bound_break = std::min(_upper_bound_break, u._time_fw + back._interval_bw.get_upper());
							}
							else {
								if (u._time_fw + back._interval_bw.get_lower() <= _upper_bound) {
									_candidate_search_ids.push_back(_context.get_search_id(u));
								}
								_upper_bound = std::min(_upper_bound, u._time_fw + back._interval_bw.get_upper());
							}
						}
					}
				}

				//If direction = backward, check if forward touched
				if (direction == BACKWARD()) {
					if (_context.touched(u._node_id, false, false)) {
						std::vector<SearchId> nobreaks = _context.get_search_ids(u._node_id, false, false);
						for (int i = 0; i < nobreaks.size(); i++) {
							if (_context.get_search_node_from_sid(nobreaks[i])._time_fw + u._interval_bw.get_lower() <= _upper_bound) {
								_candidate_search_ids.push_back(nobreaks[i]);
							}
							_upper_bound = std::min(_upper_bound, _context.get_search_node_from_sid(nobreaks[i])._time_fw + u._interval_bw.get_upper());
						}
					}
					if (_context.touched(u._node_id, true, true)) {
						std::vector<SearchId> breaktaken = _context.get_search_ids(u._node_id, true, false);
						for (int i = 0; i < breaktaken.size(); i++) {
							if (_context.get_search_node_from_sid(breaktaken[i])._time_fw + u._interval_bw.get_lower() <= _upper_bound_break) {
								_candidate_search_ids.push_back(breaktaken[i]);
							}
							_upper_bound = std::min(_upper_bound_break, _context.get_search_node_from_sid(breaktaken[i])._time_fw + u._interval_bw.get_upper());
						}
					}
				}



				//
				// relax edges
				//

				const SearchId u_sid = _context.get_search_id(u);
				const NodeId u_id = u._node_id;
				const int32_t u_time_fw = u._time_fw;
				const int32_t u_shift_fw = u._shift_time;
				const uint32_t u_dist_fw = u._dist_fw;
				const Interval u_interval_bw = u._interval_bw;
				const bool u_break = u._breakTaken;

				for (EdgeId e = _graph->edges_begin(u_id); e != _graph->edges_end(u_id); ++e)
				{
					if (direction == FORWARD() && !_graph->is_directed_upward(e)) continue;
					else if (direction == BACKWARD() && !_graph->is_directed_downward(e)) continue;

					const NodeId v_id = _graph->get_other_node(e);

					if (direction == FORWARD())
					{
						const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
						const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
						int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shift_fw; //waiting does not count for driving time, thus default time (FF) used
						if (u_break) {
							if (v_shift_fw_new < SHIFT_THRESHOLD) {
								addNewSearchNode(v_id, v_time_fw_new, v_dist_fw_new, v_shift_fw_new, e, u_sid, true, false);
							}
						}
						else {
							// ADD NODE TAKING A BREAK
							std::tuple<int32_t, uint32_t, int32_t> breakInsertion = calcBreak(e, u_id, v_id, u_time_fw, u_dist_fw, u_shift_fw); // output: <arrivaltime, distance, shiftduration>
							if (std::get<2>(breakInsertion) < SHIFT_THRESHOLD) {
								addNewSearchNode(v_id, std::get<0>(breakInsertion), std::get<1>(breakInsertion), std::get<2>(breakInsertion), e, u_sid, true, false);
							}
						
							//NORMAL RELAX
							if (v_shift_fw_new < SHIFT_THRESHOLD) {
								addNewSearchNode(v_id, v_time_fw_new, v_dist_fw_new, v_shift_fw_new, e, u_sid, false, false);
							}
						}
					}
					else
					{
						int32_t min_e, max_e;
						std::tie(min_e, max_e) = _graph->get_min_max_time(e);

						Interval v_interval_bw_new(u_interval_bw.get_lower() + min_e, u_interval_bw.get_upper() + max_e);

						SearchNode& v = _context.get_or_create_search_node_backwardsearch(v_id);

						if (v._interval_bw.is_infty())
						{
							v._interval_bw = v_interval_bw_new;
							v._predecessors_bw.emplace_back(e, u_sid);
						}
						else
						{
							if (v_interval_bw_new.get_lower() > v._interval_bw.get_upper())
								continue;

							if (v_interval_bw_new.get_upper() < v._interval_bw.get_lower())
								v._predecessors_bw.clear();

							v._predecessors_bw.emplace_back(e, u_sid);

							if (v_interval_bw_new.get_lower() >= v._interval_bw.get_lower())
								if (v_interval_bw_new.get_upper() >= v._interval_bw.get_upper())
									continue;

							v._interval_bw =
								Interval(
									std::min(v._interval_bw.get_lower(), v_interval_bw_new.get_lower()),
									std::min(v._interval_bw.get_upper(), v_interval_bw_new.get_upper())
								);
						}

						_context.pq_insert_or_update_backwardsearch(v, v._interval_bw.get_lower(), BACKWARD());
					}
				}
			}
		}

		void addNewSearchNode(NodeId v_id, int32_t v_time_fw_new, uint32_t v_dist_fw_new, int32_t v_shift_fw_new, EdgeId e, SearchId u_sid, bool breaktaken, bool down) {
			if(!_context.touched(v_id,breaktaken,down)){
				SearchNode& bl = _context.create_search_node(v_id, breaktaken, down);
				bl._time_fw = v_time_fw_new;
				bl._shift_time = v_shift_fw_new;
				bl._dist_fw = v_dist_fw_new;
				bl._breakTaken = breaktaken;
				bl._predecessor_fw = SubgraphEdge(e, u_sid);
				_context.pq_insert(bl, bl._time_fw, FORWARD());
			}
			else {
				std::vector<SearchId> nodes = _context.get_search_ids(v_id, breaktaken, down);
				bool dominated = isDominated(v_time_fw_new, v_shift_fw_new, nodes);
				if (!dominated) {
					SearchNode& bl = _context.create_search_node(v_id, breaktaken, down);
					bl._time_fw = v_time_fw_new;
					bl._shift_time = v_shift_fw_new;
					bl._dist_fw = v_dist_fw_new;
					bl._breakTaken = breaktaken;
					bl._predecessor_fw = SubgraphEdge(e, u_sid);
					_context.pq_insert(bl, bl._time_fw, FORWARD());
					removeDominated(bl, nodes);
				}
			}
		}

		bool isDominated(int32_t v_time_fw_new, int32_t v_shift_fw_new, std::vector<SearchId> & nodes) {
			int i = 0;
			while (i < nodes.size()) {
				if (v_time_fw_new >= _context.get_search_node_from_sid(nodes[i])._time_fw && v_shift_fw_new >= _context.get_search_node_from_sid(nodes[i])._shift_time) {
					return true;
				}
				i++;
			}
			return false;
		}

		void removeDominated(SearchNode& bl, std::vector<SearchId> & nodes) {
			int i = 0;
			for (int i = 0; i < nodes.size(); i++) {
				SearchNode& otherSearchNode = _context.get_search_node_from_sid(nodes[i]);
				if (bl._time_fw <= otherSearchNode._time_fw && bl._shift_time <= otherSearchNode._shift_time) {
					if (bl._time_fw < otherSearchNode._time_fw || bl._shift_time < otherSearchNode._shift_time) {
						if (otherSearchNode.is_enqueued(FORWARD())) {
							otherSearchNode.set_enqueued(false, FORWARD());
						}
					}
				}
			}
		}

		void downward_search()
		{
			util::make_unique(_candidate_search_ids);
			_context.pq_clear(FORWARD());

			for (const SearchId& x_id : _candidate_search_ids)
			{
				SearchNode& x = _context.get_search_node_from_sid(x_id);
				_context.pq_insert(x, x._time_fw, FORWARD()); //TODO?
			}

			while (!_context.pq_empty(FORWARD()))
			{
				if (!_context.pq_remove_not_enqueued(FORWARD())) {
					continue;
				}

				const SearchNode& u = _context.pq_delete_min(FORWARD());
				const SearchId u_sid = _context.get_search_id(u);
				const NodeId u_id = u._node_id;

				const int32_t u_time_fw = u._time_fw;
				const int32_t u_shift_fw = u._shift_time;
				const uint32_t u_dist_fw = u._dist_fw;
				const Interval u_interval_bw = u._interval_bw;
				const bool u_break = u._breakTaken;

				std::vector<SubgraphEdge> predecessorlist = _context.get_search_node_backwardsearch(u_id)._predecessors_bw;
				RITCH_ASSERT(predecessorlist.size() > 0 || int32_t(u._node_id) == int32_t(_dest.get_node_id(0)));

				for (const auto& pred : predecessorlist)
				{
					const EdgeId e = pred._edge_id;

					assert(_graph->is_directed_downward(e));

					const NodeId v_id2 = _graph->get_other_node(e);

					const NodeId v_id = _context.get_search_node_from_sid(pred._search_id)._node_id;


					const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
					const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
					int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shift_fw; //waiting does not count for driving time, thus default time (FF) used
					
					if (u_break) {
						if (v_shift_fw_new < SHIFT_THRESHOLD) {
							addNewSearchNode(v_id, v_time_fw_new, v_dist_fw_new, v_shift_fw_new, e, u_sid, true, true);
						}
					}
					else {
						//ADD BREAK
						std::tuple<int32_t, uint32_t, int32_t> breakInsertion = calcBreak(e, u_id, v_id, u_time_fw, u_dist_fw, u_shift_fw); // output: <time, distance, shiftduration>
						if (std::get<2>(breakInsertion) < SHIFT_THRESHOLD) {
							addNewSearchNode(v_id, std::get<0>(breakInsertion), std::get<1>(breakInsertion), std::get<2>(breakInsertion), e, u_sid, true, true);
						}
												
						//NORMAL RELAX
						if (v_shift_fw_new < SHIFT_THRESHOLD) {
							addNewSearchNode(v_id, v_time_fw_new, v_dist_fw_new, v_shift_fw_new, e, u_sid, false, true);
						}
					}
				}
			}
		}

		std::tuple<int32_t, uint32_t, int32_t> calcBreak(EdgeId e, NodeId u, NodeId v, int32_t utime, int32_t udistance, int32_t ushift) {
			std::tuple<int32_t, uint32_t, int32_t> res;
			std::tuple<int32_t, uint32_t, int32_t> res_b;
			std::pair<int32_t, int32_t> block = get_block_info_of_edge(u, v, e, utime); // blockstart, blockend
			std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> edgeblock = _graph->get_edge_block(u, v, e);
			
			int32_t block_defaulttime = get_default_time_of_edge(u, v, e, utime);
			assert(block_defaulttime - SHIFT_THRESHOLD + ushift >= 0);

			if (utime >= block.first - block_defaulttime && utime < block.second - BREAK_TIME) {
				std::get<0>(res) = utime + _graph->eval_time(e, utime);
				std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
				std::get<2>(res) = block_defaulttime - std::max(block.first - utime, 0 );
			}
			else if ( utime >= block.second - BREAK_TIME && utime < block.second ) {
				std::get<0>(res) = utime + std::min(BREAK_TIME - (block.second - utime), utime + _graph->eval_time(e,utime));
				std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
				std::get<2>(res) = block_defaulttime;
			}
			else {
				std::get<0>(res) = utime + _graph->eval_time(e, utime) + BREAK_TIME;
				std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
				std::get<2>(res) = 0;
			}

			RITCH_ASSERT(std::get<0>(res) >= utime + _graph->eval_time(e, utime));

			return res;
		}
		
		std::pair<int32_t, int32_t> get_block_info_of_edge(NodeId u, NodeId v, EdgeId e, int32_t tau) {
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block = _graph->get_edge_block(u, v, e);
			if (block.size() == 0) {
				return std::make_pair(0, 0);
			}
			else {
				for (int i = 0; i < block.size() - 1; i++) {
					if (tau > std::get<0>(block[i]) && tau <= std::get<0>(block[i + 1])) {
						return std::make_pair(std::get<1>(block[i]), std::get<2>(block[i]));
					}
				}
			}
			return std::make_pair(std::get<1>(block[block.size() - 1]), std::get<2>(block[block.size() - 1]));
		}

		int32_t get_default_time_of_edge(NodeId u, NodeId v, EdgeId e, int32_t tau) const {
			const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block = _graph->get_edge_block(u, v, e);
			int32_t  defaultime = 0;
			if (block.size() == 0) {
				defaultime = _graph->get_default_time(e);
			}
			else {
				if (tau > 604800) {
					tau = tau - 604800;
				}
				if (tau <= std::get<0>(block[0])) {
					defaultime = std::get<3>(block[0]);
				}
				for (int i = 1; i < block.size(); i++) {
					if (defaultime == 0 && tau <= std::get<0>(block[i]) && tau > std::get<0>(block[i - 1])) {
						defaultime = std::get<3>(block[i]);
					}
				}
				if (defaultime == 0) {
					if (tau >= std::get<0>(block[block.size() - 1])) {
						defaultime = std::get<3>(block[0]);
					}
					else {
						defaultime = std::get<3>(block[block.size() - 1]);
					}
				}
			}

			RITCH_ASSERT(defaultime <= _graph->eval_time(e, tau));
			return defaultime;
		}

		bool dbg_original_path_is_ok(const Path& path) const
		{


			for (size_t i = 0; i != path.get_n_edges(); ++i)
			{
				if (_graph->has_middle_node(path.get_edge_id(i)))
					return false;

				const auto edge = path.get_edge(i);
				NodeId src = std::get<0>(edge);
				EdgeId e = std::get<1>(edge);
				NodeId tgt = std::get<2>(edge);

				int32_t time_tgt_calculated = _graph->eval_time(e, path.get_arrival_time(i)) + path.get_arrival_time(i);
				int32_t dist_tgt_calculated = _graph->eval_distance(e, path.get_arrival_time(i)) + path.get_distance(i);

				if (time_tgt_calculated != path.get_arrival_time(i + 1))
					return false;

				if (dist_tgt_calculated != path.get_distance(i + 1))
					return false;
			}


			return true;
		}

		Location _start;
		Location _dest;
		int32_t _departure_time;
		int32_t _shift_start;

		TchSearchGraph const * const _graph;
		SearchContext _context;
		int32_t _upper_bound;
		int32_t _upper_bound_break;
		std::vector<SearchId> _candidate_search_ids;
		int32_t _arrival_time;
		int32_t _arrival_time_break;
		int32_t _arrival_time_nobreak;
		bool _break;
		bool _late;
		NodeId _arrival_node_id;
		uint32_t _travel_distance;

	public:

		TchQueryingML(const TchSearchGraph& graph)
			: _start(),
			_dest(),
			_departure_time(std::numeric_limits<int32_t>::max()),
			_graph(&graph),
			_context(),
			_upper_bound(std::numeric_limits<int32_t>::max()),
			_arrival_time(std::numeric_limits<uint32_t>::max()),
			_arrival_time_break(std::numeric_limits<uint32_t>::max()),
			_arrival_time_nobreak(std::numeric_limits<uint32_t>::max())

		{}

		int32_t one_to_one(const Location& start, const Location& destination, const int32_t departure_time, const int32_t shift_time, bool longbreaks)
		{
			if (longbreaks) {
				SHIFT_THRESHOLD = 32400;
				BREAK_TIME = 39600;
			}
			else {
				SHIFT_THRESHOLD = 16200;
				BREAK_TIME = 2700;
			}

			_start = start;
			_dest = destination;
			_departure_time = departure_time;
			_shift_start = shift_time;

			_arrival_time = std::numeric_limits<int32_t>::max();
			_arrival_time_nobreak = std::numeric_limits<int32_t>::max();
			_arrival_time_break = std::numeric_limits<int32_t>::max();

			_arrival_node_id = NodeId::INVALID();
			_break = false;
			_travel_distance = std::numeric_limits<uint32_t>::max();

			bidirectional_search(start, destination, departure_time, shift_time);

			if (_upper_bound == std::numeric_limits<int32_t>::max() && _upper_bound_break == std::numeric_limits<int32_t>::max())
				return std::numeric_limits<int32_t>::max();

			downward_search();

			for (size_t i = 0; i != destination.size(); ++i)
			{
				const NodeId node_id = destination.get_node_id(i);
				/* store arrivaltimes for each of the possible solutions, i.e. nobreak, breaklate or breakblock */
				std::vector<SearchId> nobreaks = std::vector<SearchId>();
				std::vector<SearchId> breaktaken = std::vector<SearchId>();
				if (_context.touched(node_id, false, true)) {
					nobreaks = _context.get_search_ids(node_id, false, true);
				}
				if (_context.touched(node_id, true, true)) {
					breaktaken = _context.get_search_ids(node_id, true, true);
				}

				for (int i = 0; i < nobreaks.size(); i++) {
					if (_context.get_search_node_from_sid(nobreaks[i])._time_fw < _arrival_time_nobreak) {
						_arrival_time_nobreak = _context.get_search_node_from_sid(nobreaks[i])._time_fw;
					}
				}
				for (int i = 0; i < breaktaken.size(); i++) {
					if (_context.get_search_node_from_sid(breaktaken[i])._time_fw < _arrival_time_break) {
						_arrival_time_break = _context.get_search_node_from_sid(breaktaken[i])._time_fw;
					}
				}
				
				_break = false;
				_arrival_time = _arrival_time_nobreak;

				if (_arrival_time > _arrival_time_break) {
					_arrival_time = _arrival_time_break;
					_break = true;
				}
				
				if (_arrival_time > 604800) {
					_arrival_time = _arrival_time - 604800;
				}

				_arrival_node_id = node_id;
			}

			return _arrival_time;
		}

		int32_t get_arrival_time() const noexcept
		{
			return _arrival_time;
		}

		uint32_t get_travel_distance() const noexcept
		{
			return _travel_distance;
		}

		bool get_break() const {
			return _break;
		}

		int32_t get_break_time() const {
			return _arrival_time_break;
		}
		int32_t get_nobreak_time() const {
			return _arrival_time_nobreak;
		}

	};



}

#endif // RITCH_QUERYING_TCH_QUERYING_H_
