#ifndef RITCH_QUERYING_TCH_QUERYING_H_
#define RITCH_QUERYING_TCH_QUERYING_H_

#include <stack>
#include <tuple>
#include <vector>

#include <boost/heap/pairing_heap.hpp>

#include "ritch/querying/tch_search_graph.h"
#include "ritch/querying/location.h"
#include "ritch/querying/path.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

class TchQuerying
{

public:

	typedef TchSearchGraph::NodeId NodeId;
	typedef TchSearchGraph::EdgeId EdgeId;
	typedef TchSearchGraph::TimeDist TimeDist;
	typedef Location_tmpl<NodeId, TimeDist> Location;
	typedef Path_tmpl<NodeId,EdgeId,Location> Path;

private:

	RITCH_DEFINE_STRONG_ID(SearchId)

	static constexpr uint32_t FORWARD() { return 0; };
	static constexpr uint32_t BACKWARD() { return 1; };
	static constexpr bool DO_STORE_PREDECESSORS = true;
	static constexpr bool DONT_STORE_PREDECESSORS = false;
	//static const int SHIFT_THRESHOLD = 16200;
	//static const int BREAK_TIME = 2700;
	int SHIFT_THRESHOLD;
	int BREAK_TIME;

	struct SubgraphEdge
	{
		EdgeId _edge_id;
		SearchId _search_id;

		SubgraphEdge()
		: _edge_id(EdgeId::INVALID()), _search_id(SearchId::INVALID())
		{}

		SubgraphEdge(EdgeId edge_id, SearchId search_id)
		: _edge_id(edge_id), _search_id(search_id)
		{}
	};

	struct Interval
	{
	private:
		int32_t _lower;
		int32_t _upper;

	public:
		Interval()
		: Interval(std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max())
		{}

		Interval(const int32_t lower, const int32_t upper)
			: _lower(lower), _upper(upper)
		{}

		static Interval INFTY() { return Interval(); }

		int32_t get_lower() const { return _lower; }
		int32_t get_upper() const { return _upper; }

		bool is_infty() const noexcept
		{
			assert(
				RITCH_EQUIV(
					_lower == std::numeric_limits<int32_t>::max(),
					_upper == std::numeric_limits<int32_t>::max()
				)
			);

			return _lower == std::numeric_limits<int32_t>::max();
		}

		friend bool operator== (const Interval& lhs, const Interval& rhs) noexcept
		{
			return lhs._lower == rhs._lower && lhs._upper == rhs._upper;
		}

		friend bool operator!= (const Interval& lhs, const Interval& rhs) noexcept
		{
			return !(lhs == rhs);
		}
	};

	template <typename HeapHandleT>
	struct SearchNode_tmpl
	{
		std::array<HeapHandleT, 2> _heap_handle;
		NodeId _node_id;
		uint32_t _flags;
		int32_t _shift_time;
		bool _breakTaken;
		bool _break_late;
		int32_t _time_fw;
		uint32_t _dist_fw;
		Interval _interval_bw;
		std::array<int32_t, 2> _upper_stall;
		SubgraphEdge _predecessor_fw;
		std::vector<SubgraphEdge> _predecessors_bw;

		SearchNode_tmpl(const SearchNode_tmpl&) = delete;
		SearchNode_tmpl& operator=(const SearchNode_tmpl&) = delete;

		SearchNode_tmpl(SearchNode_tmpl&&) = default;
		SearchNode_tmpl& operator=(SearchNode_tmpl&&) = default;

		SearchNode_tmpl()
		: _heap_handle(),
		_node_id(NodeId::INVALID()),
		_flags(0),
		_shift_time(0),
		_breakTaken(false),
		_break_late(false),

		  _time_fw(std::numeric_limits<int32_t>::max()),
		  _dist_fw(std::numeric_limits<uint32_t>::max()),
		  _interval_bw(Interval::INFTY()),
		  _upper_stall({std::numeric_limits<int32_t>::max(), std::numeric_limits<int32_t>::max()}),
		  _predecessor_fw(),
		  _predecessors_bw()
		{}

		bool is_enqueued(const size_t direction) const
		{
			assert(direction == 0 || direction == 1);
			return (_flags & (1 << direction)) != 0;
		}

		void set_enqueued(bool enqueued, uint32_t direction)
		{
			assert(direction == 0 || direction == 1);

			if ( enqueued )
				_flags |= (1 << direction);
			else
				_flags &= ~(1 << direction);
		}
	};

	class SearchContext
	{
	private:

		struct HeapItem
		{
			SearchId _search_id;
			int32_t _priority;

			HeapItem(SearchId search_id, int32_t priority)
			: _search_id(search_id), _priority(priority)
			{}

			friend inline bool operator< (const HeapItem& lhs, const HeapItem& rhs) { return rhs._priority < lhs._priority; }
		};

		struct NodeIdHash
		{
			size_t operator() (NodeId node_id) const noexcept { return std::hash<uint32_t>()(uint32_t(node_id)); }
		};

		typedef boost::heap::pairing_heap<HeapItem> HeapImpl;
		typedef HeapImpl::handle_type HeapHandle;

		std::array<HeapImpl, 2> _heap;
		std::unordered_map<NodeId, SearchId, NodeIdHash> _table;

		std::unordered_map<NodeId, SearchId, NodeIdHash> _table_breaklate;
		std::unordered_map<NodeId, SearchId, NodeIdHash> _table_breaklatedown;

		std::unordered_map<NodeId, SearchId, NodeIdHash> _table_breakblock;
		std::unordered_map<NodeId, SearchId, NodeIdHash> _table_breakblockdown;


		std::vector<SearchNode_tmpl<HeapHandle>> _search_nodes;

	public:

		typedef SearchNode_tmpl<HeapHandle> SearchNode;

		SearchContext()
			: _heap(),
			_table(),
			_table_breaklate(),
			_table_breaklatedown(),
			_table_breakblock(),
			_table_breakblockdown(),
		  _search_nodes()
		{}

		bool pq_empty(const uint32_t direction) const noexcept
		{
			assert(direction < _heap.size());
			return _heap[direction].empty();
		}

		bool pq_contains(const NodeId node_id, const uint32_t direction) const noexcept
		{
			bool nobreak = false;
			bool breakblock = false;
			bool breakblockdown = false;
			bool breaklate = false;
			bool breaklatedown = false;
			auto it = _table.find(node_id);
			if (it != _table.end()) {
				const SearchId search_id = it->second;
				nobreak = _search_nodes[search_id.the_value()].is_enqueued(direction);
			}
			it = _table_breaklate.find(node_id);
 			if (it != _table_breaklate.end()) {
				const SearchId search_id = it->second;
				breaklate = _search_nodes[search_id.the_value()].is_enqueued(direction);
			}
			it = _table_breaklatedown.find(node_id);
			if (it != _table_breaklatedown.end()) {
				const SearchId search_id = it->second;
				breaklatedown = _search_nodes[search_id.the_value()].is_enqueued(direction);
			}

			it = _table_breakblock.find(node_id);
			if (it != _table_breakblock.end()) {
				const SearchId search_id = it->second;
				breakblock = _search_nodes[search_id.the_value()].is_enqueued(direction);
			}
			it = _table_breakblockdown.find(node_id);
			if (it != _table_breakblockdown.end()) {
				const SearchId search_id = it->second;
				breakblockdown = _search_nodes[search_id.the_value()].is_enqueued(direction);
			}


			return nobreak || breakblock || breakblockdown || breaklate || breaklatedown;
		}

		bool pq_contains(const SearchNode& search_node, const uint32_t direction) const noexcept
		{
			return search_node.is_enqueued(direction);
		}

		bool touched(const NodeId node_id) const noexcept
		{

			return _table.count(node_id) > 0;
		}

		bool touched_break_late(const NodeId node_id) const noexcept
		{

			return _table_breaklate.count(node_id) > 0;
		}
		bool touched_break_late_down(const NodeId node_id) const noexcept
		{

			return _table_breaklatedown.count(node_id) > 0;
		}
		bool touched_break_late_down() const noexcept
		{

			return _table_breaklatedown.size() > 0;
		}

		bool touched_break_block(const NodeId node_id) const noexcept
		{

			return _table_breakblock.count(node_id) > 0;
		}

		bool touched_break_block_down(const NodeId node_id) const noexcept
		{

			return _table_breakblockdown.count(node_id) > 0;
		}



		const SearchNode& get_search_node(const NodeId node_id) const noexcept
		{
			assert(_table.count(node_id) > 0);
			return _search_nodes[_table.find(node_id)->second.the_value()];
		}

		const SearchNode& get_search_node_break_late(const NodeId node_id) const noexcept
		{
			assert(_table_breaklate.count(node_id) > 0);
			return _search_nodes[_table_breaklate.find(node_id)->second.the_value()];
		}

		const SearchNode& get_search_node_break_late_down(const NodeId node_id) const noexcept
		{
			assert(_table_breaklatedown.count(node_id) > 0);
			return _search_nodes[_table_breaklatedown.find(node_id)->second.the_value()];
		}


		const SearchNode& get_search_node_break_block(const NodeId node_id) const noexcept
		{
			assert(_table_breakblock.count(node_id) > 0);
			return _search_nodes[_table_breakblock.find(node_id)->second.the_value()];
		}

		const SearchNode& get_search_node_break_block_down(const NodeId node_id) const noexcept
		{
			assert(_table_breakblockdown.count(node_id) > 0);
			return _search_nodes[_table_breakblockdown.find(node_id)->second.the_value()];
		}


		SearchNode& get_search_node(const NodeId node_id) noexcept
		{
			assert(_table.count(node_id) > 0);
			return _search_nodes[_table.find(node_id)->second.the_value()];
		}

		SearchNode& get_search_node_break_late(const NodeId node_id) noexcept
		{
			assert(_table_breaklate.count(node_id) > 0);
			return _search_nodes[_table_breaklate.find(node_id)->second.the_value()];
		}

		SearchNode& get_search_node_break_late_down(const NodeId node_id) noexcept
		{
			assert(_table_breaklatedown.count(node_id) > 0);
			return _search_nodes[_table_breaklatedown.find(node_id)->second.the_value()];
		}


		SearchNode& get_search_node_break_block(const NodeId node_id) noexcept
		{
			assert(_table_breakblock.count(node_id) > 0);
			return _search_nodes[_table_breakblock.find(node_id)->second.the_value()];
		}

		SearchNode& get_search_node_break_block_down(const NodeId node_id) noexcept
		{
			assert(_table_breakblockdown.count(node_id) > 0);
			return _search_nodes[_table_breakblockdown.find(node_id)->second.the_value()];
		}

		SearchNode& get_or_create_search_node(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table[node_id];

			if ( search_id == SearchId::INVALID() )
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
			}

			return _search_nodes[search_id.the_value()];
		}

		SearchNode& get_or_create_search_node_break_late(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table_breaklate[node_id];

			if (search_id == SearchId::INVALID())
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
				_search_nodes.back()._breakTaken = true;
				_search_nodes.back()._break_late = true;
			}

			return _search_nodes[search_id.the_value()];
		}
		
		SearchNode& get_or_create_search_node_break_late_down(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table_breaklatedown[node_id];

			if (search_id == SearchId::INVALID())
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
				_search_nodes.back()._breakTaken = true;
				_search_nodes.back()._break_late = true;
			}

			return _search_nodes[search_id.the_value()];
		}

		SearchNode& get_or_create_search_node_break_block(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table_breakblock[node_id];

			if (search_id == SearchId::INVALID())
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
				_search_nodes.back()._breakTaken = true;
				_search_nodes.back()._break_late = false;
			}

			return _search_nodes[search_id.the_value()];
		}

		SearchNode& get_or_create_search_node_break_block_down(const NodeId node_id) noexcept
		{
			SearchId& search_id = _table_breakblockdown[node_id];

			if (search_id == SearchId::INVALID())
			{
				search_id = _search_nodes.size();
				_search_nodes.emplace_back();
				_search_nodes.back()._node_id = node_id;
				_search_nodes.back()._breakTaken = true;
				_search_nodes.back()._break_late = false;
			}

			return _search_nodes[search_id.the_value()];
		}



		SearchNode& get_search_node_from_sid(const SearchId search_id)
		{
			//assert(search_id < _search_nodes.size());
			return _search_nodes[search_id.the_value()];
		}

		const SearchNode& get_search_node_from_sid(const SearchId search_id) const
		{
			//assert(search_id < _search_nodes.size());
			return _search_nodes[search_id.the_value()];
		}

		SearchId get_search_id(const SearchNode& search_node) const
		{
			//assert(touched(search_node._node_id));
			const SearchId search_id(std::distance(&(_search_nodes[0]), &search_node));
			return search_id;
		}

		const SearchNode& pq_min_search_node(const uint32_t direction) const
		{
			// assert(!pq_empty(direction));
			return _search_nodes[_heap[direction].top()._search_id.the_value()];
		}

		int32_t pq_min_priority(const uint32_t direction) const
		{
			assert(!pq_empty(direction));
			return _heap[direction].top()._priority;
		}

		void pq_insert_or_update(SearchNode& search_node, const int32_t priority, const uint32_t direction)
		{
			const SearchId search_id = get_search_id(search_node);

			if ( ! search_node.is_enqueued(direction) )
			{
				const HeapHandle heap_handle = _heap[direction].push(HeapItem(search_id, priority));
				search_node._heap_handle[direction] = heap_handle;
				search_node.set_enqueued(true, direction);
			}
			else
			{
				const HeapHandle heap_handle = search_node._heap_handle[direction];
				_heap[direction].update(heap_handle, HeapItem(search_id, priority));
			}
		}

		SearchNode& pq_delete_min(const uint32_t direction)
		{
			assert(!pq_empty(direction));

			SearchNode& search_node = _search_nodes[_heap[direction].top()._search_id.the_value()];
			//assert(_table.count(search_node._node_id) > 0 || _table_breaklate.count(search_node._node_id) > 0||  _table_breaklatedown.count(search_node._node_id) > 0 || _table_breakblock.count(search_node._node_id) > 0);
			_heap[direction].pop();

			search_node.set_enqueued(false, direction);
			return search_node;
		}

		void pq_clear(const uint32_t direction)
		{
			for ( const auto& heap_item : _heap[direction] )
				_search_nodes[heap_item._search_id.the_value()].set_enqueued(false, direction);

			_heap[direction].clear();
		}

		void clear_all()
		{
			_heap[FORWARD()].clear();
			_heap[BACKWARD()].clear();
			_table.clear();
			_table_breakblock.clear();
			_table_breakblockdown.clear();
			_table_breaklate.clear();
			_table_breaklatedown.clear();
			_search_nodes.clear();
		}
	};

	typedef SearchContext::SearchNode SearchNode;

	static uint32_t inverse(const uint32_t direction)
	{
		if ( direction == FORWARD() )
			return BACKWARD();

		return FORWARD();
	}

	void bidirectional_search(const Location& start, const Location& destination, const int32_t departure_time)
	{
		_context.clear_all();

		//Add start node to forward pq
		for ( size_t i = 0 ; i != start.size() ; ++i )
		{
			const NodeId& node_id = start.get_node_id(i);
			assert(!_context.pq_contains(node_id, FORWARD()));

			const TimeDist& time_dist = start.get_time_dist(i);
			const int32_t departure_time_on_node = time_dist.eval_time(departure_time) + departure_time;
			assert(departure_time_on_node < 604800);
			const uint32_t distance_on_node = time_dist.eval_dist(departure_time);

			SearchNode& search_node = _context.get_or_create_search_node(node_id);
			_context.pq_insert_or_update(search_node, departure_time_on_node, FORWARD());
			search_node._time_fw = departure_time_on_node;
			search_node._dist_fw = distance_on_node;
		}

		//Add destination node to backward pq
		for ( size_t i = 0 ; i != destination.size() ; ++i )
		{
			const NodeId& node_id = destination.get_node_id(i);
			assert(!_context.pq_contains(node_id, BACKWARD()));

			const TimeDist& time_dist = destination.get_time_dist(i);

			SearchNode& search_node = _context.get_or_create_search_node(node_id);
			_context.pq_insert_or_update(search_node, time_dist.get_min_time(), BACKWARD());
			search_node._interval_bw = Interval(time_dist.get_min_time(), time_dist.get_max_time());

		}

		_candidate_search_ids.clear();
		_upper_bound = std::numeric_limits<int32_t>::max();
		_upper_bound_break = std::numeric_limits<int32_t>::max();


		uint32_t direction = BACKWARD();
		std::array<bool, 2> finished = {false, false};

		while ( ! _context.pq_empty(FORWARD()) || ! _context.pq_empty(BACKWARD()) )
		{
			//alternate forward/backward direction
			if ( finished[FORWARD()] && finished[BACKWARD()] ) break;
			if ( ! _context.pq_empty(inverse(direction)) ) direction = inverse(direction);
			if ( finished[FORWARD()] ) direction = BACKWARD();
			if ( finished[BACKWARD()] ) direction = FORWARD();

			if ( _context.pq_empty(direction) )
			{
				finished[direction] = true;
				continue;
			}

			//
			// take node with minimal label out of PQ
			//
			SearchNode& u = _context.pq_delete_min(direction);

			//
			// finish search if current minimal label exceeds upper bound of up-down-paths see so far
			// TODO : check if this is still valid?
			//

			/*if ( direction == FORWARD() && u._time_fw > _upper_bound )
			{
				finished[FORWARD()] = true;
				continue;
			}
			else if ( direction == BACKWARD() && u._interval_bw.get_lower() > _upper_bound )
			{
				finished[BACKWARD()] = true;
				continue;
			}*/

			//
			// store candidates and update upper bound of up-down-paths seen so far
			//
			
			//TODO: correct for if bw search > find node in fw (break)tables
			if (u._breakTaken) {
				if (_context.touched(u._node_id)) {
					SearchNode& back = _context.get_search_node(u._node_id);
					if (u._time_fw != std::numeric_limits<int32_t>::max() && !back._interval_bw.is_infty())
					{
						if (u._time_fw - BREAK_TIME + back._interval_bw.get_lower() <= _upper_bound_break) {
							_candidate_search_ids.push_back(_context.get_search_id(u));
						}
						_upper_bound_break = std::min(_upper_bound_break, u._time_fw - BREAK_TIME + back._interval_bw.get_upper());
					}
				}
			}

			if ( u._time_fw != std::numeric_limits<int32_t>::max() && !u._interval_bw.is_infty() )
			{
				if (u._time_fw + u._interval_bw.get_lower() <= _upper_bound) {
						_candidate_search_ids.push_back(_context.get_search_id(u));
				}
				_upper_bound = std::min(_upper_bound, u._time_fw + u._interval_bw.get_upper());
			}

			//
			// relax edges
			//

			const SearchId u_sid = _context.get_search_id(u);
			const NodeId u_id = u._node_id;
			const int32_t u_time_fw = u._time_fw;
			const int32_t u_shift_fw = u._shift_time;
			const uint32_t u_dist_fw = u._dist_fw;
			const Interval u_interval_bw = u._interval_bw;
			const bool u_break = u._breakTaken;
			const bool u_break_late = u._break_late;

			for ( EdgeId e = _graph->edges_begin(u_id) ; e != _graph->edges_end(u_id) ; ++e )
			{
				if ( direction == FORWARD() && ! _graph->is_directed_upward(e) ) continue;
				else if ( direction == BACKWARD() && ! _graph->is_directed_downward(e) ) continue;

				const NodeId v_id = _graph->get_other_node(e);

				if (direction == FORWARD())
				{
					if (u_break) {
						if (u_break_late) {
							SearchNode& bl = _context.get_or_create_search_node_break_late(v_id);
							const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
							const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
							int32_t v_shift_fw_new = get_default_time_of_edge(u_id,v_id,e,u_time_fw) + u_shift_fw; //waiting does not count for driving time, thus default time (FF) used
							if (v_time_fw_new < bl._time_fw || (v_time_fw_new == bl._time_fw && v_shift_fw_new < bl._shift_time))
							{
								if (v_shift_fw_new < SHIFT_THRESHOLD) {
									bl._time_fw = v_time_fw_new;
									bl._shift_time = v_shift_fw_new;
									bl._dist_fw = v_dist_fw_new;
									bl._breakTaken = true;
									bl._break_late = true;
									bl._predecessor_fw = SubgraphEdge(e, u_sid);
									_context.pq_insert_or_update(bl, bl._time_fw, FORWARD());
								}
							}
						}
						else {
							SearchNode& bb = _context.get_or_create_search_node_break_block(v_id);
							const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
							const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
							int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shift_fw; //waiting at block does not count as driving time
							if (v_time_fw_new < bb._time_fw || (v_time_fw_new == bb._time_fw && v_shift_fw_new < bb._shift_time))
							{
								if (v_shift_fw_new < SHIFT_THRESHOLD) {
									bb._time_fw = v_time_fw_new;
									bb._shift_time = v_shift_fw_new;
									bb._dist_fw = v_dist_fw_new;
									bb._breakTaken = true;
									bb._break_late = false;
									bb._predecessor_fw = SubgraphEdge(e, u_sid);
									_context.pq_insert_or_update(bb, bb._time_fw, FORWARD());
								}
							}
						}
					}
					else {
						//LATEST BREAK
						if (get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shift_fw > SHIFT_THRESHOLD) {
							//calculate Shift+Traveltime for latestBreak insertion
							SearchNode& bl = _context.get_or_create_search_node_break_late(v_id);
							std::tuple<int32_t, uint32_t, int32_t> latestBreakInsertion = calcLatestBreak(e,u_id, v_id, u_time_fw, u_dist_fw, u_shift_fw); // time, distance, shiftduration
							if (std::get<0>(latestBreakInsertion) < bl._time_fw || (std::get<0>(latestBreakInsertion) == bl._time_fw && std::get<2>(latestBreakInsertion) < bl._shift_time))
							{
								if (std::get<2>(latestBreakInsertion) < SHIFT_THRESHOLD) {
									bl._time_fw = std::get<0>(latestBreakInsertion);
									bl._shift_time = std::get<2>(latestBreakInsertion);
									bl._dist_fw = std::get<1>(latestBreakInsertion);
									bl._breakTaken = true;
									bl._break_late = true;
									bl._predecessor_fw = SubgraphEdge(e, u_sid);
									_context.pq_insert_or_update(bl, bl._time_fw, FORWARD());
								}
							}
						}
						
						//BLOCK BREAK
						else if (get_default_time_of_edge(u_id, v_id, e, u_time_fw) < _graph->eval_time(e, u_time_fw)) {
								SearchNode& bb = _context.get_or_create_search_node_break_block(v_id);
								std::tuple<int32_t, uint32_t, int32_t> latestBreakInsertion = calcBlockBreak(e, u_id, v_id, u_time_fw, u_dist_fw, u_shift_fw); // time, distance, shiftduration
								if (std::get<0>(latestBreakInsertion) < bb._time_fw || (std::get<0>(latestBreakInsertion) == bb._time_fw && std::get<2>(latestBreakInsertion) < bb._shift_time))
								{
									if (std::get<2>(latestBreakInsertion) < SHIFT_THRESHOLD) {
										bb._time_fw = std::get<0>(latestBreakInsertion);
										bb._shift_time = std::get<2>(latestBreakInsertion);
										bb._dist_fw = std::get<1>(latestBreakInsertion);
										bb._breakTaken = true;
										bb._break_late = false;
										bb._predecessor_fw = SubgraphEdge(e, u_sid);
										_context.pq_insert_or_update(bb, bb._time_fw, FORWARD());
									}
								}
							}

						//NORMAL RELAX
						SearchNode& v = _context.get_or_create_search_node(v_id);
						const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
						const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
						int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shift_fw; //waiting at block does not count as driving time
						if (v_time_fw_new < v._time_fw || (v_time_fw_new == v._time_fw && v_shift_fw_new < v._shift_time))
						{
							if (v_shift_fw_new < SHIFT_THRESHOLD) {
								v._time_fw = v_time_fw_new;
								v._shift_time = v_shift_fw_new;
								v._dist_fw = v_dist_fw_new;
								v._breakTaken = false;
								v._predecessor_fw = SubgraphEdge(e, u_sid);
								_context.pq_insert_or_update(v, v._time_fw, FORWARD());
							}
						}	
					}

				}
				else
				{
					int32_t min_e, max_e;
					std::tie(min_e, max_e) = _graph->get_min_max_time(e);

					Interval v_interval_bw_new(u_interval_bw.get_lower() + min_e , u_interval_bw.get_upper() + max_e);

					SearchNode& v = _context.get_or_create_search_node(v_id);

					if ( v._interval_bw.is_infty() )
					{
						v._interval_bw = v_interval_bw_new;
						v._predecessors_bw.emplace_back(e, u_sid);
					}
					else
					{
						if ( v_interval_bw_new.get_lower() > v._interval_bw.get_upper() )
							continue;

						if ( v_interval_bw_new.get_upper() < v._interval_bw.get_lower() )
							v._predecessors_bw.clear();

						v._predecessors_bw.emplace_back(e, u_sid);

						if ( v_interval_bw_new.get_lower() >= v._interval_bw.get_lower() )
							if ( v_interval_bw_new.get_upper() >= v._interval_bw.get_upper() )
								continue;

						v._interval_bw =
							Interval(
								std::min(v._interval_bw.get_lower(), v_interval_bw_new.get_lower()),
								std::min(v._interval_bw.get_upper(), v_interval_bw_new.get_upper())
							);
					}

					_context.pq_insert_or_update(v, v._interval_bw.get_lower(), BACKWARD());
				}
			}
		}
	}

	void downward_search()
	{
		util::make_unique(_candidate_search_ids);
		_context.pq_clear(FORWARD());

		//std::ofstream cnodes;
		//cnodes.open("D:/marieket/Documents/Maps/NBNL4pl26PPCF01/Results/cnodes-earouting.txt", std::ofstream::out | std::ofstream::app);
		//cnodes << "list: " << std::endl;
		for ( const SearchId& x_id : _candidate_search_ids )
		{
			SearchNode& x = _context.get_search_node_from_sid(x_id);
			//cnodes << x._node_id << " " << x._time_fw << std::endl;
			//get right interval_bw for break taking searchnodes!
			//if (!x._breakTaken &&  x._time_fw + x._interval_bw.get_lower() > _upper_bound_break ) continue;
			//if (x._breakTaken &&  x._time_fw + x._interval_bw.get_lower() - BREAK_TIME > _upper_bound) continue;

			//std::cout << std::endl << x._breakTaken << "," << x._break_late << std::endl;

			
			_context.pq_insert_or_update(x, x._time_fw, FORWARD());
		}
		//cnodes << std::endl;
		//cnodes.close();


		while ( ! _context.pq_empty(FORWARD()) )
		{
			const SearchNode& u = _context.pq_delete_min(FORWARD());
			const SearchId u_sid = _context.get_search_id(u);
			const NodeId u_id = u._node_id;

			const int32_t u_time_fw = u._time_fw;
			const int32_t u_shiftDuration = u._shift_time;
			const uint32_t u_dist_fw = u._dist_fw;
			const Interval u_interval_bw = u._interval_bw;
			const bool u_break = u._breakTaken;
			const bool u_break_late = u._break_late;

			std::vector<SubgraphEdge> predecessorlist = u._predecessors_bw;
			if (u._breakTaken)
			{
				RITCH_ASSERT(predecessorlist.size() == 0);
				if (_context.touched(u._node_id)) {
					SearchNode& equi = _context.get_search_node(u._node_id);
					predecessorlist = equi._predecessors_bw;
				}
			}
			RITCH_ASSERT(predecessorlist.size() > 0 || int32_t(u._node_id) == int32_t(_dest.get_node_id(0)));

			for (const auto& pred : predecessorlist)
			{
				const EdgeId e = pred._edge_id;

				assert(_graph->is_directed_downward(e));
				
				const NodeId v_id2 = _graph->get_other_node(e);
				const NodeId v_id = _context.get_search_node_from_sid(pred._search_id)._node_id;

				if (u._breakTaken) {
					if (u._break_late) {
						SearchNode& bl = _context.get_or_create_search_node_break_late_down(v_id);
						const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
						const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
						int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shiftDuration; //waiting at block does not count as driving time
						if (v_time_fw_new < bl._time_fw || (v_time_fw_new == bl._time_fw && v_shift_fw_new < bl._shift_time))
						{
							if (v_shift_fw_new < SHIFT_THRESHOLD) {
								bl._time_fw = v_time_fw_new;
								bl._shift_time = v_shift_fw_new;
								bl._dist_fw = v_dist_fw_new;
								bl._breakTaken = true;
								bl._break_late = true;
								bl._predecessor_fw = SubgraphEdge(e, u_sid);
								_context.pq_insert_or_update(bl, bl._time_fw, FORWARD());
							}
						}
					}
					else {
						SearchNode& bb = _context.get_or_create_search_node_break_block_down(v_id);

						const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u._dist_fw;
						const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
						int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shiftDuration; //waiting at block does not count as driving time
						if (v_time_fw_new < bb._time_fw || (v_time_fw_new == bb._time_fw && v_shift_fw_new < bb._shift_time))
						{
							if (v_shift_fw_new < SHIFT_THRESHOLD) {
								bb._time_fw = v_time_fw_new;
								bb._shift_time = v_shift_fw_new;
								bb._dist_fw = v_dist_fw_new;
								bb._breakTaken = true;
								bb._break_late = false;
								bb._predecessor_fw = SubgraphEdge(e, u_sid);
								_context.pq_insert_or_update(bb, bb._time_fw, FORWARD());
							}
						}
					}
				}
				else {
					//LATEST BREAK 
					if (get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shiftDuration > SHIFT_THRESHOLD) {
						SearchNode& bl = _context.get_or_create_search_node_break_late_down(v_id);

						std::tuple<int32_t, uint32_t, int32_t> latestBreakInsertion = calcLatestBreak(e, u_id, v_id, u_time_fw, u_dist_fw, u_shiftDuration); // time, distance, shiftduration
						if (std::get<0>(latestBreakInsertion) < bl._time_fw || (std::get<0>(latestBreakInsertion) == bl._time_fw && std::get<2>(latestBreakInsertion) < bl._shift_time))
						{
							if (std::get<2>(latestBreakInsertion) < SHIFT_THRESHOLD) {
								bl._time_fw = std::get<0>(latestBreakInsertion);
								bl._shift_time = std::get<2>(latestBreakInsertion);
								bl._dist_fw = std::get<1>(latestBreakInsertion);
								bl._breakTaken = true;
								bl._break_late = true;
								bl._predecessor_fw = SubgraphEdge(e, u_sid);
								_context.pq_insert_or_update(bl, bl._time_fw, FORWARD());
							}
						}
					}
					//BLOCK BREAK
					else if (_graph->eval_time(e, u._time_fw) > get_default_time_of_edge(u_id, v_id, e, u_time_fw)) {
						SearchNode& bb = _context.get_or_create_search_node_break_block_down(v_id);

						std::tuple<int32_t, uint32_t, int32_t> latestBreakInsertion = calcBlockBreak(e, u_id, v_id, u_time_fw, u_dist_fw, u_shiftDuration); // time, distance, shiftduration
						if (std::get<0>(latestBreakInsertion) < bb._time_fw || (std::get<0>(latestBreakInsertion) == bb._time_fw && std::get<2>(latestBreakInsertion) < bb._shift_time))
						{
							if (std::get<2>(latestBreakInsertion) < SHIFT_THRESHOLD) {
								bb._time_fw = std::get<0>(latestBreakInsertion);
								bb._shift_time = std::get<2>(latestBreakInsertion);
								bb._dist_fw = std::get<1>(latestBreakInsertion);
								bb._breakTaken = true;
								bb._break_late = false;
								bb._predecessor_fw = SubgraphEdge(e, u_sid);
								_context.pq_insert_or_update(bb, bb._time_fw, FORWARD());
							}
						}
					}
					
					//NORMAL RELAX
					SearchNode& v = _context.get_search_node_from_sid(pred._search_id);
					const uint32_t v_dist_fw_new = _graph->eval_distance(e, u_time_fw) + u_dist_fw;
					const int32_t v_time_fw_new = _graph->eval_time(e, u_time_fw) + u_time_fw;
					int32_t v_shift_fw_new = get_default_time_of_edge(u_id, v_id, e, u_time_fw) + u_shiftDuration; //waiting at block does not count as driving time
					if (v_time_fw_new < v._time_fw || (v_time_fw_new == v._time_fw && v_shift_fw_new < v._shift_time))
					{
						if (v_shift_fw_new < SHIFT_THRESHOLD) {
							v._time_fw = v_time_fw_new;
							v._shift_time = v_shift_fw_new;
							v._dist_fw = v_dist_fw_new;
							v._breakTaken = false;
							v._break_late = false;
							v._predecessor_fw = SubgraphEdge(e, u_sid);
							_context.pq_insert_or_update(v, v._time_fw, FORWARD());
						}
					}
				}
				
			}
		}
	}

	std::tuple<int32_t, uint32_t, int32_t> calcLatestBreak(EdgeId e, NodeId u, NodeId v, int32_t utime, int32_t udistance, int32_t ushift) {
		std::tuple<int32_t, uint32_t, int32_t> res;
		std::tuple<int32_t, uint32_t, int32_t> res_b;
		std::pair<int32_t, uint32_t> block = get_block_info_of_edge(u,v,e,utime); // blockstart, blockend
		std::pair<int32_t, uint32_t> block_b = get_block_info_of_edge(u, v,e, utime+BREAK_TIME); // blockstart, blockend
		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> edgeblock = _graph->get_edge_block(u, v, e);

		int32_t block_defaulttime = get_default_time_of_edge(u, v, e, utime);
		int32_t block_b_defaulttime = get_default_time_of_edge(u, v, e, utime+BREAK_TIME);


		assert(block_defaulttime - SHIFT_THRESHOLD + ushift >= 0);

		if (utime + SHIFT_THRESHOLD - ushift + BREAK_TIME >= block.first  && utime + SHIFT_THRESHOLD - ushift <= block.first) {
			//std::cout << "ff:" << _graph->get_default_time(e) << ",bs:" << block.first << ",be:" << block_b.second << "fe(tau)" << _graph->eval_time(e, utime);
			//std::get<0>(res) = utime + block.second + block_defaulttime - (utime + SHIFT_THRESHOLD - ushift);
			std::get<0>(res) = utime + _graph->eval_time(e, utime);
			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime - SHIFT_THRESHOLD + ushift;
		}
		else {
			std::get<0>(res) = utime + _graph->eval_time(e, utime) + BREAK_TIME;
			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime - SHIFT_THRESHOLD + ushift;
		}

		if (utime + SHIFT_THRESHOLD - ushift + BREAK_TIME >= block_b.first  && utime + SHIFT_THRESHOLD - ushift <= block_b.first) {
			//std::get<0>(res_b) = utime + block_b.second + block_b_defaulttime - (utime + SHIFT_THRESHOLD - ushift ); //klopt niet!
			std::get<0>(res_b) = utime + _graph->eval_time(e, utime+BREAK_TIME);
			std::get<1>(res_b) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res_b) = block_b_defaulttime - SHIFT_THRESHOLD + ushift;
		}
		else {
			std::get<0>(res_b) = utime + _graph->eval_time(e, utime + BREAK_TIME) + BREAK_TIME;
			std::get<1>(res_b) = _graph->eval_distance(e, utime + BREAK_TIME) + udistance;
			std::get<2>(res_b) = block_b_defaulttime - SHIFT_THRESHOLD + ushift;
		}
		if (std::get<0>(res) < utime + _graph->eval_time(e, utime)) {
			std::cout << std::get<0>(res) << " " << block.first << " " << block.second << " " << block_defaulttime << " " << _graph->eval_time(e, utime);
		}

		RITCH_ASSERT(std::get<0>(res) >= utime + _graph->eval_time(e, utime));
		RITCH_ASSERT(std::get<0>(res_b) >= utime + _graph->eval_time(e, utime + BREAK_TIME));

		if (std::get<0>(res_b) < std::get<0>(res)) {
			if (std::get<0>(res_b) < 0) {
				std::cout << std::get<0>(res_b) << " " << block_b.first << " " << block_b.second << " " << block_b_defaulttime << " " << _graph->eval_time(e, utime);
			}
			return res_b;
		}
		if (std::get<0>(res) < 0) {
			std::cout << std::get<0>(res) << " " << block.first << " " << block.second << " " << block_defaulttime << " " << _graph->eval_time(e, utime);
		}
		return res;
	}

	std::tuple<int32_t, uint32_t, int32_t> calcBlockBreak(EdgeId e, NodeId u, NodeId v, int32_t utime, int32_t udistance, int32_t ushift) {
		std::tuple<int32_t, uint32_t, int32_t> res;
		std::tuple<int32_t, uint32_t, int32_t> res_b;
		std::pair<int32_t, int32_t> block = get_block_info_of_edge(u, v, e, utime); // blockstart, blockend
		int32_t block_defaulttime = get_default_time_of_edge(u, v, e, utime);
		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> edgeblock = _graph->get_edge_block(u, v, e);

		assert(block.first == block.second && block.first == 0 || block.second > block.first);

		if (block.first == block.second) {
			std::get<0>(res) = std::numeric_limits<int32_t>::max();
			std::get<1>(res) = std::numeric_limits<uint32_t>::max();
			std::get<2>(res) = std::numeric_limits<int32_t>::max();
			return res;
		}
		if (block.second <= utime) {
			std::get<0>(res) = std::numeric_limits<int32_t>::max();
			std::get<1>(res) = std::numeric_limits<uint32_t>::max();
			std::get<2>(res) = std::numeric_limits<int32_t>::max();
			return res;
		}
		if (block.first > utime + block_defaulttime) {
			std::get<0>(res) = std::numeric_limits<int32_t>::max();
			std::get<1>(res) = std::numeric_limits<uint32_t>::max();
			std::get<2>(res) = std::numeric_limits<int32_t>::max();
			//std::cout << "error in blockbreak, ff:" << _graph->get_default_time(e) << ",bs:" << block.first << ",be:" << block.second << ", utime:" << utime << ",ushift:" << ushift << ",fe(tau)" << _graph->eval_time(e, utime) << std::endl;
			return res;
		}
		//assert(_graph->get_default_time(e) - SHIFT_THRESHOLD + ushift >= 0);

		if(utime + block_defaulttime >= block.first && utime  < block.first && block.first - utime < SHIFT_THRESHOLD - ushift ){
			//std::get<0>(res) = utime + block.second - block.first + block_defaulttime;
			std::get<0>(res) = utime + _graph->eval_time(e,utime);
			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime - (block.first-utime);
			if (std::get<0>(res) < utime + _graph->eval_time(e, utime)) {
				std::cout << "1:" << std::get<0>(res) << "," << utime + _graph->eval_time(e, utime) << std::endl;
				std::get<0>(res) = utime + _graph->eval_time(e, utime);
			}

		}
		else if (utime + block_defaulttime >= block.first && utime < block.first && block.first - utime >= SHIFT_THRESHOLD - ushift){
			//std::get<0>(res) = utime + block_defaulttime + block.second - block.first + block.second - (utime + (SHIFT_THRESHOLD - ushift));
			std::get<0>(res) = utime + (block.first - utime) - (SHIFT_THRESHOLD - ushift) + _graph->eval_time(e, utime + (block.first - utime) - (SHIFT_THRESHOLD - ushift));
			//time to block: block.first-utime
			//time to drive: threshold-ushift
			//time to break before block: block.first-utime - t-shift

			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime - SHIFT_THRESHOLD + ushift;
		}
		else if (utime >= block.first && utime <= block.second && utime + BREAK_TIME <= block.second) {
			//std::get<0>(res) = utime + block_defaulttime + block.second - utime;
			std::get<0>(res) = utime + _graph->eval_time(e,utime);
			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime;
		}
		else if (utime >= block.first && utime <= block.second && utime + BREAK_TIME > block.second){
			std::get<0>(res) = utime + BREAK_TIME + _graph->eval_time(e, utime+BREAK_TIME);
			//std::get<0>(res) = utime + BREAK_TIME bi+ _graph->eval_time(e,utime+BREAK_TIME);
			//time to block: block.first-utime
			//time to drive: threshold-ushift
			//time to break before block: block.first-utime - t-shift

			std::get<1>(res) = _graph->eval_distance(e, utime) + udistance;
			std::get<2>(res) = block_defaulttime;
		}
		else {
			std::cout << "error in blockbreak, ff:" << block_defaulttime << ",bs:" << block.first << ",be:" << block.second << ", utime:" <<utime <<",ushift:"<<ushift << ",fe(tau)" << _graph->eval_time(e, utime) <<std::endl;
		}
		
		RITCH_ASSERT(std::get<0>(res) >= utime + _graph->eval_time(e, utime));
		if (std::get<0>(res) < 0) {
			std::cout << std::get<0>(res) << " " << block.first << " " << block.second << " " << block_defaulttime << " " << _graph->eval_time(e, utime);
		}
		return res;
	}

	std::pair<int32_t, int32_t> get_block_info_of_edge(NodeId u, NodeId v, EdgeId e, int32_t tau) {
		const std::vector<std::tuple<int32_t,int32_t,int32_t, int32_t>> block = _graph->get_edge_block(u, v, e);
		if (block.size() == 0) {
			return std::make_pair(0, 0);
		}
		else {
			for (int i = 0; i < block.size() -1 ; i++) {
				if (tau > std::get<0>(block[i]) && tau <= std::get<0>(block[i + 1])){
					return std::make_pair(std::get<1>(block[i]), std::get<2>(block[i]));
				}
			}
		}
		return std::make_pair(std::get<1>(block[block.size() -1]), std::get<2>(block[block.size() - 1]));
	}

	int32_t get_default_time_of_edge(NodeId u, NodeId v, EdgeId e, int32_t tau) const{
		const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> block = _graph->get_edge_block(u, v, e);
		int32_t  defaultime = 0;
		if (block.size() == 0) {
			defaultime =  _graph->get_default_time(e);
		}
		else {
			if(tau > 604800 ){
				tau = tau - 604800;
			}
			if (tau <= std::get<0>(block[0])) {
				defaultime = std::get<3>(block[0]);
			}
			for (int i = 1; i < block.size(); i++) {
				if (defaultime == 0 && tau <= std::get<0>(block[i]) && tau > std::get<0>(block[i - 1])) {
					defaultime = std::get<3>(block[i]);
				}
			}
			if (defaultime == 0) {
				if (tau >= std::get<0>(block[block.size() - 1])) {
					defaultime = std::get<3>(block[0]);
				}
				else {
					defaultime = std::get<3>(block[block.size() - 1]);
				}
			}


			if (defaultime > _graph->eval_time(e, tau)) {
				std::cout << defaultime << " " << _graph->eval_time(e, tau) << " " << tau << std::endl;
				std::cout << u << " " << v << " " << e << " " << _graph->get_edge(e).get_other_node() << std::endl;
				std::cout << "block:" << std::endl;
				for (int i = 0; i < block.size(); i++) {
					std::cout << std::get<0>(block[i]) << " " << std::get<1>(block[i]) << " " << std::get<2>(block[i]) << " " << std::get<3>(block[i]) << std::endl;
				}
				std::cout << "traveltime:" << std::endl;
				TimeDist t = _graph->get_edge(e).get_time_dist();
				for (int i = 0; i < t.get_n_time_segments(); i++) {
					std::cout << t.get_time_segment(i, block).get_right_x() << " " << t.get_time_segment(i, block).get_right_y() << " " << t.get_time_segment(i, block).get_slope() << " " << t.get_time_segment(i, block).get_block_start() << " " << t.get_time_segment(i, block).get_block_end() << " " << t.get_time_segment(i, block).get_default_time() << std::endl;
				}

			}
		}
			
		RITCH_ASSERT(defaultime <= _graph->eval_time(e, tau));
		return defaultime;
	}

	bool extract_up_down_path_info(std::vector<NodeId>& o_reverse_node_ids, std::vector<EdgeId>& o_reverse_edge_ids) const noexcept
	{
		assert(o_reverse_node_ids.empty());
		assert(o_reverse_edge_ids.empty());

		if ( _arrival_node_id == NodeId::INVALID() )
			return false;

		SearchId search_id = _context.get_search_id(_context.get_search_node(_arrival_node_id));
		if (_context.get_search_node_from_sid(search_id)._time_fw == INT_MAX) {
			if (_context.touched_break_late_down(_arrival_node_id)) {
				search_id = _context.get_search_id(_context.get_search_node_break_late_down(_arrival_node_id));
			} 
			if (_context.touched_break_late(_arrival_node_id)) {
				search_id = _context.get_search_id(_context.get_search_node_break_late(_arrival_node_id));
			}
		}

		if (_context.get_search_id(_context.get_search_node(_dest.get_node_id(0))) == SearchId::INVALID()) 
		{
			return false;
		}

		while (true)
		{
			const SearchNode& search_node = _context.get_search_node_from_sid(search_id);
			o_reverse_node_ids.push_back(search_node._node_id);

			if (!o_reverse_edge_ids.empty() && search_node._predecessor_fw._search_id != SearchId::INVALID() && search_node._predecessor_fw._edge_id != EdgeId::INVALID() )
			{
				bool this_is_downward = _graph->is_directed_downward(o_reverse_edge_ids.back());
				bool pred_is_upward = _graph->is_directed_upward(search_node._predecessor_fw._edge_id);
			}

			//break taken during this node
			if(search_node._predecessor_fw._search_id != SearchId::INVALID() ) {
					if (search_node._node_id == _context.get_search_node_from_sid(search_node._predecessor_fw._search_id)._node_id)
					{
						search_id = search_node._predecessor_fw._search_id;
					}
					else
					{
						search_id = search_node._predecessor_fw._search_id;
						EdgeId edge_id = search_node._predecessor_fw._edge_id;
						assert(edge_id != EdgeId::INVALID());
						o_reverse_edge_ids.push_back(edge_id);
					}
				//}
			}
			else
			{
				//break; //no path possible
				return true;
			}

		}

		return true;
	}
	
	bool dbg_original_path_is_ok(const Path& path) const
	{


		for ( size_t i = 0 ; i != path.get_n_edges() ; ++i )
		{
			if ( _graph->has_middle_node(path.get_edge_id(i)) )
				return false;

			const auto edge = path.get_edge(i);
			NodeId src = std::get<0>(edge);
			EdgeId e = std::get<1>(edge);
			NodeId tgt = std::get<2>(edge);

			int32_t time_tgt_calculated = _graph->eval_time(e, path.get_arrival_time(i)) + path.get_arrival_time(i);
			int32_t dist_tgt_calculated = _graph->eval_distance(e, path.get_arrival_time(i)) + path.get_distance(i);

			if ( time_tgt_calculated != path.get_arrival_time(i+1) )
				return false;

			if ( dist_tgt_calculated != path.get_distance(i+1) )
				return false;
		}


		return true;
	}

	Location _start;
	Location _dest;
	int32_t _departure_time;

	TchSearchGraph const * const _graph;
	SearchContext _context;
	int32_t _upper_bound;
	int32_t _upper_bound_break;
	std::vector<SearchId> _candidate_search_ids;
	int32_t _arrival_time;
	int32_t _arrival_time_breaklate;
	int32_t _arrival_time_breakblock;
	int32_t _arrival_time_nobreak;
	bool _break;
	bool _late;
	NodeId _arrival_node_id;
	uint32_t _travel_distance;

public:

	TchQuerying(const TchSearchGraph& graph)
	: _start(),
	  _dest(),
	  _departure_time(std::numeric_limits<int32_t>::max()),
	  _graph(&graph),
	  _context(),
	  _upper_bound(std::numeric_limits<int32_t>::max()),
	  _arrival_time(std::numeric_limits<uint32_t>::max()),
	  _arrival_time_breaklate(std::numeric_limits<uint32_t>::max()),
		_arrival_time_breakblock(std::numeric_limits<uint32_t>::max()),
		_arrival_time_nobreak(std::numeric_limits<uint32_t>::max())

	{}

	//TODO: 
	//- use graph blocking instead of timedist-blocking
	//- test if preprocessing still works?
	//- 
	int32_t one_to_one(const Location& start, const Location& destination, const int32_t departure_time, bool longbreaks)
	{
		if (longbreaks) {
			SHIFT_THRESHOLD = 32400;
			BREAK_TIME = 39600;
		}
		else {
			SHIFT_THRESHOLD = 16200;
			BREAK_TIME = 2700;
		}
		_start = start;
		_dest = destination;
		_departure_time = departure_time;

		_arrival_time = std::numeric_limits<int32_t>::max();
		_arrival_time_nobreak = std::numeric_limits<int32_t>::max();
		_arrival_time_breakblock = std::numeric_limits<int32_t>::max();
		_arrival_time_breaklate = std::numeric_limits<int32_t>::max();

		_arrival_node_id = NodeId::INVALID();
		_break = false;
		_travel_distance = std::numeric_limits<uint32_t>::max();

		bidirectional_search(start, destination, departure_time);

		if ( _upper_bound == std::numeric_limits<int32_t>::max() )
			return std::numeric_limits<int32_t>::max();

		downward_search();

		for ( size_t i = 0 ; i != destination.size() ; ++i )
		{
			const NodeId node_id = destination.get_node_id(i);
			const SearchNode& search_node = _context.get_search_node(node_id);
			const SearchNode& search_node_bl = _context.touched_break_late_down(node_id) ? _context.get_search_node_break_late_down(node_id) : _context.get_search_node(node_id);
			const SearchNode& search_node_bb = _context.touched_break_block_down(node_id) ? _context.get_search_node_break_block_down(node_id) : _context.get_search_node(node_id);

			if (_context.touched_break_late_down(node_id)) {
				_arrival_time_breaklate = search_node_bl._time_fw + destination.get_time_dist(i).eval_time(search_node_bl._time_fw);
			}
			if (_context.touched_break_block_down(node_id)) {
				_arrival_time_breakblock = search_node_bb._time_fw + destination.get_time_dist(i).eval_time(search_node_bb._time_fw);
			}
			
			_arrival_time_nobreak = search_node._time_fw + destination.get_time_dist(i).eval_time(search_node._time_fw);
			_travel_distance = search_node._dist_fw + destination.get_time_dist(i).eval_dist(search_node._time_fw);

			_break = false;
			_late = false;
			_arrival_time = _arrival_time_nobreak;
			
			if (_arrival_time > _arrival_time_breaklate) {
				_arrival_time = _arrival_time_breaklate;
				_travel_distance = search_node_bl._dist_fw + destination.get_time_dist(i).eval_dist(search_node_bl._time_fw);
				_break = true;
				_late = true;
			}
			if (_arrival_time > _arrival_time_breakblock) {
				_arrival_time = _arrival_time_breakblock;
				_travel_distance = search_node_bb._dist_fw + destination.get_time_dist(i).eval_dist(search_node_bb._time_fw);
				_break = true;
				_late = false;
			}

			if (_arrival_time > 604800) {
				_arrival_time = _arrival_time - 604800; 
			}

			_arrival_node_id = node_id;
		}

		return _arrival_time;
	}

	int32_t get_arrival_time() const noexcept
	{
		return _arrival_time;
	}

	uint32_t get_travel_distance() const noexcept
	{
		return _travel_distance;
	}

	bool get_break() const {
		return _break;
	}
	bool get_late() const {
		return _late;
	}
	int32_t get_breaklate() const {
		return _arrival_time_breaklate;
	}
	int32_t get_breakblock() const {
		return _arrival_time_breakblock;
	}
	int32_t get_nobreak() const {
		return _arrival_time_nobreak;
	}

	Path retrieve_result_path() const
	{
		Path result_path;
		if ( _arrival_node_id == NodeId::INVALID() )
			return result_path;

		//
		// extract up-down-path from predecessor information
		//

		std::vector<NodeId> reverse_nodes;
		std::vector<EdgeId> reverse_edges;

		const bool path_found = extract_up_down_path_info(reverse_nodes, reverse_edges);

		if (path_found)
		{
			//
			// expand the up-down-path using the time-dependent middle node information computed
			// during preprocessing to obtain the result path in the original road network
			//

			std::stack<std::tuple<NodeId, EdgeId, NodeId>> S;
			const SearchNode& first_node_of_path = _context.get_search_node(reverse_nodes.back());
			int32_t current_time = first_node_of_path._time_fw;
			int32_t current_shift = first_node_of_path._shift_time;
			int32_t current_distance = first_node_of_path._dist_fw;

			std::vector<NodeId> result_nodes({ reverse_nodes.back() });
			std::vector<int32_t> result_arrival_times({ current_time });
			std::vector<int32_t> result_shift_times({ current_shift });
			std::vector<int32_t> result_distances({ current_distance });
			std::vector<EdgeId> result_edges;

			for (size_t i = reverse_edges.size(); i > 0; --i)
			{
				assert(i < reverse_nodes.size());
				// not correct: might be in break table
				//assert(current_time == _context.get_search_node(reverse_nodes[i])._time_fw);
				
				//assert(current_distance == _context.get_search_node(reverse_nodes[i])._dist_fw);

				{
					const NodeId& src = reverse_nodes[i];
					const NodeId& tgt = reverse_nodes[i - 1];
					const EdgeId& e = reverse_edges[i - 1];
					//assert(_graph->eval_time(e, current_time) + current_time == _context.get_search_node(tgt)._time_fw || BREAK_TIME + current_time == _context.get_search_node(tgt)._time_fw);

					//const uint32_t dist_e = _graph->eval_distance(e, current_time);
					//				assert(_graph->eval_distance(e, current_time) + current_distance == _context.get_search_node(tgt)._dist_fw);

					assert(S.empty());
					S.push(std::make_tuple(src, e, tgt));
				}

				while (!S.empty())
				{
					const auto& stack_top = S.top();
					NodeId src = std::get<0>(stack_top);
					EdgeId e_shortcut = std::get<1>(stack_top);
					NodeId tgt = std::get<2>(stack_top);
					S.pop();
					
					if (e_shortcut == EdgeId::INVALID()) {
						continue;
					}
					NodeId middle = _graph->get_middle_node_descr_from_departure_time(e_shortcut, current_time)._node_id;
					if (middle == NodeId::INVALID())
					{

						current_distance += _graph->eval_distance(e_shortcut, current_time);
						current_time += _graph->eval_time(e_shortcut, current_time);
						current_shift += get_default_time_of_edge(src, tgt, e_shortcut, current_time);
						result_edges.push_back(e_shortcut);
						result_nodes.push_back(tgt);
						result_arrival_times.push_back(current_time);
						result_shift_times.push_back(current_shift);
						result_distances.push_back(current_distance);

						continue;
					}
					
					EdgeId e_up;
					EdgeId e_down;
					for (EdgeId e = _graph->edges_begin(middle); e != _graph->edges_end(middle); ++e)
					{
						if (_graph->is_directed_upward(e) && _graph->get_other_node(e) == tgt) e_up = e;
						if (_graph->is_directed_downward(e) && _graph->get_other_node(e) == src) e_down = e;

						if (e_up != EdgeId::INVALID() && e_down != EdgeId::INVALID())
							break;
					}

					assert(e_up != EdgeId::INVALID());
					assert(e_down != EdgeId::INVALID());

#ifndef NDEBUG
					{
						int32_t time_at_middle = _graph->eval_time(e_down, current_time) + current_time;
						int32_t time_at_tgt_via_middle = _graph->eval_time(e_up, time_at_middle) + time_at_middle;
						int32_t time_at_tgt_via_shortcut = _graph->eval_time(e_shortcut, current_time) + current_time;
						//assert(time_at_tgt_via_middle == time_at_tgt_via_shortcut);

						int32_t dist_at_middle = _graph->eval_distance(e_down, current_time) + current_distance;
						int32_t dist_at_tgt_via_middle = _graph->eval_distance(e_up, time_at_middle) + dist_at_middle;
						int32_t dist_at_tgt_via_shortcut = _graph->eval_distance(e_shortcut, current_time) + current_distance;
						//assert(dist_at_tgt_via_middle == dist_at_tgt_via_shortcut);
					}
#endif

					S.push(std::make_tuple(middle, e_up, tgt));
					S.push(std::make_tuple(src, e_down, middle));
				}
			}

			result_path = Path(
				_start, _dest, _departure_time, _arrival_time, _travel_distance,
				std::move(result_nodes), std::move(result_edges), std::move(result_arrival_times), std::move(result_shift_times), std::move(result_distances));
			assert(dbg_original_path_is_ok(result_path));
		}
		return result_path;
	}
};



}

#endif // RITCH_QUERYING_TCH_QUERYING_H_
