#ifndef RITCH_TCH_SEARCH_GRAPH_H_
#define RITCH_TCH_SEARCH_GRAPH_H_

#include <cassert>
#include <cstdint>
#include <limits>
#include <string>
#include <vector>
#include <unordered_map>
#include <numeric>

//#include "ritch/datastr/blocking_profile.h"
#include "ritch/datastr/time_dist.h"
#include "ritch/io/ortec_format.h"
#include "ritch/io/ritch_shortcuts_format.h"
#include "ritch/util/strong_id.h"

namespace ritch
{

class TchSearchGraph
{

public:

	typedef TimeDist_tmpl<24*7*60*60> TimeDist;
	typedef std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> EdgeBlockType;


	RITCH_DEFINE_STRONG_ID(NodeId)
	RITCH_DEFINE_STRONG_ID(EdgeId)
	
	struct MiddleNodeDescr
	{
		int32_t _right_x;
		NodeId _node_id;
	};

private:
	struct NodeIdHash { size_t operator() (NodeId e) const noexcept { return std::hash<uint32_t>()(uint32_t(e)); } };

	struct NodeIdPairHash {
		size_t operator() (const std::pair<NodeId, NodeId>& p) const noexcept {
			return std::hash<uint32_t>()(uint32_t(p.first)) ^ std::hash<uint32_t>()(uint32_t(p.second));
		}
	};

	
	static int32_t calculate_default_travel_time(uint32_t speed, uint32_t distance) noexcept
	{
		return std::max(1, int32_t((36*distance) / speed + 5) / 10);
	}

	struct Node
	{
		EdgeId _first_edge;
	};

	struct Edge
	{
	public:
		static constexpr uint8_t INVALID_ROAD_TYPE() { return std::numeric_limits<uint8_t>::max(); }
		static constexpr uint16_t INVALID_BLOCKING_ID() { return std::numeric_limits<uint16_t>::max(); }

	private:

		static void do_init(Edge& edge) noexcept
		{
			edge._data = { 0xfffffffe, 0, 0, 0xffffffff };
		}

		static void do_copy(const Edge& from, Edge& to)
		{
			if ( from.has_time_dist() )
			{
				to._data[0] = from._data[0];
				new (reinterpret_cast<TimeDist*>(&(to._data[1]))) TimeDist(from.get_time_dist());
				to._data[3] = from._data[3];
			}
			else
			{
				to._data = from._data;
			}
		}

		static void do_move(Edge&& from, Edge& to) noexcept
		{
			to._data = from._data;
			do_init(from);
		}

		std::array<uint32_t, 4> _data;

	public:
		friend void swap(Edge& lhs, Edge& rhs) noexcept
		{
			if ( &lhs != &rhs )
			{
				std::array<uint32_t, 4> auxilary_data = lhs._data;
				lhs._data = rhs._data;
				rhs._data = auxilary_data;
			}
		}

		Edge() { do_init(*this); }

		Edge(const Edge& other) { do_copy(other, *this); }
		Edge(Edge&& other) { do_move(std::move(other), *this); }

		Edge& operator= (const Edge& other)
		{
			if ( this != &other )
			{
				if ( has_time_dist() )
					get_time_dist().TimeDist_tmpl::~TimeDist();

				do_copy(other, *this);
			}

			return *this;
		}

		Edge& operator= (Edge&& other)
		{
			if ( this != &other )
			{
				if ( has_time_dist() )
					get_time_dist().TimeDist_tmpl::~TimeDist();

				do_move(std::move(other), *this);
			}

			return *this;
		}

		~Edge()
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			do_init(*this);
		}

		bool is_directed_upward() const noexcept
		{
			return (_data[0] & 2) != 0;
		}

		void set_directed_upward(bool flag) noexcept
		{
			if ( flag )
				_data[0] |= 2;
			else
				_data[0] &= 0xffffffffd;
		}

		bool is_directed_downward() const noexcept
		{
			return ! is_directed_upward();
		}

		NodeId get_other_node() const noexcept
		{
			return NodeId(_data[0] >> 2);
		}

		void set_other_node(NodeId node_id) noexcept
		{
			_data[0] &= 3;
			_data[0] |= (node_id << 2);
		}

		bool has_time_dist() const noexcept
		{
			return (_data[0] & 1) != 0;
		}

		void set_time_dist(TimeDist&& time_dist) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] |= 1;
			new (reinterpret_cast<TimeDist*>(&(_data[1]))) TimeDist(std::move(time_dist));
		}

		void set_time_dist(const TimeDist& time_dist) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] |= 1;
			new (reinterpret_cast<TimeDist*>(&(_data[1]))) TimeDist(time_dist);
		}

		const TimeDist& get_time_dist() const noexcept
		{
			assert(has_time_dist());
			return (*reinterpret_cast<const TimeDist*>(&(_data[1])));
		}

		TimeDist& get_time_dist() noexcept
		{
			return const_cast<TimeDist&>(reinterpret_cast<const Edge&>(*this).get_time_dist());			
		}

		void set_blocking_index(uint16_t blocking_index)
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			reinterpret_cast<uint16_t*>(&(_data[1]))[0] = blocking_index;
		}

		uint16_t get_blocking_index() const noexcept
		{
			assert(!has_time_dist());
			return reinterpret_cast<const uint16_t*>(&(_data[1]))[0];
		}

		void set_road_type(uint8_t road_type) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			reinterpret_cast<uint16_t*>(&(_data[1]))[1] = road_type;
		}

		uint8_t get_road_type() const noexcept
		{
			assert(!has_time_dist());
			return uint8_t(reinterpret_cast<const uint16_t*>(&(_data[1]))[1]);
		}

		void set_distance(uint32_t distance) noexcept
		{
			if ( has_time_dist() )
				get_time_dist().TimeDist::~TimeDist();

			_data[0] &= 0xfffffffe;
			_data[2] = distance;
		}

		uint32_t get_distance() const noexcept
		{
			assert(!has_time_dist());
			return _data[2];
		}

		uint32_t get_middle_node_index() const noexcept
		{
			return _data[3];
		}

		void set_middle_node_index(uint32_t middle_node_index) noexcept
		{
			_data[3] = middle_node_index;
		}

		std::string to_text() const
		{
			std::stringstream ss;

			ss 
				<< "{other_node=" << get_other_node()
				<< ", up=" << util::to_text(is_directed_upward())
				<< ", down=" << util::to_text(is_directed_downward())
				<< ", middle_node_index=" << get_middle_node_index();

			if ( has_time_dist() )
				ss << ", time_dist=" << get_time_dist();
			else
				ss
					<< ", road_type=" << uint32_t(get_road_type())
					<< ", block_id=" << get_blocking_index()
					<< ", distance=" << get_distance();

			ss << "}";

			if ( !has_time_dist() )
				ss << " -> travel_time=" << calculate_default_travel_time( io::ortec_format::road_type_2_default_speed(get_road_type()), get_distance());

			return ss.str();
		}
	};

	class InputEdge
	{
	private:
		std::array<uint32_t, 5> _data;

		static void do_move(InputEdge& from, InputEdge& to) noexcept
		{
			new (&to.the_edge()) Edge(std::move(from.the_edge()));
			to._data[4] = from._data[4];
		}

	public:
		friend void swap(InputEdge& lhs, InputEdge& rhs) noexcept
		{
			swap(lhs.the_edge(), rhs.the_edge());
			std::swap(lhs._data[4], rhs._data[4]);
		}

		InputEdge()
		{
			new (&the_edge()) Edge();
			_data[4] = 0xffffffff;
		}

		InputEdge(const InputEdge& other) {
			new (&the_edge()) Edge(std::move(other.the_edge()));
			_data[4] = other._data[4];
		}

		InputEdge& operator=(const InputEdge& other){
			new (&the_edge()) Edge(std::move(other.the_edge()));
			_data[4] = other._data[4];
		}

		InputEdge(InputEdge&& other)
		{
			new (&the_edge()) Edge(std::move(other.the_edge()));
			_data[4] = other._data[4];
			//_edge_block_info = other._edge_block_info;
		}

		InputEdge& operator= (InputEdge&& other)
		{
			if ( this != &other )
			{
				the_edge() = std::move(other.the_edge());
				_data[4] = other._data[4];
				//_edge_block_info = other._edge_block_info;

			}

			return *this;
		}
 
		const Edge& the_edge() const noexcept { return *reinterpret_cast<const Edge*>(&_data); }
		Edge& the_edge() noexcept { return const_cast<Edge&>(static_cast<const InputEdge*>(this)->the_edge()); }

		NodeId get_node_id() const noexcept { return NodeId(_data[4] & ~(1 << 31)); }
		void set_node_id(NodeId node_id) noexcept { assert(node_id < ~(1 << 31)); _data[4] &= (1 << 31); _data[4] |= node_id; }
		void set_node_id(uint32_t node_id) noexcept { assert(node_id < ~(1 << 31)); _data[4] &= (1 << 31); _data[4] |= node_id; }

		bool is_original() const noexcept { return (_data[4] & (1 << 31)) != 0; }

		void set_original(bool is_original) noexcept
		{
			if ( is_original )
				_data[4] |= (1 << 31);
			else
				_data[4] &= ~(1 << 31);
		}

		std::string to_text() const
		{
			std::stringstream ss;
			ss
				<< "{node = " << get_node_id()
				<< ", orig=" << util::to_text(is_original())
				<< ", " << the_edge().to_text()
				<< "}";

			return ss.str();
		}
	};

	bool dbg_graph_is_ok() const
	{
		for ( size_t i = 1 ; i < _nodes.size() ; ++i )
			if ( _nodes[i]._first_edge < _nodes[i-1]._first_edge )
				return false;

		// TODO check more

		return true;
	}
	public:
	const Edge& get_edge(EdgeId edge_id) const noexcept
	{
		assert(edge_id != EdgeId::INVALID());
		assert(edge_id + 1 < _edges.size());
		return _edges[edge_id];
	}
	private:
	int32_t get_default_travel_time(EdgeId edge_id) const noexcept
	{
		const Edge& edge = get_edge(edge_id);

		const uint8_t road_type = edge.get_road_type();
		assert(road_type != Edge::INVALID_ROAD_TYPE());
		const uint32_t distance = edge.get_distance();
		assert(distance != Edge::INVALID_BLOCKING_ID());
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(road_type);

		return calculate_default_travel_time(speed, distance);
	}

	void load_ortec_graph_data(const std::string& graph_directory_name, std::vector<InputEdge>& io_input_edges)
	{
		io::InputFileIterator<uint32_t> target_it = io::ortec_format::get_target_iterator(graph_directory_name);
		io::InputFileIterator<uint32_t> distance_it = io::ortec_format::get_distance_iterator(graph_directory_name);
		io::InputFileIterator<uint8_t> type_it = io::ortec_format::get_road_type_iterator(graph_directory_name);

		size_t target_size_in_byte = target_it.get_size_in_byte();

		if ( distance_it.get_size_in_byte() != target_size_in_byte )
			throw std::runtime_error("Files lengtelong.net and naar_net have different number of edges.");

		if ( type_it.get_size_in_byte() != target_size_in_byte / 4 )
			throw std::runtime_error("Files type.net and naar_net have different number of edges.");

		const size_t n_edges = target_size_in_byte / 4;

		if (io_input_edges.size() < n_edges)
			io_input_edges.resize(n_edges);

		std::cerr << "STATUS: Graph directory is '" << graph_directory_name + util::PATH_SEPARATOR() + "'" << std::endl;

		//
		// read target nodes
		//

		std::cerr << "STATUS: Reading 'naar.net'... " << std::flush;
		for ( size_t pos = 0 ; ! target_it.finished() ; ++pos )
		{
			io_input_edges[pos].the_edge().set_directed_upward(true);
			assert(io_input_edges[pos].the_edge().is_directed_upward());
			assert(!io_input_edges[pos].the_edge().is_directed_downward());

			NodeId tgt = NodeId(target_it.read());
			io_input_edges[pos].the_edge().set_other_node(tgt);
			assert(io_input_edges[pos].the_edge().get_other_node() == tgt);

			io_input_edges[pos].set_original(true);
		}

		target_it.close();
		std::cerr << "OK" << std::endl;


		//
		// read distances
		//

		std::cerr << "STATUS: Reading 'lengtelong.net'... " << std::flush;
		for ( size_t pos = 0 ; ! distance_it.finished() ; ++pos )
		{
			auto distance = distance_it.read();
			io_input_edges[pos].the_edge().set_distance(distance);
			assert(io_input_edges[pos].the_edge().get_distance() == distance);
		}

		distance_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read road types
		//

		std::cerr << "STATUS: Reading 'type.net'... " << std::flush;
		for ( size_t pos = 0 ; ! type_it.finished() ; ++pos )
		{
			auto road_type = type_it.read();
			io_input_edges[pos].the_edge().set_road_type(road_type);
			assert(io_input_edges[pos].the_edge().get_road_type() == road_type);
		}

		type_it.close();
		std::cerr << "OK" << std::endl;

		//
		// read source nodes
		//

		std::vector<uint32_t> first_edge_rec_no =
			io::ortec_format::read_map_node_rec_no_2_first_edge_rec_no(graph_directory_name);

		NodeId src(0u);
		for ( EdgeId e(0u) ; e != n_edges ; ++e )
		{
			assert(src + 1 < first_edge_rec_no.size());
			while (first_edge_rec_no[src + 1] <= e)
			{
				++src;	
				if (first_edge_rec_no[src + 1] == n_edges - 1) {
					break;
				}

				//assert(src + 1 <= first_edge_rec_no.size());
				assert(src < first_edge_rec_no.size());
			}

			if ( io_input_edges.size() <= first_edge_rec_no[src] )
			{
				RITCH_ERROR("Source node index too large (see " << src << " >= " << io_input_edges.size() << ").\n");
				return;
			}

			assert(e < io_input_edges.size());
			io_input_edges[e].set_node_id(src);
			assert(io_input_edges[e].get_node_id() == src);
		}
	}

	void load_default_block_data(const std::string& graph_directory_name, const std::string& hwy_blocks_path_file_name, std::vector<InputEdge>& io_input_edges)
	{
		//
		// read and setup highway blockings, which includes reading country phone codes
		//

		if ( ! hwy_blocks_path_file_name.empty() )
		{
			std::vector<uint16_t> map_edge_rec_no_2_phone_code =
				io::ortec_format::read_map_edge_rec_no_2_country_phone_code(graph_directory_name);

			typedef io::hwy_blocks_format::BlockingDescr Blocking;
			std::vector<Blocking> blockings = io::hwy_blocks_format::read_blockings(hwy_blocks_path_file_name);

			for ( size_t i = blockings.size() ; i > 0 ; --i )
			{
				if ( blockings[i-1]._start_time == 0 && blockings[i-1]._end_time == 0 )
				{
					swap(blockings[i-1], blockings.back());
					blockings.pop_back();
				}
			}

			std::sort(blockings.begin(), blockings.end(),
				[](Blocking& lhs, Blocking& rhs)
				{
					if ( lhs._phone_code < rhs._phone_code )
						return true;

					if ( lhs._phone_code == rhs._phone_code )
						if ( lhs._start_time < rhs._start_time )
							return true;

					return false;
				}
			);

			size_t max_phone_code = 0;
			for ( const auto& blocking : blockings )
				max_phone_code = std::max(max_phone_code, size_t(blocking._phone_code));

			if ( max_phone_code >= size_t(std::numeric_limits<uint16_t>::max()) )
				throw std::runtime_error("Phone code too high to be encoded with 16 bits.");

			std::vector<uint16_t> map_phone_code_2_blocking_info_pos(max_phone_code + 1, 0xffff);
			uint16_t current_phone_code = std::numeric_limits<uint16_t>::max();
			uint16_t pos_n_blockings = 0;

			for ( size_t i = 0 ; i < blockings.size() ; ++i )
			{
				if ( blockings[i]._end_time < blockings[i]._start_time )
					blockings[i]._end_time += TimeDist::PERIOD();

				if ( ! RITCH_IMPLIES(blockings[i]._phone_code == current_phone_code, _blocking_info.size() > 1) )
					throw std::runtime_error("Something went wrong.");

				if ( blockings[i]._phone_code != current_phone_code )
				{
					current_phone_code = blockings[i]._phone_code;

					if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
						throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

					pos_n_blockings = uint16_t(_blocking_info.size());
					map_phone_code_2_blocking_info_pos[current_phone_code] = pos_n_blockings;
					util::append(_blocking_info, {0u, 0u}); // store size and max duration
				}
				else
				{
					uint32_t last_end_time = _blocking_info[_blocking_info.size() - 2] + _blocking_info[_blocking_info.size() - 1];

					if ( uint32_t(blockings[i]._start_time) <= last_end_time )
					{
						if ( i == 0 ) throw std::runtime_error("Something went wrong.");

						uint32_t duration = blockings[i]._end_time - blockings[i-1]._start_time;
						_blocking_info.back() = std::max(_blocking_info.back(), duration);

						continue;
					}
				}

				_blocking_info.push_back(blockings[i]._start_time);
				uint32_t duration = blockings[i]._end_time - blockings[i]._start_time;
				_blocking_info.push_back(duration);

				_blocking_info[pos_n_blockings] += 2; // increase size
				_blocking_info[pos_n_blockings + 1] = std::max(_blocking_info[pos_n_blockings + 1], duration); // update max duration
			}

			if ( _blocking_info.size() >= std::numeric_limits<uint16_t>::max() )
				throw std::runtime_error("Size of blocking info too high to be encoded with 16 bits.");

			for ( size_t i = 0 ; i < io_input_edges.size() ; ++i )
			{
				io_input_edges[i].the_edge().set_blocking_index(0xffff);

				uint16_t phone_code = map_edge_rec_no_2_phone_code[i];

				if ( phone_code >= map_phone_code_2_blocking_info_pos.size() )
					continue;

				if ( util::is_one_of(phone_code, { io::hwy_blocks_format::PHONE_CODE_CZ(), io::hwy_blocks_format::PHONE_CODE_SK() }) )
					if ( ! io::ortec_format::road_type_is_motorway_or_alike(io_input_edges[i].the_edge().get_road_type()) )
						continue;

				io_input_edges[i].the_edge().set_blocking_index(map_phone_code_2_blocking_info_pos[phone_code]);


//				uint16_t phone_code = 43;
//				if ( io::ortec_format::road_type_is_motorway_or_alike(io_input_edges[i].get_road_type()) )
//				io_input_edges[i].the_edge().set_blocking_index(map_phone_code_2_blocking_info_pos[phone_code]);
			}
		}
		else
		{
			for ( size_t i = 0 ; i < io_input_edges.size() ; ++i )
				io_input_edges[i].the_edge().set_blocking_index(0xffff);
		}
	}


	int32_t _period;
	std::vector<Node> _nodes;
	std::vector<Edge> _edges;
	std::vector<uint32_t> _blocking_info;
	std::vector<uint32_t> _middle_node_info;
	std::unordered_map<	std::pair<NodeId, NodeId>, std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>, NodeIdPairHash> _edges_blocks;

public:

	TchSearchGraph
	(
			const std::string& graph_directory_name,
			const std::string& hwy_blocks_path_file_name,
			const std::string& ritch_shortcuts_path_file_name,
			int32_t period
	)
	: _period(period), _nodes(), _edges(), _blocking_info(), _middle_node_info()
	{
		std::vector<InputEdge> input_edges;
		const size_t n_original_edges = io::ortec_format::read_n_edges(graph_directory_name);
		const size_t n_nodes = io::ortec_format::read_n_nodes(graph_directory_name);
		const size_t n_shortcut_edges = io::ritch_shortcuts_format::read_n_edges(ritch_shortcuts_path_file_name);


		input_edges.reserve(n_original_edges + n_shortcut_edges);

		load_ortec_graph_data(graph_directory_name, input_edges);
		load_default_block_data(graph_directory_name, hwy_blocks_path_file_name, input_edges);

		io::ritch_shortcuts_format::InputFile shortcut_input_file(ritch_shortcuts_path_file_name);
		const auto level_of = shortcut_input_file.read_map_node_id_2_level();

		for ( auto& original_edge : input_edges )
		{
			assert(original_edge.is_original());
			NodeId src = original_edge.get_node_id();
			NodeId tgt = original_edge.the_edge().get_other_node();
			
			if ( level_of[src] > level_of[tgt] )
			{
				original_edge.set_node_id(tgt);
				original_edge.the_edge().set_other_node(src);
				original_edge.the_edge().set_directed_upward(false);

				assert(original_edge.get_node_id() == tgt);
				assert(original_edge.the_edge().get_other_node() == src);
				assert(!original_edge.the_edge().is_directed_upward());
				assert(original_edge.the_edge().is_directed_downward());
			}
			//std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> temp_blockinfo;
			//get_blockinfo_original(original_edge.the_edge(), temp_blockinfo, src, tgt);
			//if (temp_blockinfo.size() != 0) {
			//	_edges_blocks.emplace(std::make_pair(NodeId(src), NodeId(tgt)), std::move(temp_blockinfo));
			//}

		}

		size_t middle_node_info_size = 0;
		std::vector<std::vector<std::pair<int32_t, uint32_t>>> temp_middle_node_data;

		std::cerr << "STATUS: Reading edges from '" << ritch_shortcuts_path_file_name << "'... " << std::flush;

		while ( shortcut_input_file.has_next_edge() )
		{
			uint32_t src;
			uint32_t tgt;
			TimeDist time_dist;
			
			input_edges.push_back(InputEdge());
			input_edges.back().set_original(false);

			std::vector<std::pair<int32_t,uint32_t>> temp_middle_node_data_of_edge;
			std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> temp_blockinfo;
			shortcut_input_file.read_next_edge(src, tgt, temp_middle_node_data_of_edge, time_dist, temp_blockinfo);

			if (temp_blockinfo.size() != 0) {
				_edges_blocks.emplace(std::make_pair(NodeId(src), NodeId(tgt)), std::move(temp_blockinfo));
			}

			if ( ! temp_middle_node_data_of_edge.empty() )
			{
				temp_middle_node_data.push_back(std::move(temp_middle_node_data_of_edge));

				if ( temp_middle_node_data.back().size() == 1 )
					middle_node_info_size += 1;
				else
					middle_node_info_size += 2 * temp_middle_node_data.back().size();

				input_edges.back().the_edge().set_middle_node_index(uint32_t(temp_middle_node_data.size() - 1));
			}


			if ( level_of[src] < level_of[tgt]  )
			{
				input_edges.back().set_node_id(src);
				input_edges.back().the_edge().set_other_node(NodeId(tgt));
				input_edges.back().the_edge().set_directed_upward(true);
			}
			else
			{
				input_edges.back().set_node_id(tgt);
				input_edges.back().the_edge().set_other_node(NodeId(src));
				input_edges.back().the_edge().set_directed_upward(false);
			}

			input_edges.back().the_edge().set_time_dist(std::move(time_dist));
			
			if ( level_of[src] < level_of[tgt] )
			{
				assert(input_edges.back().get_node_id() == src);
				assert(input_edges.back().the_edge().get_other_node() == tgt);
				assert(input_edges.back().the_edge().is_directed_upward());
				assert(!input_edges.back().the_edge().is_directed_downward());
			}
			else
			{
				assert(input_edges.back().get_node_id() == tgt);
				assert(input_edges.back().the_edge().get_other_node() == src);
				assert(!input_edges.back().the_edge().is_directed_upward());
				assert(input_edges.back().the_edge().is_directed_downward());
			}
		}

		std::cerr << " OK" << std::endl;

		if ( input_edges.empty() )
		{
			std::cerr << "ERROR: There are no edges in the RITCH_SHORTCUTS file '" << ritch_shortcuts_path_file_name << "'" << std::endl;

			_nodes.clear();
			_edges.clear();
			_blocking_info.clear();
			_middle_node_info.clear();

			return;
		}
		std::cerr << "STATUS: sorting part 1" << std::endl;
		//std::vector<size_t> index = sort_indexes(input_edges);
		//update_block(index);
		std::sort(input_edges.begin(), input_edges.end(),
			[](InputEdge lhs, InputEdge rhs) {
				if (lhs.get_node_id() < rhs.get_node_id() )
					return true;

				if ( lhs.get_node_id() == rhs.get_node_id() )
					if ( lhs.the_edge().get_other_node() < rhs.the_edge().get_other_node() )
						return true;

				if ( lhs.get_node_id() == rhs.get_node_id() )
					if ( lhs.the_edge().get_other_node() == rhs.the_edge().get_other_node() )
						if ( lhs.the_edge().is_directed_upward() && ! rhs.the_edge().is_directed_upward() )
							return true;

				if ( lhs.get_node_id() == rhs.get_node_id() )
					if ( lhs.the_edge().get_other_node() == rhs.the_edge().get_other_node() )
						if ( lhs.the_edge().is_directed_upward() == rhs.the_edge().is_directed_upward() )
							if ( !lhs.is_original() && rhs.is_original() )
								return true;

				return false;
			}
		);

		std::cerr << "STATUS: for loop sorting edges part 1" << std::endl;

		for ( size_t i = input_edges.size() - 1 ; i > 0 ; --i )
		{
			if (
				   input_edges[i].get_node_id() == input_edges[i-1].get_node_id()
				&& input_edges[i].the_edge().get_other_node() == input_edges[i-1].the_edge().get_other_node()
				&& input_edges[i].the_edge().is_directed_upward() == input_edges[i-1].the_edge().is_directed_upward()
				&& input_edges[i].the_edge().is_directed_downward() == input_edges[i-1].the_edge().is_directed_downward()
			)
			{
				assert(input_edges[i].is_original() && !input_edges[i-1].is_original());
				assert(input_edges[i].the_edge().get_middle_node_index() == std::numeric_limits<uint32_t>::max());

				swap(input_edges[i], input_edges.back());
				//_edges_blocks[i] = _edges_blocks[input_edges.size() - 1];
				//_edges_blocks.erase(input_edges.size() - 1);
				input_edges.pop_back();
			}
		}

		//std::vector<size_t> index2 = sort_indexes(input_edges);
		//update_block(index2);
		std::cerr << "STATUS: sorting part 2" << std::endl;
		std::sort(input_edges.begin(), input_edges.end(),
			[](InputEdge& lhs, InputEdge& rhs) {
				if ( lhs.get_node_id() < rhs.get_node_id() )
					return true;

				return false;
			}
		);

		_nodes.resize(n_nodes + 1);
		_edges.reserve(input_edges.size() + 1);
		_middle_node_info.reserve(middle_node_info_size);
		
		_nodes[0]._first_edge = 0;
		NodeId last_source_id_touched(0);
		std::cerr << "STATUS: for loop sorting edges part 2" << std::endl;
		for ( size_t i = 0 ; i != input_edges.size() ; ++i )
		{
			const auto& input_edge = input_edges[i];

			assert(input_edge.get_node_id() != input_edge.the_edge().get_other_node());

			while ( last_source_id_touched < input_edge.get_node_id() )
				_nodes[++last_source_id_touched]._first_edge = NodeId(_edges.size());

			_edges.push_back(std::move(input_edge.the_edge()));
			_nodes[input_edge.get_node_id() + 1]._first_edge = _edges.size();

			last_source_id_touched = input_edge.get_node_id() + 1;

			uint32_t temp_middle_node_index = _edges.back().get_middle_node_index();

			_edges.back().set_middle_node_index(uint32_t(_middle_node_info.size()));


			if ( temp_middle_node_index != std::numeric_limits<uint32_t>::max() )
			{
				const auto& middle_nodes_of_edge = temp_middle_node_data[temp_middle_node_index];

				if ( middle_nodes_of_edge.size() == 1 )
				{
					_middle_node_info.emplace_back();
					_middle_node_info.back() = middle_nodes_of_edge.back().second;
				}
				else
				{
					for ( const auto& item : middle_nodes_of_edge )
					{
						_middle_node_info.emplace_back();
						reinterpret_cast<int32_t&>(_middle_node_info.back()) = item.first;
						_middle_node_info.emplace_back();
						_middle_node_info.back() = item.second;
					}
				}
			}
		
		}

		// Sentinel needed to maintain blocking index.
		_edges.emplace_back();
		_edges.back().set_middle_node_index(uint32_t(_middle_node_info.size()));

		// Finalize
		assert(dbg_graph_is_ok());
	}

	std::vector<size_t> sort_indexes(const std::vector<InputEdge> &iedges) {
		std::vector<size_t> idx(iedges.size());
		std::iota(idx.begin(), idx.end(), 0);
		std::sort(idx.begin(), idx.end(), 
			[&iedges](size_t l, size_t r) {
			InputEdge lhs = iedges[l];
			InputEdge rhs = iedges[r];

			if (lhs.get_node_id() < rhs.get_node_id())
				return true;

			if (lhs.get_node_id() == rhs.get_node_id())
				if (lhs.the_edge().get_other_node() < rhs.the_edge().get_other_node())
					return true;

			if (lhs.get_node_id() == rhs.get_node_id())
				if (lhs.the_edge().get_other_node() == rhs.the_edge().get_other_node())
					if (lhs.the_edge().is_directed_upward() && !rhs.the_edge().is_directed_upward())
						return true;

			if (lhs.get_node_id() == rhs.get_node_id())
				if (lhs.the_edge().get_other_node() == rhs.the_edge().get_other_node())
					if (lhs.the_edge().is_directed_upward() == rhs.the_edge().is_directed_upward())
						if (!lhs.is_original() && rhs.is_original())
							return true;

			return false;
		}
		);
		std::vector<size_t> idx_correct(iedges.size());
		for (int i = 0; i < iedges.size(); i++) {
			idx_correct[idx[i]] = i;
		}

		return idx_correct;
	}

	NodeId get_n_nodes() const
	{
		if ( _nodes.empty() ) return NodeId(0u);
		return NodeId(_nodes.size() - 1);
	}

	EdgeId get_n_edges() const
	{
		if ( _edges.empty() ) return EdgeId(0u);
		return EdgeId(_edges.size() - 1);
	}

	EdgeId edges_begin(NodeId node_id) const noexcept
	{
		assert( node_id + 1 <= _nodes.size() );
		return _nodes[node_id]._first_edge;
	}

	EdgeId edges_end(NodeId node_id) const noexcept
	{
		assert( node_id + 1 <= _nodes.size() );
		if (node_id + 1 == _nodes.size()) {
			return _nodes[node_id]._first_edge;
		}
		return _nodes[node_id + 1]._first_edge;
	}

	NodeId get_other_node(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id).get_other_node();
	}

	bool is_directed_upward(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id).is_directed_upward();
	}

	bool is_directed_downward(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id).is_directed_downward();
	}

	uint32_t eval_distance(EdgeId edge_id, int32_t departure_time) const noexcept
	{
		const Edge& edge = get_edge(edge_id);

		if (edge.has_time_dist())
			return edge.get_time_dist().eval_dist(departure_time);

		return edge.get_distance();
	}

	int32_t eval_time(EdgeId edge_id, int32_t departure_time) const noexcept
	{
		const Edge& edge = get_edge(edge_id);

		if ( edge.has_time_dist() )
			return edge.get_time_dist().eval_time(departure_time);

		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);

		const uint16_t blocking_index = edge.get_blocking_index();
		if ( blocking_index == Edge::INVALID_BLOCKING_ID() )
			return default_time;
		
		departure_time = util::modulo(departure_time, _period);

		const uint32_t size = _blocking_info[blocking_index];
		const uint32_t begin = blocking_index + 2;
		const uint32_t end = begin + size;

		const int32_t last_start_time = _blocking_info[end - 2];
		const int32_t last_duration = _blocking_info[end - 1];
		if ( departure_time < last_start_time + last_duration - TimeDist::PERIOD() )
			return default_time + last_start_time +  last_duration - TimeDist::PERIOD() - departure_time;

		for ( uint32_t i = begin ; i < end ; i += 2 )
		{
			const int32_t start_time = _blocking_info[i];
			const int32_t duration = _blocking_info[i+1];

			if ( departure_time <= start_time - default_time )
				return default_time;
			if ( departure_time <= start_time )
				return default_time + duration;
			if ( departure_time <= start_time + duration )
				return default_time + start_time + duration - departure_time;
		}

		const int32_t first_start_time = _blocking_info[begin];
		const int32_t first_duration = _blocking_info[begin + 1];
		if ( departure_time > first_start_time + TimeDist::PERIOD() - default_time )
			return default_time + first_duration;

		return default_time;
	}

	int32_t get_default_time(EdgeId edge_id) const {
		const Edge& edge = get_edge(edge_id);
		if (edge.has_time_dist())
		{
			return edge.get_time_dist().get_min_time();
		}
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(edge.get_road_type());
		const uint32_t distance = edge.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);
		return default_time;
	}
	
	bool has_time_dist(EdgeId edge_id) const noexcept
	{
		return get_edge(edge_id).has_time_dist();
	}

	int32_t get_min_time(EdgeId edge_id) const noexcept
	{
		if ( get_edge(edge_id).has_time_dist() )
			return get_edge(edge_id).get_time_dist().get_min_time();

		return get_default_travel_time(edge_id);
	}

	int32_t get_max_time(EdgeId edge_id) const noexcept
	{
		if ( get_edge(edge_id).has_time_dist() )
			return get_edge(edge_id).get_time_dist().get_max_time();

		const uint16_t blocking_index = get_edge(edge_id).get_blocking_index();
		if ( blocking_index == Edge::INVALID_BLOCKING_ID() )
			return get_default_travel_time(edge_id);

		const int32_t max_duration = _blocking_info[blocking_index + 1];
		return get_default_travel_time(edge_id) + max_duration;
	}

	std::tuple<int32_t,int32_t> get_min_max_time(EdgeId edge_id) const noexcept
	{
		if ( get_edge(edge_id).has_time_dist() )
			return get_edge(edge_id).get_time_dist().get_min_max_time();

		const int32_t default_travel_time = get_default_travel_time(edge_id);
			
		const uint16_t blocking_index = get_edge(edge_id).get_blocking_index();
		if ( blocking_index == Edge::INVALID_BLOCKING_ID() )
			return std::make_tuple(default_travel_time, default_travel_time);

		const int32_t max_duration = _blocking_info[blocking_index + 1];
		return std::make_tuple(default_travel_time, default_travel_time + max_duration);
	}

	bool has_middle_node(EdgeId edge_id) const noexcept
	{
		assert(get_edge(edge_id + 1).get_middle_node_index() >= get_edge(edge_id).get_middle_node_index());
		return get_edge(edge_id + 1).get_middle_node_index() - get_edge(edge_id).get_middle_node_index() > 0;
	}

	uint32_t get_n_middle_nodes(EdgeId edge_id) const noexcept
	{
		assert(get_edge(edge_id + 1).get_middle_node_index() >= get_edge(edge_id).get_middle_node_index());
		const uint32_t size = get_edge(edge_id + 1).get_middle_node_index() - get_edge(edge_id).get_middle_node_index();

		if ( size == 0u ) return 0u;
		if ( size == 1u ) return 1u;

		return size / 2;
	}

	MiddleNodeDescr get_middle_node_descr(EdgeId edge_id, uint32_t index) const noexcept
	{
		assert(index < get_n_middle_nodes(edge_id));

		const uint32_t begin = get_edge(edge_id).get_middle_node_index();
		const uint32_t end = get_edge(edge_id + 1).get_middle_node_index();

		if ( begin == end )
			return MiddleNodeDescr{ TimeDist::PERIOD(), NodeId::INVALID() };
		if ( begin + 1 == end )
			return MiddleNodeDescr{ TimeDist::PERIOD(), NodeId(_middle_node_info[begin]) };

		assert(begin + index*2 < end);
		assert(begin + index*2 < _middle_node_info.size());

		return reinterpret_cast<const MiddleNodeDescr&>(_middle_node_info[begin + index*2]);
	}

	MiddleNodeDescr get_middle_node_descr_from_departure_time(EdgeId edge_id, int32_t departure_time) const noexcept
	{
		const uint32_t begin = get_edge(edge_id).get_middle_node_index();
		const uint32_t end = get_edge(edge_id + 1).get_middle_node_index();

		if ( begin == end )
			return MiddleNodeDescr { TimeDist::PERIOD(), NodeId::INVALID() };
		if ( begin + 1 == end )
			return MiddleNodeDescr{ TimeDist::PERIOD(), NodeId(_middle_node_info[begin]) };

		auto access_middle_node_descr = [&](uint32_t pos) {
			return reinterpret_cast<const MiddleNodeDescr&>(_middle_node_info[pos]);
		};

		departure_time = util::modulo(departure_time, TimeDist::PERIOD());

		for ( uint32_t pos = begin ; pos != end ; pos += 2 )
			if ( departure_time <= access_middle_node_descr(pos)._right_x )
				return access_middle_node_descr(pos);

		return access_middle_node_descr(begin);
	}

	const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> get_edge_block(NodeId u, NodeId v, EdgeId e) const noexcept {
		auto res = _edges_blocks.find(std::make_pair(u, v));
		if (res == _edges_blocks.end()) {
			//return std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>(); 
			if (!get_edge(e).has_time_dist()) {
				return get_blockinfo_original(e);
			}
			else {
				TimeDist t = get_edge(e).get_time_dist();
				if (t.get_n_time_segments() > 1) {
					std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> r = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();
					std::cout << t.to_text(r);
				}
				RITCH_ASSERT(t.get_n_time_segments() == 1);
				return std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();
			}
		}
		else {
			return _edges_blocks.at(std::make_pair(u, v));
		}
	}

	const std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> get_blockinfo_original(EdgeId eid) const noexcept {
		std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>> res = std::vector<std::tuple<int32_t, int32_t, int32_t, int32_t>>();
		const Edge e = get_edge(eid);
		
		const uint32_t speed = io::ortec_format::road_type_2_default_speed(e.get_road_type());
		const uint32_t distance = e.get_distance();
		const int32_t default_time = calculate_default_travel_time(speed, distance);

		const uint16_t blocking_profile_id = e.get_blocking_index();
		if (blocking_profile_id == Edge::INVALID_BLOCKING_ID())
			return res;

		const uint32_t blocking_profile_size = _blocking_info[blocking_profile_id];
		const uint32_t blocking_info_begin = blocking_profile_id + 2;
		const uint32_t blocking_info_end = blocking_info_begin + blocking_profile_size;

		//if last block start + last block duration > timedistPeriod() 
		if (_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] > uint32_t(TimeDist::PERIOD()))
		{
			if (_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD() > _blocking_info[blocking_info_end]) {
				res.push_back(std::make_tuple(_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), 0,0, default_time));
			}
			else {
				res.push_back(std::make_tuple(_blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), _blocking_info[blocking_info_end], _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD(), default_time));

			}
		}
		int32_t previous_right_x = _blocking_info[blocking_info_end - 2] + _blocking_info[blocking_info_end - 1] - TimeDist::PERIOD();
		if (previous_right_x < 0) { previous_right_x = 0; }
		for (uint32_t i = blocking_info_begin; i < blocking_info_end; i += 2)
		{
			const int32_t& start_time = _blocking_info[i];
			const int32_t& duration = _blocking_info[i + 1];

			//constant travel time
			if (start_time - default_time > 0)
			{
				//time_segments.emplace_back(0, start_time - default_time, default_time);
				if (start_time - default_time > start_time) {
					res.push_back(std::make_tuple(previous_right_x, 0, 0, default_time));
					previous_right_x = start_time - default_time;
				}
				else {
					res.push_back(std::make_tuple(previous_right_x, start_time, start_time + duration, default_time));
					previous_right_x = start_time - default_time;

				}
			}

			//full waiting time (+constant travel time)
			if (start_time > 0)
			{
				res.push_back(std::make_tuple(previous_right_x, start_time, start_time + duration, default_time));
				previous_right_x = start_time;
			}

			//slope part (during duration of block)
			if (start_time + duration > 0)
			{
				res.push_back(std::make_tuple(previous_right_x, start_time, start_time + duration, default_time));
				previous_right_x = start_time + duration;
			}
		}

		if (previous_right_x < TimeDist::PERIOD())
		{
			const int32_t& first_start_time = _blocking_info[blocking_info_begin];
			const int32_t& first_duration = _blocking_info[blocking_info_begin + 1];

			if (first_start_time - default_time <= 0)
			{
				if (first_start_time - default_time + TimeDist::PERIOD() > first_start_time) {
					res.push_back(std::make_tuple(previous_right_x, 0, 0, default_time));
					previous_right_x = first_start_time - default_time + TimeDist::PERIOD();
				}
				else {
					res.push_back(std::make_tuple(previous_right_x, first_start_time, first_start_time + first_duration, default_time));
					previous_right_x = first_start_time - default_time + TimeDist::PERIOD();

				}
			}

			if (first_start_time <= 0)
			{
				res.push_back(std::make_tuple(previous_right_x, first_start_time + TimeDist::PERIOD(), first_start_time + first_duration + TimeDist::PERIOD(), default_time));
				previous_right_x = first_start_time + TimeDist::PERIOD();
			}

			if (first_start_time + first_duration <= 0)
			{
				res.push_back(std::make_tuple(previous_right_x, first_start_time + TimeDist::PERIOD(), first_start_time + TimeDist::PERIOD() + first_duration, default_time));
			}
		}
		return res;
	}

	std::string to_text(EdgeId e) const
	{
		return get_edge(e).to_text();
	}
};

}

#endif // RITCH_TCH_SEARCH_GRAPH_H_
