#ifndef RITCH_UTIL_MISC_H_
#define RITCH_UTIL_MISC_H_

#include <array>
#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <cstdint>
#include <iostream>
#include <initializer_list>
#include <sstream>
#include <string>
#include <type_traits>
#include <unordered_set>

#define RITCH_IMPLIES(a, b) ((!(a)) || (b))
#define RITCH_EQUIV(a, b) (RITCH_IMPLIES((a),(b)), RITCH_IMPLIES((b),(a)))

#define RITCH_BE_VERBOSE true

#define RITCH_STATUS(s) if (RITCH_BE_VERBOSE) std::cout << "STATUS [" << __FILE__ << ":" << __LINE__  << "] " << s << std::flush
#define RITCH_WARNING(s) if (RITCH_BE_VERBOSE) std::cerr << "WARNING [" << __FILE__ << ":" << __LINE__  << "] " << s << std::flush
#define RITCH_ERROR(s) if (RITCH_BE_VERBOSE) std::cerr << "ERROR [" << __FILE__ << ":" << __LINE__  << "] " << s << std::flush

#define RITCH_CONTINUE_STATUS(s) if (RITCH_BE_VERBOSE) std::cout << s << std::flush
#define RITCH_CONTINUE_ERROR(s) if (RITCH_BE_VERBOSE) std::cerr << s << std::flush
#define RITCH_CONTINUE_WARNING(s) if (RITCH_BE_VERBOSE) std::cerr << s << std::flush

#ifndef NDEBUG
#define RITCH_DEBUG true
#else
#define  RITCH_DEBUG false
#endif

#define RITCH_ASSERT(expr) \
			if ( RITCH_DEBUG && !(expr) ) { \
				std::cerr << "\nRITCH_ASSERT [" << __FILE__ << ":" << __LINE__  << "] " \
					<< #expr << " FAILED" << std::flush; \
				abort(); \
			}

#define RITCH_ASSERT_MSG(expr, msg) \
			if ( RITCH_DEBUG && !(expr) ) { \
				std::cerr << "\nRITCH_ASSERT [" << __FILE__ << ":" << __LINE__  << "] " \
					<< #expr " failed --- " << msg << std::flush; \
				abort(); \
			}

#define RITCH_ASSERT_LT(lhs, rhs) \
			if ( RITCH_DEBUG && !((lhs) < (rhs)) ) { \
				std::cerr << "\nRITCH_ASSERT_LT [" << __FILE__ << ":" << __LINE__  << "] " << #lhs << " < " << #rhs " FAILED --- expected " \
					<< lhs << " < " << rhs << std::flush; \
				abort(); \
			}

#define RITCH_ASSERT_LE(lhs, rhs) \
			if ( RITCH_DEBUG && !((lhs) <= (rhs)) ) { \
				std::cerr << "\nRITCH_ASSERT_LT [" << __FILE__ << ":" << __LINE__  << "] " << #lhs << " <= " << #rhs " failed --- expected " \
					<< lhs << " <= " << rhs << std::flush; \
				abort(); \
			}

#define RITCH_ASSERT_GT(lhs, rhs) \
			if ( RITCH_DEBUG && !((lhs) > (rhs)) ) { \
				std::cerr << "\nRITCH_ASSERT_GT [" << __FILE__ << ":" << __LINE__  << "] " << #lhs << " > " << #rhs " FAILED --- expected " \
					<< lhs << " > " << rhs << std::flush; \
				abort(); \
			}

#define RITCH_PAIR_2_ARRAY(pair) (*reinterpret_cast<std::array<decltype(pair.first),2>*>(&pair))


namespace ritch
{
namespace util
{

static std::string PATH_SEPARATOR()
{
	return
		#ifdef _WIN32 
				"\\";
		#else
				"/";
		#endif
}

// computes x mod m, i.e., an r from { 0, 1, ..., m-1 } such that an integer k exists with k*m + r = x
static inline int32_t modulo(int32_t x, int32_t m) noexcept
{
	int32_t result = x % m;
	if ( result < 0 ) result += m;
	return result;
}

// return a pseudo random number lying in the interval [from, to]
static int32_t random_int32(int32_t from, int32_t to) noexcept
{
	RITCH_ASSERT_LT(from, to);
	const int32_t result = int32_t((double(rand()) / double(RAND_MAX)) * double(to - from)) + from;
	RITCH_ASSERT_LE(from, result);
	RITCH_ASSERT_LE(result, to);

	return result;
}


// return a pseudo random number lying in the interval [from, to]
static uint32_t random_uint32(uint32_t from, uint32_t to) noexcept
{
	RITCH_ASSERT_LT(from, to);
	const uint32_t result = uint32_t((double(rand()) / double(RAND_MAX)) * double(to - from)) + from;
	RITCH_ASSERT_LE(from, result);
	RITCH_ASSERT_LE(result, to);

	return result;
}

template <typename Iterator, typename BinaryPredicate>
static bool has_adjacent(Iterator begin, Iterator end, BinaryPredicate binary_predicate)
{
	return end != std::adjacent_find(begin, end, binary_predicate);
}

template <typename Sequence, typename Iterator>
static void append(Sequence& sequence, Iterator begin, Iterator end)
{
	return sequence.insert(sequence.end(), begin, end);
}

template <typename Sequence, typename SequenceToAppend>
void append(Sequence& sequence, const SequenceToAppend& sequence_to_append)
{
	sequence.insert(sequence.end(), sequence_to_append.begin(), sequence_to_append.end());
}

template <typename Sequence, typename Element>
void append(Sequence& sequence, const std::initializer_list<Element>& elements_to_append)
{
	sequence.insert(sequence.end(), elements_to_append);
}

static double rel_error(const double& wrong, const double& right)
{
	if ( fabs(right) <= 0.000001 )
		return std::numeric_limits<double>::max();

	return fabs(wrong - right) / right;
}

template <typename InitListItemType, typename ItemType>
static bool is_one_of(const ItemType& item, const std::initializer_list<InitListItemType>& list)
{
	return std::find(list.begin(), list.end(), item) != list.end();
}

template <typename T>
static bool is_in_range(const T& item, const T& lower, const T& upper)
{
	return item >= lower && item <= upper;
}

static std::string to_text(const bool b)
{
	if (b) return "true";
	else return "false";
}

static std::string to_text(const char c)
{
	std::array<char,3> text;
	text[0] = '\'';
	text[1] = c;
	text[2] = '\'';
	return std::string(text.begin(), text.end());
}

static std::string to_text(const int32_t i)
{
	return std::to_string(i);
}


static std::string to_text(const uint32_t u)
{
	return std::to_string(u);
}
template <typename T>
static std::string to_text(const std::initializer_list<T>& list)
{
	if ( list.size() == 0 )
		return "{}";

	std::stringstream ss;
	ss << "{ " << to_text(*list.begin());

	for ( auto it = list.begin() + 1 ; it < list.end() ; ++it )
		ss << ", " << to_text(*it);

	ss << " }";
	return ss.str();
}


template <typename T>
static std::string to_text(const std::vector<T>& vec)
{
	if ( vec.size() == 0 )
		return "{}";

	std::stringstream ss;
	ss << "{ " << to_text(*vec.begin());

	for ( auto it = vec.begin() + 1 ; it < vec.end() ; ++it )
		ss << ", " << to_text(*it);

	ss << " }";
	return ss.str();
}


template <typename S, typename T>
static std::string to_text(const std::pair<S,T>& item)
{
	std::stringstream ss;
	ss << "(" << to_text(item.first) << ", " << to_text(item.second) << ")";
	return ss.str();
}


template <typename Numeric>
static constexpr bool is_chain_without_gaps(const Numeric first, const Numeric second)
{
	return first + Numeric(1) == second;
}

template <typename Numeric, typename... Args>
static constexpr bool is_chain_without_gaps(const Numeric first, const Numeric second, Args... tail)
{
	return first + static_cast<Numeric>(1) == second && is_chain_without_gaps(second, tail...);
}

template <typename Iterator>
static bool sorted_sequence_is_unique(const Iterator& begin, const Iterator& end)
{
	return std::adjacent_find(begin, end) == end;
}

template <typename Iterator, typename Equal>
static bool sorted_sequence_is_unique(const Iterator& begin, const Iterator& end, Equal equal)
{
	return std::adjacent_find(begin, end, equal) == end;
}

template <typename Iterator>
static bool set_is_unique(const Iterator& begin, const Iterator& end)
{
	std::unordered_set<typename std::iterator_traits<Iterator>::value_type> elements_seen;
	for ( auto it = begin ; it != end ; ++it )
	{
		if ( elements_seen.count(*it) > 0 )
			return false;

		elements_seen.insert(*it);
	}

	return true;
}

template <typename Set>
static bool set_is_unique(const Set& set)
{
	return set_is_unique(set.begin(), set.end());
}

template <typename Sequence>
static void make_sorted_sequence_unique(Sequence& io_sequence)
{
	auto new_end = std::unique(io_sequence.begin(), io_sequence.end());
	io_sequence.resize(std::distance(io_sequence.begin(), new_end));
}

template <typename Sequence, typename Equal>
static void make_sorted_sequence_unique(Sequence& io_sequence, Equal equal)
{
	auto new_end = std::unique(io_sequence.begin(), io_sequence.end(), equal);
	io_sequence.resize(std::distance(io_sequence.begin(), new_end));
}

template <typename Sequence>
static void make_unique(Sequence& io_sequence)
{
	std::sort(io_sequence.begin(), io_sequence.end());
	auto new_end = std::unique(io_sequence.begin(), io_sequence.end());
	io_sequence.resize(std::distance(io_sequence.begin(), new_end));
}

uint16_t plus_one_till_max(uint16_t x)
{
	if ( x == std::numeric_limits<uint16_t>::max() )
		std::numeric_limits<uint16_t>::max();

	return x + 1u;
}

template <typename Numeric>
static constexpr Numeric min(const Numeric first, const Numeric second)
{
	return std::min(first, second);
}

template <typename Numeric, typename... Args>
static constexpr Numeric min(const Numeric first, const Numeric second, Args... tail)
{
	return  min(first, min(second, tail...));
}


}
}




#endif // RITCH_UTIL_MISC_H_
