#ifndef RITCH_UTIL_STRONG_ID_H_
#define RITCH_UTIL_STRONG_ID_H_

#include <cassert>
#include <cstdint>
#include <limits>
#include <ostream>

#define RITCH_DEFINE_STRONG_ID(name) \
class name \
{ \
private: \
	uint32_t _value; \
\
public: \
	/* copy/move assignment operators */ \
	name(const name&) = default; \
	name(name&&) = default; \
\
	/* copy/move assignment constructors */ \
	name& operator= (const name&) = default; \
	name& operator= (name&&) = default; \
\
	/* other constructors */ \
	explicit name() noexcept : _value(std::numeric_limits<uint32_t>::max()) {} \
	explicit name(uint32_t value) noexcept : _value(value) {} \
	explicit name(int32_t value) noexcept : _value(uint32_t(value)) { assert(0 <= value); } \
	explicit name(uint64_t value) noexcept : _value(uint32_t(value)) { assert(value < uint64_t(std::numeric_limits<uint32_t>::max())); } \
	explicit name(int64_t value) noexcept : _value(uint32_t(value)) { assert(0 <= value); assert(value < int64_t(std::numeric_limits<uint32_t>::max())); } \
\
	/* other assignment operators */ \
	name& operator= (uint32_t value) noexcept { _value = value; return *this; } \
	name& operator= (int32_t value) noexcept { assert(0 <= value); _value = uint32_t(value); return *this; } \
	name& operator= (uint64_t value) noexcept { assert(value < uint64_t(std::numeric_limits<uint32_t>::max())); _value = uint32_t(value); return *this; } \
	name& operator= (int64_t value) noexcept { assert(0 <= value); assert(value < int64_t(std::numeric_limits<uint32_t>::max())); _value = uint32_t(value); return *this; } \
\
	static name INVALID() noexcept { return name(); } \
\
	uint32_t& the_value() noexcept { return _value; } \
	const uint32_t& the_value() const noexcept { return _value; } \
\
	friend bool operator== (name lhs, name rhs) noexcept { return lhs._value == rhs._value; } \
	friend bool operator== (name lhs, uint32_t rhs) noexcept { return lhs._value == rhs; } \
	friend bool operator== (uint32_t lhs, name rhs) noexcept { return lhs == rhs._value; } \
	friend bool operator== (name lhs, int32_t rhs) noexcept { if (rhs < 0) return false; return lhs._value == uint32_t(rhs); } \
	friend bool operator== (int32_t lhs, name rhs) noexcept { if (lhs < 0) return false; return uint32_t(lhs) == rhs._value; } \
	friend bool operator== (name lhs, size_t rhs) noexcept { return lhs._value == rhs; } \
	friend bool operator== (size_t lhs, name rhs) noexcept { return lhs == rhs._value; } \
	friend bool operator== (name lhs, int64_t rhs) noexcept { return int64_t(lhs._value) == rhs; } \
	friend bool operator== (int64_t lhs, name rhs) noexcept { return lhs == int64_t(rhs._value); } \
\
	friend bool operator!= (name lhs, name rhs) noexcept { return lhs._value != rhs._value; } \
	friend bool operator!= (name lhs, uint32_t rhs) noexcept { return lhs._value != rhs; } \
	friend bool operator!= (uint32_t lhs, name rhs) noexcept { return lhs != rhs._value; } \
	friend bool operator!= (name lhs, size_t rhs) noexcept { return lhs._value != rhs; } \
	friend bool operator!= (size_t lhs, name rhs) noexcept { return lhs != rhs._value; } \
\
	friend bool operator< (name lhs, name rhs) noexcept { return lhs._value < rhs._value; } \
	friend bool operator< (name lhs, uint32_t rhs) noexcept { return lhs._value < rhs; } \
	friend bool operator< (uint32_t lhs, name rhs) noexcept { return lhs < rhs._value; } \
	friend bool operator< (name lhs, size_t rhs) noexcept { return lhs._value < rhs; } \
	friend bool operator< (size_t lhs, name rhs) noexcept { return lhs < rhs._value; } \
\
	friend bool operator< (name lhs, int32_t rhs) noexcept { \
		if (rhs < 0) return false; \
		else return lhs._value < uint32_t(rhs); \
	} \
\
	friend bool operator< (int32_t lhs, name rhs) noexcept { \
		if (lhs < 0) return true; \
		else return uint32_t(lhs) < rhs._value; \
	} \
\
	friend bool operator<= (name lhs, name rhs) noexcept { return lhs < rhs || lhs == rhs; } \
	friend bool operator<= (name lhs, uint32_t rhs) noexcept { return lhs < rhs || lhs == rhs; } \
	friend bool operator<= (uint32_t lhs, name rhs) noexcept { return lhs < rhs || lhs == rhs; } \
	friend bool operator<= (name lhs, size_t rhs) noexcept { return lhs < rhs || lhs == rhs; } \
	friend bool operator<= (size_t lhs, name rhs) noexcept { return lhs < rhs || lhs == rhs; } \
\
	friend name operator+ (name lhs, uint32_t rhs) noexcept { return name(lhs._value + rhs); } \
	friend name operator+ (name lhs, int32_t rhs) noexcept { assert(rhs >= 0 || uint32_t(-rhs) <= lhs._value); return name(lhs._value + rhs); } \
	friend name operator+ (name lhs, uint16_t rhs) noexcept { return name(lhs._value + rhs); } \
	friend name operator+ (name lhs, int16_t rhs) noexcept { assert(rhs >= 0 || uint32_t(-rhs) <= lhs._value); return name(lhs._value + rhs); } \
\
	friend name operator- (name lhs, uint32_t rhs) noexcept { assert(lhs._value >= rhs); return name(lhs._value - rhs); } \
	friend name operator- (name lhs, uint16_t rhs) noexcept { assert(lhs._value >= rhs); return name(lhs._value - rhs); } \
	friend name operator- (name lhs, int32_t rhs) noexcept { assert(rhs < 0 || lhs._value >= uint32_t(rhs)); return name(lhs._value - rhs); } \
	friend name operator- (name lhs, int16_t rhs) noexcept { assert(rhs < 0 || lhs._value >= uint32_t(rhs)); return name(lhs._value - rhs); } \
\
	name& operator++ () noexcept { ++_value; return *this; } \
	name operator++ (int) noexcept { uint32_t temp = _value; ++_value; return name(temp); } \
\
	name& operator-- () noexcept { assert(0 <= _value); --_value; return *this; } \
	name operator-- (int) noexcept { assert(0 <= _value); uint32_t temp = _value; --_value; return name(temp); } \
\
	friend uint32_t operator<< (name lhs, uint32_t rhs) noexcept { return lhs._value << rhs; } \
	friend uint32_t operator<< (name lhs, int rhs) noexcept { return lhs._value << rhs; } \
\
	friend uint32_t operator>> (name lhs, uint32_t rhs) noexcept { return lhs._value >> rhs; } \
	friend uint32_t operator>> (name lhs, int rhs) noexcept { return lhs._value >> rhs; } \
\
	friend std::ostream& operator<< (std::ostream& os, const name& id) noexcept { os << id._value; return os; } \
\
	operator uint32_t() const noexcept { return _value; } \
\
	friend void swap(name& lhs, name& rhs) noexcept { uint32_t aux_value = lhs._value; lhs._value = rhs._value; rhs._value = aux_value; } \
	friend uint32_t distance(name lhs, name rhs) noexcept { assert(!(lhs < rhs)); return uint32_t(lhs) - uint32_t(rhs); } \
};


#endif // RITCH_UTIL_STRONG_ID_H_
